#ifndef HDR_802_16_H
#define HDR_802_16_H


#include <sys/types.h>
#include <inttypes.h>
#include "ns3/header.h"
#include "ns3/trailer.h"
#include "ns3/wimax2-serviceflow.h"
#include "ns3/tag.h"
#include "ns3/mac48-address.h"
#include "ns3/wimax2-common.h"

namespace ns3 {

#define HDR_MAC802_16_CRC_SIZE 4




// struct gen_mac_header_t {
//   uint8_t ht 			 : 1;	//header type.  1 = just header, 0 = payload/subheader/CRC follows.  
//   uint8_t ec 			 : 1;	//encryption control.  1=encrypted payload, 0 = unencrypted			 
//   //Type field elements 
//   uint8_t type_mesh       : 1;  // mesh subheader: 1=present, 0=absent - note mesh not implemented
//   uint8_t type_arqfb      : 1;  // ARQ Feedback payload: 1=present, 0=absent
//   uint8_t type_ext        : 1;  // Extended type.  1=extended subheader follows generic mac header
//   uint8_t type_frag       : 1;  // Fragmentation subheader: 1=present, 0=absent
//   uint8_t type_pck        : 1;  // Packing subheader: 1=present, 0=absent
//   uint8_t type_fbgm       : 1;  // DL: FAST-FEEDBACK allocation subheader, UL: Grant Management subheader
// 
//   uint8_t rsv1			 : 1;	//reserved (and unsused) bit							
//   uint8_t ci				 : 1;	//crc indicator.  1 = crc included					
//   uint8_t eks			 : 2;	//encryption key sequence.  useful if ec=1				
//   uint8_t rsv2			 : 1;	//reserved (and unused) bit
//   uint16_t len			 : 11;	//total length of MAC PDU in bytes							
//   uint16_t cid;	//   : 16 	  unique connection identifier assigned by BS						
//   uint8_t hcs;		//	 : 8      header check sequence												
// };
/**** Defines Mobility messages (802.16e) ****/
/** Structure of physical profile ID */
struct mac802_16_phy_profile_id {
  uint8_t colocatedFA: 1;
  uint8_t FAconfig: 1;
  uint8_t timefreq_synch: 2;
  uint8_t bs_eirp: 1;
  uint8_t dcducd_ref: 1;
  uint8_t FAindex: 1;
  uint8_t trigger_ref: 1;
};

/** Structure of physical mode ID */
struct mac802_16_phy_mode_id {
  uint8_t bandwidth: 7;
  uint8_t fttsize: 3;
  uint8_t cp: 2;
  uint8_t duration_code: 4;
};


//ARQ feedback IE struct
struct arq_fb_ie {
   uint16_t   cid;                    //ID of the connection being referenced
   uint8_t      last;                   // LAST ARQ IE, false = 0, true = 1
   uint8_t      ack_type        :2;     // Type of ACK - selective =0 , cumulative = 1, cumulative with selective ACK = 2, reserved = 3
   uint16_t   fsn             :11;    // sequence number
};
#define HDR_MAC802_16_ARQFEEDBK_SIZE 5

class Frame_Prefix_Header : public Header {
public:
	uint8_t used_subchannel_bmp : 6;	/* Bitmap for 1024 FFT shown
						 * See Table 268a of 802.16e spec for otehrs
						 * bit		subchannels used by FCH and DL_MAP
						 *  0		0-5
						 *  1		6-9
						 *  2		10-15
						 *  3		16-19
						 *  4		20-25
						 *  5		26-29
						 */
	uint8_t ranging_change : 1;
	uint8_t repetition_coding: 2; //0b00 = only one transmission, 0b01 = 2 transmissions, 0b10 = 4 transmissions, 0b11 = 6 transmissions
	uint8_t coding_indication : 3;
	uint8_t DL_MAP_length; //in slots
	uint8_t reserved : 4;
	
	Frame_Prefix_Header();
	~Frame_Prefix_Header();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};



struct gen_mac_header_t {
  bool ht 			 ;	//will be 0 for GMH, 1 for CDMA_REQ or BW_REQ  
  bool ec 			 ;	//encryption control.  1=encrypted payload, 0 = unencrypted			 
  //Type field elements 
  bool type_mesh       ;  // mesh subheader: 1=present, 0=absent - note mesh not implemented
  bool type_arqfb      ;  // ARQ Feedback payload: 1=present, 0=absent
  bool type_ext        ;  // Extended type.  1=extended subheader follows generic mac header
  bool type_frag       ;  // Fragmentation subheader: 1=present, 0=absent
  bool type_pck        ;  // Packing subheader: 1=present, 0=absent
  bool type_fbgm       ;  // DL: FAST-FEEDBACK allocation subheader, UL: Grant Management subheader

  uint8_t rsv1			 : 1;	//reserved (and unsused) bit in spec, using to label mac packets						
  uint8_t ci			 : 1;	//crc indicator.  1 = crc included					
  uint8_t eks			 : 2;	//encryption key sequence.  useful if ec=1				
  uint8_t rsv2			 : 1;	//reserved (and unused) bit
  uint16_t len			 : 11;	//total length of MAC PDU in bytes							
  uint16_t cid;	//   : 16 	  unique connection identifier assigned by BS						
  uint8_t hcs;		//	 : 8      header check sequence												
};


/** Define bandwidth request header */
struct bw_req_header_t {
  uint8_t ht 	: 1;	//header type (will be 1 for this)
  uint8_t ec 	: 1;	//encryption control (will be 0 for this)
  uint8_t type 	: 3;  	//use this MSB => 00:BW_REQ, 01:CDMA-INIT, 11:CDMA-BW-REQ
  uint32_t br;//: 19;	//number of bytes of uplink bandwidth requested - note that this is serialized/deserialized as 19 bits, but is stored as 32 to prevent a binding generation error
  uint16_t cid;		//connection identifier
  uint8_t hcs;  			//header check sequence
};

/** Define cdma bandwidth request header */
// #top and #br can be increased later;
struct cdma_req_header_t {
  uint8_t ht : 1;
  uint8_t ec : 1;
  uint8_t type : 3;      //use this MSB => 00:BW_REQ, 01:CDMA-INIT, 11:CDMA-BW-REQ
  uint8_t top;           //use 8 bits so max_opportunity is 2^8 = 256
  uint16_t br : 11;    //use this "br" to indicate subcriber id so max => 2^11 = 2048
  uint8_t code;
  uint16_t cid;
};

/** Fragmentation subheader **/
struct frag_subheader_s {
  uint8_t fc : 2; 		//00:no fragmentation, 01:last, 10: first, 11: middle - see enum fragment_status
  uint8_t fsn : 3; 	//fragmentation number
  uint16_t sn:11; 	//If this is an ARQ-enabled connection, this is the BSN, otherwise it is the FSN
                    	//The size of FSN is 3 bits if the type extended is not set
  uint8_t rsv:3;
};

/** Grant Management subheader**/
//stucture for UGS service type
struct grant_map_ugs_s {
  uint8_t si:1; 		//Slipt indicator
  uint8_t pm:1; 		//Poll me indication
  uint16_t rsv:14;
};

struct grant_subheader_s { //Could be cast to grant_map_ugs_s for UGS service type
  uint16_t piggyback_req; 
};


/** Packing subheader **/
struct pack_subheader_s {
  uint8_t fc:2;	//00:no fragmentation, 01:last, 10: first, 11: middle
  uint16_t sn:11; 	//sequence number //If this is an ARQ-enabled connection, this is the BSN, otherwise it is the FSN
                    	//The size of FSN is 3 bits if the type extended is not set
  uint16_t length:11; //max fragment size is 2047 bytes (- 3B for this header = 2044 Byte payload maximum)
};

/** Fast Feedback **/
struct ffb_subheader_s {
  uint8_t alloc_offset:6;
  uint8_t type:2;
};

/** Mesh subheader **/
// not implemented




class Generic_MAC_Header : public Header
{
public:
	Generic_MAC_Header();
	~Generic_MAC_Header();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
	void setHeaderType(bool);		 //1 = just generic MAC header, 0 = subheaders following
	void setEncryptionControl(bool); //1 = encrypted payload, 0 = unencrypted payload
	void setARQFB(bool);			 // ARQ Feedback payload: 1=present, 0=absent
	void setExt(bool);		// Extended type.  1=extended subheader follows generic mac header
	void setFrag(bool);		// Fragmentation subheader: 1=present, 0=absent
	void setPck(bool);		// Packing subheader: 1=present, 0=absent
	void setFFBGM(bool);	// DL: FAST-FEEDBACK allocation subheader, UL: Grant Management subheader
	void setCI(bool); 		//set CRC indicator
	void setEks(uint8_t);	//set encryption key sequence
	void setPduLen(uint16_t); //set total length of MAC PDU in bytes
	void setCID(uint16_t);	   //set Connection ID (assigned by BS)
	void setHCS(uint8_t);	   //set header check sequence
	gen_mac_header_t 	GMH;		


private:
};

class Signaling_MAC_Header : public Header {
public:
	Signaling_MAC_Header();
	~Signaling_MAC_Header();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
	void setHeaderType(uint8_t);
	void setEncryptionControl(uint8_t);
	void setType(uint8_t);
	void setBandwidthRequested(uint32_t);
	void setCID(uint16_t);
	void setHCS(uint8_t);
	bw_req_header_t BRH;
};



/**** Defines some constant for the maximum size of messages ****/
/* When sending a packet, the message is copied for each 
 * destination, and all data must be stored in the packet
 * some of this requires arrays, which have maximum size as
 * defined here
 */
#define MAX_MAP_IE  2048
#define MAX_PROFILE 10
#define MAX_NBR     10




/** Ranging response frame */
struct mac802_16_rng_rsp_frame {
  uint8_t type; //5
  uint8_t uc_id;
  /*TLV values*/
  uint8_t pw_adjust;
  uint32_t freq_adjust;
  uint8_t rng_status; //1 = continue, 2 = abord, 3 = success, 4 = rerange
  /* byte 0: the least robust DIUC that may be used by BS for 
   * transmissions to the SS
   * byte 1: Configuration Change Count of DCD defining the burst profile
   * associated with DIUC 
   */
  uint16_t dl_op_burst_profile;
  Mac48Address ss_mac_address; //6 bytes
  uint16_t basic_cid;
  uint16_t primary_cid;
  uint8_t aas_bc_perm; 
  uint8_t top;
  uint8_t code;
};
#define RNG_RSP_SIZE 28

class RANGING_RESPONSE_FRAME : public Header {
public:
	mac802_16_rng_rsp_frame data;
	RANGING_RESPONSE_FRAME();
	~RANGING_RESPONSE_FRAME();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};


/**** Defines IEs ****/
/** Defines DL_MAP Information Element (see p462.) */
struct mac802_16_dlmap_ie {
  uint16_t cid;
  uint8_t diuc 			: 4;            //p463
  uint8_t preamble 		: 1;
  uint16_t start_time 	: 11;
  //may contain extended info
// for OFDMA RPI
  uint8_t n_cid;
  uint8_t symbol_offset;
  uint8_t subchannel_offset	: 6;
  uint8_t boosting			: 3;
  uint8_t num_of_symbols		: 7;

//Chakchai: in IE maps, #num_of_subchannels = #slots (vertical allocation within burst) so in this version we increase from 6 => 14 but IE size during transmission will be the same
  uint16_t num_of_subchannels: 14;

  uint8_t repition_coding_indicator: 2;
// rpi 
};
#define DL_MAP_IE_SIZE 10 	// byte
#define DL_MAP_HEADER_SIZE 13 	//byte

struct mac802_16_compressed_ie {
	uint8_t diuc : 4;
	uint16_t cid;
	uint8_t OFDMASymbol_offset;
	uint8_t subchannel_offset : 6;
	uint8_t boost : 3;
	uint8_t num_OFDMASymbol : 7;
	uint8_t num_subchannel : 6;
	uint8_t repetition_coding : 2;
};

class DL_MAP_IE : public Header {
public:
	mac802_16_compressed_ie ie;
	DL_MAP_IE();
	~DL_MAP_IE();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};


/** fast Ranging IE (802.16e) */
struct mac802_16_fast_ranging_ie {
  Mac48Address mac_addr;  //48 bits
  uint8_t uiuc : 4; 
  uint16_t duration : 12; //in OFDM symbols
};
#define FAST_RANGING_IE_SIZE 9 // bytes

class Fast_Ranging_IE : public Header {
public:
	mac802_16_fast_ranging_ie ie;
	Fast_Ranging_IE();
	~Fast_Ranging_IE();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

/** cdma_map_ie (use only for cdma-initial-ranging-req not for cdma-bw-req) ;change cdma-bw-req later**/
struct mac802_16_cdma_ie {
  uint8_t duration : 6;
  uint8_t uiuc : 4;
  uint8_t repetition : 2;	//00 = none, 01 = 2x, 10 = 4x, 11 = 6x
  uint8_t frame_number : 4;
  uint8_t code;				//Ranging Code
  uint8_t symbol;			//Ranging Symbol
  uint8_t subchannel : 7;	//Ranging Subchannel
  uint8_t bw_req : 1;
};
#define UL_CDMA_MAP_IE_SIZE 5 	// byte

class CDMA_MAP_IE : public Header {
public:
	mac802_16_cdma_ie ie;
	CDMA_MAP_IE();
	~CDMA_MAP_IE();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

/** Defines UL_MAP IE (see p464.) */
struct mac802_16_ulmap_ie {
  uint16_t cid;
  uint16_t start_time : 11;
  uint8_t sub_channel_index : 5;
  uint8_t uiuc : 4;
  uint16_t duration : 11;
  uint8_t midamble_rep : 2;
  //may contain additional info depending on uiuc
  uint8_t extended_uiuc : 4;
  uint8_t length : 4;
  
  Fast_Ranging_IE fast_ranging;

  CDMA_MAP_IE cdma_ie;

// added for OFDMA rpi
   uint8_t symbol_offset;
   uint8_t subchannel_offset: 7;
   uint8_t num_of_symbols: 7;

//Chakchai: in IE maps, #num_of_subchannels = #slots (vertical allocation within burst) so in this version we increase from 7 => 15 but IE size during transmission will be the same
  
   uint16_t num_of_subchannels: 15;
   uint8_t ranging_method: 2;
// rpi
};
#define UL_MAP_IE_SIZE 4 	// byte
#define UL_MAP_HEADER_SIZE 8 	//byte


//as defined in the 802.16e spec Table 287
struct mac802_16_ofdma_ulmap_ie {
	uint16_t cid;
	uint8_t uiuc:4;
	uint16_t duration:10;		//in slots
	uint8_t repetition_coding:2; //only should be used for QPSK burst profile
	
	CDMA_MAP_IE cdma_ie;
};

class UL_MAP_IE : public Header {
public:
	mac802_16_ofdma_ulmap_ie ie;
	UL_MAP_IE();
	~UL_MAP_IE();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};



/**** Defines burst profiles ****/

/* Burst profiles are TLV encoded...we just pick the info we 
 * are interested in.
 */
/** Defines DCD profile */
struct mac802_16_dcd_profile {
  uint8_t diuc : 4;
  uint32_t frequency;
  uint8_t fec;
  //may have more info
};
#define DCD_PROFILE_SIZE 12

/** Defines UCD profile */
struct mac802_16_ucd_profile {
  uint8_t uiuc : 4;
  uint8_t fec;
  //may have more info
};
#define UCD_PROFILE_SIZE 6

/**** Defines frames ****/
/** DCD frame */
struct mac802_16_dcd_frame {
  uint8_t type;
  uint8_t dcid;
  uint8_t config_change_count;
  //info for all channels: TLV encoded
  uint8_t frame_duration_code;
  uint32_t frame_number;// : 24; //note this is a full 32 bits here becuase otherwise there is a binding gen error - only 24 are Serialized/Deserialized
  uint8_t ttg;
  uint8_t rtg;
  uint32_t frequency;
  //downlink burst profiles
  uint32_t nb_prof;
  mac802_16_dcd_profile profiles[MAX_PROFILE];
};

class DCD_FRAME_HEADER : public Header {
public:
	mac802_16_dcd_frame mf;
	DCD_FRAME_HEADER();
	~DCD_FRAME_HEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};


//+3 for the end of map profile
#define GET_DCD_SIZE(X) 22+X*DCD_PROFILE_SIZE+3

/** DL_MAP frame */
struct mac802_16_dl_map_frame {
  uint8_t type;
  uint8_t dcd_count;
  Mac48Address bsid; 
// added extra for OFDMA -rpi
  uint8_t num_of_ofdma_symbols;
// rpi
  //DL_MAP IEs
  uint32_t nb_ies;
  DL_MAP_IE ies[MAX_MAP_IE]; 
};

//there is X IEs 
//#define GET_DL_MAP_SIZE(X) 8+X*DL_MAP_IE_SIZE
#define GET_DL_MAP_SIZE(X) 13+X*DL_MAP_IE_SIZE


class DL_MAP_FRAME_HEADER : public Header {
public:
	mac802_16_dl_map_frame mf;
	DL_MAP_FRAME_HEADER();
	~DL_MAP_FRAME_HEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};



/**** Defines Dynamic Service Addition messages ****/
/** DSA request frame */
struct mac802_16_dsa_req_frame {
  uint8_t type; //11
  uint16_t transaction_id;
  /*TLV values*/
  bool uplink; //direction of the flow, normaly coded in TLV with
               //type 145 or 146 (see p647).
  Wimax2ServiceFlow * staticflow ;
  uint16_t cid;
};
//parameter X indicates if cid is present (i.e request from BS)
#define GET_DSA_REQ_SIZE(X) 6+4*X

class DSA_REQ_FRAME_HEADER : public Header {
public:
	mac802_16_dsa_req_frame data;
	DSA_REQ_FRAME_HEADER();
	~DSA_REQ_FRAME_HEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};


/** DSA response frame */
struct mac802_16_dsa_rsp_frame {
  uint8_t type; //12
  uint16_t transaction_id;
  uint8_t confirmation_code;
  /*TLV values*/
  bool uplink; //direction of the flow, normaly coded in TLV with
               //type 145 or 146 (see p647).
  Wimax2ServiceFlow * staticflow ;
  uint16_t cid;
};
//parameter X indicates if cid is present (i.e response from BS)
#define GET_DSA_RSP_SIZE(X) 6+4*X

class DSA_RSP_FRAME_HEADER : public Header {
public:
	mac802_16_dsa_rsp_frame data;
	DSA_RSP_FRAME_HEADER();
	~DSA_RSP_FRAME_HEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

/** DSA Acknowledgement frame */
struct mac802_16_dsa_ack_frame {
  uint8_t type; //13
  uint16_t transaction_id;
  uint8_t confirmation_code;
  /*TLV values*/
  bool uplink; //direction of the flow, normaly coded in TLV with
               //type 145 or 146 (see p647).
};
#define DSA_ACK_SIZE 6

class DSA_ACK_FRAME_HEADER : public Header {
public:
	mac802_16_dsa_ack_frame data;
	DSA_ACK_FRAME_HEADER();
	~DSA_ACK_FRAME_HEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};


/** UCD frame */
struct mac802_16_ucd_frame {
  uint8_t type;
  uint8_t config_change_count;
  uint8_t rng_backoff_start;
  uint8_t rng_backoff_end;
  uint8_t req_backoff_start;
  uint8_t req_backoff_end;
  //info for overall channel
  /*
  uint16_t rsv_timeout;
  */
  uint16_t bw_req_size;
  uint16_t rng_req_size;
  //uplink burst profile
  uint32_t nb_prof;
  mac802_16_ucd_profile profiles[MAX_PROFILE];
};

class UCD_FRAME_DATA : public Header {
public:
	mac802_16_ucd_frame ucdf;
	UCD_FRAME_DATA();
	~UCD_FRAME_DATA();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

//+3 for the end of map profile
#define GET_UCD_SIZE(X) 14+X*UCD_PROFILE_SIZE+3

/** UL_MAP frame */
struct mac802_16_ul_map_frame {
  uint8_t type;
  uint8_t ucid;
  uint8_t ucd_count;
  uint32_t allocation_start;
//added extra for OFDMA - rpi
  uint8_t num_ofdma_symbols;
// rpi end

  //UL_MAP IEs
  uint32_t nb_ies;
  UL_MAP_IE ies[MAX_MAP_IE];
};

class UL_MAP_FRAME_HEADER : public Header {
public:
	mac802_16_ul_map_frame mf;
	UL_MAP_FRAME_HEADER();
	~UL_MAP_FRAME_HEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

//#define GET_UL_MAP_SIZE(X) 7+X*UL_MAP_IE_SIZE
#define GET_UL_MAP_SIZE(X) 8+X*UL_MAP_IE_SIZE



/** Registration response frame */
struct mac802_16_reg_rsp_frame {
  uint8_t type;
  uint8_t response;
  /*TLV values*/
  uint8_t ss_mngmt_support;
  uint16_t sec_mngmt_cid;
};
#define REG_RSP_SIZE 5


class REGISTRATION_RESPONSE_FRAME : public Header {
public:
	mac802_16_reg_rsp_frame data;
	REGISTRATION_RESPONSE_FRAME();
	~REGISTRATION_RESPONSE_FRAME();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};


/**  request frame */
struct mac802_16_reg_req_frame {
  uint8_t type;
  /*TLV values*/
  uint8_t ss_mngmt_support;
  uint8_t ip_mngmt_support;
  uint16_t uplink_cid_support;
};
#define REG_REQ_SIZE 5


class REGISTRATION_REQUEST_FRAME : public Header {
public:
	mac802_16_reg_req_frame data;
	REGISTRATION_REQUEST_FRAME();
	~REGISTRATION_REQUEST_FRAME();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

/**** Defines ranging messages ****/
/** Ranging request frame */
struct mac802_16_rng_req_frame {
  uint8_t type;
  uint8_t dc_id;
  /*TLV values*/
  /* Requested Downlink Burst Profile
   * bits 0-3: DIUC of the downlink burst profile
   * bits 4-7: LSB of Configuration Change Count value
   * of DCD defining the burst profile assocciated with DIUC
   */
  uint8_t req_dl_burst_profile;
  Mac48Address ss_mac_address;
  uint32_t channel_num; //rpi - added to pass over the propagation model physical channel from MN to BS.  
  //uint8_t aas_bc_cap; //broadcast capability. optional
};
#define RNG_REQ_SIZE 13 //max value

class RANGING_REQUEST_FRAME : public Header {
public:
	mac802_16_rng_req_frame data;
	RANGING_REQUEST_FRAME();
	~RANGING_REQUEST_FRAME();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

class CDMA_REQUEST_HEADER : public Header {
public:
	cdma_req_header_t data;
	CDMA_REQUEST_HEADER();
	~CDMA_REQUEST_HEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

// /** Structure containing physical layer information */
// struct phy_info_t {
//   double freq_;              //the frequency at which it is sent
//   Ofdm_mod_rate modulation_; //modulation at which the packet was sent
//   double g_; //The cyclic prefix information. Used for synchronization
//   double subcarrier_lbound;
//   double subcarrier_ubound;
//   double time_lbound;
//   double time_ubound;
//   int num_subchannels;
//   int subchannel_offset;
//   int num_OFDMSymbol;
//   int OFDMSymbol_offset;
//   int channel_index;
//   int direction;   // 0 = DL 1= UL // rpi , just one bit needed, so eithr use :1 or use enum or bool  
// };

class PHY_INFO_Header: public Header {
public: 
  PHY_INFO_Header();
  ~PHY_INFO_Header();
  double freq_;              //the frequency at which it is sent
  Ofdm_mod_rate modulation_; //modulation at which the packet was/will be sent
  double g_; //The cyclic prefix information. Used for synchronization
  double subcarrier_lbound;
  double subcarrier_ubound;
  double time_lbound;
  double time_ubound;
  uint32_t num_subchannels;
  uint32_t subchannel_offset;
  uint32_t num_OFDMSymbol;
  uint32_t OFDMSymbol_offset;
  int32_t channel_index;
  int32_t direction;   // 0 = DL 1= UL 
  virtual uint32_t GetSerializedSize() const;
  virtual void Serialize(Buffer::Iterator start) const;
  virtual uint32_t Deserialize(Buffer::Iterator start);
  virtual void Print (std::ostream &os) const;
  static TypeId GetTypeId(void);
  virtual TypeId GetInstanceTypeId (void) const;

};

class FRAG_SUBHEADER : public Header {
public:
	frag_subheader_s data;
	FRAG_SUBHEADER();
	~FRAG_SUBHEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

//note that this should be improved to allow for ugs header type
class GRANT_MAP_SUBHEADER : public Header {
public:
	grant_subheader_s data;
	GRANT_MAP_SUBHEADER();
	~GRANT_MAP_SUBHEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

class PACKING_SUBHEADER : public Header {
public:
	pack_subheader_s data;
	PACKING_SUBHEADER();
	~PACKING_SUBHEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

class FFB_SUBHEADER : public Header {
public:
	ffb_subheader_s data;
	FFB_SUBHEADER();
	~FFB_SUBHEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};




/** ARQ Feedback **/
#define MAX_NUM_ACK 100


class ARQ_FB_IE_HEADER : public Header {
public:
	ARQ_FB_IE_HEADER();
	~ARQ_FB_IE_HEADER();
	static TypeId GetTypeId(void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
	void last(bool l);
	bool last_;
	arq_fb_ie arqfbie;
};


/** Code BS using index in scan request */
struct mac802_16_mob_scn_req_bs_index {
  uint8_t neighbor_bs_index;
  uint8_t scanning_type: 3; //0: scanning without association
                           //1: scanning+assoc level 0
                           //2: scanning+assoc level 1
                           //3: scanning+assoc level 2
                           //4-7: reserved
};
 
/** Code BS using full address in scan request */
struct mac802_16_mob_scn_req_bs_full {
  Mac48Address recommended_bs_id;   /* 6 bytes */
  uint8_t scanning_type: 3; //0: scanning without association
                           //1: scanning+assoc level 0
                           //2: scanning+assoc level 1
                           //3: scanning+assoc level 2
                           //4-7: reserved
};

/** Scan request frame */
class mac802_16_mob_scn_req_frame : public Header {
public:
  uint8_t type; //54
  uint8_t scan_duration; //units of frames
  uint8_t interleaving_interval; //units of frames
  uint8_t scan_iteration; //in frame
  uint8_t n_recommended_bs_index; //number of BS recommended
  uint8_t ccc; //present if n_recommended_bs_index!= 0
  mac802_16_mob_scn_req_bs_index rec_bs_index[MAX_NBR];  //MAX_NBR defined as 10 above
  uint8_t n_recommended_bs_full;
  mac802_16_mob_scn_req_bs_full rec_bs_full[MAX_NBR];
  
    
    mac802_16_mob_scn_req_frame();
    ~mac802_16_mob_scn_req_frame();
  	static TypeId GetTypeId(void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

/** Code BS using index in scan response */
struct mac802_16_mob_scn_rsp_bs_index {
  uint8_t neighbor_bs_index;
  uint8_t scanning_type: 3; //0: scanning without association
                           //1: scanning+assoc level 0
                           //2: scanning+assoc level 1
                           //3: scanning+assoc level 2
                           //4-7: reserved
  /* next present if scanning is 2 or 3*/
  uint8_t rdv_time;
  uint8_t cdma_code;
  uint8_t transmission_opp_offset;
};

/** Code BS using full address in scan response */
struct mac802_16_mob_scn_rsp_bs_full {
  Mac48Address recommended_bs_id;   /* 6 bytes */
  uint8_t scanning_type: 3; //0: scanning without association
                           //1: scanning+assoc level 0
                           //2: scanning+assoc level 1
                           //3: scanning+assoc level 2
                           //4-7: reserved
  /* next present if scanning is 2 or 3*/
  uint8_t rdv_time;
  uint8_t cdma_code;
  uint8_t transmission_opp_offset;
};

/** Scan response frame */
class mac802_16_mob_scn_rsp_frame : public Header{
public:
  uint8_t type; //55
  uint8_t scan_duration; //units of frames
  uint8_t report_mode: 2;
  uint8_t report_period;
  uint8_t report_metric;
  /*next information present only if scan duration !=0*/
  uint8_t start_frame: 4;
  uint8_t interleaving_interval;
  uint8_t scan_iteration;
  uint8_t n_recommended_bs_index;
  //next if n_recommended_bs_index !=0
  uint8_t ccc_mob_nbr_adv;
  mac802_16_mob_scn_rsp_bs_index rec_bs_index[MAX_NBR];
  uint8_t n_recommended_bs_full;
  mac802_16_mob_scn_rsp_bs_full rec_bs_full[MAX_NBR];
  /* end if scan duration !=0 */
  /* TLV information */
  
  
    mac802_16_mob_scn_rsp_frame();
    ~mac802_16_mob_scn_rsp_frame();
  	static TypeId GetTypeId(void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};





/** Structure for Handover indication */
struct mac802_16_mob_ho_ind_bs {
  uint8_t temp_bsid: 3;
};


/** Structure for BSHO request */
struct mac802_16_mob_bsho_req_mode_new_bs {
  Mac48Address neighbor_bsid; /* 6 bytes */
  uint8_t temp_bsid: 3;
 
};

/** Structure for BSHO request */
struct mac802_16_mob_bsho_req_mode_current_bs {
  uint8_t temp_bsid: 3;
};


/** Structure for BSHO response */
struct mac802_16_mob_bsho_rsp_mode_new_bs2 {
  Mac48Address neighbor_bsid; /* 6 bytes */
  uint8_t temp_bsid: 3;
  uint16_t new_cid[];
};



/** Information about a neighbor BS */
struct mac802_16_nbr_adv_info {
  uint8_t length;
  mac802_16_phy_profile_id phy_profile_id;
  uint8_t fa_index; //if FA index indicator=1 in phy_profile_id
  uint8_t bs_eirp;
  Mac48Address nbr_bsid;    
  uint8_t preamble_index; //in OFDM, 5lsb=
  uint8_t ho_process_opt;
  uint8_t sched_srv_supported;
  uint8_t dcd_ccc: 4;
  uint8_t ucd_ccc: 4;
  /* other TLV information */
  bool dcd_included; //tell if the dcd is included
  DCD_FRAME_HEADER dcd_settings;
  bool ucd_included; //tell if the ucd is included
  UCD_FRAME_DATA ucd_settings;
  bool phy_included; //tell if the phy mode ID is included
  mac802_16_phy_mode_id phy_mode_id;
};


/** Neighbor advertisment frame */
struct mac802_16_mob_nbr_adv_frame {
  uint8_t type; //53
  uint8_t skip_opt_field; //if 0 include next field, if 1 skip next field
  uint32_t operatorID;//: 24; //note that for some reason bit sets on a uint32_t are causing binding generation errors, this is still serialized and deserialized as a 24 bit number, but not declared as such
  uint8_t ccc;
  uint8_t frag_index: 4;
  uint8_t total_frag: 4;
  uint8_t n_neighbors; //number of neighbors
  mac802_16_nbr_adv_info nbr_info[MAX_NBR];
};

class NEIGHBOR_ADVERTISEMENT_FRAME_HEADER : public Header {
public:
	mac802_16_mob_nbr_adv_frame data;
	NEIGHBOR_ADVERTISEMENT_FRAME_HEADER();
	~NEIGHBOR_ADVERTISEMENT_FRAME_HEADER();
	static TypeId GetTypeId(void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};


/** HO indication frame */
struct mac802_16_mob_ho_ind_frame {
  uint8_t type; //59
  uint8_t mode: 2;
  //if mode==0b00
  uint8_t ho_ind_type: 2;
  uint8_t rng_param_valid_ind: 2;
  //next valid if ho_ind_type==0b00
  Mac48Address target_bsid; /* 6 bytes */
  
  //62 bits
  
  //if mode==0b01
  uint8_t mdhofbss_ind_type: 2;
  //next valid if mdhofbss_ind_type==0b00
  uint8_t bsid: 3;
  uint8_t action_time;
  //end valid if mdhofbss_ind_type==0b00
  
  //74 bits

  //if mode==0b10
  //uint8_t mdhofbss_ind_type: 2;
  //next valid if mdhofbss_ind_type==0b00
  uint8_t diversity_set_included;
  //next valid if diversity_set_included==1
  uint8_t anchor_bsid: 3;
  uint8_t n_bs;
  mac802_16_mob_ho_ind_bs bs[MAX_NBR];
  //end valid if diversity_set_included==1
  //uint8_t action_time;
  //end valid if mdhofbss_ind_type==0b00
  uint8_t preamble_index;
  
  /* TLV information */
};

class MOB_HO_IND_FRAME_HEADER : public Header {
public:
	mac802_16_mob_ho_ind_frame data;
	MOB_HO_IND_FRAME_HEADER();
	~MOB_HO_IND_FRAME_HEADER();
	static TypeId GetTypeId(void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};


/** Structure for BSHO response */
struct mac802_16_mob_bsho_rsp_rec {
  Mac48Address neighbor_bsid; /* 6 bytes */
  uint8_t preamble_index;
  uint8_t service_level_prediction;
  uint8_t ho_process_optimization;
  uint8_t net_assisted_ho_supported: 1;
  uint8_t ho_id_included_indicator: 1;
  //if ho_id_included_indicator==1
  uint8_t ho_id;
  //end if
  uint8_t ho_autho_policy_indicator: 1;
  //if ho_autho_policy_indicator==1
  uint8_t ho_autho_policy_support;
  //end if
  
};

/** Structure for BSHO request */
struct mac802_16_mob_bsho_req_mode_new_bs3 {
  int neighbor_bsid; /* 6 bytes */
  uint8_t temp_bsid: 3;
  uint16_t new_cid[MAX_NBR];
  uint16_t new_said[MAX_NBR];
  int cqich_id; //variable 
  uint8_t feedback_channel_off: 6;
  uint8_t period: 2;
  uint8_t frame_offset: 3;
  uint8_t duration: 3;
  uint8_t mimo_permutation_feedback: 2; 
};



/** BSHO response frame */
struct mac802_16_mob_bsho_rsp_frame {
  uint8_t type; //58
  uint8_t mode: 3;
  
  //if mode == 0b000
  uint8_t ho_operation_mode: 1;
  uint8_t n_recommended;
  uint8_t resource_retain_flag: 1;
  mac802_16_mob_bsho_rsp_rec n_rec[MAX_NBR];

  //if mode == 0b001
  uint8_t tmp_bsid: 3;
  uint8_t ak_change_indicator: 1;
  uint8_t n_cids;
  uint16_t cids[MAX_NBR];
  uint8_t n_saids;
  uint16_t saids[MAX_NBR];

  //if mode == 0b011
  uint8_t n_new_bs: 3;
  mac802_16_mob_bsho_req_mode_new_bs new_bs[MAX_NBR];
  uint8_t n_current_bs: 3;
  mac802_16_mob_bsho_req_mode_current_bs current_bs[MAX_NBR];
  mac802_16_mob_bsho_rsp_mode_new_bs2 new_bs2[MAX_NBR];

  mac802_16_mob_bsho_req_mode_new_bs3 new_bs3[MAX_NBR]; //same as request
  uint8_t action_time: 7;
  /* TLV information */

};


class BSHO_RESPONCE_FRAME_HEADER : public Header {
public:
	mac802_16_mob_bsho_rsp_frame data;
	BSHO_RESPONCE_FRAME_HEADER();
	~BSHO_RESPONCE_FRAME_HEADER();
	static TypeId GetTypeId(void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};




/** Structure for MSHO request */
struct mac802_16_mob_msho_req_bs_index {
  uint8_t neighbor_bs_index;
  uint8_t preamble_index;
  uint8_t bs_cinr_mean; //if report_metric[0]==1
  uint8_t bs_rssi_mean; //if report_metric[1]==1
  uint8_t relative_delay; //if report_metric[2]==1
  uint8_t service_level_prediction: 3;
  uint8_t arrival_time_diff_ind: 1;
  //next if arrival_time_diff_ind==1
  uint8_t arrival_time_diff: 4;
};


/** Structure for MSHO request */
struct mac802_16_mob_msho_req_current_bs {
  uint8_t temp_bsid: 4;
  uint8_t bs_cinr_mean; //if report_metric[0]==1
  uint8_t bs_rssi_mean; //if report_metric[1]==1
  uint8_t relative_delay; //if report_metric[2]==1
  uint8_t bs_rtd; //if report_metric[3]==1
};



/** MSHO request frame */
struct mac802_16_mob_msho_req_frame {
  uint8_t type; //57
  uint8_t report_metric;
  uint8_t n_new_bs_index;
  //next if n_recommended_bs_index !=0
  uint8_t ccc_mob_nbr_adv;
  mac802_16_mob_msho_req_bs_index bs_index[MAX_NBR];
  //end 
  uint8_t n_new_bs_full;
  mac802_16_mob_msho_req_bs_index bs_full[MAX_NBR];
  uint8_t n_current_bs;
  mac802_16_mob_msho_req_current_bs bs_current[MAX_NBR];
  /* other TLV information */
};

class MSHO_REQUEST_FRAME_HEADER : public Header {
public:
	mac802_16_mob_msho_req_frame data;
	MSHO_REQUEST_FRAME_HEADER();
	~MSHO_REQUEST_FRAME_HEADER();
	static TypeId GetTypeId(void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};



/** This class contains helpers for manipulating 802.16 messages
 *  and getting the packet size 
 */
class Mac802_16pkt {
 public:
  /**
   * Return the size of the MOB_NBR-ADV frame
   * @param frame The frame 
   */
  static int getMOB_NBR_ADV_size(mac802_16_mob_nbr_adv_frame *frame);

  /**
   * Return the size of the MOB_SCN-REQ 
   * @param frame The frame 
   */
  static int getMOB_SCN_REQ_size(mac802_16_mob_scn_req_frame *frame);

  /**
   * Return the size of the MOB_SCN-RSP 
   * @param frame The frame 
   */
  static int getMOB_SCN_RSP_size(mac802_16_mob_scn_rsp_frame *frame);

  /**
   * Return the size of the MOB_MSHO-REQ 
   * @param frame The frame 
   */
  static int getMOB_MSHO_REQ_size(mac802_16_mob_msho_req_frame *frame);

  /**
   * Return the size of the MOB_BSHO-RSP 
   * @param frame The frame 
   */
  static int getMOB_BSHO_RSP_size(mac802_16_mob_bsho_rsp_frame *frame);

  /**
   * Return the size of the MOB_HO-IND 
   * @param frame The frame 
   */
  static int getMOB_HO_IND_size(mac802_16_mob_ho_ind_frame *frame);
};

class MISC_OTHER_HEADER : public Header {
public:
	MISC_OTHER_HEADER();
	~MISC_OTHER_HEADER();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
	
	uint16_t num_of_acks;
	struct arq_fb_ie arq_ie[MAX_NUM_ACK]; 		//if the type_arqfb is true, we have at least 1 element
                                                       	//we use the last bit to check if it is the last element
	uint8_t gm_sm; // 0 = SMH, 1 = GMH, 2 = CRH
	

	bool type_arqfb;  

	uint8_t cdma;
	
	dir_t direction;
	double txtime;
	double timestamp;
	Mac48Address dest;
	bool error; //specifies whether a recieved packet came in with a transmision error
	bool mac_packet; //specifies whether the packet should be processed by the MAC layer or passed up
};

class Type_Check : public Header {
public:
	uint8_t type;
	Type_Check();
	~Type_Check();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

class MAC_TAG : public Tag {
public: 
	bool isMAC;
	MAC_TAG();
	MAC_TAG(bool);
	~MAC_TAG();
	
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (TagBuffer i) const;
	virtual void Deserialize (TagBuffer i);
	virtual TypeId GetInstanceTypeId (void) const;


};

class UL_Protocol_Header : public Header {
public:
	uint16_t protocol;
	UL_Protocol_Header();
	UL_Protocol_Header(uint16_t ptcl);
	~UL_Protocol_Header();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};

class Copy_Trailer: public Trailer{
public:
	uint16_t length;
	uint8_t *d;
	Copy_Trailer(uint16_t len);
	Copy_Trailer();
	~Copy_Trailer();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
};


/**
 * This header is used to track assorted information as packets are passed
 * around between classes, but is not transmitted.  Just before transmission
 * this header is removed from a packet and the appropriate set of headers
 * is added
 */

class All_Headers : public Header {
public:
	All_Headers();
	~All_Headers();
	static TypeId GetTypeId (void);
	virtual TypeId GetInstanceTypeId (void) const;
	virtual void Print (std::ostream &os) const;
	virtual uint32_t GetSerializedSize (void) const;
	virtual void Serialize (Buffer::Iterator start) const;
	virtual uint32_t Deserialize (Buffer::Iterator start);
	//uint16_t& size() {return GMH.GMH.len;}
	
	Generic_MAC_Header 	 GMH;
	Signaling_MAC_Header SMH;
	CDMA_REQUEST_HEADER  CRH;
	
	PHY_INFO_Header		PIH;
	
	FRAG_SUBHEADER		FSH;
	GRANT_MAP_SUBHEADER	GMSH;
	PACKING_SUBHEADER	PSH;
	FFB_SUBHEADER		FFSH;
		
	
	
	uint16_t num_of_acks;
	struct arq_fb_ie arq_ie[MAX_NUM_ACK]; 		//if the type_arqfb is true, we have at least 1 element
                                                       	//we use the last bit to check if it is the last element
	uint8_t gm_sm; // 0 = SMH, 1 = GMH, 2 = CRH, 3 = DL_MAP or FCH (no MAC Header)
	

	bool *type_arqfb;
	bool *type_ext  ;  // Extended type.  1=extended subheader follows generic mac header
 	bool *type_frag ;  // Fragmentation subheader: 1=present, 0=absent
 	bool *type_pck  ; // Packing subheader: 1=present, 0=absent
 	bool *type_fbgm ;  // DL: FAST-FEEDBACK allocation subheader, UL: Grant Management subheader
  

	uint8_t cdma;
	
	dir_t direction;
	double txtime;	  //duration of transmission
	double timestamp; //time departed
	Mac48Address dest;
	bool error; //specifies whether a recieved packet came in with a transmision error
	bool mac_packet; //specifies whether the packet should be processed by the MAC layer or passed up
};




} //namesapce ns3

#endif //802_16HDR_H
