/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  ksshar
 */
#include "arqstatus.h"
#include "ns3/packet.h"
#include "connection.h"
//#include "mac802_16.h"

namespace ns3 {



/* Maximum Sequence number for ARQ*/
#define MAX_SEQ 2048 

/* 
 * Create an Arq Status object  
 * @param 
 */
Arqstatus::Arqstatus (): arq_retrans_time_(0), ack_seq_(0), ack_counter_(0),
			 arq_curr_seq_(0), ack_period_(0), arq_max_seq_(0),
			 arq_max_window_(0), arq_curr_window_(0), last_ack_sent_(0) 
{
  arq_trans_queue_ = new HeadQueue;
  arq_retrans_queue_ = new HeadQueue;
  arq_feedback_queue_ = new HeadQueue;
  arq_max_seq_ = MAX_SEQ;
	
}




/* 
 * Delete an Arq Status object  
 * @param 
 */
Arqstatus::~Arqstatus ()
{
	//don't need to explicetly delete, the smart pointers will deal with this for us
  //if (arq_trans_queue_) delete arq_trans_queue_;
  //if (arq_retrans_queue_) delete arq_retrans_queue_;
  //if (arq_feedback_queue_) delete arq_feedback_queue_;

  delete arqRetransTimer;  /*RPI*/
}

void Arqstatus::arqTimerHandler ()
{
  Ptr<Packet> ph, p, pac;
  All_Headers ph_head, p_head, pac_head;
  bool dont_queue = false, dont_queue_retrans = false;
  Connection* connection = arqRetransTimer-> getConnection ();
  debug2("ARQ Timer Expire for Connection Id: %d \n", connection->get_cid());
  arq_trans_queue_->resetIterator ();
  // send all packets with active timers to retransmission queue
  ph = arq_trans_queue_->getNext();
  ph->PeekHeader(ph_head);

  debug2 ("  ARQ: con %d qlength=%d arqTx=%d arqRtx=%d\n", connection->get_cid(), connection->get_queue ()->GetNPackets(), arq_trans_queue_->GetNPackets(), arq_retrans_queue_->GetNPackets());

  while (ph) {
    connection->get_queue ()->resetIterator ();
    p = connection->get_queue ()->getNext ();
    p->PeekHeader(p_head);
    while(p)
      {
	if(connection->isPackingEnable ())
	  {
	    if(p_head.PSH.data.sn == ph_head.PSH.data.sn){
	      debug2("  ARQ Timer handler, NOT Enqueueing packet in Retransmission queue Seq No : %d\n ",p_head.PSH.data.sn);	
	      dont_queue = true;
	      break;	
	    }		
	  }
	else 
	  {
	    if(p_head.FSH.data.sn == ph_head.FSH.data.sn){
	      dont_queue = true;
	      break;	
	    }
	  }
	p = connection->get_queue ()->getNext (); 
      }

    arq_retrans_queue_->resetIterator ();
    pac = arq_retrans_queue_->getNext();
    pac->PeekHeader(pac_head);
    while(pac)
      {
	if(connection->isPackingEnable ())
	  {
	    if(pac_head.PSH.data.sn == ph_head.PSH.data.sn){
	      debug2("  ARQ Timer handler, NOT Enqueueing packet in Retransmission queue, already in Retrans queue Seq No : %d\n ", pac_head.PSH.data.sn);	
	      dont_queue_retrans = true;
	      break;	
	    }		
	  }
	else 
	  {
	    if(pac_head.FSH.data.sn == ph_head.FSH.data.sn){
	      dont_queue_retrans = true;
	      break;	
	    }
	  }
	pac =  arq_retrans_queue_->getNext (); 
      }
		
    if((!dont_queue) && (!dont_queue_retrans)){
      debug2("  ARQ Timer handler, Enqueueing packet in Retransmission queue Seq No : %d \n",p_head.PSH.data.sn);
      arq_retrans_queue_->Enqueue(ph->Copy());
    }
    dont_queue = false;
    dont_queue_retrans = false;
    ph = arq_trans_queue_->getNext();
  }
  //schedule timer based on retransmission time setting
  debug2 ("  ARQ: con %d qlength=%d arqTx=%d arqRtx=%d\n", connection->get_cid(), connection->get_queue ()->GetNPackets(), arq_trans_queue_->GetNPackets(), arq_retrans_queue_->GetNPackets());
  arqRetransTimer->Schedule(Seconds(arq_retrans_time_));
  debug2("  ARQ Timer Re-Set\n");			
}


void  Arqstatus::arqRecvFeedback(Ptr<Packet> p, uint16_t i, Connection* connection) 
{
  // Read the header to find the flow the feedback is for
  All_Headers *mh = new All_Headers();
  p->RemoveHeader(*mh);
  
  arq_fb_ie ie = mh->arq_ie[i];
  uint16_t flowno = ie.cid;
  if (flowno); //to silence compiler warning about unused variable
  uint8_t  ack_type = ie.ack_type;
  uint32_t seq_recd = ie.fsn;
  uint32_t sn;
  //debug2("ARQ Feedback: feedback recd for flow:%d and seq:%d ack_type : %d\n",flowno,seq_recd, ack_type);
  debug2 ("  ARQ Feedback: feedback recd for con %d qlength=%d arqTx=%d arqRtx=%d\n", connection->get_cid(), connection->get_queue ()->GetNPackets(), arq_trans_queue_->GetNPackets(), arq_retrans_queue_->GetNPackets());

  if(ack_type == 1)
    ack_seq_ = seq_recd;
  // deque packets for which feedback received from transmission queue, free packet
  arq_trans_queue_->resetIterator ();
  while (arq_trans_queue_->GetNPackets() != 0) {
    Ptr<Packet> ph = (arq_trans_queue_)->getNext ();
    if (!ph) {
      // If one complete iteration is over, we can break the loop  
      break;
    }
    All_Headers *gh = new All_Headers();
    ph->PeekHeader(*gh);
    if(connection->isPackingEnable () == true){ 
      sn = gh->PSH.data.sn;
    }
    else {
      sn = gh->FSH.data.sn;
    }
    debug2("ARQ Feedback: packet has seq:%d\t and recd seq:%d (arq_curr_window_=%d, arq_max_window_=%d, MAX_SEQ=%d)\n",sn,seq_recd, arq_curr_window_, arq_max_window_, MAX_SEQ);
    //if ((ack_type == 1) && ((((seq_recd - sn) >= 0) && ((seq_recd - sn) < arq_max_window_)) || ((seq_recd >= 0 && seq_recd < arq_max_window_) && (sn < MAX_SEQ && (sn >= (MAX_SEQ - arq_max_window_)))))){	    
    if ((ack_type == 1) && 
	( (((seq_recd - sn) >= 0) && ((seq_recd - sn) < arq_max_window_))
	  || ((seq_recd - sn < 0) && (seq_recd+MAX_SEQ - sn < arq_max_window_))) ){	
      arq_trans_queue_->Remove(ph);
      ph = 0;	
      debug2("ARQ Feedback:dequeued1 packet with seq:%d\n",sn);
    } 
    else if((ack_type == 2) && (seq_recd == sn)) {
      arq_trans_queue_->Remove(ph);
      ph = 0;	
      debug2("ARQ Feedback:dequeued2 packet with seq:%d\n",sn);
    }
  }

  // deque packets for which feedback received from retransmission queue, free packet- This case can happen sometimes
  arq_retrans_queue_->resetIterator ();
  while (arq_retrans_queue_->GetNPackets() != 0) {
    Ptr<Packet> ph = (arq_retrans_queue_)->getNext ();
    if (!ph) {
      // If one complete iteration is over, we can break the loop  
      break;
    }
    All_Headers *gh = new All_Headers();
    ph->PeekHeader(*gh);
    if(connection->isPackingEnable () == true){ 
      sn = gh->PSH.data.sn;
    }
    else {
      sn = gh->FSH.data.sn;
    }
    debug2("ARQ Feedback: packet has seq:%d\t and recd seq:%d\n",sn,seq_recd);
    //if ((ack_type == 1) && ((((seq_recd - sn) >= 0) && ((seq_recd - sn) < arq_max_window_))  || ((seq_recd >= 0 && seq_recd < arq_max_window_) && (sn < MAX_SEQ && (sn >= (MAX_SEQ - arq_max_window_)))))){	
    if ((ack_type == 1) && 
	( (((seq_recd - sn) >= 0) && ((seq_recd - sn) < arq_max_window_))
	  || ((seq_recd - sn < 0) && (seq_recd+MAX_SEQ - sn < arq_max_window_))) ){	
      arq_retrans_queue_->Remove(ph);
      ph = 0;  	
      debug2("ARQ Feedback:dequeued packet from retrans queue with seq:%d\n",sn);
    } 
    else if((ack_type == 2) && (seq_recd == sn)) {
      arq_retrans_queue_->Remove(ph);
      ph = 0; 	
      debug2("ARQ Feedback:dequeued packet from retrans queue with seq:%d\n",sn);
    }
  }

}

}
