/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#include "connectionmanager.h"
//#include "mac802_16.h"
#include "wimax2-net-device.h"
#include <sys/queue.h>

namespace ns3 {

/**
 * Create the manager for the given mac
 * @param mac The Mac where the manager is located
 */
ConnectionManager::ConnectionManager (Ptr<Wimax2NetDevice> mac) 
{
  assert (mac!=NULL);
  mac_ = mac;

  //init list
  LIST_INIT (&i_con_list_);
  LIST_INIT (&o_con_list_);
}



/**
 * Add a connection to the list
 * @param con The connection to add
 * @param out true if it is an outgoing connection
 */
void ConnectionManager::add_connection (Connection* con, bool out) {
  assert (con!=NULL);
  assert (!get_connection (con->get_cid(), out)); //check duplicate
//  mac_->debug ("At %f in %d adding %s connection %d\n", NOW, mac_->addr(), out?"out":"in", con->get_cid());
  if (out)
    con->insert_entry (&o_con_list_);
  else
    con->insert_entry (&i_con_list_);

  con->setManager(this);
  if (out) debug10("Outgoing connection added with cid %d\n", con->get_cid());
  else debug10("Incoming connection added with cid %d\n", con->get_cid());
}



/**
 * Remove a connection
 * @param The connection to remove
 */
void ConnectionManager::remove_connection (Connection* con) {
  assert (con !=NULL);
//  mac_->debug ("At %f in %d removing connection %d\n",  NOW, mac_->addr(), con->get_cid());
  con->remove_entry ();
}


/**
 * Remove connection by CID, both directions.
 * @param cid The connection id
 */
void ConnectionManager::remove_connection (int cid)
{
  Connection *con = get_connection (cid, IN_CONNECTION);
  if (con)
    remove_connection (con);
  con = get_connection (cid, OUT_CONNECTION);
  if (con)
    remove_connection (con);
}


/**
 * Return the connection that has the given CID
 * @param cid The connection ID
 * @param out The direction
 * @return the connection or NULL
 */
Connection* ConnectionManager::get_connection (int cid, bool out) {
  //search throught the list
  for (Connection *n=out?o_con_list_.lh_first:i_con_list_.lh_first; n; n=n->next_entry()) {
    	if (n->get_cid ()==cid) {
		debug10("%s connection found with cid %d\n", out ? "outoing" : "incoming", cid);
        	return n;
      }
  }
  debug10("no %s connection found with cid %d\n", out ? "outoing" : "incoming", cid);
  return NULL;
}

void ConnectionManager::print_connections(bool out) {
	int counter = 0;
	if (out) {
		for (Connection *n=o_con_list_.lh_first; n; n= n->next_entry())
			printf("Outgoing Connection %d has cid %d\n", counter++, n->get_cid());
	} else {
		for (Connection *n=i_con_list_.lh_first; n; n= n->next_entry())
			printf("Outgoing Connection %d has cid %d\n", counter++, n->get_cid());
	}
}

//Will return a connection of this type, used for a (pretty hacky) solution to
//ARP sending on ff:ff:ff:ff:ff:ff
int ConnectionManager::getConOfType(ConnectionType_t ct, bool out) {
	if (out) {
		for (Connection *n=o_con_list_.lh_first; n; n= n->next_entry())
			if (n->getType() == ct) return n->get_cid();
	} else {
		for (Connection *n=i_con_list_.lh_first; n; n= n->next_entry())
			if (n->getType() == ct) return n->get_cid();
	}
	return -1;
}

Ptr<Wimax2NetDevice> ConnectionManager::getMac() {return mac_;}

/**
 * Flush the queues. This can be called after switching BS.
 */
void ConnectionManager::flush_queues () {
//  mac_->debug ("At %f in %d Flushing queues\n", NOW, mac_->addr());
  for (Connection *n=o_con_list_.lh_first; n; n=n->next_entry()) {
    int i = n->flush_queue();
    cout<<i<<std::endl;
//   mac_->debug ("\tFreed %d packet in queue for connection %d\n", i, n->get_cid());
  }
}


}
