#include "wimax2-net-device.h"
#include "destclassifier.h"
//#include "mac802_16.h"
#include "ns3/wimax2-scheduler.h"
//#include "peernode.h"

namespace ns3 {

DestClassifier::DestClassifier(): SDUClassifier() {}

int DestClassifier::classify(Ptr<Packet> p, const Address & adr) {
	Mac48Address to = Mac48Address::ConvertFrom(adr);
	return classify (p, to);
}

void DestClassifier::DoDispose() {
	SDUClassifier::DoDispose();
}
	


int DestClassifier::classify(Ptr<Packet> p, Mac48Address & addr) {
	 if (addr.IsBroadcast()) { //broadcast...
		return BROADCAST_CID;
	}

	//if not broadcast...
	for (PeerNode *n = mac_->getPeerNode_head(); n; n=n->next_entry()) {
		if (addr == n->getAddr()) {
			MAC_TAG t;
			//if this is a packet bound for higher layers, send it on the data connection
			if (p->PeekPacketTag(t) && !t.isMAC) {
				if (n->getOutData()) {
					debug10("classified packet, cid = %d\n", n->getOutData()->get_cid());
					return n->getOutData()->get_cid();
				}
			}


			//we should be classifying to the secondary connection if it's an ARP message
			//this is not currently implemented, and the code below will not work for that
			//if (temp->GetInstanceTypeId() == arp.GetTypeId()) { //if it's an ARP message
			//if (n->getSecondary(OUT_CONNECTION)) 
			//	return n->getSecondary(OUT_CONNECTION)->get_cid();
			//}
			//else  //if it's not an Arp Message

			//otherwise send it on the Primary connection
			else if (n->getPrimary(true)) { //we're not ready to send data yet, but this is useful for sending registration messages
				debug10("classified primary management connection for packet (not ready to send data yet) cid: %d\n",n->getPrimary(true)->get_cid()); //true = outgoing
				return n->getPrimary(true)->get_cid();
			}
		}
		
	}
	debug10("unable to classify packet, returning -1\n");
	return -1; // if we don't find a peernode with the right address...
}

Mac48Address DestClassifier::declassify(int cid) {
	for (PeerNode *n = mac_->getPeerNode_head(); n; n=n->next_entry()) {
		if (n->getOutData() && n->getOutData()->get_cid() == cid) return n->getAddr();
		if (n->getInData() && n->getInData()->get_cid() == cid)  return n->getAddr();
	}
	for (PeerNode *n = mac_->getPeerNode_head(); n; n=n->next_entry()) {
		if (n->getPrimary(true) && n->getPrimary(true)->get_cid() == cid) return n->getAddr();
		if (n->getPrimary(false) && n->getPrimary(false)->get_cid() == cid) return n->getAddr();
	}
	for (PeerNode *n = mac_->getPeerNode_head(); n; n=n->next_entry()) {
		if (n->getBasic(true) && n->getBasic(true)->get_cid() == cid) return n->getAddr();
		if (n->getBasic(false) && n->getBasic(false)->get_cid() == cid)  return n->getAddr();
	}
	for (PeerNode *n = mac_->getPeerNode_head(); n; n=n->next_entry()) {
		if (n->getSecondary(true) && n->getSecondary(true)->get_cid() == cid) return n->getAddr();
		if (n->getSecondary(false) && n->getSecondary(false)->get_cid() == cid) return n->getAddr();
	}
	return Mac48Address::GetBroadcast();
}

bool DestClassifier::isFromUpperLayer(int cid) {
	for (PeerNode *n = mac_->getPeerNode_head(); n; n=n->next_entry()) {
		if (n->getOutData() && n->getOutData()->get_cid() == cid) return true;
		if (n->getInData() && n->getInData()->get_cid() == cid)  return true;
	}
	return false;
}

}

