/**
	A note on serializing nested Headers:
	if you have a subclass of Header that contains another header, you can
	serialize and deserialize that nested header by calling Serialize(i) and 
	Deserialize(i) directly, but Serialize(i) will not properly increment the
	itterator (i), so you must call i.Next on the size of what was just written.
	This size can be obtained from the GetSerializedSize function of that nested
	Header or simply as the return value of the Serialize function.
*/


#include "802_16headers.h"
#include <iostream>
#include "ns3/address-utils.h"
#include "ns3/ulburst.h"
#include "ns3/simulator.h"

namespace ns3{	
	
Generic_MAC_Header::Generic_MAC_Header() {
	GMH.ht = 0;
	GMH.ec = 0;
	GMH.type_mesh = 0;
	GMH.type_arqfb = 0;
	GMH.type_ext = 0;
	GMH.type_frag = 0;
	GMH.type_pck = 0;
	GMH.type_fbgm = 0;
	GMH.rsv1 = 0;
	GMH.ci = 0;
	GMH.eks = 0;
	GMH.rsv2 = 0;
	GMH.len = 6;
	GMH.cid = 0;
	GMH.hcs = 0;
}

Generic_MAC_Header::~Generic_MAC_Header() {
}

Signaling_MAC_Header::Signaling_MAC_Header() {
	BRH.ht = 0;
	BRH.ec = 0;
	BRH.type = 0;
	BRH.br = 0;
	BRH.cid = 0;
	BRH.hcs = 0;
}

Signaling_MAC_Header::~Signaling_MAC_Header() {
}

TypeId Signaling_MAC_Header::GetTypeId (void) {
	static TypeId tid = TypeId("ns3::Signaling_MAC_Header")
	.SetParent<Header> ()
	.AddConstructor<Signaling_MAC_Header>()
	;
	return tid;
}

TypeId Generic_MAC_Header::GetTypeId (void) {
 static TypeId tid = TypeId ("ns3::Generic_MAC_Header")
 .SetParent<Header> ()
 .AddConstructor<Generic_MAC_Header> ()
 ;
 return tid;
}

TypeId Signaling_MAC_Header::GetInstanceTypeId(void) const {
	return GetTypeId();
}

TypeId Generic_MAC_Header::GetInstanceTypeId (void) const {
	return GetTypeId();
}

void Signaling_MAC_Header::Print(std::ostream &os) const {
	os << "this is a test printing for a Signaling_MAC_Header";
}

//uint32_t Generic_MAC_Header::getVal(uint16_t data, uint16_t bits) const {
//	uint32_t max = 1;
//	uint32_t current = 0;
//	for (int i = 1; i <= bits; i++) {
//		max *= 2;
//		while (current <= max-1) {
//			cout << current << "<=" << max << std::endl;
//			if (data == current) return data;
//			else current++;
//		}
//	}
//	return max+1;
//}

void Generic_MAC_Header::Print (std::ostream &os) const {
	os<< "\nht: " <<static_cast<uint16_t>(GMH.ht) << std::endl
	<<"ec: " 	  <<static_cast<uint16_t>(GMH.ec) << std::endl
	<<"type_mesh: " << static_cast<uint16_t>(GMH.type_mesh) << std::endl
	<<"type_arqfb: "<<static_cast<uint16_t>(GMH.type_arqfb) << std::endl
	<<"type_ext: " 	<< static_cast<uint16_t>(GMH.type_ext) << std::endl
	<<"type_frag " 	<< static_cast<uint16_t>(GMH.type_frag) << std::endl
	<<"type_pck " 	<< static_cast<uint16_t>(GMH.type_pck) <<std::endl
	<<"type_fbgm: " <<static_cast<uint16_t>(GMH.type_fbgm)<<std::endl
	<<"ci: "  <<static_cast<uint16_t>(GMH.ci) << std::endl
	<<"eks: " <<static_cast<uint16_t>(GMH.eks) << std::endl
	<<"len: " <<static_cast<uint16_t>(GMH.len)<<std::endl
	<<"cid: " << static_cast<uint16_t>(GMH.cid) << std::endl
	<<"hcs: " <<static_cast<uint16_t>(GMH.hcs) << std::endl;
}

uint32_t Signaling_MAC_Header::GetSerializedSize(void) const {
	return 6;
}

uint32_t Generic_MAC_Header::GetSerializedSize (void) const {
	return 6;
}

void Signaling_MAC_Header::Serialize (Buffer::Iterator start) const {
	Buffer::Iterator i = start;	
	
	uint32_t temp = 0, temp2 = 0;
	uint16_t temp3 = 0;
	temp = temp | BRH.ht;
	temp = temp | BRH.ec << 1;
	temp = temp | BRH.type << 2;
	temp = temp | (BRH.br & 0x7FFFF)<< 5;
	temp2 = 255 << 8;
	temp = temp | (temp2 & BRH.cid) << 16;
	i.WriteHtonU32(temp);
	
	temp3 = BRH.cid << 8;
	temp3 = temp3 | BRH.hcs;
	i.WriteHtonU16(temp3);
	
}

void Generic_MAC_Header::Serialize (Buffer::Iterator start) const {
	  Buffer::Iterator i = start;
	  uint32_t write1 = 0; 
	  uint16_t write2 = 0;
	  //it's much easier to work with everything if we encode all the headers the same way
	  //so we pack the bits into conveinent denominations, then repack those into a 32 and 16 bit uint
	  
//	  break the gen_mac_header_t into appropriately sized chunks
	  uint8_t temp = 0;
	  temp = temp | GMH.ht;
	  temp = temp | GMH.ec << 1;
	  temp = temp | GMH.type_mesh << 2;
	  temp = temp | GMH.type_arqfb << 3;
	  temp = temp | GMH.type_ext << 4;
	  temp = temp | GMH.type_frag << 5;
	  temp = temp | GMH.type_pck << 6;
	  temp = temp | GMH.type_fbgm << 7;
	  write1 |= temp;
	  
	  uint16_t temp2 = 0;
	  temp2 = temp2 | GMH.rsv1;
	  temp2 = temp2 | GMH.ci << 1;
	  temp2 = temp2 | GMH.eks << 2;
	  temp2 = temp2 | GMH.rsv2 << 4;
	  temp2 = temp2 | GMH.len << 5;
	  write1 |= (temp2 << 8);
	  write1 |= (GMH.cid << 24);
	  write2 |= (GMH.cid >> 8);
	  write2 |= (GMH.hcs << 8);
	  i.WriteHtonU32(write1);
	  i.WriteHtonU16(write2);
}

uint32_t Signaling_MAC_Header::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing Signaling_MAC_Header\n");

	Buffer::Iterator i = start;
	uint32_t temp = i.ReadNtohU32(), mask = 1;
	uint16_t temp3 = 0;
	uint8_t temp4 = 0;
	BRH.ht = temp & mask;
	mask = 2;
	BRH.ec = (temp & mask) >> 1;
	
	mask = 7 << 2;
	BRH.type = (temp & mask) >> 2;
	mask = 0x7FFFF << 5;// 2^19 - 1
	BRH.br = (temp & mask) >> 5;
	
	mask = 255 << 24;
	temp3 = (temp & mask) >> 16;
	temp4 = i.ReadU8();
	BRH.cid = temp3 | temp4;
	BRH.hcs = i.ReadU8();
	return 6;
}

uint32_t Generic_MAC_Header::Deserialize (Buffer::Iterator start) {
	debugHeader("Deserializing Generic_MAC_Header\n");
	Buffer::Iterator i = start;
	
	//and now we unpack this a bit strangely to match the packing
	uint32_t read1;
	uint16_t read2;
	
	read1 = i.ReadNtohU32();
	read2 = i.ReadNtohU16();
	
	uint32_t mask_big = 255, temp = 0;
	
	temp = mask_big & read1;
	
	uint8_t mask = 1;
	GMH.ht = temp & mask;

	mask = mask << 1;
	GMH.ec = (temp & mask) >> 1;
	
	mask = mask << 1; 
	GMH.type_mesh = (temp & mask) >> 2;	
	mask = mask << 1; 
	GMH.type_arqfb = (temp & mask) >> 3;	
	mask = mask << 1; 
	GMH.type_ext = (temp & mask) >> 4;	
	mask = mask << 1; 
	GMH.type_frag = (temp & mask) >> 5;	
	mask = mask << 1; 
	GMH.type_pck = (temp & mask) >> 6;	
	mask = mask << 1; 
	GMH.type_fbgm = (temp & mask) >> 7;
	
	mask_big = 65535 << 8;
	uint16_t temp2 = (mask_big & read1) >> 8;
	uint16_t mask2 = 1;
	GMH.rsv1 = temp2 & mask2;
	mask2 = 2;
	GMH.ci = (temp2 & mask2) >> 1;
	
	mask2 = 3 << 2;
	GMH.eks = (temp2 & mask2) >> 2;
	
	mask2 = 1 << 4;
	GMH.rsv2 = (temp2 & mask2) >> 4;
	
	mask2 = 2047 << 5;
	GMH.len = (temp2 & mask2) >> 5;
	GMH.cid = (read1 >> 24) | (read2 << 8);
	GMH.hcs = read2 >> 8;
	return 6;
}

void Generic_MAC_Header::setHeaderType(bool input) {
	GMH.ht=input;
}
void Generic_MAC_Header::setEncryptionControl(bool input) {
	GMH.ec=input;
}
void Generic_MAC_Header::setARQFB(bool input){
	GMH.type_arqfb = input;
}	
void Generic_MAC_Header::setExt(bool input){
	GMH.type_ext = input;
}	
void Generic_MAC_Header::setFrag(bool input){
	GMH.type_frag = input;
}	
void Generic_MAC_Header::setPck(bool input){
	GMH.type_pck = input;
}	
void Generic_MAC_Header::setFFBGM(bool input){
	GMH.type_fbgm = input;
}	
void Generic_MAC_Header::setCI(bool input){
	GMH.ci = input;
} 	
void Generic_MAC_Header::setEks(uint8_t input) {
	GMH.eks = input;
}
void Generic_MAC_Header::setPduLen(uint16_t input){
	GMH.len = input;
}
void Generic_MAC_Header::setCID(uint16_t input){
	GMH.cid = input;
}	
void Generic_MAC_Header::setHCS(uint8_t input){
	GMH.hcs = input;
}	 




void Signaling_MAC_Header::setHeaderType(uint8_t input) {BRH.ht = input;}
void Signaling_MAC_Header::setEncryptionControl(uint8_t input) {BRH.ec = input;}
void Signaling_MAC_Header::setType(uint8_t input) {BRH.type = input;}
void Signaling_MAC_Header::setBandwidthRequested(uint32_t input) {BRH.br = input;}
void Signaling_MAC_Header::setCID(uint16_t input) {BRH.cid = input;}
void Signaling_MAC_Header::setHCS(uint8_t input) {BRH.hcs = input;}


DL_MAP_IE::DL_MAP_IE() {
	ie.diuc = 0;
	ie.cid = 0;
	ie.OFDMASymbol_offset = 0;
	ie.subchannel_offset = 0;
	ie.boost = 0;
	ie.num_OFDMASymbol = 0;
	ie.num_subchannel = 0;
	ie.repetition_coding = 0;
}
DL_MAP_IE::~DL_MAP_IE() {}
TypeId DL_MAP_IE::GetTypeId() {
	static TypeId tid = TypeId("ns3::DL_MAP_IE")
	.SetParent<Header> ()
	.AddConstructor<DL_MAP_IE>()
	;
	return tid;
}
TypeId DL_MAP_IE::GetInstanceTypeId() const{
	return GetTypeId();
}

void DL_MAP_IE::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE DL_MAP_IE Header\n";	
}

uint32_t DL_MAP_IE::GetSerializedSize() const{
	return 7;
}
void DL_MAP_IE::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	
	uint32_t temp = 0;
	temp = temp | (ie.diuc << 28);
	temp = temp | (ie.cid << 12);
	temp = temp | (ie.subchannel_offset << 6);
	temp = temp | (ie.boost << 3);
	i.WriteHtonU32(temp);
	i.WriteU8(ie.OFDMASymbol_offset);
	uint16_t t2 = 0;
	t2 = t2 | (ie.num_OFDMASymbol << 9);
	t2 = t2 | (ie.num_subchannel << 3);
	t2 = t2 | (ie.repetition_coding << 1);
	i.WriteHtonU16(t2);
}

uint32_t DL_MAP_IE::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing DL_MAP_IE\n");

	Buffer::Iterator i = start;
	
	uint32_t mask = 0, temp = 0;
	temp = i.ReadNtohU32();
	mask = 15 << 28;
	ie.diuc = (temp & mask) >> 28;;

	mask = 65535 << 12;
	ie.cid  = (temp & mask) >> 12;

	mask = 63 << 6;
	ie.subchannel_offset = (temp & mask) >> 6;

	mask = 7 <<3;
	ie.boost = (mask & temp) >>3;
	ie.OFDMASymbol_offset = i.ReadU8();
	uint16_t mask2 = 127 << 9;
	uint16_t temp2 = i.ReadNtohU16();
	ie.num_OFDMASymbol = (temp2 & mask2) >> 9;
	mask2 = 63 << 3;
	ie.num_subchannel = (temp2 & mask2) >> 3;
	mask2 = 3 << 1;
	ie.repetition_coding =(temp2 & mask2) >> 1;

	return 6;
}

Fast_Ranging_IE::Fast_Ranging_IE() {
	ie.mac_addr = Mac48Address::GetBroadcast();
	ie.uiuc = 0;
	ie.duration = 0;
}
Fast_Ranging_IE::~Fast_Ranging_IE() {}
TypeId Fast_Ranging_IE::GetTypeId() {
	static TypeId tid = TypeId("ns3::Fast_Ranging_IE")
	.SetParent<Header> ()
	.AddConstructor<Signaling_MAC_Header>()
	;
	return tid;
}
TypeId Fast_Ranging_IE::GetInstanceTypeId() const{
	return GetTypeId();
}

void Fast_Ranging_IE::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE Fast_Ranging_IE Header\n";	
}

uint32_t Fast_Ranging_IE::GetSerializedSize() const {
	return 8;
}

void Fast_Ranging_IE::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint8_t adr[6];
	ie.mac_addr.CopyTo(adr);
	for (int j = 0; j < 6; j++)
		i.WriteU8(adr[j]);
	uint16_t temp = 0;
	temp = temp | ie.uiuc;
	temp = temp | ie.duration << 4;
	i.WriteHtonU16(temp);
}

uint32_t Fast_Ranging_IE::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing Fast_Ranging_IE\n");

	Buffer::Iterator i = start;
	uint8_t adr[6];
	for (int j = 0; j < 6; j++)
		adr[j] = i.ReadU8();
	ie.mac_addr.CopyFrom(adr);
	uint16_t temp = i.ReadNtohU16();
	uint16_t mask = 15;
	ie.uiuc = temp & mask;
	mask = 4095 << 4;
	ie.duration = (temp & mask) >> 4;
	return 8;
}


CDMA_MAP_IE::CDMA_MAP_IE() {
	ie.duration = 0;
	ie.uiuc = 0;
	ie.repetition = 0;
	ie.frame_number = 0;
	ie.code = 0;
	ie.symbol = 0;
	ie.subchannel = 0;
	ie.bw_req = 0;
}
CDMA_MAP_IE::~CDMA_MAP_IE() {}
TypeId CDMA_MAP_IE::GetTypeId() {
	static TypeId tid = TypeId("ns3::CDMA_MAP_IE")
	.SetParent<Header> ()
	.AddConstructor<Signaling_MAC_Header>()
	;
	return tid;
}
TypeId CDMA_MAP_IE::GetInstanceTypeId() const{
	return GetTypeId();
}

void CDMA_MAP_IE::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE CDMA_MAP_IE Header\n";	
}

uint32_t CDMA_MAP_IE::GetSerializedSize() const {
	return 5;
}

void CDMA_MAP_IE::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint16_t temp = 0;
	temp = temp | ie.duration;
	temp = temp | ie.uiuc << 6;
	temp = temp | ie.repetition << 10;
	temp = temp | ie.frame_number << 12;
	i.WriteHtonU16(temp);
	i.WriteU8(ie.code);
	i.WriteU8(ie.symbol);
	uint8_t temp2 = 0;
	temp2 = temp2 | ie.subchannel;
	temp2 = temp2 | ie.bw_req << 7;
	i.WriteU8(temp2);	
}

uint32_t CDMA_MAP_IE::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing CDMA_MAP_IE\n");

	Buffer::Iterator i = start;
	uint16_t temp = i.ReadNtohU16();
	uint16_t mask = 63;
	ie.duration = temp & mask;
	mask = 15 << 6;
	ie.uiuc = (temp & mask) >> 6;
	mask = 3 << 10;
	ie.repetition = (temp & mask) >> 10;
	mask = 15 << 12;
	ie.frame_number = (temp & mask) >> 12;
	ie.code = i.ReadU8();
	ie.symbol = i.ReadU8();
	uint8_t temp2 = i.ReadU8(), mask2 = 127;
	ie.subchannel = temp2 & mask2;
	mask2 = 1 << 7;
	ie.bw_req = (temp2 & mask2) >> 7;
	
	return 5;
}



UL_MAP_IE::UL_MAP_IE() {
	ie.cid = 0;
	ie.uiuc = 0;
	ie.duration = 0;
	ie.repetition_coding = 0;
}
UL_MAP_IE::~UL_MAP_IE() {}
TypeId UL_MAP_IE::GetTypeId() {
	static TypeId tid = TypeId("ns3::UL_MAP_IE")
	.SetParent<Header> ()
	.AddConstructor<UL_MAP_IE>()
	;
	return tid;
}
TypeId UL_MAP_IE::GetInstanceTypeId() const{
	return GetTypeId();
}

void UL_MAP_IE::Print(std::ostream &os) const{
	os << "PRINTING UL_MAP_IE\n";	
	os << " cid: " << (int)ie.cid << std::endl;
	os << " uiuc: " << (int)ie.uiuc << std::endl;
	os << " duration: " << (int)ie.duration << std::endl;
	os << " repetition coding: " << (int)ie.repetition_coding << std::endl;
	os << "DONE PRINTING UL_MAP_IE\n";
}

uint32_t UL_MAP_IE::GetSerializedSize() const {
	if (ie.uiuc == UIUC_CDMA_ALLOCATION_IE) return 4 + ie.cdma_ie.GetSerializedSize();
	return 4;
}

void UL_MAP_IE::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint16_t temp = 0;
	
	i.WriteHtonU16(ie.cid);
	temp = temp | ie.uiuc;
	temp = temp | ie.duration << 4;
	temp = temp | ie.repetition_coding << 14;
	i.WriteHtonU16(temp);
	
	if (ie.uiuc == UIUC_CDMA_ALLOCATION_IE) {
		ie.cdma_ie.Serialize(i);
		i.Next(ie.cdma_ie.GetSerializedSize());	
	}
	
	
	// temp = 0;
// 	uint32_t temp2 = 0;
// 	temp2 = temp2 | ie.uiuc;
// 	temp2 = temp2 | ie.duration << 4;
// 	temp2 = temp2 | ie.midamble_rep << 15;
// 	temp2 = temp2 | ie.extended_uiuc << 17;
// 	temp2 = temp2 | ie.length << 21;  //at 25/32 bits
// 	
// 	//the problem here is that we have 25 bits of data, followed by a fast ranging IE and 
// 	//a cdma IE, then 39 more bits.  Since we can only serialize in bytes, we have to pull
// 	//the serialized form of these sub-IEs out, then repack everything to avoid transmitting
// 	//filler
// 	
// 	uint32_t ranging_and_cdma_size = ie.fast_ranging.GetSerializedSize() + ie.cdma_ie.GetSerializedSize();
// 	uint8_t mask = 127, temp3 = 0;
// 	uint8_t *tbuf = new uint8_t[ranging_and_cdma_size];
// 
// 	
// 	//to get everything in a convienent buffer of bytes, we serialize the nested headers
// 	//then pull them out so we can go byte-by-byte
// 	ie.fast_ranging.Serialize(i); i.Next(ie.fast_ranging.GetSerializedSize());
// 	ie.cdma_ie.Serialize(i);	  i.Next(ie.cdma_ie.GetSerializedSize());
// 	i.Prev(ranging_and_cdma_size); //so the itterator is in the right place to read it out
// 	i.Read(tbuf, ranging_and_cdma_size); 
// 	i.Prev(ranging_and_cdma_size);
// 	
// 	temp2 = temp2 | (tbuf[0] & mask) << 25;
// 	i.WriteHtonU32(temp2);
// 	
// 	for (uint32_t j = 1; j< ranging_and_cdma_size; j++) {
// 		temp3 = 0;
// 		temp3 = temp3 | (tbuf[j-1] >> 7);
// 		temp3 = temp3 | (tbuf[j] << 1);
// 		i.WriteU8(temp3);
// 
// 	}
// 	//we now have the last bit of the CDMA_MAP_IE
// 	temp3 = 0;
// 	temp3 = temp3 | (tbuf[ranging_and_cdma_size-1] >> 7);
// 	temp3 = temp3 | (ie.symbol_offset << 1);
// 	i.WriteU8(temp3);
// 
// 	temp3 = 0;
// 	temp3 = temp3 | (ie.symbol_offset >> 7);
// 	temp3 = temp3 | ie.subchannel_offset;
// 	i.WriteU8(temp3);
// 
// 	temp3 = 0;
// 	uint16_t subchan = ie.num_of_subchannels;
// 	subchan = (subchan << 15) >> 8;
// 	temp3 = subchan;
// 	temp3 = temp3 | ie.num_of_symbols;
// 	i.WriteU8(temp3);
// 
// 	
// 	temp = 0;
// 	temp = temp | (ie.num_of_subchannels >> 1);
// 	temp = temp | (ie.ranging_method << 14);
// 	i.WriteHtonU16(temp);

}

//note that because this is called from the Deserialize function for UL_MAP_FRAME_HEADER
//start will not be the start of the buffer, so there is no easy way to find out how far
//in you already are, and how much of the buffer you can safely access
uint32_t UL_MAP_IE::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing UL_MAP_IE\n");

	Buffer::Iterator i = start;
	ie.cid = i.ReadNtohU16();
	uint16_t temp = i.ReadNtohU16(), mask = 0;
	mask = 15;
	ie.uiuc = mask & temp;
	mask = 1023 << 4;
	ie.duration = (mask & temp) >> 4;
	mask = 3 << 14;
	ie.repetition_coding = (mask & temp) >> 14;
	
	if (ie.uiuc == UIUC_CDMA_ALLOCATION_IE) {
		ie.cdma_ie.Deserialize(i);
		i.Next(ie.cdma_ie.GetSerializedSize());	
	}
	

	
// 	
// 	uint32_t temp2 = i.ReadNtohU32(), mask2 = 15;
// 	ie.uiuc = mask2 & temp2;
// 	mask2 = 2047 << 4;
// 	ie.duration = (mask2 & temp2) >> 4;
// 	mask2 = 3 << 15;
// 	ie.midamble_rep = (mask2 & temp2) >> 15;
// 	mask2 = 15 << 17;
// 	ie.extended_uiuc = (mask2 & temp2) >> 17;
// 	mask2 = 15 << 21;
// 	ie.length = (mask2 & temp2) >> 21; //have used 25/32 bits in temp2
// 	
// 	//we now have to pull out the fast ranging IE and CDMA map IE, but since we can only
// 	//pull data out of the buffer in bytes (or multiple whole bytes), and this doesn't line up
// 	//we're going to pull out some extra data, push the appropriate data back into the buffer
// 	//and then send the buffer off to the deserialize functions for those IEs
// 	uint32_t ranging_and_cdma_size = ie.fast_ranging.GetSerializedSize() + ie.cdma_ie.GetSerializedSize();
// 	uint8_t *tbuf = new uint8_t[ranging_and_cdma_size], temp3 = 0, mask3;
// 	for (uint32_t j = 0; j < ranging_and_cdma_size; j++) tbuf[j] = 0;
// 	mask2 = 127 << 25;
// 	tbuf[0] = (mask2 & temp2) >> 25;
// 	for (uint32_t j = 1; j < ranging_and_cdma_size; j++) {
// 		temp3 = i.ReadU8();
// 		tbuf[j-1] = tbuf[j-1] | (temp3 << 7);
// 		tbuf[j] = tbuf[j] | (temp3 >> 1);
// 	}
// 	temp3 = i.ReadU8();
// 	tbuf[ranging_and_cdma_size-1] = tbuf[ranging_and_cdma_size-1] | (temp3 << 7);
// 	//tbuf now contains the data for both ie.fast_ranging and ie.cdma_ie, which will 
// 	//be pushed back into the buffer then decoded below
// 	
// 	ie.symbol_offset = temp3 >> 1;
// 	temp3 = i.ReadU8();
// 	ie.symbol_offset = ie.symbol_offset | (temp3 << 7);
// 	ie.subchannel_offset = temp3 >> 1;
// 	
// 	temp3 = i.ReadU8();
// 	mask3 = 127;
// 	ie.num_of_symbols = temp3 & mask3;
// 	temp = temp3 >> 7;
// 	ie.num_of_subchannels = temp;
// 	
// 	temp = i.ReadNtohU16();
// 	mask = 16383;
// 	ie.num_of_subchannels = ie.num_of_subchannels | ((mask & temp) << 1);
// 	mask = 3 << 14;
// 	ie.ranging_method = (temp & mask) >> 14;
// 	//now we're going to deal with tbuf and the cdma and fast_ranging IEs
// 	i.Prev(ranging_and_cdma_size);		//so we have room to push things onto the buffer
// 	i.Write(tbuf, ranging_and_cdma_size);
// 	i.Prev(ranging_and_cdma_size);		//so we're pulling the right data off the buffer
// 	ie.fast_ranging.Deserialize(i); i.Next(ie.fast_ranging.GetSerializedSize());
// 	ie.cdma_ie.Deserialize(i);		i.Next(ie.cdma_ie.GetSerializedSize());
	return GetSerializedSize();
}


DL_MAP_FRAME_HEADER::DL_MAP_FRAME_HEADER() {
	mf.type = 0;
	mf.dcd_count = 0;
	mf.bsid = Mac48Address::GetBroadcast();
	mf.num_of_ofdma_symbols = 0;
	mf.nb_ies = 0;
}
DL_MAP_FRAME_HEADER::~DL_MAP_FRAME_HEADER() {}
TypeId DL_MAP_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::DL_MAP_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<DL_MAP_FRAME_HEADER>()
	;
	return tid;
}
TypeId DL_MAP_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void DL_MAP_FRAME_HEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE DL_MAP_FRAME_HEADER Header\n";	
}

uint32_t DL_MAP_FRAME_HEADER::GetSerializedSize() const {
	if (mf.nb_ies > 0) return 13 + mf.ies[0].GetSerializedSize() * mf.nb_ies;
	else return 13;
}

void DL_MAP_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(mf.type);
	i.WriteU8(mf.dcd_count);
	
	//i.WriteHtonU32(mf.bsid);
	
	uint8_t adr[6];
	mf.bsid.CopyTo(adr);
	for (int j = 0; j < 6; j++)
		i.WriteU8(adr[j]);
	
	i.WriteU8(mf.num_of_ofdma_symbols);
	i.WriteHtonU32(mf.nb_ies);
	for (uint32_t j = 0; j < mf.nb_ies; j++) {
		mf.ies[j].Serialize(i);
		i.Next(mf.ies[j].GetSerializedSize());
	}
}

uint32_t DL_MAP_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing DL_MAP_FRAME_HEADER\n");

	Buffer::Iterator i = start;
	mf.type = i.ReadU8();
	mf.dcd_count = i.ReadU8();
	
	//mf.bsid = i.ReadNtohU32();
	
	uint8_t adr[6];
	for (int j = 0; j < 6; j++)
		adr[j] = i.ReadU8();
	mf.bsid.CopyFrom(adr);
	
	mf.num_of_ofdma_symbols = i.ReadU8();
	mf.nb_ies = i.ReadNtohU32();
	for (uint32_t j = 0; j < mf.nb_ies; j++) {
		mf.ies[j].Deserialize(i);
		i.Next(mf.ies[j].GetSerializedSize());
	}
	return GetSerializedSize();
}





UCD_FRAME_DATA::UCD_FRAME_DATA() {
	ucdf.type = 0;
	ucdf.config_change_count = 0;
	ucdf.rng_backoff_start = 0;
	ucdf.rng_backoff_end = 0;
	ucdf.req_backoff_start = 0;
	ucdf.req_backoff_end = 0;
	ucdf.bw_req_size = 0;
	ucdf.rng_req_size = 0;
	ucdf.nb_prof = 0;
}
UCD_FRAME_DATA::~UCD_FRAME_DATA() {}
TypeId UCD_FRAME_DATA::GetTypeId() {
	static TypeId tid = TypeId("ns3::UCD_FRAME_DATA")
	.SetParent<Header> ()
	.AddConstructor<UCD_FRAME_DATA>()
	;
	return tid;
}
TypeId UCD_FRAME_DATA::GetInstanceTypeId() const{
	return GetTypeId();
}

void UCD_FRAME_DATA::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE UCD_FRAME_DATA Header\n";	
}

uint32_t UCD_FRAME_DATA::GetSerializedSize()  const{
	return 14 + (int)(1.5*(double)ucdf.nb_prof + 0.5);
}

void UCD_FRAME_DATA::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(ucdf.type);
	i.WriteU8(ucdf.config_change_count);
	i.WriteU8(ucdf.rng_backoff_start);
	i.WriteU8(ucdf.rng_backoff_end);
	i.WriteU8(ucdf.req_backoff_start);
	i.WriteU8(ucdf.req_backoff_end);
	
	
	i.WriteHtonU16(ucdf.bw_req_size);
	i.WriteHtonU16(ucdf.rng_req_size);
	i.WriteHtonU32(ucdf.nb_prof);
	uint8_t temp = 0;
	bool last_half = false;
	for (uint32_t j = 0; j < ucdf.nb_prof; j++) {
		if (!last_half) {
			temp = ucdf.profiles[j].uiuc;
			temp = temp | (ucdf.profiles[j].fec << 4);
			i.WriteU8(temp);
			temp = ucdf.profiles[j].fec >> 4;
			last_half = true;
		}
		else {
			temp = temp | (ucdf.profiles[j].uiuc << 4);
			i.WriteU8(temp);
			i.WriteU8(ucdf.profiles[j].fec);
			last_half = false;
		}
	}
	if (last_half) i.WriteU8(temp); //write the 0-padded last half-byte if necessary
}

uint32_t UCD_FRAME_DATA::Deserialize(Buffer::Iterator start) {
	Buffer::Iterator i = start;
	ucdf.type = i.ReadU8();
	ucdf.config_change_count = i.ReadU8();
	ucdf.rng_backoff_start = i.ReadU8();
	ucdf.rng_backoff_end = i.ReadU8();
	ucdf.req_backoff_start = i.ReadU8();
	ucdf.req_backoff_end = i.ReadU8();
	
	ucdf.bw_req_size = i.ReadNtohU16();
	ucdf.rng_req_size = i.ReadNtohU16();
	ucdf.nb_prof = i.ReadNtohU32();
	
	uint8_t temp = 0;
	bool end_even = true;
	for (uint32_t j = 0; j < ucdf.nb_prof; j++) {
	if (end_even) {
		temp = i.ReadU8();
		ucdf.profiles[j].uiuc = (temp << 4) >> 4; //clears out upper bits
		ucdf.profiles[j].fec = (temp >> 4);
		temp = i.ReadU8();
		ucdf.profiles[j].fec = ucdf.profiles[j].fec | (temp << 4);
		end_even = false;
	}
	else {
		ucdf.profiles[j].uiuc = (temp >> 4);
		ucdf.profiles[j].fec = i.ReadU8();
		end_even = true;
	}
	
	}
	return GetSerializedSize();
}



UL_MAP_FRAME_HEADER::UL_MAP_FRAME_HEADER() {
	mf.type = 0;
	mf.ucid = 0;
	mf.ucd_count = 0;
	mf.allocation_start = 0;
	mf.num_ofdma_symbols = 0;
	mf.nb_ies = 0;
}
UL_MAP_FRAME_HEADER::~UL_MAP_FRAME_HEADER() {}
TypeId UL_MAP_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::UL_MAP_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<UL_MAP_FRAME_HEADER>()
	;
	return tid;
}
TypeId UL_MAP_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void UL_MAP_FRAME_HEADER::Print(std::ostream &os) const{
	os << "PRINTING UL_MAP_FRAME_HEADER\n";
	os << " type: " << (int)mf.type << "\n";
	os << " ucid: " << (int)mf.ucid << "\n";
	os << " ucd_count: " << (int)mf.ucd_count << "\n";
	os << " allocation_start: " << (int)mf.allocation_start << "\n";
	os << " num_ofdma_symbols: " << (int)mf.num_ofdma_symbols << "\n";
	os << " nb_ies: " << (int) mf.nb_ies << "\n";
	for (uint16_t i = 0; i < mf.nb_ies; i++) {
		os << "  IE " << i << "\n";
		mf.ies[i].Print(os);
	}
	os << "FINISHED PRINTING UL_MAP_FRAME_HEADER\n";
//	os << "PRINTING SAMPLE UL_MAP_FRAME_HEADER Header\n";	
}

uint32_t UL_MAP_FRAME_HEADER::GetSerializedSize() const {
	int accum = 12;
	for (uint16_t i = 0; i < mf.nb_ies; i++) accum += mf.ies[i].GetSerializedSize();
	return accum;
}

void UL_MAP_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	//debug10("SerializedSize: %d\n", GetSerializedSize());
	//debug10("in UL_MAP_FRAME_HEADER, 
	i.WriteU8(mf.type);
	i.WriteU8(mf.ucid);
	i.WriteU8(mf.ucd_count);
	i.WriteHtonU32(mf.allocation_start);
	i.WriteU8(mf.num_ofdma_symbols);
	i.WriteHtonU32(mf.nb_ies);
	//debug10("so far added 12 bytes\n");
	for (uint32_t j = 0; j < mf.nb_ies; j++) {
		mf.ies[j].Serialize(i);
		i.Next(mf.ies[j].GetSerializedSize());		
	}
}

uint32_t UL_MAP_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing UL_MAP_FRAME_HEADER\n");

	Buffer::Iterator i = start;
	mf.type = i.ReadU8();
	mf.ucid = i.ReadU8();
	mf.ucd_count = i.ReadU8();
	mf.allocation_start = i.ReadNtohU32();
	mf.num_ofdma_symbols = i.ReadU8();
	mf.nb_ies = i.ReadNtohU32();
	for (uint32_t j = 0; j < mf.nb_ies; j++) {
		mf.ies[j].Deserialize(i);
		i.Next(mf.ies[j].GetSerializedSize());		
	}
	return GetSerializedSize();
}


PHY_INFO_Header::PHY_INFO_Header() {
  freq_ = 0;              //the frequency at which it is sent
  modulation_ = OFDM_BPSK_1_2; //default modulation, may be changed later
  g_ = 0; //The cyclic prefix information. Used for synchronization
  subcarrier_lbound = 0;
  subcarrier_ubound = 0;
  time_lbound = 0;
  time_ubound = 0;
  num_subchannels = 0;
  subchannel_offset = 0;
  num_OFDMSymbol = 0;
  OFDMSymbol_offset = 0;
  channel_index = 0;
  direction = 0;   // 0 = DL 1= UL 
}

PHY_INFO_Header::~PHY_INFO_Header() {}

uint32_t PHY_INFO_Header::GetSerializedSize() const{
	int size = 0;
	size += sizeof(double) * 6;
	size += sizeof(uint32_t) * 6;
	size += 1; //ofdm_mod_rate
	//size += 1;
	//cout<<"TOTAL SIZE: " << size << "\n" << std::flush;
	return size;
}

//becuase we need to write doubles to the buffer from a const function, we need
//to do some strange casting to make sure the anti-aliasing rules aren't violated
void PHY_INFO_Header::Serialize(Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint8_t temp = modulation_;
	const double *conversion_holder = &freq_;
	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
	//i.WriteHtonU64(*reinterpret_cast<uint64_t *>(&freq_));
	i.WriteU8(temp);
	conversion_holder = &g_;
	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
	conversion_holder = &subcarrier_lbound;
	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
	conversion_holder = &subcarrier_ubound;
	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
	conversion_holder = &time_lbound;
	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
	conversion_holder = &time_ubound;
	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
	i.WriteHtonU32(num_subchannels);
	i.WriteHtonU32(subchannel_offset);
	i.WriteHtonU32(num_OFDMSymbol);
	i.WriteHtonU32(OFDMSymbol_offset);
	i.WriteHtonU32(channel_index);
	i.WriteHtonU32(direction);
}


//becuase we need to read doubles from the buffer, we need
//to do some strange casting to make sure the anti-aliasing rules aren't violated
uint32_t PHY_INFO_Header::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing PHY_INFO_HEADER\n");

	Buffer::Iterator i = start;
	uint8_t temp;
	
	uint64_t temp2 = i.ReadNtohU64();
	uint64_t *conversion_holder = &temp2;
	freq_ = *reinterpret_cast<double *>(conversion_holder);
	
	temp  = i.ReadU8();
	modulation_ = (Ofdm_mod_rate) temp;
	
	temp2 = i.ReadNtohU64();
	g_ =  *reinterpret_cast<double *>(conversion_holder);
	temp2 = i.ReadNtohU64();
	subcarrier_lbound =  *reinterpret_cast<double *>(conversion_holder);
	temp2 = i.ReadNtohU64();
	subcarrier_ubound = *reinterpret_cast<double *>(conversion_holder);
	temp2 = i.ReadNtohU64();
	time_lbound =  *reinterpret_cast<double *>(conversion_holder);
	temp2 = i.ReadNtohU64();
	time_ubound = *reinterpret_cast<double *>(conversion_holder);
	num_subchannels = i.ReadNtohU32();
	subchannel_offset = i.ReadNtohU32();
	num_OFDMSymbol = i.ReadNtohU32();
	OFDMSymbol_offset = i.ReadNtohU32();
	channel_index = i.ReadNtohU32();
	direction = i.ReadNtohU32();
	return GetSerializedSize();
}
void PHY_INFO_Header::Print(std::ostream &os) const{
	os << "PHY_INFO_Header: " << std::endl;
	os << "freq: " << freq_<< std::endl;
	os << "modulation: " << modulation_<< std::endl;
	os << "g (cyclic prefix information): " << g_<< std::endl;
	os << "subcarrier_lbound: " << subcarrier_lbound<< std::endl;
	os << "subcarrier_ubound: " << subcarrier_ubound<< std::endl;
	os << "time_lbound: " << time_lbound<< std::endl;
	os << "time_ubound: " << time_ubound<< std::endl;
	os << "num_subchannels: "<< num_subchannels<< std::endl;
	os << "subchannel_offset: " << subchannel_offset<< std::endl;
	os << "num_OFDMSymbol: " << num_OFDMSymbol<< std::endl;
	os << "OFDMSymbol_offset: " << OFDMSymbol_offset<< std::endl;
	os << "channel_index: " << channel_index<< std::endl;
	os << "direction: ";  if (direction) os << "UL\n"; else os << "DL\n";
}

TypeId PHY_INFO_Header::GetTypeId() {
	static TypeId tid = TypeId("ns3::PHY_INFO_Header")
	.SetParent<Tag> ()
	.AddConstructor<PHY_INFO_Header>()
	;
	return tid;
}
TypeId PHY_INFO_Header::GetInstanceTypeId() const{
	return GetTypeId();
}


All_Headers::All_Headers() {
	 gm_sm = 1;
	 type_arqfb = &(GMH.GMH.type_arqfb);
	 type_ext = &(GMH.GMH.type_ext);
	 type_frag = &(GMH.GMH.type_frag);
	 type_pck = &(GMH.GMH.type_pck);
	 type_fbgm = &(GMH.GMH.type_fbgm);
	 
	 *type_arqfb = false;
	 *type_ext = false;
	 *type_frag = false;
	 *type_pck = false;
	 *type_fbgm = false;
	 
	 num_of_acks = 0;
	 cdma = 0;
	 direction = NONE;
	 txtime = 0;
	 timestamp = 0;
	 dest = Mac48Address::GetBroadcast();
	 error = false;
	 mac_packet = false;

	 GMH.GMH.ht = 0;
	 GMH.GMH.ec = 0;
	 GMH.GMH.type_mesh = false;
	 GMH.GMH.type_arqfb = false;
	 GMH.GMH.type_ext = false;
	 GMH.GMH.type_frag = false;
	 GMH.GMH.type_pck = false;
	 GMH.GMH.type_pck = false;
	 GMH.GMH.rsv1 = 0;
	 GMH.GMH.ci = 0;
	 GMH.GMH.eks = 0;
	 GMH.GMH.rsv2 = 0;
	 GMH.GMH.len = 0;
	 GMH.GMH.cid = 0;
	 GMH.GMH.hcs = 0;

	 SMH.BRH.ht = 0;
	 SMH.BRH.ec = 0;
	 SMH.BRH.type = 0;
	 SMH.BRH.br = 0;
	 SMH.BRH.cid = 0;
	 SMH.BRH.hcs = 0;

	 CRH.data.ht = 0;
	 CRH.data.ec = 0;
	 CRH.data.type = 0;
	 CRH.data.top = 0;
	 CRH.data.br = 0;
	 CRH.data.code = 0;
	 CRH.data.cid = 0;
}

All_Headers::~All_Headers() {}

TypeId All_Headers::GetTypeId() {
	static TypeId tid = TypeId("ns3::All_Headers")
	.SetParent<Header> ()
	.AddConstructor<All_Headers>()
	;
	return tid;
}
TypeId All_Headers::GetInstanceTypeId() const{
	return GetTypeId();
}

void All_Headers::Print(std::ostream &os) const{
	os << "PRINTING INSTANCE OF All_Headers:\n";
	if 		(gm_sm == 0) SMH.Print(os);
	else if (gm_sm == 1) GMH.Print(os);
	else CRH.Print(os);
	
	if (type_frag) FSH.Print(os);
	if (type_fbgm && PIH.direction == 1) GMSH.Print(os);
	if (type_pck) PSH.Print(os);
	if (type_fbgm && PIH.direction == 0) FFSH.Print(os);
	PIH.Print(os);
	os << "All_Headers.cdma = " << cdma
	   << "\nAll_Headers.num_of_acks = " << num_of_acks << std::endl;
	
}

uint32_t All_Headers::GetSerializedSize() const {
	uint32_t size = GMH.GetSerializedSize();
	size += SMH.GetSerializedSize();
	size += CRH.GetSerializedSize();
	
	size += PIH.GetSerializedSize();
	
	if (type_frag) size += FSH.GetSerializedSize();
	if (type_fbgm && PIH.direction == 1) size += GMSH.GetSerializedSize();
	if (type_pck) size += PSH.GetSerializedSize();
	if (type_fbgm && PIH.direction == 0) size += FFSH.GetSerializedSize();
	
	if (GMH.GMH.type_arqfb) {
		size += 2; //num_of_acks
		size += 5*num_of_acks; //arq_ie
	}
	
	size += 28; //gm_sm, moth.type_arqfb, cdma, direction, txtime, timestamp, dest, error, mac_packet
	return size;
}

void All_Headers::Serialize (Buffer::Iterator start) const{
	debugHeader("serializing all_headers\n");

	Buffer::Iterator i = start;
	
	GMH.Serialize(i);
	i.Next(GMH.GetSerializedSize());	
	SMH.Serialize(i);
	i.Next(SMH.GetSerializedSize());
	CRH.Serialize(i);
	i.Next(CRH.GetSerializedSize());
	PIH.Serialize(i);
	i.Next(PIH.GetSerializedSize());

	if (type_frag) {
		FSH.Serialize(i);
		i.Next(FSH.GetSerializedSize());
	}
	if (type_fbgm && PIH.direction == 1) {
		GMSH.Serialize(i);
		i.Next(GMSH.GetSerializedSize());
	}
	if (type_pck){
		PSH.Serialize(i);
		i.Next(PSH.GetSerializedSize());
	}
	if (type_fbgm && PIH.direction == 0) {
		FFSH.Serialize(i);
		i.Next(FFSH.GetSerializedSize());
	}


	MISC_OTHER_HEADER moth;		 

	if (gm_sm == 1) moth.type_arqfb = GMH.GMH.type_arqfb;	//1
	else moth.type_arqfb = false;

	moth.num_of_acks = num_of_acks;

	for (int j = 0; j < num_of_acks; j++) {
		moth.arq_ie[j] = arq_ie[j];
	}

	moth.gm_sm = gm_sm;			//1
	moth.cdma = cdma;			//2
	moth.direction = direction;	//3
	moth.txtime = txtime;		//11
	moth.timestamp = timestamp;	//19
	moth.dest = dest;			//25
	moth.error = error;			//26
	moth.Serialize(i);	
	i.Next(moth.GetSerializedSize());





	// 
// 	
// 	if (GMH.GMH.type_arqfb) {
// 		i.WriteHtonU16(num_of_acks);
// 		for (int j = 0; j < num_of_acks; j++) {
// 			i.WriteHtonU16(arq_ie[j].cid);
// 			i.WriteU8(arq_ie[j].last);
// 			uint16_t temp = 0;
// 			temp = temp & arq_ie[j].ack_type;
// 			temp = temp & (arq_ie[j].fsn << 2);
// 			i.WriteHtonU16(temp);
// 		}
// 	}
// 	
// 	i.WriteU8(gm_sm);
// 	i.WriteU8(cdma);
// 	uint8_t temp = direction;
// 	i.WriteU8(temp);
// 	
// 	const double *conversion_holder = &txtime;
// 	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
// 	conversion_holder = &timestamp;
// 	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
// 	
// 	
// 	uint8_t adr[6];
// 	dest.CopyTo(adr);
// 	for (int j = 0; j < 6; j++)
// 		i.WriteU8(adr[j]);
// 	
// 	i.WriteU8(error);
}

uint32_t All_Headers::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing All_Headers\n");
	Buffer::Iterator i = start;
	
	GMH.Deserialize(i); i.Next(GMH.GetSerializedSize());
	SMH.Deserialize(i); i.Next(SMH.GetSerializedSize());
	CRH.Deserialize(i); i.Next(CRH.GetSerializedSize());
	
	PIH.Deserialize(i); i.Next(PIH.GetSerializedSize());

	if (type_frag) {
		FSH.Deserialize(i);
		i.Next(FSH.GetSerializedSize());
	}
	if (type_fbgm && PIH.direction == 1) { 
		GMSH.Deserialize(i);
		i.Next(GMSH.GetSerializedSize());
	}
	if (type_pck) { 
		PSH.Deserialize(i);
		i.Next(PSH.GetSerializedSize());
	}
	if (type_fbgm && PIH.direction == 0) {
		FFSH.Deserialize(i);
		i.Next(FFSH.GetSerializedSize());
	}
	
	MISC_OTHER_HEADER moth;
	
	moth.Deserialize(i);
	i.Next(moth.GetSerializedSize());
	
	num_of_acks = moth.num_of_acks;
	for (int j = 0; j < num_of_acks; j++) arq_ie[j] = moth.arq_ie[j];
	gm_sm = moth.gm_sm;
	cdma = moth.cdma;
	direction = moth.direction;
	txtime = moth.txtime;
	timestamp = moth.timestamp;
	dest = moth.dest;
	error = moth.error;
	

// 
// 	
// 	if (GMH.GMH.type_arqfb) {
// 		num_of_acks = i.ReadNtohU16();
// 		for (int j = 0; j < num_of_acks; j++) {
// 			arq_ie[j].cid = i.ReadNtohU16();
// 			arq_ie[j].last = i.ReadU8();
// 			uint16_t temp = i.ReadNtohU16(), mask = 3;
// 			arq_ie[j].ack_type = mask & temp;
// 			mask = 2047 << 2;
// 			arq_ie[j].fsn = (mask & temp) >> 2;
// 		}
// 	}
// 	
// 	gm_sm = i.ReadU8();
// 	cdma = i.ReadU8();
// 	direction = static_cast<dir_t>(i.ReadU8());
// 	
// 	
// 	uint64_t temp2 = i.ReadNtohU64();
// 	uint64_t *conversion_holder = &temp2;
// 	txtime = *reinterpret_cast<double *>(conversion_holder);
// 	
// 	temp2 = i.ReadNtohU64();
// 	conversion_holder = &temp2;
// 	timestamp = *reinterpret_cast<double *>(conversion_holder);
// 	
// 	uint8_t adr[6];
// 	for (int j = 0; j < 6; j++)
// 		adr[j] = i.ReadU8();
// 	dest.CopyFrom(adr);
// 	
// 	error = i.ReadU8();
// 	
// 	type_arqfb = &GMH.GMH.type_arqfb;
// 	type_ext = &GMH.GMH.type_ext;
// 	type_frag = &GMH.GMH.type_frag;
// 	type_pck = &GMH.GMH.type_pck;
// 	type_fbgm = &GMH.GMH.type_fbgm;
	
	return GetSerializedSize();
}




MISC_OTHER_HEADER::MISC_OTHER_HEADER() {
	gm_sm = 0;
	type_arqfb = false;
	cdma = 0;
	direction = NONE;
	txtime = 0;
	timestamp = 0;
	dest = Mac48Address::GetBroadcast();
	error = false;
	mac_packet = false;
}
MISC_OTHER_HEADER::~MISC_OTHER_HEADER() {}
TypeId MISC_OTHER_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::FRAG_SUBHEADER")
	.SetParent<Header> ()
	.AddConstructor<MISC_OTHER_HEADER>()
	;
	return tid;
}
TypeId MISC_OTHER_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void MISC_OTHER_HEADER::Print(std::ostream &os) const{
	os << "PRINTING MISC_OTHER_HEADER\n";	
	os << "\tgm_sm: " 		<< (int)gm_sm 		<< std::endl;
	os << "\ttype_arqfb: " 	<< (int)type_arqfb 	<< std::endl;
	os << "\tcmda: " 		<<(int) cdma 		<< std::endl;
	os << "\tdirection: " 	<< direction 		<< std::endl;
	os << "\ttxtime: " 		<< txtime 			<< std::endl;
	os << "\ttimestamp: " 	<< timestamp 		<< std::endl;
	os << "\tdest: "		<< dest 			<< std::endl;
	os << "\terror: " 		<< error 			<< std::endl;
	os << "\tmac_packet: " 	<< mac_packet		<< std::endl;
	os << "\tnum_of_acks: " << num_of_acks 		<< std::endl;
	if (type_arqfb) {
		for (int i = 0; i < num_of_acks; i++) {
			os << "\t\tPrinting ARQ_IE " << i << std::endl;
			os << "\t\tcid: " << arq_ie[i].cid << std::endl;
			os << "\t\tlast: " <<(int)arq_ie[i].last << std::endl;
			os << "\t\tack_type: " <<(int)arq_ie[i].ack_type << std::endl;
			os << "\t\tfsn: " <<arq_ie[i].fsn << std::endl;
		}
	}
}

uint32_t MISC_OTHER_HEADER::GetSerializedSize() const {
	int size = 28; //gm_sm, type_arqfb, cdma, direction, txtime, timestamp, dest, error
	if (type_arqfb) {
		size += 2; //num_of_acks
		size += 5*num_of_acks; //arq_ie
	}
	
	
	return size;
}

uint32_t MISC_OTHER_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing MISC_OTHER_HEADER\n");

	Buffer::Iterator i = start;
	
	gm_sm = i.ReadU8();
	if (i.ReadU8() == 1) type_arqfb = true;
	else type_arqfb = false;
	
	cdma = i.ReadU8();
	direction = static_cast<dir_t>(i.ReadU8());
	
	
	uint64_t temp2 = i.ReadNtohU64();
	uint64_t *conversion_holder = &temp2;
	txtime = *reinterpret_cast<double *>(conversion_holder);
	
	temp2 = i.ReadNtohU64();
	conversion_holder = &temp2;
	timestamp = *reinterpret_cast<double *>(conversion_holder);
	
	uint8_t adr[6];
	for (int j = 0; j < 6; j++)
		adr[j] = i.ReadU8();
	dest.CopyFrom(adr);
	
	error = i.ReadU8();
	mac_packet = i.ReadU8();
		
	if (type_arqfb) {
		num_of_acks = i.ReadNtohU16();
		for (int j = 0; j < num_of_acks; j++) {
			arq_ie[j].cid = i.ReadNtohU16();
			arq_ie[j].last = i.ReadU8();
			uint16_t temp = i.ReadNtohU16(), mask = 3;
			arq_ie[j].ack_type = mask & temp;
			mask = 2047 << 2;
			arq_ie[j].fsn = (mask & temp) >> 2;
		}
	}
	else num_of_acks = 0;
	
	//cout << "Deserializing MISC_OTHER_HEADER\n" <<std::endl;
	//Print(cout);
	
	return GetSerializedSize();
}


void MISC_OTHER_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	
	//cout << "Serializing MISC_OTHER_HEADER\n" <<std::endl;
	//Print(cout);
	
		
	i.WriteU8(gm_sm);
	if (type_arqfb) i.WriteU8(1);
	else i.WriteU8(0);
	i.WriteU8(cdma);
	uint8_t temp = direction;
	i.WriteU8(temp);
	
	const double *conversion_holder = &txtime;
	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
	conversion_holder = &timestamp;
	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
	
	
	uint8_t adr[6];
	dest.CopyTo(adr);
	for (int j = 0; j < 6; j++)
		i.WriteU8(adr[j]);
	
	i.WriteU8(error);
	i.WriteU8(mac_packet);
	
	if (type_arqfb) {
		i.WriteHtonU16(num_of_acks);
		for (int j = 0; j < num_of_acks; j++) {
			i.WriteHtonU16(arq_ie[j].cid);
			i.WriteU8(arq_ie[j].last);
			uint16_t temp = 0;
			temp = temp & arq_ie[j].ack_type;
			temp = temp & (arq_ie[j].fsn << 2);
			i.WriteHtonU16(temp);
		}
	}

}
	





FRAG_SUBHEADER::FRAG_SUBHEADER() {
	data.fc = 0;
	data.fsn = 0;
	data.sn = 0;
	data.rsv = 0;
}
FRAG_SUBHEADER::~FRAG_SUBHEADER() {}
TypeId FRAG_SUBHEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::FRAG_SUBHEADER")
	.SetParent<Header> ()
	.AddConstructor<FRAG_SUBHEADER>()
	;
	return tid;
}
TypeId FRAG_SUBHEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void FRAG_SUBHEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE FRAG_SUBHEADER Header\n";	
}

uint32_t FRAG_SUBHEADER::GetSerializedSize() const {
	return 3;
}

void FRAG_SUBHEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint16_t temp = 0, mask = 3;
	temp = temp | (mask & data.fc);
	mask = 7;
	temp = temp | ((mask & data.fsn) << 2);
	mask = 2047;
	temp = temp | ((mask & data.sn) << 5);
	i.WriteHtonU16(temp);
	uint8_t temp2 = 0, mask2 = 7;
	temp2 = temp2 | (mask2 & data.rsv);
	i.WriteU8(temp2);
}

uint32_t FRAG_SUBHEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing FRAG_SUBHEADER\n");

	Buffer::Iterator i = start;
	uint16_t temp = i.ReadNtohU16(), mask = 3;
	data.fc = (mask & temp);
	mask = 7 << 2;
	data.fsn = (mask & temp) >> 2;
	mask = 2047 << 5;
	data.sn = (mask & temp) >> 5;
	uint8_t temp2 = i.ReadU8(), mask2 = 7;
	data.rsv = (mask2 & temp2);
	
	return GetSerializedSize();
}



GRANT_MAP_SUBHEADER::GRANT_MAP_SUBHEADER() {
	data.piggyback_req = 0;
}
GRANT_MAP_SUBHEADER::~GRANT_MAP_SUBHEADER() {}
TypeId GRANT_MAP_SUBHEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::GRANT_MAP_SUBHEADER")
	.SetParent<Header> ()
	.AddConstructor<GRANT_MAP_SUBHEADER>()
	;
	return tid;
}
TypeId GRANT_MAP_SUBHEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void GRANT_MAP_SUBHEADER::Print(std::ostream &os) const{
	os << "Piggybacked Bandwidth Request of Size: " << data.piggyback_req << " Bytes\n";
}

uint32_t GRANT_MAP_SUBHEADER::GetSerializedSize() const {
	return 2;
}

void GRANT_MAP_SUBHEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteHtonU16(data.piggyback_req);
}

uint32_t GRANT_MAP_SUBHEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing GRANT_MAP_SUBHEADER\n");
	Buffer::Iterator i = start;
	data.piggyback_req = i.ReadNtohU16();
	return GetSerializedSize();
}


PACKING_SUBHEADER::PACKING_SUBHEADER() {
	data.fc = 0;
	data.sn = 0;
	data.length = 0;
}
PACKING_SUBHEADER::~PACKING_SUBHEADER() {}
TypeId PACKING_SUBHEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::PACKING_SUBHEADER")
	.SetParent<Header> ()
	.AddConstructor<PACKING_SUBHEADER>()
	;
	return tid;
}
TypeId PACKING_SUBHEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void PACKING_SUBHEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE PACKING_SUBHEADER Header\n";	
}

uint32_t PACKING_SUBHEADER::GetSerializedSize() const {
	return 3;
}

void PACKING_SUBHEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint16_t temp = 0;
	temp = temp | (data.fc);
	temp = temp | (data.sn << 2);
	temp = temp | (data.length << 13);
	i.WriteHtonU16(temp);
	
	uint8_t temp2 = (data.length >> 3);
	i.WriteU8(temp2);
}

uint32_t PACKING_SUBHEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing PACKING_SUBHEADER\n");

	Buffer::Iterator i = start;
	uint16_t temp = 0, mask = 3;
	temp = i.ReadNtohU16();
	data.fc = temp & mask;
	mask = 2047 << 2;
	data.sn = (temp & mask) >> 2;
	mask = 7 << 13;
	data.length = 0;
	data.length = data.length | ((mask & temp) >> 13);
	data.length = data.length | (i.ReadU8() << 3);
	return GetSerializedSize();
}


FFB_SUBHEADER::FFB_SUBHEADER() {
	data.alloc_offset = 0;
	data.type = 0;
}
FFB_SUBHEADER::~FFB_SUBHEADER() {}
TypeId FFB_SUBHEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::FFB_SUBHEADER")
	.SetParent<Header> ()
	.AddConstructor<FFB_SUBHEADER>()
	;
	return tid;
}
TypeId FFB_SUBHEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void FFB_SUBHEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE FFB_SUBHEADER Header\n";	
}

uint32_t FFB_SUBHEADER::GetSerializedSize() const {
	return 1;
}

void FFB_SUBHEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint8_t temp = 0;
	temp = temp & data.alloc_offset;
	temp = temp & (data.type << 6);
	i.WriteU8(temp);
}

uint32_t FFB_SUBHEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing FFB_SUBHEADER\n");

	Buffer::Iterator i = start;
	uint8_t temp = i.ReadU8(), mask = 63;
	data.alloc_offset = temp & mask;
	mask = 3 << 6;
	data.type = (temp & mask) >> 6;
	return GetSerializedSize();
}


RANGING_REQUEST_FRAME::RANGING_REQUEST_FRAME() {
	data.type = 0;
	data.dc_id = 0;
	data.req_dl_burst_profile = 0;
	data.ss_mac_address = Mac48Address::GetBroadcast();
	data.channel_num = 0;
}
RANGING_REQUEST_FRAME::~RANGING_REQUEST_FRAME() {}
TypeId RANGING_REQUEST_FRAME::GetTypeId() {
	static TypeId tid = TypeId("ns3::RANGING_REQUEST_FRAME")
	.SetParent<Header> ()
	.AddConstructor<RANGING_REQUEST_FRAME>()
	;
	return tid;
}
TypeId RANGING_REQUEST_FRAME::GetInstanceTypeId() const{
	return GetTypeId();
}

void RANGING_REQUEST_FRAME::Print(std::ostream &os) const{
	os << "PRINTING RANGING_REQUEST_FRAME Header\n";	
	os << "\tTYPE: " << (int)data.type;
	os << "\n\tDC_ID: " << (int)data.dc_id;
	os << "\n\tReq_burst_profile: " << (int)data.req_dl_burst_profile;
	os << "\n\tSS_mac_address: " << data.ss_mac_address;
	os << "\n\tchannel_num: " << data.channel_num << std::endl;
}

uint32_t RANGING_REQUEST_FRAME::GetSerializedSize() const {
	return 13;
}

void RANGING_REQUEST_FRAME::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	i.WriteU8(data.dc_id);
	i.WriteU8(data.req_dl_burst_profile);
	uint8_t adr[6];
	data.ss_mac_address.CopyTo(adr);
	for (int j = 0; j < 6; j++)
		i.WriteU8(adr[j]);
	i.WriteHtonU32(data.channel_num);
}

uint32_t RANGING_REQUEST_FRAME::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing RANGING_REQUEST_FRAME\n");

	Buffer::Iterator i = start;
	data.type = i.ReadU8();
	data.dc_id = i.ReadU8();
	data.req_dl_burst_profile = i.ReadU8();
	
	uint8_t adr[6];
	for (int j = 0; j < 6; j++)
		adr[j] = i.ReadU8();
	data.ss_mac_address.CopyFrom(adr);
	
	data.channel_num = i.ReadNtohU32();
	return GetSerializedSize();
}



CDMA_REQUEST_HEADER::CDMA_REQUEST_HEADER() {
	data.ht = 0;
	data.ec = 0;
	data.type = 0;
	data.top = 0;
	data.br = 0;
	data.code = 0;
	data.cid = 0;
}
CDMA_REQUEST_HEADER::~CDMA_REQUEST_HEADER() {}
TypeId CDMA_REQUEST_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::CDMA_REQUEST_HEADER")
	.SetParent<Header> ()
	.AddConstructor<CDMA_REQUEST_HEADER>()
	;
	return tid;
}
TypeId CDMA_REQUEST_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void CDMA_REQUEST_HEADER::Print(std::ostream &os) const{
	os << "PRINTING CDMA_REQUEST_HEADER \n";	
	os << "ht: " 	<< (int)data.ht << std::endl;
	os << "ec: " 	<< (int)data.ec << std::endl;
	os << "type: " 	<< (int)data.type << std::endl;
	os << "top: " 	<< (int)data.top << std::endl;
	os << "br: " 	<< (int)data.br << std::endl;
	os << "code: " 	<< (int)data.code << std::endl;
	os << "cid: " 	<< (int)data.cid << std::endl;
}

uint32_t CDMA_REQUEST_HEADER::GetSerializedSize() const {
	return 6;
}

void CDMA_REQUEST_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint32_t temp = 0;
	temp = temp | data.ht;
	temp = temp | (data.ec   << 1);
	temp = temp | (data.type << 2);
	temp = temp | (data.top  << 5);
	temp = temp | (data.br   << 13);
	temp = temp | (data.code << 24);
// 	for (int j = 0; j < 4; j++) {
// 		uint8_t temp2 = 0;
// 		temp2 = temp2 | temp >> (8*j);
// 		i.WriteU8(temp2);
// 
// 	}
	i.WriteHtonU32(temp);
	i.WriteHtonU16(data.cid);
}

uint32_t CDMA_REQUEST_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing CDMA_REQUEST_HEADER\n");

	Buffer::Iterator i = start;
	uint32_t temp = 0, mask = 1;
//	uint8_t temp2;
// 	for (int j = 0; j < 4; j++) {
// 		temp2 = i.ReadU8();
// 		temp = temp | (temp2 << 8*j);
// 	}
	temp = i.ReadNtohU32();


	data.ht = temp & mask;
	mask = 2;
	data.ec = (temp & mask) >> 1;
	mask = 7 << 2;
	data.type = (temp & mask) >> 2;
	mask = 255 << 5;
	data.top = (temp & mask) >> 5;
	mask = 2047 << 13;
	data.br = (temp & mask) >> 13;
	mask = 255 << 24;
	data.code = (temp & mask) >> 24;
	data.cid = i.ReadNtohU16();
	return GetSerializedSize();
}


DCD_FRAME_HEADER::DCD_FRAME_HEADER() {
	mf.type = 0;
	mf.dcid = 0;
	mf.config_change_count = 0;
	mf.frame_duration_code = 0;
	mf.frame_number = 0;
	mf.ttg = 0;
	mf.rtg = 0;
	mf.frequency = 0;
	mf.nb_prof = 0;
}
DCD_FRAME_HEADER::~DCD_FRAME_HEADER() {}
TypeId DCD_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::DCD_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<DCD_FRAME_HEADER>()
	;
	return tid;
}
TypeId DCD_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void DCD_FRAME_HEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE DCD_FRAME_HEADER Header\n";	
}

uint32_t DCD_FRAME_HEADER::GetSerializedSize() const {
	return 17 + mf.nb_prof * 6;
}

void DCD_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(mf.type);
	i.WriteU8(mf.dcid);
	i.WriteU8(mf.config_change_count);
	i.WriteU8(mf.frame_duration_code);
	
	uint32_t temp = 0;
	temp = temp | (mf.frame_number & 0xFFFFFF);
	temp = temp | (mf.ttg << 24);
	i.WriteHtonU32(temp);
	i.WriteU8(mf.rtg);
	i.WriteHtonU32(mf.frequency);
	i.WriteHtonU32(mf.nb_prof);
	
	for (uint32_t j = 0; j < mf.nb_prof; j++) {
		uint8_t temp2 = mf.profiles[j].diuc;
		i.WriteU8(temp2);
		i.WriteHtonU32(mf.profiles[j].frequency);
		i.WriteU8(mf.profiles[j].fec);
	}
}

uint32_t DCD_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing DCD_FRAME_HEADER\n");

	Buffer::Iterator i = start;
	mf.type = i.ReadU8();
	mf.dcid = i.ReadU8();
	mf.config_change_count = i.ReadU8();
	mf.frame_duration_code = i.ReadU8();
	uint32_t temp = i.ReadNtohU32(), mask = 0xFFFFFF; //pow = 2^24 - 1
	mf.frame_number = mask & temp;
	mask = 255 << 24;
	mf.ttg = (mask & temp) >> 24;
	mf.rtg = i.ReadU8();
	mf.frequency = i.ReadNtohU32();
	mf.nb_prof = i.ReadNtohU32();
	
	for (uint32_t j = 0; j < mf.nb_prof; j++) {
		uint8_t temp2 = i.ReadU8(), mask2 = 15;
		mf.profiles[j].diuc = mask2&temp2;
		mf.profiles[j].frequency = i.ReadNtohU32();
		mf.profiles[j].fec = i.ReadU8();
	}
	
	return GetSerializedSize();
}



ARQ_FB_IE_HEADER::ARQ_FB_IE_HEADER() {
	arqfbie.cid = 0;
	arqfbie.last = false;
	arqfbie.ack_type = 0;
	arqfbie.fsn = 0;
	last_ = false;
}
ARQ_FB_IE_HEADER::~ARQ_FB_IE_HEADER() {}
TypeId ARQ_FB_IE_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::ARQ_FB_IE_HEADER")
	.SetParent<Header> ()
	;
	return tid;
}
TypeId ARQ_FB_IE_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void ARQ_FB_IE_HEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE ARQ_FB_IE_HEADER Header\n";	
}

uint32_t ARQ_FB_IE_HEADER::GetSerializedSize() const {
	return 5;
}

void ARQ_FB_IE_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteHtonU16(arqfbie.cid);
	i.WriteU8(arqfbie.last);
	uint16_t temp = 0;
	temp = temp | arqfbie.ack_type;
	temp = temp | (arqfbie.fsn << 2);
	if (last_) temp = temp | (1 << 15);
	else temp = temp | (0 << 15);
	i.WriteHtonU16(temp);
}

uint32_t ARQ_FB_IE_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing ARQ_FB_IE_HEADER\n");

	Buffer::Iterator i = start;
	//arqfbie = new arq_fb_ie;
	arqfbie.cid = i.ReadNtohU16();
	arqfbie.last = i.ReadU8();
	uint16_t temp = i.ReadNtohU16(), mask = 3;
	arqfbie.ack_type = temp & mask;
	mask = 2047 << 2;
	arqfbie.fsn = (temp & mask) >> 2;
	last_ = (temp & (1 << 15))>>15;
	return GetSerializedSize();
}

void ARQ_FB_IE_HEADER::last(bool l) {
	last_ = l;
}


mac802_16_mob_scn_req_frame::mac802_16_mob_scn_req_frame() {
	type = 54;
	scan_duration = 0;
	interleaving_interval = 0;
	scan_iteration = 0;
	n_recommended_bs_index = 0;
	ccc = 0;
	n_recommended_bs_full = 0;
}
mac802_16_mob_scn_req_frame::~mac802_16_mob_scn_req_frame() {}
TypeId mac802_16_mob_scn_req_frame::GetTypeId() {
	static TypeId tid = TypeId("ns3::mac802_16_mob_scn_req_frame")
	.SetParent<Header> ()
	;
	return tid;
}
TypeId mac802_16_mob_scn_req_frame::GetInstanceTypeId() const{
	return GetTypeId();
}

void mac802_16_mob_scn_req_frame::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE mac802_16_mob_scn_req_frame Header\n";	
}

uint32_t mac802_16_mob_scn_req_frame::GetSerializedSize() const {
	return 7 + 2*n_recommended_bs_index + 7 * n_recommended_bs_full;
}

void mac802_16_mob_scn_req_frame::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint8_t temp;
	i.WriteU8(type);
	i.WriteU8(scan_duration);
	i.WriteU8(interleaving_interval);
	i.WriteU8(scan_iteration);
	i.WriteU8(n_recommended_bs_index);
	i.WriteU8(ccc);
	
	for (int j = 0; j < n_recommended_bs_index; j++) {
		i.WriteU8(rec_bs_index[j].neighbor_bs_index);
		temp = rec_bs_index[j].scanning_type;
		i.WriteU8(temp);
	}
	
	i.WriteU8(n_recommended_bs_full);
	
	for (int j = 0; j < n_recommended_bs_full; j++) {
		uint8_t adr[6];
		rec_bs_full[j].recommended_bs_id.CopyTo(adr);
		for (int k = 0; k < 6; k++)
			i.WriteU8(adr[k]);
		temp = rec_bs_full[j].scanning_type;
		i.WriteU8(temp);
	}
}

uint32_t mac802_16_mob_scn_req_frame::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing mac802_16_mob_scn_req_frame\n");

	Buffer::Iterator i = start;
	uint8_t temp;
	type = i.ReadU8();
	scan_duration  = i.ReadU8();
	interleaving_interval = i.ReadU8();
	scan_iteration = i.ReadU8();
	n_recommended_bs_index = i.ReadU8();
	ccc = i.ReadU8();
	
	for (int j = 0; j < n_recommended_bs_index; j++) {
		rec_bs_index[j].neighbor_bs_index  = i.ReadU8();
		temp  = i.ReadU8();
		rec_bs_index[j].scanning_type = temp;
	}
	n_recommended_bs_full = i.ReadU8();
	for (int j = 0; j < n_recommended_bs_full; j++) {
		uint8_t adr[6];
		for (int k = 0; k < 6; k++)
			adr[k] = i.ReadU8();
		rec_bs_full[j].recommended_bs_id.CopyFrom(adr);
		temp = i.ReadU8();
		rec_bs_full[j].scanning_type = temp;
	}
	return GetSerializedSize();
}





mac802_16_mob_scn_rsp_frame::mac802_16_mob_scn_rsp_frame() {
	type = 55;
	scan_duration = 0;
	report_mode = 0;
	report_period = 0;
	report_metric = 0;
	start_frame = 0;
	interleaving_interval = 0;
	scan_iteration = 0;
	n_recommended_bs_index = 0;
	ccc_mob_nbr_adv = 0;
	n_recommended_bs_full = 0;
}
mac802_16_mob_scn_rsp_frame::~mac802_16_mob_scn_rsp_frame() {}
TypeId mac802_16_mob_scn_rsp_frame::GetTypeId() {
	static TypeId tid = TypeId("ns3::mac802_16_mob_scn_rsp_frame")
	.SetParent<Header> ()
	;
	return tid;
}
TypeId mac802_16_mob_scn_rsp_frame::GetInstanceTypeId() const{
	return GetTypeId();
}

void mac802_16_mob_scn_rsp_frame::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE mac802_16_mob_scn_rsp_frame Header\n";	
}

uint32_t mac802_16_mob_scn_rsp_frame::GetSerializedSize() const {
	int len = 5;
	if (scan_duration != 0) {
		len += 4;
		if (n_recommended_bs_index != 0) {
			len += 2;
			for (int i = 0; i < n_recommended_bs_index; i++) {
				len += 2;
				if (rec_bs_index[i].scanning_type == 2 || rec_bs_index[i].scanning_type == 3)
					len += 3;
			}
			for (int i = 0; i < n_recommended_bs_full; i++) {
				len +=7;
				if (rec_bs_full[i].scanning_type == 2 || rec_bs_full[i].scanning_type == 3)
					len += 3;
			}
		}
	}
	return len;
}

void mac802_16_mob_scn_rsp_frame::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint8_t temp;
	i.WriteU8(type);
	i.WriteU8(scan_duration);
	temp = report_mode;
	i.WriteU8(temp);
	i.WriteU8(report_period);
	i.WriteU8(report_metric);
	
	if (scan_duration != 0) {
		temp = start_frame;
		i.WriteU8(temp);
		i.WriteU8(interleaving_interval);
		i.WriteU8(scan_iteration);
		i.WriteU8(n_recommended_bs_index);
		
		if (n_recommended_bs_index != 0) {
			i.WriteU8(ccc_mob_nbr_adv);
			for (int j = 0; j < n_recommended_bs_index; j++) {
				i.WriteU8(rec_bs_index[j].neighbor_bs_index);
				temp = rec_bs_index[j].scanning_type;
				i.WriteU8(temp);
				if (rec_bs_index[j].scanning_type == 2 || rec_bs_index[j].scanning_type == 3) {
					i.WriteU8(rec_bs_index[j].rdv_time);
					i.WriteU8(rec_bs_index[j].cdma_code);
					i.WriteU8(rec_bs_index[j].transmission_opp_offset);
				}
			}
			
			i.WriteU8(n_recommended_bs_full);
			
			for (int j = 0; j < n_recommended_bs_full; j++) {
				uint8_t adr[6];
				rec_bs_full[j].recommended_bs_id.CopyTo(adr);
				for (int k = 0; k < 6; k++)
					i.WriteU8(adr[k]);
				temp = rec_bs_full[j].scanning_type;
				i.WriteU8(temp);
				if (rec_bs_full[j].scanning_type == 2 || rec_bs_full[j].scanning_type == 3) {
					i.WriteU8(rec_bs_full[j].rdv_time);
					i.WriteU8(rec_bs_full[j].cdma_code);
					i.WriteU8(rec_bs_full[j].transmission_opp_offset);
				}

			}
		}
	}
}

uint32_t mac802_16_mob_scn_rsp_frame::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing mac802_16_mob_scn_rsp_frame\n");

	Buffer::Iterator i = start;
	uint8_t temp;
	type = i.ReadU8();
	scan_duration  = i.ReadU8();
	temp = i.ReadU8();
	report_mode = temp;
	report_period = i.ReadU8();
	report_metric = i.ReadU8();
	
	if (scan_duration != 0) {
		temp = i.ReadU8();
		start_frame = temp;
		interleaving_interval = i.ReadU8();
		scan_iteration = i.ReadU8();
		n_recommended_bs_index = i.ReadU8();
		
		if (n_recommended_bs_index != 0) {
			ccc_mob_nbr_adv = i.ReadU8();
			
			for (int j = 0; j < n_recommended_bs_index; j++) {
				rec_bs_index[j].neighbor_bs_index  = i.ReadU8();
				temp  = i.ReadU8();
				rec_bs_index[j].scanning_type = temp;
				if (rec_bs_index[j].scanning_type == 2 || rec_bs_index[j].scanning_type == 3) {
					rec_bs_index[j].rdv_time = i.ReadU8();
					rec_bs_index[j].cdma_code = i.ReadU8();
					rec_bs_index[j].transmission_opp_offset = i.ReadU8();
				}
			}
			n_recommended_bs_full = i.ReadU8();
			for (int j = 0; j < n_recommended_bs_full; j++) {
				uint8_t adr[6];
				for (int k = 0; k < 6; k++)
					adr[k] = i.ReadU8();
				rec_bs_full[j].recommended_bs_id.CopyFrom(adr);
				rec_bs_full[j].scanning_type = temp;
				if (rec_bs_full[j].scanning_type == 2 || rec_bs_full[j].scanning_type == 3) {
					rec_bs_full[j].rdv_time = i.ReadU8();
					rec_bs_full[j].cdma_code = i.ReadU8();
					rec_bs_full[j].transmission_opp_offset = i.ReadU8();
				}
			}
		}
	
	}
	return GetSerializedSize();
}


/**
 * Return the size of the MOB_NBR-ADV frame
 * @param frame The frame 
 */
int Mac802_16pkt::getMOB_NBR_ADV_size(mac802_16_mob_nbr_adv_frame *frame)
{
  int size = 4; //min size
  if ((frame->skip_opt_field & 0x1) == 0)
    size +=3; //add operator id
  
  for (int i = 0 ; i < frame->n_neighbors ; i++) {
    size += 4; //min size for neighbor info
    if (frame->nbr_info[i].phy_profile_id.FAindex & 0x1)
      size++;
    if (frame->nbr_info[i].phy_profile_id.bs_eirp & 0x1)
      size++;
    if ((frame->skip_opt_field & 0x2) == 0)
      size +=3; //contain neighbor bs id
    if ((frame->skip_opt_field & 0x4) == 0)
      size ++; //contain HO process optimization
    if ((frame->skip_opt_field & 0x8) == 0)
      size ++; //contain neighbor bs id
    if (frame->nbr_info[i].dcd_included)
      size += 2+GET_DCD_SIZE (frame->nbr_info[i].dcd_settings.mf.nb_prof); 
    if (frame->nbr_info[i].ucd_included)
      size += 2+GET_UCD_SIZE (frame->nbr_info[i].ucd_settings.ucdf.nb_prof); 
    if (frame->nbr_info[i].phy_included)
      size += 2;
  }

  return size;
}

/**
 * Return the size of the MOB_SCN-REQ 
 * @param frame The frame 
 */
int Mac802_16pkt::getMOB_SCN_REQ_size(mac802_16_mob_scn_req_frame *frame)
{
  int size=6;

  if (frame->n_recommended_bs_index != 0)
    size ++;
  int tmp = 11*(frame->n_recommended_bs_index+frame->n_recommended_bs_full);
  size += tmp/8;
  if ((tmp%8)!=0)
    size ++;

  return size;
}

/**
 * Return the size of the MOB_SCN-RSP 
 * @param frame The frame 
 */
int Mac802_16pkt::getMOB_SCN_RSP_size(mac802_16_mob_scn_rsp_frame *frame)
{
  int size=6;

  if (frame->scan_duration!=0) {
    size += 3;
    if (frame->n_recommended_bs_index!=0)
      size ++;
    int tmp = 0;
    for (int i = 0 ; i < frame->n_recommended_bs_index ; i++) {
      tmp+=11;
      if (frame->rec_bs_index[i].scanning_type==2 ||
	  frame->rec_bs_index[i].scanning_type==3)
	tmp+=24;
    }
    for (int i = 0 ; i < frame->n_recommended_bs_index ; i++) {
      tmp+=51;
      if (frame->rec_bs_index[i].scanning_type==2 ||
	  frame->rec_bs_index[i].scanning_type==3)
	tmp+=24;
    }
    size += tmp/8;
    if ((tmp%8)!=0)
      size ++;  
  }

  return size;
}

/**
 * Return the size of the MOB_MSHO-REQ 
 * @param frame The frame 
 */
int Mac802_16pkt::getMOB_MSHO_REQ_size(mac802_16_mob_msho_req_frame *frame)
{
  int size=4;
  int tmp, tmpB;

  tmp = 0;
  tmpB = 0;

  if (frame->n_new_bs_index !=0) 
    size++;

  if (frame->report_metric & 0x1) tmp++;
  if (frame->report_metric & 0x2) tmp++;
  if (frame->report_metric & 0x4) tmp++;

  for (int i = 0 ; i < frame->n_new_bs_index ; i++) {
    tmpB += 20 + 8*tmp;    
    if (frame->bs_index[i].arrival_time_diff_ind & 0x1)
      tmpB += 4;
  }
  //n_new_bs_full
  for (int i = 0 ; i < frame->n_new_bs_full ; i++) {
    tmpB += 20 + 8*tmp;
    if (frame->bs_full[i].arrival_time_diff_ind & 0x1)
      tmpB += 4;
  }
  tmpB += 4;
  //N_current
  if (frame->report_metric & 0x8) tmp++;
  for (int i = 0 ; i < frame->n_current_bs ; i++) {
    tmpB += 4 + 8*tmp;
  }
  //increase size
  size += tmp/8;
  if ((tmp%8)!=0)
    size ++;  //includes padding
  
  return size;
}

/**
 * Return the size of the MOB_MSHO-REQ 
 * @param frame The frame 
 */
int Mac802_16pkt::getMOB_BSHO_RSP_size(mac802_16_mob_bsho_rsp_frame *frame)
{
  int size=4;

  return size;
}

/**
 * Return the size of the MOB_HO-IND 
 * @param frame The frame 
 */
int Mac802_16pkt::getMOB_HO_IND_size(mac802_16_mob_ho_ind_frame *frame)
{
  int size=4;

  return size;
}



RANGING_RESPONSE_FRAME::RANGING_RESPONSE_FRAME() {
	data.type = 5;
	data.uc_id = 0;
	data.pw_adjust = 0;
	data.freq_adjust = 0;
	data.rng_status = 0;
	data.dl_op_burst_profile = 0;
	data.ss_mac_address = Mac48Address::GetBroadcast();
	data.basic_cid = 0;
	data.primary_cid = 0;
	data.aas_bc_perm = 0;
	data.top = 0;
	data.code = 0;
}
RANGING_RESPONSE_FRAME::~RANGING_RESPONSE_FRAME() {}
TypeId RANGING_RESPONSE_FRAME::GetTypeId() {
	static TypeId tid = TypeId("ns3::RANGING_RESPONSE_FRAME")
	.SetParent<Header> ()
	.AddConstructor<RANGING_RESPONSE_FRAME>()
	;
	return tid;
}
TypeId RANGING_RESPONSE_FRAME::GetInstanceTypeId() const{
	return GetTypeId();
}

void RANGING_RESPONSE_FRAME::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE RANGING_RESPONSE_FRAME Header\n";	
}

uint32_t RANGING_RESPONSE_FRAME::GetSerializedSize() const {
	return 23;
}

void RANGING_RESPONSE_FRAME::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	i.WriteU8(data.uc_id);
	i.WriteU8(data.pw_adjust);
	i.WriteHtonU32(data.freq_adjust);
	i.WriteU8(data.rng_status);
	
	i.WriteHtonU16(data.dl_op_burst_profile);
	
	
	uint8_t adr[6];
	data.ss_mac_address.CopyTo(adr);
	for (int j = 0; j < 6; j++)
		i.WriteU8(adr[j]);

	i.WriteHtonU16(data.basic_cid);
	i.WriteHtonU16(data.primary_cid);
	i.WriteU8(data.aas_bc_perm);
	i.WriteU8(data.top);
	i.WriteU8(data.code);
}

uint32_t RANGING_RESPONSE_FRAME::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing RANGING_RESPONSE_FRAME\n");

	Buffer::Iterator i = start;
	data.type = i.ReadU8();
	data.uc_id = i.ReadU8();
	data.pw_adjust = i.ReadU8();
	data.freq_adjust = i.ReadNtohU32();
	data.rng_status = i.ReadU8();
	data.dl_op_burst_profile = i.ReadNtohU16();
	
	uint8_t adr[6];
	for (int j = 0; j < 6; j++)
		adr[j] = i.ReadU8();
	data.ss_mac_address.CopyFrom(adr);
	
	data.basic_cid = i.ReadNtohU16();
	data.primary_cid = i.ReadNtohU16();
	data.aas_bc_perm = i.ReadU8();
	data.top = i.ReadU8();
	data.code = i.ReadU8();
	
	return GetSerializedSize();
}



DSA_REQ_FRAME_HEADER::DSA_REQ_FRAME_HEADER() {
	data.type = 0;
	data.transaction_id = 0;
	data.uplink = false;
	data.staticflow = NULL;
	data.cid = 0;
}
DSA_REQ_FRAME_HEADER::~DSA_REQ_FRAME_HEADER() {}
TypeId DSA_REQ_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::DSA_REQ_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<DSA_REQ_FRAME_HEADER>()
	;
	return tid;
}
TypeId DSA_REQ_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void DSA_REQ_FRAME_HEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE DSA_REQ_FRAME_HEADER Header\n";	
}

uint32_t DSA_REQ_FRAME_HEADER::GetSerializedSize() const {

#ifdef ENVIRONMENT32
	return 10;
#else
	return 14;
#endif
	
}

void DSA_REQ_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	i.WriteHtonU16(data.transaction_id);
	uint8_t temp = 0;
	if (data.uplink) temp = 1;
	i.WriteU8(temp);
#ifdef ENVIRONMENT32
	i.WriteHtonU32 ( reinterpret_cast<uint32_t>( data.staticflow));
#else
	i.WriteHtonU64 ( reinterpret_cast<uint64_t>( data.staticflow));
#endif
	i.WriteHtonU16(data.cid);
}

uint32_t DSA_REQ_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing DSA_REQ_FRAME_HEADER\n");

	Buffer::Iterator i = start;
	data.type = i.ReadU8();
	data.transaction_id = i.ReadNtohU16();
	uint8_t temp = i.ReadU8();
	if (temp) data.uplink = true;
	else data.uplink = false;
#ifdef ENVIRONMENT32
	uint32_t temp2 = i.ReadNtohU32();
	data.staticflow = reinterpret_cast<Wimax2ServiceFlow*>( temp2);
#else		
	uint64_t temp2 = i.ReadNtohU64();
	data.staticflow = reinterpret_cast<Wimax2ServiceFlow*> (temp2);
#endif
	data.cid = i.ReadNtohU16();
	
	return GetSerializedSize();
}



DSA_RSP_FRAME_HEADER::DSA_RSP_FRAME_HEADER() {
	data.type = 12;
	data.transaction_id = 0;
	data.confirmation_code = 0;
	data.uplink = false;
	data.staticflow = NULL;
	data.cid = 0;
}
DSA_RSP_FRAME_HEADER::~DSA_RSP_FRAME_HEADER() {}
TypeId DSA_RSP_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::DSA_RSP_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<DSA_RSP_FRAME_HEADER>()
	;
	return tid;
}
TypeId DSA_RSP_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void DSA_RSP_FRAME_HEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE DSA_RSP_FRAME_HEADER Header\n";	
}

uint32_t DSA_RSP_FRAME_HEADER::GetSerializedSize() const {
#ifdef ENVIRONMENT32
	return 11;
#else
	return 15;
#endif
}

void DSA_RSP_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	i.WriteHtonU16(data.transaction_id);
	i.WriteU8(data.confirmation_code);
	uint8_t temp = 0;
	if (data.uplink) temp = 1;
	i.WriteU8(temp);
	
#ifdef ENVIRONMENT32
	i.WriteHtonU32 ( reinterpret_cast<uint32_t>( data.staticflow));
#else
	i.WriteHtonU64( reinterpret_cast<uint64_t>( data.staticflow));
#endif
	i.WriteHtonU16(data.cid);
}

uint32_t DSA_RSP_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing DSA_RSP_FRAME_HEADER\n");

	Buffer::Iterator i = start;
	data.type = i.ReadU8();
	data.transaction_id = i.ReadNtohU16();
	data.confirmation_code = i.ReadU8();
	uint8_t temp = i.ReadU8();
	if (temp) data.uplink = true;
	else data.uplink = false;
	
#ifdef ENVIRONMENT32
	uint32_t temp2 = i.ReadNtohU32();
	data.staticflow = reinterpret_cast<Wimax2ServiceFlow*>( temp2);
#else
	uint64_t temp2 = i.ReadNtohU64();
	data.staticflow = reinterpret_cast<Wimax2ServiceFlow*>( temp2);
#endif
	data.cid = i.ReadNtohU16();
	
	return GetSerializedSize();
}


DSA_ACK_FRAME_HEADER::DSA_ACK_FRAME_HEADER() {
	data.type = 13;
	data.transaction_id = 0;
	data.confirmation_code = 0;
	data.uplink = false;
}
DSA_ACK_FRAME_HEADER::~DSA_ACK_FRAME_HEADER() {}
TypeId DSA_ACK_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::DSA_ACK_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<DSA_ACK_FRAME_HEADER>()
	;
	return tid;
}
TypeId DSA_ACK_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void DSA_ACK_FRAME_HEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE DSA_ACK_FRAME_HEADER Header\n";	
}

uint32_t DSA_ACK_FRAME_HEADER::GetSerializedSize() const {
	return 5;
}

void DSA_ACK_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	i.WriteHtonU16(data.transaction_id);
	i.WriteU8(data.confirmation_code);
	uint8_t temp = 0;
	if (data.uplink) temp = 1;
	i.WriteU8(temp);
}

uint32_t DSA_ACK_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing DSA_ACK_FRAME_HEADER\n");

	Buffer::Iterator i = start;
	data.type = i.ReadU8();
	data.transaction_id = i.ReadNtohU16();
	data.confirmation_code = i.ReadNtohU16();
	uint8_t temp = i.ReadU8();
	if (temp) data.uplink = true;
	else data.uplink = false;
	
	return GetSerializedSize();
}



REGISTRATION_RESPONSE_FRAME::REGISTRATION_RESPONSE_FRAME() {
	data.type = 0;
	data.response = 0;
	data.ss_mngmt_support = 0;
	data.sec_mngmt_cid = 0;
}
REGISTRATION_RESPONSE_FRAME::~REGISTRATION_RESPONSE_FRAME() {}
TypeId REGISTRATION_RESPONSE_FRAME::GetTypeId() {
	static TypeId tid = TypeId("ns3::REGISTRATION_RESPONSE_FRAME")
	.SetParent<Header> ()
	.AddConstructor<REGISTRATION_RESPONSE_FRAME>()
	;
	return tid;
}
TypeId REGISTRATION_RESPONSE_FRAME::GetInstanceTypeId() const{
	return GetTypeId();
}

void REGISTRATION_RESPONSE_FRAME::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE REGISTRATION_RESPONSE_FRAME Header\n";	
}

uint32_t REGISTRATION_RESPONSE_FRAME::GetSerializedSize() const {
	return 5;
}

void REGISTRATION_RESPONSE_FRAME::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	i.WriteU8(data.response);
	i.WriteU8(data.ss_mngmt_support);
	i.WriteHtonU16(data.sec_mngmt_cid);
}

uint32_t REGISTRATION_RESPONSE_FRAME::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing REGISTRATION_RESPONCE_FRAME\n");

	Buffer::Iterator i = start;
	data.type = i.ReadU8();
	data.response = i.ReadU8();
	data.ss_mngmt_support = i.ReadU8();
	data.sec_mngmt_cid = i.ReadNtohU16();
	
	return GetSerializedSize();
}




NEIGHBOR_ADVERTISEMENT_FRAME_HEADER::NEIGHBOR_ADVERTISEMENT_FRAME_HEADER() {
	data.type = 53;
	data.skip_opt_field = 0;
	data.operatorID = 0;
	data.ccc = 0;
	data.frag_index = 0;
	data.total_frag = 0;
	data.n_neighbors = 0;
}
NEIGHBOR_ADVERTISEMENT_FRAME_HEADER::~NEIGHBOR_ADVERTISEMENT_FRAME_HEADER() {}
TypeId NEIGHBOR_ADVERTISEMENT_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::NEIGHBOR_ADVERTISEMENT_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<NEIGHBOR_ADVERTISEMENT_FRAME_HEADER>()
	;
	return tid;
}
TypeId NEIGHBOR_ADVERTISEMENT_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void NEIGHBOR_ADVERTISEMENT_FRAME_HEADER::Print(std::ostream &os) const{
	os << "\tPRINTING  NEIGHBOR_ADVERTISEMENT_FRAME_HEADER \n";	
	os << "Type " << (int)data.type << "\nskip_opt_field (1 = ignore next)" << data.skip_opt_field 
	   << "\noperatorID " << data.operatorID << "\nccc " << data.ccc << "\nfrag_index " << data.frag_index
	   << "ntotal_frag " << data.total_frag << "\nn_neighbors " << data.n_neighbors << " and the nbr_info array would be spilled out here\n";
}

uint32_t NEIGHBOR_ADVERTISEMENT_FRAME_HEADER::GetSerializedSize() const {
	int len =  5;
	if ((data.skip_opt_field & 0x1) == 0) len += 3;
	len += data.n_neighbors*15;
	for (int i = 0; i < data.n_neighbors; i++) {
		if (data.nbr_info[i].dcd_included) len += data.nbr_info[i].dcd_settings.GetSerializedSize();
		if (data.nbr_info[i].ucd_included) len += data.nbr_info[i].ucd_settings.GetSerializedSize();
		if (data.nbr_info[i].phy_included) len += 2;
	}
	return len;
}

void NEIGHBOR_ADVERTISEMENT_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	i.WriteU8(data.skip_opt_field);
	
	if ((data.skip_opt_field & 0x1) == 0) {
		uint32_t tmp = 0;
		tmp = tmp | (data.operatorID & 0xFFFFFF);
		tmp = tmp | (data.ccc << 24);
		i.WriteHtonU32(tmp);
	}
	else i.WriteU8(data.ccc);
	
	uint8_t temp = 0;
	temp = (data.frag_index & 15);
	temp = temp | ((data.total_frag & 15) << 4);
	i.WriteU8(temp);
	i.WriteU8(data.n_neighbors);
	
	for (int j = 0; j < data.n_neighbors; j++) {
		i.WriteU8(data.nbr_info[j].length);
		temp = 0;
		temp = temp | data.nbr_info[j].phy_profile_id.colocatedFA;
		temp = temp | (data.nbr_info[j].phy_profile_id.FAconfig << 1);
		temp = temp | (data.nbr_info[j].phy_profile_id.timefreq_synch << 2);
		temp = temp | (data.nbr_info[j].phy_profile_id.bs_eirp << 4);
		temp = temp | (data.nbr_info[j].phy_profile_id.dcducd_ref << 5);
		temp = temp | (data.nbr_info[j].phy_profile_id.FAindex << 6);
		temp = temp | (data.nbr_info[j].phy_profile_id.trigger_ref << 7);
		i.WriteU8(temp);
		
		i.WriteU8(data.nbr_info[j].fa_index);
		i.WriteU8(data.nbr_info[j].bs_eirp);
		
		uint8_t adr[6];
		data.nbr_info[j].nbr_bsid.CopyTo(adr);
		for (int k = 0; k < 6; k++)
			i.WriteU8(adr[k]);

		i.WriteU8(data.nbr_info[j].preamble_index);
		i.WriteU8(data.nbr_info[j].ho_process_opt);
		i.WriteU8(data.nbr_info[j].sched_srv_supported);
		
		temp = 0;
		temp = temp | data.nbr_info[j].dcd_ccc;
		temp = temp | (data.nbr_info[j].ucd_ccc << 4);
		i.WriteU8(temp);
		
		temp = 0;
		if (data.nbr_info[j].dcd_included) temp = 1;
		if (data.nbr_info[j].ucd_included) temp = (temp | 2);
		if (data.nbr_info[j].phy_included) temp = (temp | 4);
		i.WriteU8(temp);
		
		if (data.nbr_info[j].dcd_included) {
			data.nbr_info[j].dcd_settings.Serialize(i);
			i.Next(data.nbr_info[j].dcd_settings.GetSerializedSize());
		}
		if (data.nbr_info[j].ucd_included) {
			data.nbr_info[j].ucd_settings.Serialize(i);
			i.Next(data.nbr_info[j].ucd_settings.GetSerializedSize());
		}
		if (data.nbr_info[j].phy_included) {
			uint16_t temp2 = 0;
			temp2 = data.nbr_info[j].phy_mode_id.bandwidth;
			temp2 = temp2 | (data.nbr_info[j].phy_mode_id.fttsize << 7);
			temp2 = temp2 | (data.nbr_info[j].phy_mode_id.cp << 10);
			temp2 = temp2 | (data.nbr_info[j].phy_mode_id.duration_code << 12);
			i.WriteHtonU16(temp2);
		}
	}
}

uint32_t NEIGHBOR_ADVERTISEMENT_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
		debugHeader("Deserializing NEIGHBOR_ADVERTISEMENT_FRAME_HEADER\n");

	Buffer::Iterator i = start;
	
	data.type = i.ReadU8();
	data.skip_opt_field = i.ReadU8();
	
	if ((data.skip_opt_field & 0x1) == 0) {
		uint32_t temp2 = i.ReadNtohU32();
		data.operatorID = temp2 & 0xFFFFFF;
		data.ccc = (temp2 >> 24) & 0xFF;
	} else {
		data.operatorID = 0;
		data.ccc = i.ReadU8();
	}
	

	
	uint8_t temp = i.ReadU8();
	data.frag_index = temp & 15;
	data.total_frag = (temp >> 4) && 15;
	
	data.n_neighbors = i.ReadU8();
	for (int j = 0; j < data.n_neighbors; j++) {
		data.nbr_info[j].length = i.ReadU8();
		
		
		data.nbr_info[j].fa_index = i.ReadU8();
		data.nbr_info[j].bs_eirp = i.ReadU8();
		
		uint8_t adr[6];
		for (int k = 0; k < 6; k++)
			adr[k] = i.ReadU8();
		data.nbr_info[j].nbr_bsid.CopyFrom(adr);
	
		data.nbr_info[j].preamble_index = i.ReadU8();
		data.nbr_info[j].ho_process_opt = i.ReadU8();
		data.nbr_info[j].sched_srv_supported = i.ReadU8();
		temp = i.ReadU8();
		data.nbr_info[j].dcd_ccc = temp & 15;
		data.nbr_info[j].ucd_ccc = (temp >> 4) & 15;
		temp = i.ReadU8();
		data.nbr_info[j].dcd_included = temp & 1;
		data.nbr_info[j].ucd_included = temp & 2;
		data.nbr_info[j].phy_included = temp & 4;
		
		if (data.nbr_info[j].dcd_included) {
			data.nbr_info[j].dcd_settings.Deserialize(i);
			i.Next(data.nbr_info[j].dcd_settings.GetSerializedSize());
		}
		if (data.nbr_info[j].ucd_included) {
			data.nbr_info[j].ucd_settings.Deserialize(i);
			i.Next(data.nbr_info[j].ucd_settings.GetSerializedSize());
		}
		if (data.nbr_info[j].phy_included) {
			uint16_t temp = i.ReadNtohU16();
			data.nbr_info[j].phy_mode_id.bandwidth = temp & 127;
			data.nbr_info[j].phy_mode_id.fttsize = (temp >> 7) & 7;
			data.nbr_info[j].phy_mode_id.cp = (temp >> 10) & 3;
			data.nbr_info[j].phy_mode_id.duration_code = (temp >> 12) & 15;
		}
	}
	return GetSerializedSize();
}




MSHO_REQUEST_FRAME_HEADER::MSHO_REQUEST_FRAME_HEADER() {
	data.type = 57;
	data.report_metric = 0;
	data.n_new_bs_index = 0;
	data.ccc_mob_nbr_adv = 0;
	data.n_new_bs_full = 0;
	data.n_current_bs = 0;
}
MSHO_REQUEST_FRAME_HEADER::~MSHO_REQUEST_FRAME_HEADER() {}
TypeId MSHO_REQUEST_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::MSHO_REQUEST_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<MSHO_REQUEST_FRAME_HEADER>()
	;
	return tid;
}
TypeId MSHO_REQUEST_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void MSHO_REQUEST_FRAME_HEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE MSHO_REQUEST_FRAME_HEADER Header\n";	
}

uint32_t MSHO_REQUEST_FRAME_HEADER::GetSerializedSize() const {
	int len = 6;
	len += 6*data.n_new_bs_index;
	len += 6*data.n_new_bs_full;
	len += 5*data.n_current_bs;
	return len;
}

void MSHO_REQUEST_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	i.WriteU8(data.report_metric);
	i.WriteU8(data.n_new_bs_index);
	i.WriteU8(data.ccc_mob_nbr_adv);
	for (int j = 0; j < data.n_new_bs_index; j++) {
		i.WriteU8(data.bs_index[j].neighbor_bs_index);
		i.WriteU8(data.bs_index[j].preamble_index);
		i.WriteU8(data.bs_index[j].bs_cinr_mean);
		i.WriteU8(data.bs_index[j].bs_rssi_mean);
		i.WriteU8(data.bs_index[j].relative_delay);
		uint8_t temp = 0;
		temp = temp & (data.bs_index[j].service_level_prediction & 7);
		temp = temp & ( (data.bs_index[j].arrival_time_diff_ind & 1) << 3);
		temp = temp & ( (data.bs_index[j].arrival_time_diff & 15) << 4);
		i.WriteU8(temp);
	}
	i.WriteU8(data.n_new_bs_full);
	for (int j = 0; j < data.n_new_bs_full; j++) {
		i.WriteU8(data.bs_full[j].neighbor_bs_index);
		i.WriteU8(data.bs_full[j].preamble_index);
		i.WriteU8(data.bs_full[j].bs_cinr_mean);
		i.WriteU8(data.bs_full[j].bs_rssi_mean);
		i.WriteU8(data.bs_full[j].relative_delay);
		uint8_t temp = 0;
		temp = temp & (data.bs_full[j].service_level_prediction & 7);
		temp = temp & ( (data.bs_full[j].arrival_time_diff_ind & 1) << 3);
		temp = temp & ( (data.bs_full[j].arrival_time_diff & 15) << 4);
		i.WriteU8(temp);
	}
	i.WriteU8(data.n_current_bs);
	for (int j = 0; j < data.n_current_bs; j++) {
		uint8_t tmp = data.bs_current[j].temp_bsid;
		i.WriteU8(tmp);
		i.WriteU8(data.bs_current[j].bs_cinr_mean);
		i.WriteU8(data.bs_current[j].bs_rssi_mean);
		i.WriteU8(data.bs_current[j].relative_delay);
		i.WriteU8(data.bs_current[j].bs_rtd);
	}
}

uint32_t MSHO_REQUEST_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing MSHO_REQUEST_FRAME_HEADER\n");

	Buffer::Iterator i = start;
	data.type = i.ReadU8();
	data.report_metric = i.ReadU8();
	data.n_new_bs_index = i.ReadU8();
	data.ccc_mob_nbr_adv = i.ReadU8();
	for (int j = 0; j < data.n_new_bs_index; j++) {
		data.bs_index[j].neighbor_bs_index = i.ReadU8();
		data.bs_index[j].preamble_index = i.ReadU8();
		data.bs_index[j].bs_cinr_mean = i.ReadU8();
		data.bs_index[j].bs_rssi_mean = i.ReadU8();
		data.bs_index[j].relative_delay = i.ReadU8();
		uint8_t temp = i.ReadU8();
		data.bs_index[j].service_level_prediction = temp & 7;
		data.bs_index[j].arrival_time_diff_ind = (temp >> 3) & 1;
		data.bs_index[j].arrival_time_diff = (temp >> 4) & 15;
	}
	data.n_new_bs_full = i.ReadU8();
	for (int j = 0; j < data.n_new_bs_full; j++) {
		data.bs_full[j].neighbor_bs_index = i.ReadU8();
		data.bs_full[j].preamble_index = i.ReadU8();
		data.bs_full[j].bs_cinr_mean = i.ReadU8();
		data.bs_full[j].bs_rssi_mean = i.ReadU8();
		data.bs_full[j].relative_delay = i.ReadU8();
		uint8_t temp = i.ReadU8();
		data.bs_full[j].service_level_prediction = temp & 7;
		data.bs_full[j].arrival_time_diff_ind = (temp >> 3) & 1;
		data.bs_full[j].arrival_time_diff = (temp >> 4) & 15;
	}
	data.n_current_bs = i.ReadU8();
	for (int j = 0; j < data.n_current_bs; j++) {
		data.bs_current[j].temp_bsid = i.ReadU8() & 15;
		data.bs_current[j].bs_cinr_mean = i.ReadU8();
		data.bs_current[j].bs_rssi_mean = i.ReadU8();
		data.bs_current[j].relative_delay = i.ReadU8();
		data.bs_current[j].bs_rtd = i.ReadU8();
	}
	return GetSerializedSize();
}


// NOT IMPLEMENTED!
BSHO_RESPONCE_FRAME_HEADER::BSHO_RESPONCE_FRAME_HEADER() {
	data.type = 58;
	data.mode = 0;
	data.ho_operation_mode = 0;
	data.n_recommended = 0;
	data.resource_retain_flag = 0;
	data.tmp_bsid = 0;
	data.ak_change_indicator = 0;
	data.n_cids = 0;
	data.n_saids = 0;
	data.n_new_bs = 0;
	data.n_current_bs = 0;
	data.action_time = 0;
}
BSHO_RESPONCE_FRAME_HEADER::~BSHO_RESPONCE_FRAME_HEADER() {}
TypeId BSHO_RESPONCE_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::BSHO_RESPONCE_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<BSHO_RESPONCE_FRAME_HEADER>()
	;
	return tid;
}
TypeId BSHO_RESPONCE_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void BSHO_RESPONCE_FRAME_HEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE BSHO_RESPONCE_FRAME_HEADER Header\n";	
}

uint32_t BSHO_RESPONCE_FRAME_HEADER::GetSerializedSize() const {
	
	
	return 4;
}

void BSHO_RESPONCE_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	
	debugHeader("NOT IMPLEMENTED!\n");
	
}

uint32_t BSHO_RESPONCE_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing BSHO+RESPONCE_FRAME_HEADER\n");
	debugHeader("NOT IMPLEMENTED!\n");

	Buffer::Iterator i = start;
	
	return GetSerializedSize();
}



REGISTRATION_REQUEST_FRAME::REGISTRATION_REQUEST_FRAME() {
	data.type = 0;
	data.ss_mngmt_support = 0;
	data.ip_mngmt_support = 0;
	data.uplink_cid_support = 0;
}
REGISTRATION_REQUEST_FRAME::~REGISTRATION_REQUEST_FRAME() {}
TypeId REGISTRATION_REQUEST_FRAME::GetTypeId() {
	static TypeId tid = TypeId("ns3::REGISTRATION_REQUEST_FRAME")
	.SetParent<Header> ()
	.AddConstructor<REGISTRATION_REQUEST_FRAME>()
	;
	return tid;
}
TypeId REGISTRATION_REQUEST_FRAME::GetInstanceTypeId() const{
	return GetTypeId();
}

void REGISTRATION_REQUEST_FRAME::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE REGISTRATION_REQUEST_FRAME Header\n";	
}

uint32_t REGISTRATION_REQUEST_FRAME::GetSerializedSize() const {
	return 5;
}

void REGISTRATION_REQUEST_FRAME::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	i.WriteU8(data.ss_mngmt_support);
	i.WriteU8(data.ip_mngmt_support);
	i.WriteHtonU16(data.uplink_cid_support);
}

uint32_t REGISTRATION_REQUEST_FRAME::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing REGISTRATION_REQUEST_FRAME\n");

	Buffer::Iterator i = start;
	data.type = i.ReadU8();
	data.ss_mngmt_support = i.ReadU8();
	data.ip_mngmt_support = i.ReadU8();
	data.uplink_cid_support = i.ReadNtohU16();
	
	return GetSerializedSize();
}


MOB_HO_IND_FRAME_HEADER::MOB_HO_IND_FRAME_HEADER() {
	data.type = 59;
	data.mode = 0;
	data.ho_ind_type = 0;
	data.rng_param_valid_ind = 0;
	data.target_bsid = Mac48Address::GetBroadcast();
	data.mdhofbss_ind_type = 0;
	data.bsid = 0;
	data.action_time = 0;
	data.diversity_set_included = 0;
	data.anchor_bsid = 0;
	data.n_bs = 0;
	data.preamble_index = 0;
}
MOB_HO_IND_FRAME_HEADER::~MOB_HO_IND_FRAME_HEADER() {}
TypeId MOB_HO_IND_FRAME_HEADER::GetTypeId() {
	static TypeId tid = TypeId("ns3::MOB_HO_IND_FRAME_HEADER")
	.SetParent<Header> ()
	.AddConstructor<MOB_HO_IND_FRAME_HEADER>()
	;
	return tid;
}
TypeId MOB_HO_IND_FRAME_HEADER::GetInstanceTypeId() const{
	return GetTypeId();
}

void MOB_HO_IND_FRAME_HEADER::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE MOB_HO_IND_FRAME_HEADER Header\n";	
}

uint32_t MOB_HO_IND_FRAME_HEADER::GetSerializedSize() const {
	if (data.mode == 1) return 10;
	if (data.mode == 2) {
		int temp = 3 + 3*data.n_bs;
		if (temp%8 == 0) temp /= 8;
		else {
			temp /= 8;
			temp++;
		}
		return 11 + temp;
	}
	return 8;
}

void MOB_HO_IND_FRAME_HEADER::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(data.type);
	uint8_t temp = data.mode;
	temp = temp | (data.ho_ind_type << 2);
	temp = temp | (data.rng_param_valid_ind << 2);
	i.WriteU8(temp);
	
	uint8_t adr[6];
	data.target_bsid.CopyTo(adr);
	for (int j = 0; j < 6; j++)
		i.WriteU8(adr[j]);
	
	if (data.mode == 1) {
		temp = 0;
		temp = temp | data.mdhofbss_ind_type;
		temp = temp | (data.bsid << 2);
		i.WriteU8(temp);
		i.WriteU8(data.action_time);
	}
	
	if (data.mode == 2) {
		i.WriteU8(data.diversity_set_included);
		i.WriteU8(data.n_bs);
		int temp2 = 3 + 3 * data.n_bs;
		if (temp2 % 8 == 0) temp2 /= 8;
		else {temp2 /= 8; temp2 ++;}
		
		temp = data.anchor_bsid;
		uint8_t temp_filled = 3;
		for (int j = 1; j < temp2; j++) {
			bool added = false;
			if (temp_filled <= 5) {
				uint8_t temp3 = data.bs[j].temp_bsid;
				temp = temp | (temp3 << temp_filled);
				temp_filled += 3;
				added = true;
				if (temp_filled == 8) {
					i.WriteU8(temp);
					temp_filled = 0;
					added = true;
				}
			}
			else if (!added && temp_filled == 6) {
				uint8_t temp3 = data.bs[j].temp_bsid;
				temp = temp | (3 & (temp3 << temp_filled));
				i.WriteU8(temp);
				temp = 0;
				temp = temp | (temp3 >> 2);
				temp_filled = 1;
				added = true;
			}
			else if (!added && temp_filled == 7) {
				uint8_t temp3 = data.bs[j].temp_bsid;
				temp = temp | (1 & (temp3 << temp_filled));
				i.WriteU8(temp);
				temp = 0;
				temp = temp | (temp3 >> 1);
				temp_filled = 2;
				added = true;
			}

		}
		i.WriteU8(data.preamble_index);
	}
}

uint32_t MOB_HO_IND_FRAME_HEADER::Deserialize(Buffer::Iterator start) {
	debugHeader("Deserializing MOB_HO_IND_FRAME_HEADER\n");
	Buffer::Iterator i = start;
	data.type = i.ReadU8();
	uint8_t temp = i.ReadU8();
	data.mode = temp & 3;
	data.ho_ind_type = temp & 12;
	data.rng_param_valid_ind = temp & 48;
	
	uint8_t adr[6];
	for (int k = 0; k < 6; k++)
		adr[k] = i.ReadU8();
	data.target_bsid.CopyFrom(adr);
	
	if (data.mode == 1) {
		temp = i.ReadU8();
		data.mdhofbss_ind_type = temp & 3;
		data.bsid = (temp & 28) >> 3;
		data.action_time = i.ReadU8();
	}
	if (data.mode == 2) {
		data.diversity_set_included = i.ReadU8();
		data.n_bs = i.ReadU8();
		temp = i.ReadU8();
		data.anchor_bsid = temp & 7;
		uint8_t temp_processed = 3;
		for (int j = 0; j < data.n_bs; j++) {
			if (temp_processed <= 5) {
				data.bs[j].temp_bsid = (temp & (7 << temp_processed))>>temp_processed;
				temp_processed += 3;
				if (temp_processed == 8) temp = i.ReadU8();
			}
			else if (temp_processed == 6) {
				data.bs[j].temp_bsid = temp >> 6;
				temp = i.ReadU8();
				data.bs[j].temp_bsid = (data.bs[j].temp_bsid & 1) << 2;
				temp_processed = 1;
			}
			else if (temp_processed == 7) {
				data.bs[j].temp_bsid = temp >> 7;
				temp = i.ReadU8();
				data.bs[j].temp_bsid = (data.bs[j].temp_bsid & 2) << 1;
				temp_processed = 2;
			}
		}
		data.preamble_index = i.ReadU8();
	}
	return GetSerializedSize();
}


Type_Check::Type_Check() {
	type = 0;
}
Type_Check::~Type_Check() {}
TypeId Type_Check::GetTypeId() {
	static TypeId tid = TypeId("ns3::Type_Check")
	.SetParent<Header> ()
	.AddConstructor<DL_MAP_FRAME_HEADER>()
	;
	return tid;
}
TypeId Type_Check::GetInstanceTypeId() const{
	return GetTypeId();
}

void Type_Check::Print(std::ostream &os) const{
	os << "Frame Type: " << type << "\n";
}

uint32_t Type_Check::GetSerializedSize() const {
	return 1;
}

void Type_Check::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	i.WriteU8(type);
}

uint32_t Type_Check::Deserialize(Buffer::Iterator start) {
	Buffer::Iterator i = start;
	type = i.ReadU8();
	return GetSerializedSize();
}

MAC_TAG::MAC_TAG() {
	isMAC = true;
}
MAC_TAG::MAC_TAG(bool m) {
	isMAC = m;
}
MAC_TAG::~MAC_TAG(){}
void MAC_TAG::Print(std::ostream &os) const{
	if(isMAC) os<<"This is a mac Packet\n";
	else os << "This is not a mac Packet\n";
}

uint32_t MAC_TAG::GetSerializedSize() const {
	return 1;
}

void MAC_TAG::Serialize(TagBuffer i) const{
	if (isMAC) i.WriteU8(1);
	else i.WriteU8(0);
}
void MAC_TAG::Deserialize(TagBuffer i) {
	uint8_t uc = i.ReadU8();
	if (uc == 0) isMAC = false;
	else isMAC = true;
}

TypeId MAC_TAG::GetInstanceTypeId() const{
	static TypeId tid = TypeId("ns3::MAC_TAG")
	.SetParent<Header> ()
	.AddConstructor<MAC_TAG>()
	;
	return tid;
}

UL_Protocol_Header::UL_Protocol_Header() {
	protocol = 0;
}

UL_Protocol_Header::UL_Protocol_Header(uint16_t ptcl) {
	protocol = ptcl;
}

UL_Protocol_Header::~UL_Protocol_Header() {}

TypeId UL_Protocol_Header::GetTypeId() {
	static TypeId tid = TypeId("ns3::UL_Protocol_Header")
	.SetParent<Header> ()
	.AddConstructor<UL_Protocol_Header>()
	;
	return tid;
}

TypeId UL_Protocol_Header::GetInstanceTypeId() const {
	return GetTypeId();
}

void UL_Protocol_Header::Print (std::ostream &os) const { 
	os << "UL_Protocol_Header:\n\t Protocol: " << protocol << "\n";
}

uint32_t UL_Protocol_Header::GetSerializedSize() const {
	return 2;
}

void UL_Protocol_Header::Serialize(Buffer::Iterator start) const {
	Buffer::Iterator i = start;
	i.WriteHtonU16(protocol);
}

uint32_t UL_Protocol_Header::Deserialize(Buffer::Iterator start) {
	Buffer::Iterator i = start;
	protocol = i.ReadNtohU16();
	return GetSerializedSize();
}

Frame_Prefix_Header::Frame_Prefix_Header() {
	used_subchannel_bmp = 0;
	ranging_change = 0;
	repetition_coding = 0;
	coding_indication = 0;
	DL_MAP_length = 0;
	reserved = 0;
}
Frame_Prefix_Header::~Frame_Prefix_Header() {}

void Frame_Prefix_Header::Print (std::ostream &os) const {
	os << "Frame Prefix Header\n";
	os << "Used Subchannel Bitmap: " << used_subchannel_bmp << std::endl;
	os << "Ranging Change: "; if (ranging_change) os << "TRUE\n"; else os << "FALSE\n";
	os << "Repetition COding: " << repetition_coding << std::endl;
	os << "Coding Indication: " <<coding_indication << std::endl;
	os << "DL_MAP Length: " << DL_MAP_length << std::endl;
	os << "resrved (should be 0): " << reserved << std::endl;
}

TypeId Frame_Prefix_Header::GetTypeId() {
	static TypeId tid = TypeId("ns3::Frame_Prefix_Header")
	.SetParent<Header> ()
	.AddConstructor<Frame_Prefix_Header>()
	;
	return tid;
}

TypeId Frame_Prefix_Header::GetInstanceTypeId() const {
	return GetTypeId();
}

uint32_t Frame_Prefix_Header::GetSerializedSize() const {
	return 6;
}

void Frame_Prefix_Header::Serialize(Buffer::Iterator start) const {
	Buffer::Iterator i = start;
	uint32_t temp = 0;
	temp = temp | used_subchannel_bmp;
	temp = temp | (ranging_change << 6);
	temp = temp | (repetition_coding << 7);
	temp = temp | (coding_indication << 9);
	temp = temp | (DL_MAP_length << 12);
	temp = temp | (reserved << 20);
	
	uint16_t a = (temp << 16) >> 16;
	uint8_t b = temp >> 16;
	i.WriteHtonU16(a);
	i.WriteU8(b);
	//since this frame is alwyas repeated, we just serialize it twice here
	i.WriteHtonU16(a);
	i.WriteU8(b);
}

uint32_t Frame_Prefix_Header::Deserialize(Buffer::Iterator start) {
	Buffer::Iterator i = start;
	uint16_t a = i.ReadNtohU16();
	uint8_t b = i.ReadU8();
	uint16_t a2 = i.ReadNtohU16();
	uint8_t b2 = i.ReadU8();

	if (a != a2 || b != b2) {
		std::cerr << "there has been a transmission error on the FCH.  This is not a case that's currently well handled by the simulator, expect potential odd behavior from one (or more) of the SS during this frame\n";
	}

	uint32_t temp = 0;
	temp = temp | a;
	temp = temp | (b << 16);

	uint32_t mask = 63;
	used_subchannel_bmp = temp & mask;
	ranging_change = ((1 << 6) & temp ) >> 6;
	repetition_coding = ((3 << 7) & temp ) >> 7;
	coding_indication = ((7 << 9) & temp ) >> 9;
	DL_MAP_length = ((255 << 12) & temp ) >> 12;
	reserved = ((15 << 20) & temp) >> 20;

	return GetSerializedSize();
}


Copy_Trailer::Copy_Trailer() {
	length = 1;
	d = new uint8_t[1];
}

Copy_Trailer::Copy_Trailer(uint16_t len) {
	length = len;
	d = new uint8_t[len];
}

Copy_Trailer::~Copy_Trailer() {
	delete [] d;
}

TypeId Copy_Trailer::GetTypeId() {
	static TypeId tid = TypeId("ns3::Copy_Trailer")
	.SetParent<Header> ()
	.AddConstructor<Copy_Trailer>()
	;
	return tid;
}

TypeId Copy_Trailer::GetInstanceTypeId() const {
	return GetTypeId();
}

void Copy_Trailer::Print (std::ostream &os) const { 
	os << "This is a workaround for a bug in ns3.10's Packet::AddAtEnd function.  If you are using ns3.11 or later, it's probably been fixed and this can be removed.\n";
}

uint32_t Copy_Trailer::GetSerializedSize() const {
	return length;
}

void Copy_Trailer::Serialize(Buffer::Iterator start) const {
	Buffer::Iterator i = start;
	for (uint16_t j = 0; j < length; j++) {
		i.WriteU8(d[j]);
	}
}

uint32_t Copy_Trailer::Deserialize(Buffer::Iterator start) {
	Buffer::Iterator i = start;
	for (uint16_t j = 0; j < length; j++) {
		d[j] = i.ReadU8();
	}
	return GetSerializedSize();
}






} // namespace ns3
