/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 File * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */


#include "framemap.h"
#include "ns3/contentionslot.h"
#include "ns3/wimax2-scheduler.h"
#include "ns3/ulburst.h"
#include "ns3/dlburst.h"
#include "ns3/wimax2-net-device.h"

namespace ns3 {

/*
 * Creates a map of the frame
 * @param mac Pointer to the mac layer
 */
FrameMap::FrameMap (Ptr<Wimax2NetDevice> mac): dlsubframe_(this), ulsubframe_(this)
{
  assert (mac);
  mac_ = mac; 

  //retreive information from mac
  duration_ = mac_->getFrameDuration();
  rtg_ = mac_->phymib_.rtg;
  ttg_ = mac_->phymib_.ttg;
}

/**
 * Compute the DL_MAP packet based on the information contained in the structure
 */
Ptr<Packet> FrameMap::getDL_MAP( )
{
  Ptr<Packet> p = mac_->getPacket();
  All_Headers ch;
  p->RemoveHeader(ch);
  //hdr_cmn* ch = HDR_CMN(p);
  
  //we use 1 less than the number of bursts becuase the DL_MAP itself doesn't get an entry
  int nbies = dlsubframe_.getPdu()->getNbBurst() - 1; 

  //allocate data for DL_MAP
  //p->allocdata (sizeof (struct mac802_16_dl_map_frame));
  DL_MAP_FRAME_HEADER frame;
  //mac802_16_dl_map_frame *frame = (mac802_16_dl_map_frame*) p->accessdata();

  frame.mf.type = MAC_DL_MAP;
  frame.mf.bsid = mac_->addr();
  frame.mf.nb_ies = nbies;
  
    debug2("nbies=%d\n",nbies);

  
  //allocate IEs
  DL_MAP_IE *ies = frame.mf.ies;
  //mac802_16_dlmap_ie *ies = frame.mf.ies;

  int bc_ie = 0;
  int skip_count = 0; //used to compensate for a bug involving bursts not getting properly removed after transmission sometimes
 		      //and them to show up again with width and/or height of 0 in the next frame
  for (int i = 0 ; i < nbies ; i++) {
    Burst *b = dlsubframe_.getPdu()->getBurst(i+1); //i+1 here to skip the DL_MAP's entry
    if (b->getnumSubchannels() == 0 || b->getDuration() == 0) {	
    	skip_count++;
    	continue;
    }
    ies[i - skip_count].ie.diuc = b->getIUC();
//rpi added for including subchannels for OFDMA
    ies[i - skip_count].ie.subchannel_offset=b->getSubchannelOffset ();
    
    ies[i - skip_count].ie.num_subchannel = b->getnumSubchannels();
    
    ies[i - skip_count].ie.num_OFDMASymbol=b->getDuration();

    ies[i - skip_count].ie.OFDMASymbol_offset=b->getStarttime();

//rpi end    
    debug2("DL_MAP_IE index :%d, cid :%d, bc_ie :%d, diuc :%d, starttime :%d, #subchannels :%d, #symbols :%d, subchannel offset: %d\n", i, b->getCid(), bc_ie, ies[i - skip_count].ie.diuc, ies[i - skip_count].ie.OFDMASymbol_offset, ies[i - skip_count].ie.num_subchannel, ies[i - skip_count].ie.num_OFDMASymbol, ies[i - skip_count].ie.subchannel_offset);
    if (b->getIUC()!=DIUC_END_OF_MAP) {
      ies[i - skip_count].ie.cid =  b->getCid();
    }
  }

  frame.mf.nb_ies -= skip_count;
  
//#ie -> we exclude dl-map but include ul-map
//In next version, the way to calcuate dl-map will be investigated.
  int dl_ie = nbies - bc_ie - 1;
  	if (dl_ie); //quiet, compiler
  ch.GMH.GMH.len = frame.GetSerializedSize();
  ch.GMH.GMH.ht = 0;
  ch.GMH.GMH.cid = BROADCAST_CID;
  ch.gm_sm = 3;
  ch.cdma = 0;
  debug2("Frame_map.DL_MAP before size => chsize :%d, ie (include ulmap) :%d, after ie_size+fix :%d\n", ch.GMH.GMH.len, nbies-bc_ie-1, int(ceil(GET_DL_MAP_SIZE(dl_ie))));
//  ch->size() = ch->size() + int (GET_DL_MAP_SIZE(dl_ie));
	
  //ch.GMH.GMH.len += ceil (GET_DL_MAP_SIZE(dl_ie));
  //ch.GMH.GMH.ht = 0;  
  p->AddHeader(frame);
  p->AddHeader(ch);

  return p;
}

/**
 * Compute and return the DCD frame
 */
Ptr<Packet> FrameMap::getDCD( )
{
  debug10("Getting the DCD Frame\n");
  Ptr<Packet> p = mac_->getPacket ();
  All_Headers ah;
  p->RemoveHeader(ah);

  //in the future when we support FCH we may not need GMH on DCD/UCD
  //at which point we should set ah.GMH.GMH.len to 0 here and modify
  //Wimax2BSNetDevice::transmit appropriately

  //allocate data for DL_MAP
  //printf ("getDCD...nbprofile=%d\n", dlsubframe_.getNbProfile());
  DCD_FRAME_HEADER frame;

  uint8_t adr_pull[6];
  mac_->addr().CopyTo(adr_pull);


  frame.mf.type = MAC_DCD;
  frame.mf.dcid = adr_pull[6]; //to check if needs to be different from ucid
  frame.mf.config_change_count = dlsubframe_.getCCC(); 
  frame.mf.frame_duration_code = mac_->getFrameDurationCode ();
  frame.mf.frame_number = mac_->frame_number_;
  frame.mf.nb_prof = dlsubframe_.getNbProfile();
  frame.mf.ttg = mac_->phymib_.ttg;
  frame.mf.rtg = mac_->phymib_.rtg;
  frame.mf.frequency = (int) (mac_->getPhy()->getFreq()/1000); 
  


  //allocate IEs
  mac802_16_dcd_profile *profiles = frame.mf.profiles;


  int i=0;
  for (Profile *p = dlsubframe_.getFirstProfile() ; p ; p=p->next_entry()) {
    //set data for first burst
    profiles[i].diuc = p->getIUC(); 
    profiles[i].frequency = p->getFrequency();
    profiles[i].fec = p->getEncoding(); 
    i++;
  }
  //the end of map is already included in the frame length
  ah.GMH.GMH.len += frame.GetSerializedSize();
  
  p->AddHeader(frame);
  p->AddHeader(ah);
  return p;
}

Ptr<Wimax2NetDevice> FrameMap::getMac() { return mac_; }

/**
 * Compute and return the UL_MAP frame
 */
Ptr<Packet> FrameMap::getUL_MAP( )
{
	//we need to pull out the 

  Ptr<Packet> p = mac_->getPacket ();
  All_Headers ch;
  p->RemoveHeader(ch);

  int nbies = ulsubframe_.getNbPdu(); //there is one UL phy PDU per burst
  UL_MAP_FRAME_HEADER frame;
   
  //here we store the 48 bit MAC address in a byte array
  uint8_t adr_pull[6];
  mac_->addr().CopyTo(adr_pull);

  //set up the overall frame information
  frame.mf.type = MAC_UL_MAP;
  frame.mf.ucid = adr_pull[6];  //set channel ID to index_ to be unique.  since we're using sequential MAC addresses, the lower 6 bits of them works just as well
  frame.mf.ucd_count = ulsubframe_.getCCC(); 
  frame.mf.allocation_start = ulsubframe_.getStarttime(); //the subframe starts with the contention slot
  frame.mf.nb_ies = nbies; 
  
  //allocate IEs
  UL_MAP_IE *ies = frame.mf.ies;

  int bc_ie = 0;
  int sum_ie_size = 0;
  int i_normal_ie = 0;
  int i_cdma_ie = 0;
  int i=0;
  
  //in this loop we pull a burst out of each pdu and set up the IE for it.  This assumes one burst per pdu
//  for (PhyPdu *pdu = ulsubframe_.getFirstPdu(); pdu ; pdu = pdu ->next_entry()) {
  for (std::list<PhyPdu*>::iterator pdu = ulsubframe_.getPduList()->begin(); *pdu != NULL && pdu != ulsubframe_.getPduList()->end(); pdu++) {
    UlBurst *b = (UlBurst*) (*pdu)->getBurst(0);
    
    ies[i].ie.uiuc = b->getIUC();
    if (ies[i].ie.uiuc == UIUC_INITIAL_RANGING || ies[i].ie.uiuc == UIUC_REQ_REGION_FULL) 
	    ies[i].ie.duration = b->getnumSubchannels();
    else ies[i].ie.duration = b->getDuration() / 3; //because this is slots, not raw symbols
    ies[i].ie.cid = b->getCid();
    ies[i].ie.repetition_coding = 0;
    
    ies[i].ie.cdma_ie.ie.code=b->getB_CDMA_CODE();
    ies[i].ie.cdma_ie.ie.subchannel=b->getB_CDMA_TOP();

    

    if (b->getCid() == -1) bc_ie++;

    debug10("Frame_map.UL_MAP_IE index :%d, cid :%d, bc_ie :%d, uiuc :%d, duration in slots: %d\n", i, (uint16_t)b->getCid(), bc_ie, ies[i].ie.uiuc, ies[i].ie.duration);
//rpi end 

    if (b->getCid() >= 0) {
       if ( (ies[i].ie.cdma_ie.ie.code==0) && (ies[i].ie.cdma_ie.ie.subchannel==0) ) {
		i_normal_ie++;
       } else {
		i_cdma_ie++;
		ies[i].ie.cdma_ie.ie.uiuc = b->getExtendedUIUC();
		ies[i].ie.uiuc = UIUC_CDMA_ALLOCATION_IE;
       }
       sum_ie_size += ies[i].GetSerializedSize();
    }

    i++;
  }

	

  ch.GMH.GMH.ht = 0;  
  ch.GMH.GMH.len += frame.GetSerializedSize();

  debug2("Frame_map.UL_MAP_IE index :%d, #normal_ie :%d, #cdma_ie :%d, ie_size :%d, ul_map_size :%d\n", i, i_normal_ie, i_cdma_ie, sum_ie_size, int(GET_UL_MAP_SIZE(0))+sum_ie_size);
  //ch.GMH.GMH.len = int(GET_UL_MAP_SIZE(0))+sum_ie_size;
//  ch->size() += GET_UL_MAP_SIZE(nbies);

  p->AddHeader(frame);
  p->AddHeader(ch);
  return p;
}

/**
 * Compute and return the UCD frame
 */
Ptr<Packet> FrameMap::getUCD( )
{
  Ptr<Packet> p = mac_->getPacket ();
  All_Headers ah;
  p->RemoveHeader(ah);

  UCD_FRAME_DATA frame;

  frame.ucdf.type = MAC_UCD;
  frame.ucdf.config_change_count = 0; //changed by scheduler
  frame.ucdf.rng_backoff_start = ulsubframe_.getRanging()->getBackoff_start();
  frame.ucdf.rng_backoff_end = ulsubframe_.getRanging()->getBackoff_stop();
  frame.ucdf.rng_req_size = ulsubframe_.getRanging()->getSize();
  frame.ucdf.req_backoff_start = ulsubframe_.getBw_req()->getBackoff_start();
  frame.ucdf.req_backoff_end = ulsubframe_.getBw_req()->getBackoff_stop()+1;
  frame.ucdf.bw_req_size = ulsubframe_.getBw_req()->getSize();

  frame.ucdf.nb_prof = ulsubframe_.getNbProfile();
  //allocate IEs
  mac802_16_ucd_profile *profiles = frame.ucdf.profiles;

  int i=0;
  for (Profile *p = ulsubframe_.getFirstProfile() ; p ; p=p->next_entry()) {
    //set data for first burst
    profiles[i].uiuc = p->getIUC(); 
    profiles[i].fec = p->getEncoding(); 
    i++;
  }
  
  ah.GMH.GMH.len += frame.GetSerializedSize();
  
  

  //the end of map is already included in the frame length
  //printf ("Frame_map.UCD before size => chsize :%d, after size :%d\n",ch.GMH.GMH.len, int(GET_UCD_SIZE(ulsubframe_.getNbProfile())));
  debug10("Frame_map.UCD size = %d\n", ah.GMH.GMH.len);
//  ch->size() += int(GET_UCD_SIZE(ulsubframe_.getNbProfile()));
  p->AddHeader(frame);
  p->AddHeader(ah);
  return p;
}

/**
 * Parse a DL_MAP message and create the data structure
 * @param frame The DL frame information
 */
void FrameMap::parseDLMAPframe (DL_MAP_FRAME_HEADER *frame)
{
  //printf ("parse DL-MAP in %d\n", mac_->addr());
  // Clear previous information  
  while (dlsubframe_.getPdu()->getNbBurst()>0) {
    Burst *b = dlsubframe_.getPdu()->getBurst (0);
    dlsubframe_.getPdu()->removeBurst (b);
    delete b;
  }

  int nbies = frame->mf.nb_ies;
  DL_MAP_IE *ies = frame->mf.ies;

  for (int i = 0 ; i < nbies ; i++) {
    Burst *b = dlsubframe_.getPdu()->addBurst(i);
    b->setIUC(ies[i].ie.diuc);
    b->setStarttime(ies[i].ie.OFDMASymbol_offset);

    //rpi added for including subchannels for OFDMA
    b->setnumSubchannels(ies[i].ie.num_subchannel);//=ies[i].num_of_subchannels;ies[i].subchannel_offset=;
    b->setSubchannelOffset (ies[i].ie.subchannel_offset);//ies[i].num_of_subchannels=b->getSubchannelOffset (); 
    b->setDuration(ies[i].ie.num_OFDMASymbol);//ies[i].num_of_symbols=b->getDuration();
//rpi end 

  }
}

/**
 * Parse a DCD message and create the data structure
 * @param frame The DL frame information
 */
void FrameMap::parseDCDframe (mac802_16_dcd_frame *frame)
{
  //clear previous profiles
  dlsubframe_.removeProfiles();

  int nb_prof = frame->nb_prof;
  mac_->frame_number_ = frame->frame_number;
  mac802_16_dcd_profile *profiles = frame->profiles;
  mac_->setFrameDurationCode (frame->frame_duration_code);

  for (int i = 0 ; i < nb_prof ; i++) {
    Profile *p = dlsubframe_.addProfile (profiles[i].frequency, (Ofdm_mod_rate)profiles[i].fec);
    p->setIUC (profiles[i].diuc);
    //printf ("\t Adding dl profile %i: f=%d, rate=%d, iuc=%d\n", i, p->getFrequency(), p->getEncoding(), p->getIUC());
  }
}

/**
 * Parse a UL_MAP message and create the data structure
 * @param frame The UL frame information
 */
void FrameMap::parseULMAPframe (UL_MAP_FRAME_HEADER *frame, double txtime) {
  // Clear previous information

  for (std::list<PhyPdu*>::iterator p = ulsubframe_.getPduList()->begin(); p != ulsubframe_.getPduList()->end(); p = ulsubframe_.getPduList()->begin()) {
	delete *p;
	ulsubframe_.getPduList()->erase(p);
  }

  
  int nbies = frame->mf.nb_ies;
  UL_MAP_IE *ies = frame->mf.ies;
  
  debug10("At %f, Parsing UL_MAP Frame in SS %s\n", NOW, mac_->addr_s());
  Time ul_start =  mac_->start_of_frame + Seconds(frame->mf.allocation_start * mac_->getPhy()->getPS());
  debug10("Frame start time (absolute): %f\n", mac_->start_of_frame.GetSeconds());
  debug10("UL Subframe start time offset from frame start = %f, absolute = %f\n",frame->mf.allocation_start * mac_->getPhy()->getPS(), ul_start.GetSeconds());
  ulsubframe_.setStarttime(frame->mf.allocation_start);
//  setStarttime(ul_start.GetSeconds());
  

//  debug10("Total frame start time %f\n",getStarttime());
//  debug10("UL Subframe delta: %f\n", frame->mf.allocation_start * mac_->getPhy()->getPS());
//  debug10("Correction Factor (s): %f\n", NOW-getStarttime());
//  int correction_factor = (NOW-getStarttime()) / mac_->getPhy()->getPS();
//  debug10("correction_factor: %d\n", correction_factor);
//  debug10("Now in PS: %d\n", (int) (NOW * mac_->getPhy()->getPS()));
//  ulsubframe_.setStarttime( /*NOW * mac_->getPhy()->getPS() +*/ frame->mf.allocation_start - correction_factor);
//  debug2 ("\tul start time = %f\n", ulsubframe_.getStarttime()*mac_->getPhy()->getPS());
//  duration_ =  mac_->getFrameDuration() - frame->mf.allocation_start*mac_->getPhy()->getPS();
//  debug2("\t duration = %f\n",duration_);
//  setStarttime(getStarttime() + frame->mf.allocation_start*mac_->getPhy()->getPS());
 // mac_->setStartUlduration(duration_ - 3*mac_->getPhy()->getPS()); // 3 = starting frames (initial ranging + req_region_full)

  int running_subchannel_offset = 0;
  int running_symbol_offset 	= 0;
  int contention_width  = mac_->getPhy()->getSlotLength(UL_); //number of columns at least partially used
  int subframe_height = mac_->getPhy()->getNumsubchannels( UL_ ); //in units of subchannel
  int subframe_width = mac_->getMaxUlduration(); //in units of symbol (1/3 slot) 
  
  
  //note all allocations start from TOP LEFT corner of the allocation grid
  //and that contention regions have the entire first slot-column reserved
  

  for (int i = 0 ; i < nbies ; i++) {
    UlBurst *b = (UlBurst*)(ulsubframe_.addPhyPdu(i,0))->addBurst(0);
  	b->setCid(ies[i].ie.cid);
  
  	if (i == 0 && ies[i].ie.uiuc == UIUC_INITIAL_RANGING) {
  		b->setIUC(ies[i].ie.uiuc);
  		b->setStarttime(0);
  		b->setDuration(mac_->getPhy()->getSlotLength(UL_)); //3 symbols / slot
  		b->setB_CDMA_CODE(0);
  		b->setB_CDMA_TOP(0);
  		
		b->setnumSubchannels(ies[i].ie.duration); //this many subchannels by 1 slot (so by 3 symbols)
  		
  		continue;
  	} else if (i == 0) {
  		debug10("The UL MAP appears to be poorly constructed (does not start with INIT_RANGING region)\n");
  		return;
  	}
  	
  	if (i == 1 && ies[i].ie.uiuc == UIUC_REQ_REGION_FULL) { 
		b->setIUC(ies[i].ie.uiuc);
  		b->setDuration(mac_->getPhy()->getSlotLength(UL_)); //3 subchannels / slot
  		b->setB_CDMA_CODE(0);
  		b->setB_CDMA_TOP(0);
  		
		b->setnumSubchannels(ies[i].ie.duration);
  		
		running_symbol_offset = mac_->getPhy()->getSlotLength(UL_);
		debug10("The vertically mapped contention region takes the top %d subchannels of the first slot-column, withn the rest reserved for CQI & ACK\n", ies[0].ie.duration + ies[1].ie.duration);
  		
  		continue;
  	} else if (i == 1) {
		debug10("The UL MAP appears to be poorly constructed (does not have cdma_bw_req region as second element)\n");
		return;
  	}
  
    if (ies[i].ie.uiuc == UIUC_CDMA_ALLOCATION_IE) {
    	b->setIUC(ies[i].ie.uiuc);
	b->setExtendedUIUC(ies[i].ie.cdma_ie.ie.uiuc);
    	b->setB_CDMA_CODE(ies[i].ie.cdma_ie.ie.code);
	b->setB_CDMA_TOP (ies[i].ie.cdma_ie.ie.subchannel);
    } else {
    	b->setIUC(ies[i].ie.uiuc);
    	b->setB_CDMA_CODE(0);
    	b->setB_CDMA_TOP (0);
    }

    
    
    debug10("processing ie %d of UL_MAP_FRAME, uiuc = %d, starttime (OFDMA Symbol-times) = %d, Subchannel Offset = %d, duration (OFDMA Symbol-times)= %d\n", i, b->getIUC(), running_symbol_offset, running_subchannel_offset, ies[i].ie.duration * mac_->getPhy()->getSlotLength(UL_));

    //we now define the start point in terms of slot x subchannel and the duration
    b->setStarttime(running_symbol_offset); //symbol
    b->setSubchannelOffset(running_subchannel_offset); //subchannel

    b->setDuration(ies[i].ie.duration * mac_->getPhy()->getSlotLength(UL_));

    running_symbol_offset += ies[i].ie.duration * mac_->getPhy()->getSlotLength(UL_);
    running_subchannel_offset += floor ( (double)(running_symbol_offset - contention_width) /  (double) (subframe_width - contention_width));
    running_symbol_offset = (running_symbol_offset - contention_width) % (subframe_width - contention_width) + contention_width;

    if (running_subchannel_offset > subframe_height) {
	debug10("ERROR: UL_MAP describes subframe that needs too many slots\n");
	return;
    }


   //NOTE: The following code was written assuming the contention region took only part of the first column, and the rest would have the
   //horizontally-wrapping allocations wrap into it.  This is not the current implementation, but the basic code structure was left in case
   //that become possible at a later revision of the WiMAX standard.
    
    //update the running_symbol_offset and running_subchannel_offset variables to point to the first
    //position still available.  Depending on what height thte contention region ends at, some rows in
    //the map are going to be slightly longer
    
    //if we're entirely above the cutoff for the partially-used-for-contention column, symbol offset = (running_symbol_offset+ (duration * 3)) % (subframe_width - (contention_width -1) )
    //if we're entirely below the cutoff for that column,
    //symbol offset = (running_symbol_offset+ (duration * 3)) % (subframe_width - (contention_width) )
    //if we overlap it, calculate the appropriate positions
    
//    int remaining_space = subframe_width - running_symbol_offset;
//    int remaining_duration = b->getDuration();
    
    //we handle first the case where we tack this allocation entirely on to the same line as the last
    //if this is not possible, we figure out how far down to go, and handle the split in usable row width
    //at the top of the contention regions if necessary
//    if (remaining_duration <= remaining_space) {
//    	running_symbol_offset += remaining_duration;
//   		//since we stay on the same row, remaining_subchannel_offset is unchanged
//    } else while (remaining_duration > 0) {
//    	 //if we're at the end of the current line, assign values and stop looping
//		 if (remaining_duration < remaining_space) { 
//			running_symbol_offset = remaining_duration + contention_width;
//			//if we're over the break point for contention region, put that slot back in
//			break;	
//		}
//		//otherwise update the duration and current row
//		remaining_duration -= remaining_space;
//		running_subchannel_offset++;
//		remaining_space = subframe_width - contention_width;
//		if (running_subchannel_offset >= subframe_height) {
//			debug10("ERROR: UL_MAP describes subframe that needs too many slots\n");
//			return;
//
//		}
//	} 
  }
}

/**
 * Parse a UCD message and create the data structure
 * @param frame The DL frame information
 */
void FrameMap::parseUCDframe (mac802_16_ucd_frame *frame)
{
  assert (frame);
  //clear previous profiles
  ulsubframe_.removeProfiles();

  debug10("Parsing UCD Frame: \n");
  
  /*printf ("parse UCD..rng_start=%d, rng_stop=%d, req_start=%d, req_stop=%d\n",\
    frame->rng_backoff_start, frame->rng_backoff_end, frame->req_backoff_start,
    frame->req_backoff_end);*/
  ulsubframe_.getRanging()->setBackoff_start(frame->rng_backoff_start);
  ulsubframe_.getRanging()->setBackoff_stop(frame->rng_backoff_end);
  ulsubframe_.getRanging()->setSize(frame->rng_req_size);
  ulsubframe_.getBw_req()->setBackoff_start(frame->req_backoff_start);
  ulsubframe_.getBw_req()->setBackoff_stop(frame->req_backoff_end);  
  ulsubframe_.getBw_req()->setSize(frame->bw_req_size);

  int nb_prof = frame->nb_prof;
  mac802_16_ucd_profile *profiles = frame->profiles;
  for (int i = 0 ; i < nb_prof ; i++) {
    debug10("In %s, adding profile with IUC %d and Ofdm_mod_rate %d\n", mac_->addr_s(), profiles[i].uiuc, profiles[i].fec);
    Profile *p = ulsubframe_.addProfile (0, (Ofdm_mod_rate)(profiles[i].fec));
    p->setIUC (profiles[i].uiuc);
    //printf ("\t Adding ul profile %i: f=%d, rate=%d, iuc=%d\n", i, p->getFrequency(), p->getEncoding(), p->getIUC());
  }
}


/**
 * Print a snapshot of the frame
 */
void FrameMap::print_frame ()
{
	if (mac_->getNodeType() == STA_MN) {
		printf("Frame %d from this MN.  duration = %f total frame start = %f UL-Subframe start %f stop = %f\n", mac_->getFrameNumber(), duration_, starttime_, ulsubframe_.getStarttime() * mac_->getPhy()->getPS() + starttime_, starttime_+duration_);	  
		printf ("\tUplink:\n");  
	  int nbies = ulsubframe_.getNbPdu(); //there is one burst per UL phy PDU
	  	if (nbies); //quiet, compiler
	  int i=0;
  	  for (std::list<PhyPdu*>::iterator p = ulsubframe_.getPduList()->begin(); *p != NULL && p != ulsubframe_.getPduList()->end(); p++) {
		UlBurst *b = (UlBurst*) (*p)->getBurst(0);
// 		printf ("\t\tBurst %d: start=%d (%f)", i, b->getStarttime(), starttime_+b->getStarttime()*mac_->getPhy()->getSymbolTime());
// 		printf(" num subchanels = %d subchannel offset =%d code =%d top =%d\n",b->getnumSubchannels(),b->getSubchannelOffset (), b->getB_CDMA_CODE(), b->getB_CDMA_TOP() );
// 		printf(" num subchanels = %d subchannel offset =%d\n",b->getnumSubchannels(),b->getSubchannelOffset () );
// 	
		printf (" UIUC=%d",  b->getIUC());
	
		switch (b->getIUC()) {
		case UIUC_INITIAL_RANGING:
		  printf (" (INITIAL_RANGING)");
		  break;
		case UIUC_REQ_REGION_FULL:
		  printf (" (REQ_REGION_FULL)");
		  break;
		case UIUC_REQ_REGION_FOCUSED:
		  printf (" (REQ_REGION_FOCUSED)");
		  break;
		case UIUC_FOCUSED_CONTENTION_IE:
		  printf (" (FOCUSED_CONTENTION_IE)");
		  break;
		case UIUC_PROFILE_1:
		case UIUC_PROFILE_2:
		case UIUC_PROFILE_3:
		case UIUC_PROFILE_4:
		case UIUC_PROFILE_5:
		case UIUC_PROFILE_6:
		case UIUC_PROFILE_7:
		case UIUC_PROFILE_8:
		  printf (" (PROFILE %d)", b->getIUC() - UIUC_PROFILE_1 + 1);
		  break;
		case UIUC_SUBCH_NET_ENTRY:
		  printf (" (SUBCH_NET_ENTRY)");
		  break;
		case UIUC_CDMA_ALLOCATION_IE:
		  printf (" (CDMA_ALLOCATION)");
		  break;
		case UIUC_EXT_UIUC:
		  printf (" (EXT_UIUC)");
		  if (b->getExtendedUIUC ()== UIUC_FAST_RANGING) {
		printf (" ExtUIUC=FAST_RANGING addr=%s FR_UIUC=%d", HelperFunctions::getMacAddrString(b->getFastRangingMacAddr ()).c_str(), b->getFastRangingUIUC ());
		  }      
		  break;
		}
	  	
	  	printf (" CID = %u duration = %d ", (uint16_t)b->getCid(), b->getDuration());
	  	if (b->getIUC() == UIUC_INITIAL_RANGING || b->getIUC() == UIUC_REQ_REGION_FULL)
	  		printf("vertically mapped subchannels ");
		else printf ("OFDMA Symbols ");
		if (b->getIUC() == UIUC_CDMA_ALLOCATION_IE)
			printf(" extended-iuc = %d code = %d top = %d", b->getExtendedUIUC(), b->getB_CDMA_CODE(), b->getB_CDMA_TOP());
	  	printf("\n");
		i++;
	  }
	}
	if (mac_->getNodeType() == STA_BS) {
	  printf ("Frame %d: duration=%f start=%f stop=%f\n", mac_->getFrameNumber (), duration_, starttime_, starttime_+duration_);
	  printf ("\tDownlink:\n");
	  
	  int nbies = dlsubframe_.getPdu()->getNbBurst();
	  for (int i = 0 ; i < nbies ; i++) {
		Burst *b = dlsubframe_.getPdu()->getBurst(i);
		printf ("\t\tBurst %d: start=%d (%f)", i, b->getStarttime(), starttime_+b->getStarttime()*mac_->getPhy()->getSymbolTime());
		printf (" DIUC=%d",  b->getIUC());
		printf(" num subchanels = %d subchannel offset = %d symbol duration = %d ",b->getnumSubchannels(),b->getSubchannelOffset (), b->getDuration());
		if (b->getIUC()==DIUC_END_OF_MAP) {
		  printf (" (END_OF_MAP)\n");
		} else {
		  printf (" CID=%d\n",b->getCid());
		}
	  }
	
	  printf ("\tUplink:\n");  
	  nbies = ulsubframe_.getNbPdu(); //there is one burst per UL phy PDU
	  int i=0;

  	  for (std::list<PhyPdu*>::iterator p = ulsubframe_.getPduList()->begin();/* *p != NULL && p != ulsubframe_.getPduList()->end();*/true; p++) {
	  	if (*p == NULL) {
			printf("well, it's null %p\n", *p);
//			break;
		}
		else if (p == ulsubframe_.getPduList()->end()) {
			break;
		}
		UlBurst *b = (UlBurst*) (*p)->getBurst(0);
		printf ("\t\tBurst %d: UIUC: %d", i, b->getIUC());
		//printf(" num subchanels = %d subchannel offset =%d\n",b->getnumSubchannels(),b->getSubchannelOffset () );
	
		switch (b->getIUC()) {
		case UIUC_INITIAL_RANGING:
		  printf (" (INITIAL_RANGING)");
		  break;
		case UIUC_REQ_REGION_FULL:
		  printf (" (REQ_REGION_FULL)");
		  break;
		case UIUC_REQ_REGION_FOCUSED:
		  printf (" (REQ_REGION_FOCUSED)");
		  break;
		case UIUC_FOCUSED_CONTENTION_IE:
		  printf (" (FOCUSED_CONTENTION_IE)");
		  break;
		case UIUC_PROFILE_1:
		case UIUC_PROFILE_2:
		case UIUC_PROFILE_3:
		case UIUC_PROFILE_4:
		case UIUC_PROFILE_5:
		case UIUC_PROFILE_6:
		case UIUC_PROFILE_7:
		case UIUC_PROFILE_8:
		  printf (" (PROFILE %d)", b->getIUC() - UIUC_PROFILE_1 + 1);
		  break;
		case UIUC_SUBCH_NET_ENTRY:
		  printf (" (SUBCH_NET_ENTRY)");
		  break;
		case UIUC_CDMA_ALLOCATION_IE:
		  printf (" (CDMA_ALLOCATION)");
		  break;
		case UIUC_EXT_UIUC:
		  printf (" (EXT_UIUC)");
		  if (b->getExtendedUIUC ()== UIUC_FAST_RANGING) {
		printf (" ExtUIUC=FAST_RANGING addr=%s FR_UIUC=%d", HelperFunctions::getMacAddrString(b->getFastRangingMacAddr ()).c_str(), b->getFastRangingUIUC ());
		  }      
		  break;
		}
		
		printf (" CID = %u duration = %d ", (uint16_t)b->getCid(), b->getDuration());
	  	if (b->getIUC() == UIUC_INITIAL_RANGING || b->getIUC() == UIUC_REQ_REGION_FULL)
	  		printf("vertically mapped subchannels ");
		else printf ("OFDMA Symbols ");
	  	if (b->getIUC() == UIUC_CDMA_ALLOCATION_IE) {
	  		printf(" code = %d, top = %d", b->getB_CDMA_CODE(), b->getB_CDMA_TOP() );
	  	}
	  	printf("\n");
		i++;
	  }

  }

}

} //namespace ns3
