/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef BSSCHEDULER_H
#define BSSCHEDULER_H

#include "ns3/wimax2-scheduler.h"
//#include "scanningstation.h"

#define INIT_DL_DURATION 20 //enough for DL_MAP, UL_MAP, DCD, UCD and some RNG-RSP
#define MIN_CONTENTION_SIZE 5 //minimum number of opportunity for allocation

//#define DEFAULT_DL_RATIO 0.3 //default ratio for downlink subframe

#define NUM_REALIZATIONS1 2000 
#define OTHER_NUM_DL_BURST 2 //including 1st DL burst and END_OF_MAP burst.
#define OTHER_NUM_UL_BURST 3 //including initial ranging, bw req and END_OF_MAP burst


namespace ns3 {
// This structure is used for deficit round robin (next version)
struct con_data_alloc {
  int cid;
  int direction;
  Ofdm_mod_rate mod_rate;
  int req_slots;
  int req_bytes;
  int grant_slots;
  int grant_bytes;
  double weight;
  int counter;
};

struct con_drr {
  int cid;
  int quantum;
  int counter;
};

enum downlink_mapping_type {
	OFDM,
	OCSA,
	eOCSA,
	mOCSA
};


class Wimax2BSNetDevice;
class WimaxCtrlAgent;
/**
 * Class BSScheduler
 * Implement the packet scheduler on the BS side
 */ 
class Wimax2BSScheduler : public Wimax2Scheduler
{

	
	
  //friend class SendTimer;
 public:
  /*
   * Create a scheduler
   */
  Wimax2BSScheduler ();
  ~Wimax2BSScheduler();

//sam

  int Repetition_code_;


  int channel_gain[NUM_REALIZATIONS1] ;

//sam



  /**
   * Initializes the scheduler
   */
  virtual void init ();
 
  /**
   * This function is used to schedule bursts/packets
   */
  virtual void schedule ();

// protected: uncomment if python binding generator can handle protected variables in the future

  /**
   * Default modulation 
   */
  Ofdm_mod_rate default_mod_;

  /**
   * Number of transmission opportunity for initial ranging
   * and bw request (i.e contention slots)
   */
  int contention_size_; 
  int init_contention_size_; 
  int bw_req_contention_size_; 
  /**
   * Compute and return the bandwidth request opportunity size
   * @return The bandwidth request opportunity size
   */

   inline uint16_t getMaximumNumberDlBursts() {return max_dl_bursts;}
   inline uint64_t getTotalOverallocations(int32_t index) {return total_overallocations[index];}
   inline uint64_t getTotalUnmappedBlocks(int32_t index) {return total_unmapped_blocks[index];}
   inline uint64_t getTotalUnmappedBursts(int32_t index) {return total_unmapped_bursts[index];}
   inline uint64_t getTotalEmptyNeededBlocks(int32_t index) {return total_empty_needed_blocks[index];}
   inline uint64_t getNumberFramesWithNumberBursts(int32_t index) {return total_number_of_frames[index];}
   
protected:
   
  int getBWopportunity ();

  /**
   * Compute and return the initial ranging opportunity size
   * @return The initial ranging opportunity size
   */
  int getInitRangingopportunity ();  

  /**
   * Add a downlink burst with the given information
   * @param burstid The burst number
   * @param c The connection to add
   * @param iuc The profile to use
   * @param dlduration current allocation status
   * @param the new allocation status
   */
  //int addDlBurst (int burstid, Connection * , int iuc, int dlduration, int maxdlduration);

  /**
   * Given number of bursts, calculate total number of occupied slots in downlink subframe
   * @return total number of occupied slots
   */
  int check_overallocation (int num_of_entries);

  /**
   * Check if there is a existing ie_map for the particular connection
   * @return connection id or -1 if not exist
   */
  int doesvirtual_allocexist(int num_of_entries, int cid);

  /**
   * Check number of available slots given additional slots for the particular allocation 
   * @return available slots (= requested slots if there is enough slots in downlink subframe)
   */
  int overallocation_withoutdlmap(int num_of_entries, int totalslots, int ownslots);

  /**
   * Check number of available slots given additional slots for the particular allocation with increase of dl_map_ie
   * @return available slots (= requested slots if there is enough slots in downlink subframe)
   */
  int overallocation_withdlmap(int num_of_entries, int totalslots, int ownslots);

  /**
   * Add granted slots to downlink subframe given cid
   * @return positive number if granted slots are added else 0
   */
  int addslots_withoutdlmap(int number_of_entries, int byte, int slots, int cid);

  /**
   * Find out the maximum burst the downlink subframe can support (the increase of dl_map_ie)
   * @return total number of available dl_map_ie
   */
  int max_conn_withdlmap(int num_of_entries, int totalslots);

  /**
   * Bubble Sort (sort_field = 0 or 1 low to high or high to low)
   */
 // -#- @array(transfer_ownership=true) -#-
  void bubble_sort (int arrayLength, con_data_alloc array[], int sort_field);

  /**
   * Compute number of available slots with the increase of #conn*dl_map_ie size
   * @return total number of availableo slots
   */
  int freeslots_withdlmap_given_conn(int num_of_entries, int totalslots, int newconn);

  /**
   * Add dl_map_ie size to dl_map
   * @return total number of occupied slots including new dl_map_ie
   */
  int increase_dl_map_ie(int num_of_entries, int totalslots, int num_ie);


//added rpi for OFDMA -------  
/**
 * Add a downlink burst with the given information
 * @param burstid The burst number
 * @param c The connection to add
 * @param iuc The profile to use
 * @param ofdmsymboloffset 
 * @param numofdmsymbol 
 * @param subchanneloffset
 * @param numsubchannels
 */
void addDlBurst (int burstid, int cid, int iuc, int ofdmsymboloffset, int numofdmsymbols, int subchanneloffset, int numsubchannels);

/* returns the DIUC profile associated with a current modulation and coding scheme */
diuc_t getDIUCProfile(Ofdm_mod_rate rate);

uiuc_t getUIUCProfile(Ofdm_mod_rate rate);

//ritun

//mac802_16_dl_map_frame * dl_stage2(Connection *head, int total_subchannels, int total_symbols, int symbol_start, int stripping);

 // -#- @head(transfer_ownership=false) -#-
 mac802_16_dl_map_frame * dl_stage2(Connection *head, int input_subchannel_offset,  int total_subchannels, int total_symbols, int symbol_start, int  stripping, int total_dl_slots_pusc);

 // -#- @head(transfer_ownership=false) -#-
 mac802_16_ul_map_frame * ul_stage2(Connection *head, int total_subchannels, int total_symbols, int symbol_start, int stripping);

 // -#- @cid_list(transfer_ownership=false) -#-
 int doesMapExist(int cid, int* cid_list, int num_of_entries);

//ritun

//int  GetInitialChannel();

// added rpi for ofdma - end


/**
 * get the dlmap. 
 * @param out connection list , number of ofdm symbols tht can be allocated, permutation scheme, CQI , ofdm symbol offset
 */

//mac802_16_dl_map_frame GetDLMap(mac_->getCManager()->get_out_connection (), int numofdmsymbols,int numsubchannels,Permutationscheme_, int CQI, int symboloffset);

/**
 * get the ulmap. 
 * @param in connection list , number of ofdm symbols tht can be allocated, permutation scheme, CQI , ofdm symbol offset
 */

//mac802_16_ul_map_frame GetULMap(mac_->getCManager()->get_in_connection (), int numofdmsymbols,int numsubchannels,Permutationscheme_, int CQI, int symboloffset);

// added rpi for ofdma 

 private:

  /**
   * This signals the algorithm used for OFDMA Downlink mapping 
   */
  downlink_mapping_type  mt_;

  /** 
   * Transfer packets from the connection to the burst
   * @param c The Connection
   * @param b The Burst
   */
 uint32_t transfer_packets_for_dl_subframe(Connection *c, Burst *b);


  /**
   * Return the MAC casted to BSScheduler
   * @return The MAC casted to BSScheduler
   */
  Ptr<Wimax2BSNetDevice> getMac();
   
  /**
   * The ratio for downlink subframe
   */
//  double dlratio_;   rpi removed , moved to wimax scheudler

  /**
   * The address of the next node for DL allocation
   */
  int nextDL_;
  
  /**
   * The address of the next node for UL allocation
   */
  int nextUL_;


  /**
   * The output file stream used to output data regarding mapping efficiency
   */
  std::ofstream allocation_file;

  //these are used to track the efficency of mapping
  uint16_t maximum_number_dl_bursts; //determined by max number of IE's in DL map based on size
  uint16_t max_dl_bursts; //max actually observed
  uint64_t *total_overallocations;
  uint64_t *total_unmapped_blocks;
  uint64_t *total_unmapped_bursts;
  uint64_t *total_empty_needed_blocks;
  uint64_t *total_number_of_frames;
};

}//namepsace ns3

#endif //BSSCHEDULER_H

