/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#include "wimax2-ss-scheduler.h"
#include "ns3/burst.h"
#include "ns3/wimax2-ss-net-device.h"

namespace ns3 {

int frame_no = 0;


/*
 * Create a scheduler
 */
Wimax2SSScheduler::Wimax2SSScheduler () : Wimax2Scheduler ()
{
  debug2 ("Wimax2SSScheduler created\n");
}

/**
 * Initializes the scheduler
 */
void Wimax2SSScheduler::init ()
{
  Wimax2Scheduler::init();
}


/**
 * Schedule bursts/packets
 */
void Wimax2SSScheduler::schedule ()
{
  debug10("entering (SS-S) %s schedule\n", mac_->addr_s());

  int b_data;
  Connection *c;
  Burst *b, *b_tmp, *my_burst, *my_burst_rng;
  FrameMap *map = mac_->getMap();
  PeerNode *peer = mac_->getPeerNode_head(); //this is the BS
  bool gotAllocation = false;
  #ifdef DEBUG_WIMAX
	  assert (peer!=NULL);
  #endif

//  rpi - increment the propagation channel 
  
 
  if(peer)
    peer->setchannel((peer->getchannel())+1);


if(peer->getchannel()>999) peer->setchannel(2);  
 
//rpi
// Begin RPI   
 // We will try to Fill in the ARQ Feedback Information now...
  Ptr<Packet>  ph = 0;
  PeerNode * peernode;
  Connection * basic ;
  Connection * OutData;
  Ptr<Packet>  pfb = 0;
  All_Headers ah_map;
  All_Headers ah_ph;
//  hdr_mac802_16 *wimaxHdrMap ;
  u_int16_t   temp_num_of_acks = 0;
  bool out_datacnx_exists = false; 
  for (Connection *n= mac_->getCManager ()->get_in_connection (); n; n=n->next_entry()) {
   if(n->getArqStatus () != NULL && n->getArqStatus ()->isArqEnabled() == 1)
   {
     if(!(n->getArqStatus ()->arq_feedback_queue_) || (n->getArqStatus ()->arq_feedback_queue_->GetNPackets() == 0))
     {
		continue;
     } 
     else 
     {
		   peernode = n->getPeerNode ();
		   if(peernode->getOutData () != NULL && peernode->getOutData ()->queueLength () != 0 && getMac()->arqfb_in_ul_data_) {
		      out_datacnx_exists = true;
	   }
	   if(out_datacnx_exists == false) 
	   {
		   debug2("ARQ SS: Feedback in Basic Cid \n");
		   basic = peernode->getBasic (OUT_CONNECTION);
       	   pfb = mac_->getPacket ();
      	   pfb->RemoveHeader(ah_map);
//         wimaxHdrMap= HDR_MAC802_16(pfb);
	       ah_map.GMH.GMH.cid = basic->get_cid ();	
           ah_map.num_of_acks = 0;
	   }
	   else {
		   debug2("ARQ SS : Feedback in data Cid \n");
		   OutData = peernode->getOutData ();
		   pfb = OutData->dequeue ();
		   pfb->RemoveHeader(ah_map);
//		   wimaxHdrMap= HDR_MAC802_16(pfb);
		   if(ah_map.GMH.GMH.type_arqfb == 1)
		   {
				debug2("ARQ SS: Feedback already present, do nothing \n");
				pfb->AddHeader(ah_map);
				OutData->enqueue_head (pfb);
				continue;
		   }
	   ah_map.num_of_acks = 0;
	   }
          //Dequeue is only once as one feedback in 5ms frame is permitted 
	  ph = n->getArqStatus ()->arq_feedback_queue_->Dequeue();
      ah_map.GMH.GMH.type_arqfb = 1;
      ph->RemoveHeader(ah_ph);
//	  hdr_mac802_16 *wimaxHdr= HDR_MAC802_16(ph);
	  for (temp_num_of_acks = 0; temp_num_of_acks < ah_ph.num_of_acks; temp_num_of_acks++)
	  {	
            ah_map.arq_ie[temp_num_of_acks].cid  	 = ah_ph.arq_ie[temp_num_of_acks].cid;
            ah_map.arq_ie[temp_num_of_acks].last 	 = ah_ph.arq_ie[temp_num_of_acks].last;
            ah_map.arq_ie[temp_num_of_acks].ack_type = ah_ph.arq_ie[temp_num_of_acks].ack_type;
            ah_map.arq_ie[temp_num_of_acks].fsn 	 = ah_ph.arq_ie[temp_num_of_acks].fsn;
          }    
      ah_map.num_of_acks = ah_ph.num_of_acks;
	  ah_map.GMH.GMH.len += (ah_ph.num_of_acks * HDR_MAC802_16_ARQFEEDBK_SIZE);  
	  debug2("ARQ : In Wimax2SSScheduler: Enqueueing an feedback cid: %d arq_ie->fsn:%d \n", ah_ph.arq_ie[0].cid,  ah_ph.arq_ie[0].fsn);
	  if(out_datacnx_exists == false) {	
		// If I am here then the Ack packet has been created, so we will enqueue it in the Basic Cid
		pfb->AddHeader(ah_map);
		basic->enqueue (pfb);
	  }
	  else
	  	pfb->AddHeader(ah_map);
		OutData->enqueue_head (pfb);  	
	  }
    }
  }
// End RPI
//	debug10("=SS= frame number :%d\n", frame_no++);

//#ifdef DEBUG_WIMAX
	debug2("\n==========MAC SS %s subframe =============\n", mac_->addr_s());
	mac_->getMap()->print_frame();
	debug2("============================================================\n");
//#endif


  //we start off by checking if we have available allocations for anything in the queues
  //and if we find anything, we transfer to bursts
  for (uint32_t index = 0 ; index < map->getUlSubframe()->getNbPdu (); index++) 
  {
    b_tmp = map->getUlSubframe()->getPhyPdu (index)->getBurst (0);
    if (b_tmp->getIUC()==UIUC_INITIAL_RANGING) continue;
    else if (b_tmp->getIUC()==UIUC_REQ_REGION_FULL) continue;
    else {

	int allocationfound = 0; 
	Connection *c_burst = mac_->getCManager ()->get_connection (b_tmp->getCid(), OUT_CONNECTION);

    	if (!c_burst) {
      		continue;
        } else{
            my_burst = b_tmp;
      		allocationfound = 1;
    	}

		b_data = 0;
    	if (allocationfound == 1) {
       		uint8_t b_code = b_tmp->getB_CDMA_CODE ();
       		uint8_t b_top = b_tmp->getB_CDMA_TOP ();
       		int init_code = mac_->getMap()->getUlSubframe()->getRanging()->getCODE (mac_->addr());
       		int init_top = mac_->getMap()->getUlSubframe()->getRanging()->getTOP (mac_->addr());       

       		int bw_code = mac_->getMap()->getUlSubframe()->getBw_req()->getCODE (b_tmp->getCid());
       		int bw_top = mac_->getMap()->getUlSubframe()->getBw_req()->getTOP (b_tmp->getCid());
       			if (bw_code || bw_top);//quiet, compiler

			debug10 ("\tStart SS addr: %s, CID :%d, b_code :%d, b_top :%d, ssid_code :%d, ssid_top :%d, bw_code :%d, bw_top :%d\n", mac_->addr_s(), b_tmp->getCid(), b_code, b_top, init_code, init_top, bw_code, bw_top);       
			debug10 ("\tBurst start time %d (Symbols), Burst start subchannel offset %d, Burst duration %d, IUC %d, CID %d\n", my_burst->getStarttime(), my_burst->getnumSubchannels(),  my_burst->getDuration(), my_burst->getIUC(), my_burst->getCid());
			
			//if it's an intial ranging message allocation, we have to do a lot less before sending
			if (b_tmp->getCid() == 0) {
				if ( (b_code == init_code) && (b_top == init_top) ) {          
					debug10 ("\tFound initial_ranging_msg opportunity (remove enqueued cdma_init_req), Burst start time %d, Burst duration %d, IUC %d, subchannel offset %d, num of subchannels %d, CID %d\n", my_burst->getStarttime(), my_burst->getDuration(), my_burst->getIUC(),  my_burst->getSubchannelOffset(),  my_burst->getnumSubchannels(), my_burst->getCid());
					b_data = transfer_packets_for_ul_subframe(c_burst, my_burst);
				}
			} else {//it's not an initial ranging message allocation
				 
				 ConnectionType_t contype;
				 Connection *head = mac_->getCManager()->get_out_connection();
				 Connection *c_tmp1;
		
				 for(int i=0; i<4; ++i){
					c_tmp1 = head;
					if(i==0) 	contype = CONN_BASIC;
					else if(i==1)   contype = CONN_PRIMARY;
					else if(i==2)   contype = CONN_SECONDARY;
					else 	    	contype = CONN_DATA;
		
					b_data = 0;
						while(c_tmp1!=NULL){
							if(c_tmp1->get_category()==contype){
								int bw_code = mac_->getMap()->getUlSubframe()->getBw_req()->getCODE (c_tmp1->get_cid());
								int bw_top = mac_->getMap()->getUlSubframe()->getBw_req()->getTOP (c_tmp1->get_cid());
								
								if ( (b_code == bw_code) && (b_top == bw_top) ) {
									gotAllocation = true;
									debug10 ("\tFound bw_req opportunity (do not enqueue a new cdma_bw_req), Burst start time %d, Burst duration %d, IUC %d, subchannel offset %d, num of subchannels %d, CID %d\n", my_burst->getStarttime(), my_burst->getDuration(), my_burst->getIUC(),  my_burst->getSubchannelOffset(),  my_burst->getnumSubchannels(), my_burst->getCid());
									debug10 ("\tCID :%d, b_code :%d, b_top :%d, bw_code :%d, bw_top :%d\n", c_tmp1->get_cid(), b_code, b_top, bw_code, bw_top);
									
									mac_->getMap()->getUlSubframe()->getBw_req()->setFLAGNOWTRANSMIT (c_tmp1->get_cid(), 1);
									b_data = transfer_packets_for_ul_subframe(c_tmp1, my_burst);
								} else if (b_tmp->getCid() == c_tmp1->get_cid()) {
									//we have an allocation, so we don't want to send a new contention based bw req
									//note this won't work well if multiple connections need BW
									gotAllocation = true;
								}
							}
			
						c_tmp1 = c_tmp1->next_entry();
						}//end while
					}//end for
			  }

    	} 

    }//end else all IUC

  }//end for all NbPU


  int allocationfound = 0;
  debug2 ("in ss schedule function SS has %d ul bursts\n", map->getUlSubframe()->getNbPdu ());
  
  //go through the list of UL Bursts to find allocations for what we want to send
  for (uint32_t index = 0 ; index < map->getUlSubframe()->getNbPdu (); index++) 
  {
    b = map->getUlSubframe()->getPhyPdu (index)->getBurst (0);
 

if (b->getIUC()==UIUC_INITIAL_RANGING && (mac_->GetObject<Wimax2SSNetDevice>()->getMacState() == MAC802_16_RANGING || mac_->GetObject<Wimax2SSNetDevice>()->getMacState() == MAC802_16_WAIT_ANON_RNG_RSP)) {

	bool t_bool = false;
        Connection *c_tmp = mac_->getCManager()->get_connection(0, t_bool );
//        debug10 ("CON :%d\n", c_tmp->get_cid());

        my_burst_rng = b;
        Mac48Address tmp_mac = mac_->addr(); 
        const char* tmp_mac_s = mac_->addr_s();
        	if (tmp_mac_s); //quiet, compiler
        ContentionSlot *slot1 = getMac()->getMap()->getUlSubframe()->getRanging ();        
		debug10 ("Start CDMA-INIT Ranging Region: GET init-req-backoff-start :%d, stop :%d, contention size :%d, nbretry :%d\n", slot1->getBackoff_start(), slot1->getBackoff_stop(), slot1->getSize(), getMac()->macmib_->request_retry);

        int s_init_contention_size = slot1->getSize(); //0 to 4
        int s_nbretry = getMac()->macmib_->reg_req_retry;
        int s_backoff_start = slot1->getBackoff_start();
        	if (s_backoff_start); //quiet, compiler
        int s_backoff_stop = slot1->getBackoff_stop();
        if (mac_->getMap()->getUlSubframe()->getRanging()->getRequest_mac (tmp_mac) != NULL) {
             int tmp_cid_mac = mac_->getMap()->getUlSubframe()->getRanging()->getCID (tmp_mac); if (tmp_cid_mac); //quiet, compiler
             debug10 ("Init Ranging: Found request for SS Mac addr :%s\n", tmp_mac_s);
             int tmp_backoff = mac_->getMap()->getUlSubframe()->getRanging()->getBACKOFF (tmp_mac);
             int tmp_timeout = mac_->getMap()->getUlSubframe()->getRanging()->getTIMEOUT (tmp_mac);
             int tmp_nbretry = mac_->getMap()->getUlSubframe()->getRanging()->getNBRETRY (tmp_mac);
             int tmp_win = mac_->getMap()->getUlSubframe()->getRanging()->getWINDOW (tmp_mac);
             int tmp_code = mac_->getMap()->getUlSubframe()->getRanging()->getCODE (tmp_mac);
             int tmp_top = mac_->getMap()->getUlSubframe()->getRanging()->getTOP (tmp_mac);
             int tmp_flagtransmit = mac_->getMap()->getUlSubframe()->getRanging()->getFLAGTRANSMIT (tmp_mac);
             debug10 ("\tINIT.1 SSID :%d, backoff :%d, timeout :%d, nbretry :%d, win :%d, code :%d, top :%d, flagtransmit :%d\n", mac_->getIdIndex(), tmp_backoff, tmp_timeout, tmp_nbretry, tmp_win, tmp_code, tmp_top, tmp_flagtransmit);

             int update_w        = tmp_win;
             int update_retry    = tmp_nbretry;             
	     int newbackoff      = tmp_backoff;             
	     int newtop          = tmp_top;             
	     int newcode         = tmp_code;             
	     int newtimeout      = tmp_timeout;
             int newflagtransmit = tmp_flagtransmit;

             Ptr<Packet> p_tmp = NULL;
             if (PeekPointer((p_tmp = mac_->getMap()->getUlSubframe()->getRanging ()->getPacket_P_mac (tmp_mac))) != NULL) {
	     	p_tmp = p_tmp->Copy(); //in case we don't send and this gets reused, it will still have the All_Headers on it this way
                newtimeout--;
                debug10 ("Init Ranging: Enqueue request for SS Mac addr :%s timeout %d\n", tmp_mac_s, newtimeout);

                if (newtimeout==0) {
                   update_retry++;
                   if (update_retry > s_nbretry) {
                      debug10 ("\tINIT.3.removeRequest (>retry) SSID :%d, Max_nbretry :%d, current_nbretry :%d\n", mac_->getIdIndex(), s_nbretry, update_retry);
                      mac_->getMap()->getUlSubframe()->getRanging()->removeRequest_mac (tmp_mac);
		      mac_->GetObject<Wimax2SSNetDevice>()->setMacState(MAC802_16_RANGING);
                      continue;
                   } else {
                     if (tmp_win<s_backoff_stop) {
                        update_w =tmp_win+1;
                        mac_->getMap()->getUlSubframe()->getRanging()->setWINDOW (tmp_mac, update_w);
                     } else {
                       update_w = s_backoff_stop;
                       mac_->getMap()->getUlSubframe()->getRanging()->setWINDOW (tmp_mac, update_w);
                     }

                     int result = rand() % ((int)(pow (2, update_w)+1));
                     newbackoff = result;
		     newtop = s_init_contention_size == 0 ? 0 : result % s_init_contention_size;

                     int code_range = getMac()->macmib_->cdma_code_bw_start - getMac()->macmib_->cdma_code_bw_stop + 1;
                     int c_rand = rand() % code_range;
                     newcode  = (uint8_t)getMac()->macmib_->cdma_code_bw_start + (uint8_t)c_rand;

                     newtimeout = CDMA_TIMEOUT-1;
                     mac_->getMap()->getUlSubframe()->getRanging()->setBACKOFF (tmp_mac, newbackoff);
                     mac_->getMap()->getUlSubframe()->getRanging()->setTIMEOUT (tmp_mac, newtimeout);
                     mac_->getMap()->getUlSubframe()->getRanging()->setNBRETRY (tmp_mac, update_retry);
                     mac_->getMap()->getUlSubframe()->getRanging()->setFLAGTRANSMIT (tmp_mac, 0);
                     mac_->getMap()->getUlSubframe()->getRanging()->setTOP (tmp_mac, newtop);
                     mac_->getMap()->getUlSubframe()->getRanging()->setCODE (tmp_mac, newcode);

                     debug10 ("\tINIT.3.updatebackoff (0 TIMEOUT) SSID :%d, oldbackoff :%d, newbackoff :%d, old_timeout :%d, new_timeout :%d, old_nbretry :%d, new_nbretry :%d, old_win :%d, new_win :%d, oldcode :%d, newcode :%d\n", mac_->getIdIndex(), tmp_backoff, newbackoff, tmp_timeout-1, newtimeout, tmp_nbretry, update_retry, tmp_win, update_w, tmp_code, newcode);
                     debug10 ("\t   start_backoff :%d, stop_backoff :%d, contention_size :%d, NBRETRY :%d, TIMEOUT :%d\n", s_backoff_start, s_backoff_stop, s_init_contention_size, s_nbretry, CDMA_TIMEOUT);
                     newflagtransmit = 0;

		     All_Headers ah_t;
		     p_tmp->RemoveHeader(ah_t);
                     cdma_req_header_t *header_t = &ah_t.CRH.data;
		     if (s_init_contention_size != 0) {
			ah_t.PIH.subchannel_offset = (newbackoff % s_init_contention_size)*6;
			debug10 ("\tINIT.3.updateOFFSET SSID :%d, #old_subchannel_offset :%d, #subchannel_offset :%d, oldtop :%d, newtop :%d\n", mac_->getIdIndex(), ah_t.PIH.subchannel_offset, (newbackoff%s_init_contention_size)*6, header_t->top, (newbackoff%s_init_contention_size));
		     } else {
			ah_t.PIH.subchannel_offset = 0;
			debug10 ("\tINIT.3.updateOFFSET SSID :%d, #old_subchannel_offset :%d, #subchannel_offset :%d, oldtop :%d, newtop :%d\n", mac_->getIdIndex(), ah_t.PIH.subchannel_offset, 0, header_t->top, 0);
		    }
                    

		     header_t->top = newtop;
		     header_t->code = newcode;
				 
		     p_tmp->AddHeader(ah_t);
                     c_tmp->setINIT_REQ_QUEUE(mac_->getIdIndex(), 0);
                   }

                } else {
                  mac_->getMap()->getUlSubframe()->getRanging()->setTIMEOUT (tmp_mac, newtimeout);
                  debug10 ("\tINIT.3.updatetimeout_perframe SSID :%d, oldtimeout :%d, newtimeout :%d\n", mac_->getIdIndex(), tmp_timeout, newtimeout);
                }

                int newbackoff_bak = newbackoff;
                	if (newbackoff_bak);//quiet, compiler

                newbackoff -= s_init_contention_size;
                if (newbackoff < 0) newbackoff = 0;
                mac_->getMap()->getUlSubframe()->getRanging()->setBACKOFF (tmp_mac, newbackoff);

                debug10 ("\tINIT.4.updatebackoff_perframe SSID :%d, oldbackoff :%d, newbackoff :%d (ori_new_backoff :%d), newtimeout :%d, newnbretry :%d, newwin :%d, top :%d, code :%d\n", mac_->getIdIndex(), tmp_backoff, newbackoff, newbackoff_bak, newtimeout, update_retry, update_w, newtop, newcode);


                if (newbackoff == 0 ) {
                   All_Headers ah_tmp;
                   p_tmp->RemoveHeader(ah_tmp);
                   ah_tmp.gm_sm = 3;
                   ah_tmp.GMH.GMH.ht = 1;
                   ah_tmp.cdma = 1;
                   
                   
                   cdma_req_header_t *header_s = &ah_tmp.CRH.data;
                   debug10 ("\tINIT.4.backoff=0 may_transmit (if flagtransmit == 0) SSID :%d, #symbol :%d, #symbol_offset :%d, #subchannel :%d, #subchannel_offset :%d, code :%d, top :%d, channel_index :%d, direction :%d\n", mac_->getIdIndex(), ah_tmp.PIH.num_OFDMSymbol, ah_tmp.PIH.OFDMSymbol_offset, ah_tmp.PIH.num_subchannels, ah_tmp.PIH.subchannel_offset, header_s->code, header_s->top, ah_tmp.PIH.channel_index , ah_tmp.PIH.direction);
                   debug10 ("\t   SSID :%d, size :%d, time :%f, direction :%d, flagtransmit :%d\n", mac_->getIdIndex(), ah_tmp.GMH.GMH.len, ah_tmp.txtime, ah_tmp.direction, newflagtransmit);
				   p_tmp->AddHeader(ah_tmp);
                   
                   if (newflagtransmit == 0) {
                      newtimeout = CDMA_TIMEOUT-1;
                      debug10 ("\tINIT.5.transmit SSID :%d, #old_flagtransmit :%d, #new_flagtransmit :%d reset timeout to :%d\n", mac_->getIdIndex(), newflagtransmit, 1, newtimeout);
                      newflagtransmit = 1;
		      p_tmp->RemoveHeader(ah_tmp);
		      p_tmp->AddHeader(ah_tmp.CRH);
                      mac_->getMap()->getUlSubframe()->getRanging()->setFLAGTRANSMIT (tmp_mac, newflagtransmit);
                      mac_->getMap()->getUlSubframe()->getRanging()->setTIMEOUT (tmp_mac, newtimeout);
                      mac_->getMap()->getUlSubframe()->getRanging()->setFLAGNOWTRANSMIT (tmp_mac, 1);
                      mac_->getMap()->getUlSubframe()->getRanging()->setCODE(tmp_mac, header_s->code);
                      mac_->getMap()->getUlSubframe()->getRanging()->setTOP(tmp_mac, header_s->top);
                      my_burst_rng->enqueue(p_tmp->Copy());
                      mac_->GetObject<Wimax2SSNetDevice>()->setMacState(MAC802_16_WAIT_ANON_RNG_RSP);
                   }
                }
             }
        }
    }
///////////End CDMA_INITIAL_RANGING COUNTER-BASED

    if (b->getIUC()==UIUC_REQ_REGION_FULL) {
    	if (gotAllocation) {
		debug10("testing...\n");
    		//do nothing
    	} else {
			my_burst_rng = b;       
			Connection *head_tmp = mac_->getCManager()->get_out_connection();
			Connection *con_tmp = head_tmp;
			int t_i = 0;
			ContentionSlot *slot1 = getMac()->getMap()->getUlSubframe()->getBw_req ();
			debug10 ("Start CDMA-BW-REQ Region: GET bw-req-backoff-start :%d, stop :%d, contention size :%d, nbretry :%d\n", slot1->getBackoff_start(), slot1->getBackoff_stop(), slot1->getSize(), getMac()->macmib_->request_retry);
	
			int s_bw_contention_size = slot1->getSize(); //0 to 4
			int s_nbretry = getMac()->macmib_->request_retry;
			int s_backoff_start = slot1->getBackoff_start();
				if (s_backoff_start); //quiet, compiler
			int s_backoff_stop = slot1->getBackoff_stop();
	
			while (con_tmp!=NULL){
	
			  int tmp_cid = con_tmp->get_cid();
			  if (con_tmp->queueByteLength() == 0 && (con_tmp->getArqStatus () == NULL || con_tmp->getArqStatus ()->arq_retrans_queue_->GetNPackets() == 0))  {
					debug10 ("\tCID %d, queueBytes = 0 so do not need to create CDMA_BW_REQ (remove if exist), c->queueBytes :%d\n", tmp_cid, con_tmp->queueByteLength());
					mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (tmp_cid);
					con_tmp = con_tmp->next_entry();
					continue;
			  }
	
	//We always create cdma_requst but we remove it later if we can send somepacket or bandwidth request out
			  if(con_tmp->get_category()==CONN_BASIC || con_tmp->get_category()==CONN_PRIMARY || con_tmp->get_category()==CONN_SECONDARY || con_tmp->get_category()==CONN_DATA){
				if(con_tmp->get_category()==CONN_DATA) {
					if( (con_tmp->get_serviceflow()->getScheduling()==SERVICE_UGS) || (con_tmp->get_serviceflow()->getScheduling()==SERVICE_rtPS) ){
						//we're going to get a bw grant without polling
					} else {
					  create_cdma_request(con_tmp);
					}
				} else {
					create_cdma_request(con_tmp);
				}
			  }
	
			  debug10 ("CDMA-BW_REQ REGION %d CID(BC) :%d, con_CID :%d, con_category :%d, con_qlength :%d\n", t_i, my_burst_rng->getCid(), con_tmp->get_cid(), con_tmp->get_category(), con_tmp->queueByteLength());
	
			  if (con_tmp->queueLength()>0 || (con_tmp->getArqStatus () != NULL && con_tmp->getArqStatus ()->arq_retrans_queue_->GetNPackets() > 0) ) {
				 int tmp_backoff = mac_->getMap()->getUlSubframe()->getBw_req()->getBACKOFF (tmp_cid);
				 int tmp_timeout = mac_->getMap()->getUlSubframe()->getBw_req()->getTIMEOUT (tmp_cid);
				 int tmp_nbretry = mac_->getMap()->getUlSubframe()->getBw_req()->getNBRETRY (tmp_cid);
				 int tmp_win = mac_->getMap()->getUlSubframe()->getBw_req()->getWINDOW (tmp_cid);
				 int tmp_code = mac_->getMap()->getUlSubframe()->getBw_req()->getCODE (tmp_cid);
				 int tmp_top = mac_->getMap()->getUlSubframe()->getBw_req()->getTOP (tmp_cid);
				 int tmp_flagtransmit = mac_->getMap()->getUlSubframe()->getBw_req()->getFLAGTRANSMIT (tmp_cid);
				 debug10 ("\tBW-REQ.1 CID :%d, backoff :%d, timeout :%d, nbretry :%d, win :%d, code :%d, top :%d, flagtransmit :%d\n", tmp_cid, tmp_backoff, tmp_timeout, tmp_nbretry, tmp_win, tmp_code, tmp_top, tmp_flagtransmit);
	
				 int update_w 			= tmp_win;
				 int update_retry 		= tmp_nbretry;
				 int newbackoff 		= tmp_backoff;
				 int newtop 			= tmp_top;
				 int newcode 			= tmp_code;
				 int newtimeout 		= tmp_timeout;
				 int newflagtransmit 	= tmp_flagtransmit;
	
				 Ptr<Packet> p_tmp = NULL;
				 if ((p_tmp = mac_->getMap()->getUlSubframe()->getBw_req()->getPacket_P (tmp_cid)) != NULL) {
					All_Headers ah_tmp;
					p_tmp->RemoveHeader(ah_tmp);
					cdma_req_header_t *header_tmp = &ah_tmp.CRH.data;
						if (header_tmp); //quiet, compiler
			
					mac_->getMap()->getUlSubframe()->getBw_req()->setFLAGNOWTRANSMIT (tmp_cid, 0);
	
					int n_sub = ah_tmp.PIH.num_subchannels; if (n_sub); //quiet, compiler
					int sub_off = ah_tmp.PIH.subchannel_offset; if (sub_off); //quiet, compiler
					debug10 ("\tBW-REQ.2.packetinfo CID :%d, #symbol :%d, #symbol_offset :%d, #subchannel :%d, #subchannel_offset :%d, code :%d, top :%d\n", tmp_cid, ah_tmp.PIH.num_OFDMSymbol, ah_tmp.PIH.OFDMSymbol_offset, ah_tmp.PIH.num_subchannels, ah_tmp.PIH.subchannel_offset, header_tmp->code, header_tmp->top);
					debug10 ("\tBW-REQ.2.burstinfo  #b_duration :%d, #b_subchannel :%d, #b_subchannel_offset :%d, b_starttime (in units of symbol duration) :%d\n", my_burst_rng->getDuration(), my_burst_rng->getnumSubchannels(), my_burst_rng->getSubchannelOffset(), my_burst_rng->getStarttime());
	
					newtimeout--;
					p_tmp->AddHeader(ah_tmp);
	
					if (newtimeout==0) {
					   update_retry++;
					   //if we've tried this too many times without getting a bw-grant, remove the request
					   if (update_retry > s_nbretry) {
						  debug10 ("\tBW-REQ.3.removeRequest CID :%d, Max_nbretry :%d, current_nbretry :%d\n", tmp_cid, s_nbretry, update_retry);
						  mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (tmp_cid);
						  con_tmp = con_tmp->next_entry();
						  continue;
					   } else {
	
						 if (tmp_win<s_backoff_stop) {
							update_w = tmp_win+1;
							mac_->getMap()->getUlSubframe()->getBw_req()->setWINDOW (tmp_cid, update_w);
						 } else {
						   update_w = s_backoff_stop;
						   mac_->getMap()->getUlSubframe()->getBw_req()->setWINDOW (tmp_cid, update_w);
						 }
	
						 int result = rand() % ((int)(pow (2, update_w)+1));
						 newbackoff = result;
						 newtop     = result % s_bw_contention_size;
						 newtimeout = CDMA_TIMEOUT-1;
	
						 int code_range = getMac()->macmib_->cdma_code_bw_start - getMac()->macmib_->cdma_code_bw_stop + 1;
						 int c_rand = rand() % code_range;
						 newcode  = (uint8_t)getMac()->macmib_->cdma_code_bw_start + (uint8_t)c_rand;
						 //if we get a code of 0 randomly it can cause issues below, this should really only loop once
	
						 mac_->getMap()->getUlSubframe()->getBw_req()->setBACKOFF (tmp_cid, newbackoff);
						 mac_->getMap()->getUlSubframe()->getBw_req()->setTOP (tmp_cid, newtop);
						 mac_->getMap()->getUlSubframe()->getBw_req()->setCODE (tmp_cid, newcode);
						 mac_->getMap()->getUlSubframe()->getBw_req()->setTIMEOUT (tmp_cid, newtimeout);
						 mac_->getMap()->getUlSubframe()->getBw_req()->setNBRETRY (tmp_cid, update_retry);
						 mac_->getMap()->getUlSubframe()->getBw_req()->setFLAGTRANSMIT (tmp_cid, 0);
	
						 debug10 ("\tBW-REQ.3.updatebackoff (0 TIMEOUT) CID :%d, oldbackoff :%d, newbackoff :%d, old_timeout :%d, new_timeout :%d, old_nbretry :%d, new_nbretry :%d, old_win :%d, new_win :%d, code :%d\n", tmp_cid, tmp_backoff, newbackoff, tmp_timeout-1, newtimeout, tmp_nbretry, update_retry, tmp_win, update_w, tmp_code);
						 debug10 ("\t   start_backoff :%d, stop_backoff :%d, contention_size :%d, NBRETRY :%d, TIMEOUT :%d\n", s_backoff_start, s_backoff_stop, s_bw_contention_size, s_nbretry, CDMA_TIMEOUT);
						 newflagtransmit = 0;
	
						 All_Headers ah_t;
						 p_tmp->RemoveHeader(ah_t);
						 cdma_req_header_t *header_t = &ah_t.CRH.data;
	
						 ah_t.PIH.subchannel_offset = (newbackoff % s_bw_contention_size)*6;
	
						 debug10 ("\tBW-REQ.3.updateOFFSET CID :%d, #old_subchannel_offset :%d, #subchannel_offset :%d, oldtop :%d, newtop :%d, oldcode :%d, newcode :%d\n", tmp_cid, ah_t.PIH.subchannel_offset, (newbackoff%s_bw_contention_size)*6, header_t->top, (newbackoff%s_bw_contention_size), header_t->code, newcode);
						 header_t->top = newtop;
						 header_t->code = newcode;
						 p_tmp->AddHeader(ah_t);
	
	//                     tmp_top = header_t->top = newbackoff % s_bw_contention_size;
	//             	     mac_->getMap()->getUlSubframe()->getBw_req()->setTOP (tmp_cid, (int)header_t->top);
	
						 //mac_->getMap()->getUlSubframe()->getBw_req()->setPacket_P (tmp_cid, p_tmp);
						 con_tmp->setBW_REQ_QUEUE(0);
					   }
	
					} else {
					  mac_->getMap()->getUlSubframe()->getBw_req()->setTIMEOUT (tmp_cid, newtimeout);
					  debug10 ("\tBW-REQ.3.updatetimeout_perframe CID :%d, oldtimeout :%d, newtimeout :%d\n", tmp_cid, tmp_timeout, newtimeout);
					}
	
					int newbackoff_bak = newbackoff;
						if (newbackoff_bak); //quiet, compiler
					newbackoff = newbackoff - s_bw_contention_size;
					if (newbackoff < 0) newbackoff = 0;
					mac_->getMap()->getUlSubframe()->getBw_req()->setBACKOFF (tmp_cid, newbackoff);
					debug10 ("\tBW-REQ.4.updatebackoff_perframe CID :%d, oldbackoff :%d, newbackoff :%d (ori_newbackoff :%d), newtimeout :%d, newnbretry :%d, newwin :%d, top :%d, code :%d\n", tmp_cid, tmp_backoff, newbackoff, newbackoff_bak, newtimeout, update_retry, update_w, newtop, newcode);
	
					if (newbackoff == 0 ) {
					   p_tmp->RemoveHeader(ah_tmp);
					   cdma_req_header_t *header_s = &ah_tmp.CRH.data;
					   
					   header_s->code = newcode;
					   header_s->top = newtop;
					   header_s->type = 3;
					   debug10 ("\tBW-REQ.4.backoff=0 may_transmit (if flagtransmit == 0) CID :%d, #symbol :%d, #symbol_offset :%d, #subchannel :%d, #subchannel_offset :%d, code :%d, top :%d, channel_index :%d, direction :%d\n", tmp_cid, ah_tmp.PIH.num_OFDMSymbol, ah_tmp.PIH.OFDMSymbol_offset, ah_tmp.PIH.num_subchannels, ah_tmp.PIH.subchannel_offset, header_s->code, header_s->top, ah_tmp.PIH.channel_index , ah_tmp.PIH.direction);
					   debug10 ("\t   CID :%d, size :%d, time :%f, direction :%d, flagtransmit :%d\n", tmp_cid, ah_tmp.GMH.GMH.len, ah_tmp.txtime, ah_tmp.direction, newflagtransmit);
					   p_tmp->AddHeader(ah_tmp);
	
	
					   if (newflagtransmit == 0) {
						  newtimeout = CDMA_TIMEOUT-1;
						  debug10 ("\tBW-REQ.5.transmit CID :%d, #old_flagtransmit :%d, #new_flagtransmit :%d reset timeout to :%d\n", tmp_cid, newflagtransmit, 1, newtimeout);
						  newflagtransmit = 1;
						  p_tmp = p_tmp->Copy(); //otherwise next time around when we do early checks, this won't have an All_Headers on it and we'll segfault
						  p_tmp->RemoveHeader(ah_tmp);
						  p_tmp->AddHeader(ah_tmp.CRH);
						  mac_->getMap()->getUlSubframe()->getBw_req()->setFLAGTRANSMIT (tmp_cid, newflagtransmit);
						  mac_->getMap()->getUlSubframe()->getBw_req()->setTIMEOUT (tmp_cid, newtimeout);
						  mac_->getMap()->getUlSubframe()->getBw_req()->setFLAGNOWTRANSMIT (tmp_cid, 1);
						  mac_->getMap()->getUlSubframe()->getBw_req()->setCODE(tmp_cid, header_s->code);
						  mac_->getMap()->getUlSubframe()->getBw_req()->setTOP(tmp_cid, header_s->top);
						  my_burst_rng->enqueue(p_tmp->Copy());
					   }
					}
				}
	
			   }
			   t_i++;
			   con_tmp = con_tmp->next_entry();
		  }
		}
   }
///////////End CDMA_BW_REQ COUNTER-BASED

    //get the packets from the connection with the same CID
    //debug2 ("\tBurst CID=%d\n", b->getCid());
    c = mac_->getCManager ()->get_connection (b->getCid(), true);
    
    if (!c) {
      continue;
    } else{
      my_burst = b;
      allocationfound = 1;
    }
    

  }
  
  Connection *head = mac_->getCManager()->get_out_connection();
  Connection *con = head;


  //transfer packets that are not CDMA BW REQs or CDMA INIT RANGING REQs to bursts
  if ( (allocationfound == 1) && (my_burst->getB_CDMA_CODE()==0) && (my_burst->getB_CDMA_TOP()==0) ) {
    b_data = 0;
    debug10 ("In SS, Found data opportunity, Burst start time %d, Burst duration %d OFDMA Symbols, IUC %d, subchannel offset %d, CID %d\n", my_burst->getStarttime(), my_burst->getDuration(), my_burst->getIUC(),  my_burst->getSubchannelOffset(), my_burst->getCid());
    
    ConnectionType_t contype;

    for(int i=0; i<4; ++i){
      c = head;
      if(i==0) 	    contype = CONN_BASIC;
      else if(i==1) contype = CONN_PRIMARY;
      else if(i==2) contype = CONN_SECONDARY;
      else 	    	contype = CONN_DATA;

      while(c!=NULL){
	    if(c->get_category()==contype){
		  if(c->queueLength()>0 || (con->getArqStatus () != NULL && con->getArqStatus ()->arq_retrans_queue_->GetNPackets() > 0)){
		  
			debug10 ("In SS, Entering transfer_packets_for_ul_subframe, CID :%d, c->queueBytes :%d, c->queueLength :%d\n", c->get_cid(), con->queueByteLength(), c->queueLength());
			debug10 ("Check FRAG/PACK/ARG: Frag : %d, Pack :%d, ARQ: %p\n",c->isFragEnable(),c->isPackingEnable(), c->getArqStatus ()); 	

			b_data = transfer_packets_for_ul_subframe(c,my_burst);
			
			debug10 ("In SS, after transfer_packets_for_ul_subframe, c->queueBytes :%d, c-queueLength :%d, b_data :%d\n", con->queueByteLength(), c->queueLength(), b_data);
		  }
		}
	
		c = c->next_entry();
      }//end while
    }//end for
  }

}

/**
 * Create a cdma bw-req request for the given connection
 * This function may need to be updated to handle
 * incremental and aggregate requests
 * @param con The connection to check
 */
void Wimax2SSScheduler::create_cdma_request (Connection *con)
{ 
	debug10("entering (SS-S) %s create_cdma_request\n", mac_->addr_s());

  if (con->queueLength()==0 && (con->getArqStatus () == NULL || con->getArqStatus ()->arq_retrans_queue_->GetNPackets() == 0)){
    debug2 ("Create CDMA-BW-REQ, contype :%d CID :%d -- At %f in Mac %s but con->queuelength is empty => return\n", con->get_category(), con->get_cid(), NOW, mac_->addr_s());
    return; 
  }
  if (mac_->getMap()->getUlSubframe()->getBw_req()->getRequest (con->get_cid())!=NULL) {
    debug2 ("Create CDMA-BW-REQ, contype :%d CID %d -- At %f in Mac %s already pending requests => return\n", con->get_category(), con->get_cid(), NOW, mac_->addr_s());    
    return; 
  }

  debug2("Create CDMA-BW-REQ, contype :%d CID :%d -- (queue is not empty :%d) in Wimax2SSScheduler for basic :5, pri :6, sec :7, data :8\n", con->get_category(), con->get_cid(), con->queueByteLength());
    
  ContentionSlot *slot1 = getMac()->getMap()->getUlSubframe()->getBw_req ();

  int s_bw_contention_size = slot1->getSize(); //0 to 4
  if (s_bw_contention_size == 0) {
  	debug10("unable to create_cdma_request, no space allocated in UL_MAP for CDMA Requests\n");
  	return;
  }
  
  int s_nbretry = 0;
  int s_backoff_start = slot1->getBackoff_start();
  int s_backoff_stop = slot1->getBackoff_stop(); if (s_backoff_stop);//quiet, compiler
  int s_window = s_backoff_start;
  int fix_six_subchannel = 6;
//  int result = rand() % ((int)(pow (2, s_window)+1));
  int result = rand() % s_bw_contention_size;
//  int s_backoff = floor(result / s_bw_contention_size);
//BW_REQ; no backoff
  int s_backoff = 0;
  	if (s_backoff); //quiet, compiler
  int sub_channel_off = result % s_bw_contention_size;
  int s_flagtransmit = 0;
  int s_flagnowtransmit = 0;

  Ptr<Packet> p = mac_->getPacket();
  All_Headers ah;
  p->RemoveHeader(ah);

  ah.cdma = 1; //mark the packet
  ah.gm_sm = 2;
  ah.GMH.GMH.len = 6;
  ah.PIH.num_subchannels = fix_six_subchannel;
  ah.PIH.subchannel_offset = sub_channel_off*fix_six_subchannel;
  ah.PIH.num_OFDMSymbol = 1;
  ah.PIH.OFDMSymbol_offset = 0; //initial_offset;
  ah.PIH.channel_index = con->getPeerNode()->getchannel();
  ah.PIH.direction = 1;
  uint8_t header_top = sub_channel_off;

  int code_range = getMac()->macmib_->cdma_code_bw_start - getMac()->macmib_->cdma_code_bw_stop + 1;
  int c_rand = rand() % code_range;
  uint8_t header_code = (uint8_t)getMac()->macmib_->cdma_code_bw_start + (uint8_t)c_rand;
  
  debug10 (" CODE init start :%d, stop :%d, rand :%d, final code :%d\n", getMac()->macmib_->cdma_code_bw_start, getMac()->macmib_->cdma_code_bw_stop, c_rand, header_code);
  

  cdma_req_header_t *header = &ah.CRH.data;
  header->ht=1;
  header->ec=1;
  header->type = 3; //CDMA_BW_REQ
  header->top = header_top;
  header->br = getMac()->id_index;
  header->code = header_code;
  header->cid = con->get_cid();

  //in case one of the other headers is used to check cid at some point
  ah.GMH.GMH.cid = con->get_cid();
  ah.SMH.BRH.cid = con->get_cid();
  

  int req_size = con->queueByteLength();

  if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1) {
    assert ((con->getArqStatus ()->arq_retrans_queue_->GetNPackets() == 0 && con->getArqStatus ()->arq_retrans_queue_->GetNBytes() == 0)
	    || (con->getArqStatus ()->arq_retrans_queue_->GetNPackets() > 0 && con->getArqStatus ()->arq_retrans_queue_->GetNBytes() > 0));
     if ((con->getArqStatus ()->arq_retrans_queue_) && (con->getArqStatus ()->arq_retrans_queue_->GetNPackets() > 0)) {
        debug2(" ARQ Create Request: Picking retransmission queue too for sending request\n");
        req_size += con->getArqStatus ()->arq_retrans_queue_->GetNBytes ();
     }
  }

  double txtime = mac_->getPhy()->getSymbolTime();
  ah.txtime = txtime;
  p->AddHeader(ah);

  mac_->getMap()->getUlSubframe()->getBw_req()->addRequest (p, con->get_cid(), req_size, result, CDMA_TIMEOUT, s_nbretry, s_window, (int)header_code, (int)header_top, s_flagtransmit, s_flagnowtransmit);

  debug10 (" Wimax2SSScheduler enqueued cdma_bw request for cid :%d, q-len :%d, (nbPacket :%d), code :%d, top :%d, ss_id :%s, backoff :%d, window :%d, nbretry :%d, flagtransmit :%d, flagnowtransmit :%d\n", con->get_cid(), req_size, con->queueLength(), header_code, header_top, mac_->addr_s(), s_backoff, s_window, s_nbretry, s_flagtransmit, s_flagnowtransmit);

  debug10 (" symbol_offset :%d, symbol :%d, subchannel_offset :%d, subchannel :%d, channel_index :%d, direction :%d, txtime :%e\n", ah.PIH.OFDMSymbol_offset, ah.PIH.num_OFDMSymbol, ah.PIH.subchannel_offset, ah.PIH.num_subchannels, ah.PIH.channel_index , ah.PIH.direction, ah.txtime);
}




/*
	This function transfers packets from the queue in a connection to a burst.
	If that connection supports fragmentation and/or packing, those are employed
	to maximize utilization of granted bandwidth.

	Upon completion of this function, any transfered packets will have been removed
	from the Connection's queue and added in final form (all headers correct) to the
	Burst.  The return value is the total number of bytes transfered.

	The following is an outline of what specifically happens:

	1. If there are no packets in the Connection's queue, return
		1.A If there is only room for one MAC header and more than that in the queue, send a BRH


	2. If Fragmentation and Packing are both disabled, add the first packet in
	   the connection queue if possible, then return
	

	3. If Fragmentation is Enabled and Packing is not:
		3.A If the entire first packet will fit, add it and return
		3.B If the entire first packet will not fit, add the part that does
	
	4. If Packing is Enabled and Fragmentation is not:
		4.A If the entire first packet will not fit, return
		4.B Add the first packet
		4.C remove that packet from the connection, so the next is now in position 1
		4.D Goto 4.A
	
	5.  If Packing and Fragmentation are BOTH enabled:
		5.A If the entire first packet will not fit, add what will and return
		5.B If the entire first packet fits and there is not enough froom for another, add it and return
		5.C Add the first packet entirely and remove it from the connection queue, so the next is now in position 1
		5.D Goto 5.A

	Note that this function is identical to the DL Subframe equivilent, except that if all
	data in the Connection's queue cannot be transfered, we add a grant management subheader
	to request enough bandwidth to tranmit the remaining data
*/

uint32_t Wimax2SSScheduler::transfer_packets_for_ul_subframe(Connection *c, Burst *b) {
	//1. If there are no packets in the Connection's queue, return
	if (c->queueLength() == 0) return 0;


	Ofdm_mod_rate  modulation;
	if (b->getIUC() == UIUC_CDMA_ALLOCATION_IE)
		modulation = getMac()->getMap()->getUlSubframe()->getProfile(((UlBurst*)b)->getExtendedUIUC())->getEncoding();
	else 
		modulation = getMac()->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding();
	//in UL_MAP decoded bursts, the duration field is total number of subchannels (which may be mapped accross many rows in the allocation grid)
	uint32_t remaining_bytes_in_burst = getMac()->getPhy()->getMaxPktSize( b->getDuration(), modulation, UL_);

	All_Headers ah_br;
	
	//if we only have an allocation for one MAC Header, and the contents of the queue are longer than that, send it instead
	if (remaining_bytes_in_burst >= ah_br.SMH.GetSerializedSize() && remaining_bytes_in_burst < ah_br.GMH.GetSerializedSize() + ah_br.GMSH.GetSerializedSize() + 1) {
		//we make sure what's in the queue isn't just this short
		Ptr<Packet> p = c->dequeue();
		p->RemoveHeader(ah_br);
		//if there's nothing left in the packet other than the headers, and we have a signaling mac header of some kind (since a GMH is meaningless with no data)...
		if (p->GetSize() == 0 && ah_br.GMH.GMH.ht == 1) {
			if (ah_br.SMH.BRH.type == 0) { //if we already have a BRH in the connection queue for some reason...
				p->AddHeader(ah_br.SMH);
				b->enqueue(p);
				return ah_br.SMH.GetSerializedSize();
			} else { //it's a CDMA packet of some kind...
				p->AddHeader(ah_br.CRH);
				b->enqueue(p);
				return ah_br.CRH.GetSerializedSize();
			}
		} else { //otherwise shove a BW-REQ header in and send it
			p->AddHeader(ah_br);
			c->enqueue_head(p);
			ah_br.SMH.BRH.ht = 1;
			ah_br.SMH.BRH.ec = 0;
			ah_br.SMH.BRH.type = 0;
			ah_br.SMH.BRH.br = calculate_needed_bandwidth(c);
			ah_br.SMH.BRH.cid = c->get_cid();
			ah_br.SMH.BRH.hcs = 0;
			Ptr<Packet> req = Create<Packet>();
			req->AddHeader(ah_br.SMH);
			b->enqueue(req);
			return ah_br.SMH.GetSerializedSize();

		}
		
		
	}

	//2. If Fragmentation and Packing are both disabled, add the first packet in
	//   the connection queue if possible, then return
	
	if ((! c->isFragEnable() ) &&  (! c->isPackingEnable() )) {
		Ptr<Packet> p = c->dequeue();
		All_Headers ah;
		p->RemoveHeader(ah);
		
		//if our one packet won't fit...
		if (ah.GMH.GMH.len > remaining_bytes_in_burst) {
			p->AddHeader(ah);
			c->enqueue_head(p);
			Signaling_MAC_Header smh;
			smh.BRH.ht = 1;
			smh.BRH.ec = 0;
			smh.BRH.type = 0; //to define as a BW_REQ packet
			smh.BRH.br = calculate_needed_bandwidth(c);
			smh.BRH.cid = c->get_cid();
			smh.BRH.hcs = 0;
			Ptr<Packet> req = Create<Packet>();
			req->AddHeader(smh);
			b->enqueue(req);

			return smh.GetSerializedSize();
		}
		//otherwise, add it
	
		//since Fragmentation is disabled, this cannot be a later portion of a packet
		p->AddHeader(ah.GMH);
		b->enqueue(p);
		return ah.GMH.GMH.len;
	}


	//3. If Fragmentation is Enabled and Packing is not:
	//	3.A If the entire first packet will fit, add it and return
	//	3.B If the entire first packet will not fit, add the part that does

	if (( c->isFragEnable() ) &&  (! c->isPackingEnable() )) {
		Ptr<Packet> p = c->dequeue();
		All_Headers ah;
		p->RemoveHeader(ah);
		

		//if we've already fragmented this packet, we'll need a fragmentation subheader
		//this is taken into account for total packet size here to ensure we accurately
		//check if the packet can fit unfragmented or not
		if (ah.GMH.GMH.type_frag == 1) {
			ah.GMH.GMH.len += ah.FSH.GetSerializedSize();
		}
		//if we have more packets in the queue we're going to need more bw next frame
		if (c->queueLength() > 0) {
			ah.GMH.GMH.type_fbgm = 1;
			remaining_bytes_in_burst -= ah.GMSH.GetSerializedSize();
		}
		
		//if our one packet won't fit, fragment it
		if (ah.GMH.GMH.len > remaining_bytes_in_burst) {
			//fix the remaining_bytes_in_burst to the accurate measure, we take the size of the header into account in the logic below
			if (c->queueLength() > 0) remaining_bytes_in_burst += ah.GMSH.GetSerializedSize();
			//if we're fragmenting, we're going to need more bandwidth next time
			ah.GMH.GMH.type_fbgm = 1;
			//set up the fragmentation header

			//if we're already fragmented
			if (ah.GMH.GMH.type_frag == 1) {
				ah.FSH.data.fc = FRAG_CONT;
			} else {
				ah.FSH.data.fc = FRAG_FIRST;
			}
			ah.FSH.data.fsn = 0; //used by ARQ, not currently supported
			ah.FSH.data.sn = c->incrementFragNumber();

			//Fragment the packet
			uint32_t new_packet_size = remaining_bytes_in_burst - ah.FSH.GetSerializedSize() - ah.GMH.GetSerializedSize() - ah.GMSH.GetSerializedSize();
			Ptr<Packet> transmission_fragment = p->CreateFragment(0, new_packet_size);

			//set the fields in the GMH to the correct values
			ah.GMH.GMH.type_frag = 1;
			ah.GMH.GMH.len = remaining_bytes_in_burst;

			//add the leftover fragment back to the queue
			p->RemoveAtStart(new_packet_size);
			ah.GMH.GMH.len = p->GetSize() + ah.GMH.GetSerializedSize();
			p->AddHeader(ah);
			c->enqueue_head(p);

			//calculate bw needed to finish this packet next frame
			ah.GMSH.data.piggyback_req = calculate_needed_bandwidth(c);

			//add the headers and enqueue
			transmission_fragment->AddHeader(ah.FSH);
			if (ah.GMH.GMH.type_fbgm == 1) transmission_fragment->AddHeader(ah.GMSH);
			ah.GMH.GMH.len = transmission_fragment->GetSize() + ah.GMH.GetSerializedSize();
			ah.GMH.GMH.type_frag = true;
			ah.GMH.GMH.type_pck = false;
			transmission_fragment->AddHeader(ah.GMH);
			b->enqueue(transmission_fragment);
			return remaining_bytes_in_burst;
		}
		//otherwise, add it whole


		 //if we're already fragmented, but we can fit the rest of the packet here
		if (ah.GMH.GMH.type_frag == 1) {
			ah.FSH.data.fc = FRAG_LAST;
			ah.FSH.data.fsn = 0;
			ah.FSH.data.sn = c->getFragNumber();
			c->incrementFragNumber();
			p->AddHeader(ah.FSH);
		}
		
		//if we're piggybacking a bw_request, add that here
		if (ah.GMH.GMH.type_fbgm == 1) {
			ah.GMSH.data.piggyback_req = calculate_needed_bandwidth(c);
			p->AddHeader(ah.GMSH);
		}
		
		p->AddHeader(ah.GMH);
		b->enqueue(p);
		return ah.GMH.GMH.len;
	}

	//4. If Packing is Enabled and Fragmentation is not:
	//	4.A If the entire first packet will not fit, return
	//	4.B Add the first packet
	//	4.C remove that packet from the connection, so the next is now in position 1
	//	4.D Goto 4.A

	if ((! c->isFragEnable() ) &&  ( c->isPackingEnable() )) {
		Ptr<Packet> p = c->dequeue();
		All_Headers ah;
		p->RemoveHeader(ah);
		
		//if our first packet won't fit,  request bandwidth.  can't piggyback here since there's nothing else being transmitted
		if (ah.GMH.GMH.len > remaining_bytes_in_burst) {
			p->AddHeader(ah);
			c->enqueue_head(p);
			Signaling_MAC_Header smh;
			smh.BRH.ht = 1;
			smh.BRH.ec = 0;
			smh.BRH.type = 0; //to define as a BW_REQ packet
			smh.BRH.br = calculate_needed_bandwidth(c);
			smh.BRH.cid = c->get_cid();
			smh.BRH.hcs = 0;
			Ptr<Packet> req = Create<Packet>();
			req->AddHeader(smh);
			b->enqueue(req);

			return smh.GetSerializedSize();
		}

		//we can fit at least one packet in, determine if we can fit more than one, so we know to take Packing Subheaders into account for size
		Ptr<Packet> p_next = c->dequeue();
		All_Headers ah_next;
		if (p_next != 0) {
		      p_next->PeekHeader(ah_next);
		      c->enqueue_head(p_next);
		}
	
		//if there isn't another packet, transfer this one to the burst and do not request more bandwidth
		if (p_next == 0) {
			ah.GMH.GMH.type_pck = 0;
			p->AddHeader(ah.GMH);
			b->enqueue(p);
			return ah.GMH.GMH.len;
		}

		//if we can only fit our first packet in the burst, add it and return
		if (ah.GMH.GMH.len + ah.GMH.GMH.len + 2 * ah.PSH.GetSerializedSize() > remaining_bytes_in_burst - ah.GMSH.GetSerializedSize()) {
			ah.GMH.GMH.type_pck = 0;
			ah.GMH.GMH.type_fbgm = 1;
			ah.GMSH.data.piggyback_req = calculate_needed_bandwidth(c);
			p->AddHeader(ah.GMSH);
			p->AddHeader(ah.GMH);
			b->enqueue(p);
			return p->GetSize();
		}
	
		//we are going to pack two or more packets into this burst
		Ptr<Packet> p_burst= getMac()->getPacket();
		All_Headers ah_burst;
		p_burst->RemoveHeader(ah_burst);

		ah_burst.GMH.GMH.type_pck = 1;
		remaining_bytes_in_burst -= ah.GMH.GetSerializedSize();
		

		while ( ah.GMH.GMH.len - ah.GMH.GetSerializedSize() + ah.PSH.GetSerializedSize() <= remaining_bytes_in_burst) {
			//set up the packing subheader
			ah.PSH.data.fc = FRAG_NOFRAG;
			ah.PSH.data.sn = c->incrementPackNumber();
			ah.PSH.data.length = ah.GMH.GMH.len - ah.GMH.GetSerializedSize() + ah.PSH.GetSerializedSize();
			p->AddHeader(ah.PSH);
			p_burst->AddAtEnd(p);
			remaining_bytes_in_burst -= ah.PSH.data.length;


			//get the next packet to examine
			p= c->dequeue();
			//if we're out of packets, stop looping
			if (p== 0) break;
			//otherwise, pull off the All_Headers to get size information for next loop
			p->RemoveHeader(ah);
		}

		//if p_next isn't null, then the last packet dequeued from the connection couldn't fit, so we need to add it back on
		//so that it will be sent in a future frame
		if (p != 0) {
			p->AddHeader(ah);
			c->enqueue_head(p);
			ah_burst.GMH.GMH.type_fbgm = 1;
			ah_burst.GMSH.data.piggyback_req = calculate_needed_bandwidth(c);
		}


		//Generate the GMH for the accumulator packet that we've been copying everything into
		ah_burst.GMH.GMH.type_pck = 1;
		ah_burst.GMH.GMH.len = p_burst->GetSize() + ah_burst.GMH.GetSerializedSize();
		ah_burst.GMH.GMH.cid = c->get_cid();
		if (ah_burst.GMH.GMH.type_fbgm == 1) p_burst->AddHeader(ah.GMSH);
		p_burst->AddHeader(ah_burst.GMH);
		b->enqueue(p);
		return p_burst->GetSize();
	}


	//5.  If Packing and Fragmentation are BOTH enabled:
	//	5.A If the entire first packet will not fit, add what will and return
	//	5.B If the entire first packet fits, add it and remove from the queue so we have a new first, go to 5.A
	//	5.C If part of the first packet fits, add what will
	//	5.D Set up the accumulator packet's header (GMH for the whole burst), and return

	//note that if we apply both fragmentation and packing to the same burst, we only use packing subheaders.  the fc field on these
	//works the same as the fc field on the fragmentation subheaders, so using both is not necessary.
	if (c->isFragEnable() && c->isPackingEnable()) {
		Ptr<Packet> p = c->dequeue();
		All_Headers ah;
		p->RemoveHeader(ah);
		

		//cases for only including part of the first packet:
		//  1. The burst is too small for even the first packet
		//  2. The burst is just large enough for that whole packet, but there are more and we need room for a piggybacked bw_req
		//if our first packet won't fit, we'll fragment it
		//since part of the first fragment takes the whole burst, we won't have room to pack more in
		if (ah.GMH.GMH.len > remaining_bytes_in_burst || (c->queueLength() > 1 && ah.GMH.GMH.len + ah.GMSH.GetSerializedSize() > remaining_bytes_in_burst )) {
			//set up the fragmentation header

			//if we're already fragmented
			if (ah.GMH.GMH.type_frag == 1) {
				ah.FSH.data.fc = FRAG_CONT;
			} else {
				ah.FSH.data.fc = FRAG_FIRST;
			}
			ah.FSH.data.fsn = 0; //used by ARQ, not currently supported
			ah.FSH.data.sn = c->incrementFragNumber();

			//Fragment the packet
			int new_packet_size =  remaining_bytes_in_burst - ah.FSH.GetSerializedSize() - ah.GMH.GetSerializedSize() - ah.GMSH.GetSerializedSize();
			Ptr<Packet> transmission_fragment = p->CreateFragment(0, new_packet_size);

			//set the fields in the GMH to the correct values
			ah.GMH.GMH.type_frag = 1;
			ah.GMH.GMH.type_pck = 0;
			ah.GMH.GMH.type_fbgm = 1; //we know we need more BW, we can't even fit this whole packet...
			ah.GMH.GMH.len = remaining_bytes_in_burst;
			
			//add the leftover part of this packet back to the queue
			p->RemoveAtStart(new_packet_size);
			ah.GMH.GMH.len = p->GetSize() + ah.GMH.GetSerializedSize();
			p->AddHeader(ah);
			c->enqueue_head(p);
			
			//calculate the bw needed for next frame
			ah.GMSH.data.piggyback_req = calculate_needed_bandwidth(c);

			//add the headers and enqueue
			transmission_fragment->AddHeader(ah.FSH);
			transmission_fragment->AddHeader(ah.GMSH);
			transmission_fragment->AddHeader(ah.GMH);
			b->enqueue(transmission_fragment);

			return p->GetSize();
		}

		//if we make it this far, we know that we can fit at least one full packet
		//we now want to loop on adding new packets to the burst until either we 
		//run out of room in the burst or out of packets in the Connection's queue

		//before we loop we check how much data we do have.  if it's enough for the first packet and no 
		//others (or if there's only one packet, which has already been dequeued), we add it and return
		//without adding packing subheader(s)
		uint32_t bytes_to_next_payload = ah.GMH.GMH.len + ah.PSH.GetSerializedSize() * 2;
		if (remaining_bytes_in_burst <= bytes_to_next_payload || c->queueLength() == 0) {
			debug10("There is only room for one packet in the burst, so we're adding that\n");
			//just add the first packet
			if (ah.GMH.GMH.type_frag == 1) {
				ah.FSH.data.fc = FRAG_LAST;
				ah.FSH.data.fsn = 0;
				ah.FSH.data.sn = c->incrementFragNumber();
				p->AddHeader(ah.FSH);
			}
			
			if (c->queueLength() > 0) {
			    ah.GMH.GMH.type_fbgm = 1;
			    ah.GMSH.data.piggyback_req = calculate_needed_bandwidth(c);
			    p->AddHeader(ah.GMSH);
			    ah.GMH.GMH.len += 2;
			}

			p->AddHeader(ah.GMH);
			b->enqueue(p);
			return p->GetSize();
		}

		//looks like we're going to pack a few in

		Ptr<Packet> p_burst = getMac()->getPacket();
		All_Headers ah_burst;
		p_burst->RemoveHeader(ah_burst);

		ah_burst.GMH.GMH.type_pck = 1;
		ah_burst.GMH.GMH.type_frag = 0;
		remaining_bytes_in_burst -= ah.GMH.GetSerializedSize();
		bytes_to_next_payload -= ah.GMH.GetSerializedSize();
		bool partial_packet_left = true;
		bool first_loop = true;
		debug10("Packing several packets into burst\n");
		while (remaining_bytes_in_burst > bytes_to_next_payload || (c->queueLength() > 0 && remaining_bytes_in_burst > bytes_to_next_payload + ah.GMSH.GetSerializedSize())) {
			if (first_loop && ah.GMH.GMH.type_frag == 1) {
				ah.PSH.data.fc = FRAG_LAST;
				ah_burst.GMH.GMH.type_frag = 1;
			} else {
				ah.PSH.data.fc = FRAG_NOFRAG;
			}
			first_loop = false;

			ah.PSH.data.sn = c->incrementPackNumber();
			ah.PSH.data.length = p->GetSize() + ah.PSH.GetSerializedSize();
			p->AddHeader(ah.PSH);
			p_burst->AddAtEnd(p);
			debug10("\t Added packet of size %d (including packing subheader) to burst\n", ah.PSH.data.length);
			remaining_bytes_in_burst -= p->GetSize();

			if (c->queueLength() > 0) p = c->dequeue();
			else {
				partial_packet_left = false;
				break;
			}

			p->RemoveHeader(ah);
			bytes_to_next_payload = p->GetSize() + ah.PSH.GetSerializedSize();
		}

		//we've added as many full packets as we can.  check if we can add part of this one still
		if (partial_packet_left && p != 0) {
			if (remaining_bytes_in_burst > ah.PSH.GetSerializedSize()) {
				ah_burst.GMH.GMH.type_fbgm = 1; //if we're fragmenting we need more data next time to finish this packet (at least)
				if (ah.GMH.GMH.type_frag == 1) {
					ah.PSH.data.fc = FRAG_CONT;
					ah_burst.GMH.GMH.type_frag = 1;
				} else {
					ah.GMH.GMH.type_frag = 1;
					ah.PSH.data.fc = FRAG_FIRST;
					ah_burst.GMH.GMH.type_frag = 1;
				}
				ah.PSH.data.sn = c->incrementFragNumber();
				ah.PSH.data.length = remaining_bytes_in_burst - ah.GMSH.GetSerializedSize(); //as the bw-req still has to be added


				int new_packet_size = remaining_bytes_in_burst - ah.PSH.GetSerializedSize() - ah.GMSH.GetSerializedSize();
				if (new_packet_size > 0) {
					Ptr<Packet> transmission_fragment = p->CreateFragment(0, new_packet_size);

					transmission_fragment->AddHeader(ah.PSH);
					p_burst->AddAtEnd(transmission_fragment);
					debug10("\t Added packet fragment of size %d (of %d) (including packing subheader) to burst\n", ah.PSH.data.length, p->GetSize());


					p->RemoveAtStart(new_packet_size);
					ah.GMH.GMH.len = p->GetSize() + ah.GMH.GetSerializedSize();
					p->AddHeader(ah);
					c->enqueue_head(p);
				}
			} else {
				//we can't fit any of the packet, so just shove it back in the queue
				debug10("\t Could not fit even a fragment of the next packet into the burst.  Placing that packet at the head of the queue\n");
				p->AddHeader(ah);
				c->enqueue_head(p);
			}
		}
		
		if (c->queueLength() > 0) ah_burst.GMH.GMH.type_fbgm = 1; //if there's still data on the Connection's queue, we need more data next frame
		if (ah_burst.GMH.GMH.type_fbgm == 1) {
		    ah_burst.GMSH.data.piggyback_req = calculate_needed_bandwidth(c);
		    p_burst->AddHeader(ah_burst.GMSH);
		}
		//set up the GMH for the accumulator packet

		ah_burst.GMH.GMH.len = p_burst->GetSize() + ah_burst.GMH.GetSerializedSize();
		ah_burst.GMH.GMH.cid = c->get_cid();
		p_burst->AddHeader(ah_burst.GMH);
		b->enqueue(p_burst);
		return p_burst->GetSize();
	}
	//should really never get here...
	return 0;
	

}

//This is a helper function to calculate the amount of banwdith to request if we are short
//It will always put an extra two bytes on the end of the request to fit in a subsequent
//piggybacked bandwidth request
uint16_t Wimax2SSScheduler::calculate_needed_bandwidth(Connection *c) {
	if (c->queueLength() == 0) return 0;
	Ptr<Packet> first = c->dequeue();
	c->enqueue_head(first);
	first = first->Copy();
	All_Headers ah;
	first->RemoveHeader(ah);

	if (c->queueLength() == 1 || !c->isPackingEnable()) {
		if (c->isFragEnable() && ah.GMH.GMH.type_frag == 1) return ah.GMH.GMH.len + ah.FSH.GetSerializedSize() + ah.GMSH.GetSerializedSize();
		else return ah.GMH.GMH.len + ah.GMSH.GetSerializedSize();
	} else { //we'll need packing subheaders on each packet
		uint16_t total_required = ah.GMH.GetSerializedSize() + ah.GMSH.GetSerializedSize();
		for (int i = 0; i < c->queueLength(); i++) {
		     first = c->queueLookup(i)->Copy();
		     first->RemoveHeader(ah);
		     total_required += first->GetSize() + ah.PSH.GetSerializedSize();
		}
		return total_required;
	}
	
	//if we somehow get here...
	return 0;
}

}//namespace
