/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef SSSCHEDULER_H
#define SSSCHEDULER_H

#include "wimax2-scheduler.h"

#define MAXCODE 256
#define MIN_CONTENTION_SIZE 5
#define CDMA_TIMEOUT 2

namespace ns3 {
/**
 * Class SSscheduler
 * Scheduler for SSs
 */ 
class Wimax2SSScheduler : public Wimax2Scheduler {
  friend class Wimax2SSNetDevice;
 public:
  /**
   * Create a scheduler
   */
  Wimax2SSScheduler ();



  /**
   * Initializes the scheduler
   */
  virtual void init ();

  /**
   * This function is used to schedule bursts/packets
   */
  virtual void schedule ();

 protected:

  /**
   * Create a request for the given connection
   * @param con The connection to check
   */
  void create_request (Connection *con);

//Chakchai setnd cdma_request
  void create_cdma_request (Connection *con);

  /**
   * Transfer packets from the connection to the burst
   * This is specific to the UL Subframe becuase if there isn't enough room in
   * the burst, it will add a grant management subheader (piggybacked bw request)
   * @param c The connection packets are currently enqueued to
   * @param b The burst to which we move the packets as we dequeue them from c
   */
uint32_t transfer_packets_for_ul_subframe(Connection *c, Burst *b);

 private:

 //helper function to determine how much bandwidth to request for transmitting packets from the given connection
uint16_t calculate_needed_bandwidth(Connection *c);
};

}//namespace

#endif //SSSCHEDULER_H

