/**
	SDUClassifier class serves as a base class from which to derive flow classifiers
	note that this class is abstract, see DestClassifier for a simple implementation
*/

#ifndef SDUCLASSIFIER_H
#define SDUCLASSIFIER_H

#include "ns3/packet.h"
#include "ns3/address.h"
#include "ns3/mac48-address.h"
#include "ns3/wimax2-common.h"
#include <sys/queue.h>
//#include "wimax2-net-device.h"

namespace ns3 {

class SDUClassifier;
class Wimax2NetDevice;
LIST_HEAD (sduClassifier_list, SDUClassifier);

class mac802_16;

struct sdulink {
		struct SDUClassifier *le_next;
		struct SDUClassifier **le_prev;
};

class SDUClassifier: public ns3::Object {
	friend class Wimax2NetDevice;

public:	
	SDUClassifier();
	 // -#- @head(transfer_ownership=true) -#-
	SDUClassifier(SDUClassifier *head);
	inline int getPriority() {return priority_;}
	inline void setPriority(int pri) {priority_ = pri;}
	
	//get the CID for a given MAC address
	virtual int classify(Ptr<Packet> p, const Address & dest) = 0;
	//get the MAC address for a given CID
	virtual Mac48Address declassify(int cid) = 0;
	
	
	 // -#- @head(transfer_ownership=true) -#-
	inline void insert_entry_head(struct sduClassifier_list *head) { LIST_INSERT_HEAD(head, this, link); }
	inline void insert_entry(struct SDUClassifier *elem) {LIST_INSERT_AFTER(elem, this, link);}
	SDUClassifier* next_entry(void) const {return link.le_next;}
	void remove_entry() {LIST_REMOVE(this, link);}

//mac_, LIST_ENTRY macro, and priority_ are public instead of protected so python bindings scanning works
	Ptr<Wimax2NetDevice> mac_;
	int priority_;
//	LIST_ENTRY(SDUClassifier) link;
	sdulink link;
protected:
	void setMac(Ptr<Wimax2NetDevice> mac);
	virtual void DoDispose();
};

}
#endif
