
#ifndef WIMAX2_CHANNEL_H
#define WIMAX2_CHANNEL_H

#include "ns3/channel.h"
#include "ns3/mac48-address.h"
#include <vector>
#include "wimax2-net-device.h"
#include "wimax2-phy.h"
#include "ns3/cost231-propagation-loss-model.h"
#include "ns3/propagation-loss-model.h"

namespace ns3 {

//class Wimax2NetDevice;
class Packet;

/**
 * \ingroup channel
 * \brief A wimax channel, for wimax things and testing
 */
class Wimax2Channel : public Channel
{
public:
  static TypeId GetTypeId (void);
  Wimax2Channel ();

  //This transmits a single block in the OFDMA allocation grid (1 subchannl by 1 slot
  //of 2 or 3 symbol durations dependin on whether this is happening in the UL or DL
  //subframe
  void Send (Ptr<Packet> p, Ptr<Wimax2Phy> sender, uint8_t subchannel);

  void Add (Ptr<Wimax2Phy> device);
  inline void Add (Ptr<Wimax2NetDevice> device) {Add(device->getPhy());}

  // inherited from ns3::Channel
  virtual uint32_t GetNDevices (void) const;
  virtual Ptr<NetDevice> GetDevice (uint32_t i) const;
  virtual Ptr<Wimax2Phy> GetPhy(uint32_t i) const;

  virtual void resetCollisionMonitor();
  virtual void initializeCollisionDetection();

  virtual void sync_preamble();

protected:
  virtual void DoDispose();

private:
  std::vector<Ptr<Wimax2Phy> > m_devices;
  double freq_, lambda_;

  Ptr<PropagationLossModel> m_loss;
  /**
   * The frequency bandwidth (Hz)
   */
  double fbandwidth_;

  bool **collisions_;
  uint32_t num_dl_symbol, num_ul_symbol, num_dl_subch, num_ul_subch;
  Time frame_duration;
  bool isInitialized;
  
};

} // namespace ns3

#endif /* WIMAX_CHANNEL_H */
