#include <stdlib.h>
#include "ns3/mac48-address.h"
#include "ns3/log.h"
#include <sstream>
#include <sys/types.h>
#include <inttypes.h>
#include <stdio.h>
#include <cassert>
#include "ns3/object.h"


#ifndef WIMAX2_CMN
#define WIMAX2_CMN


namespace ns3 {
#define BS_NOT_CONNECTED -1 //bs_id when MN is not connected

#define DL_PREAMBLE 1  					//preamble
#define INIT_RNG_PREAMBLE 0         	// no preamble for ranging
#define BW_REQ_PREAMBLE 0           	// no preamble for bw req

#define BASE_POWER  1e-18  //rpi
#define DOUBLE_INT_ZERO 1e-30

#define HDR_MAC802_16_SIZE 6
#define HDR_MAC802_16_FRAGSUB_SIZE 2


#define DEBUG_WIMAX
//#define DEBUG_WIMAX_EXT
//#define DEBUG_WIMAX_HEADER_SERIALIZATIONS
//#define DEBUG_WIMAX_TRANS


//Define new debug function for cleaner code
#ifdef DEBUG_WIMAX
#define debug2 printf 
#else
#define debug2(arg1,...) 
#endif

//Define additional debug; level 10 (scheduling and ranging details)
#ifdef DEBUG_WIMAX_EXT
#define debug10 printf
#else
#define debug10(arg1,...) 
#endif

//Define yet another debug, level 15 (transmission details)
#ifdef DEBUG_WIMAX_TRANS
#define debug15 printf
#else
#define debug15(arg1,...) 
#endif

#ifdef DEBUG_WIMAX_HEADER_SERIALIZATIONS
#define debugHeader printf
#else
#define debugHeader(arg1,...)
#endif

#define 	TIMER_IDLE   0
#define 	TIMER_PENDING   1

#define NOW Simulator::Now ().GetSeconds ()
//#define cout std::cout
using std::cout;

#define VERTICAL_STRIPPING 1
#define HORIZONTAL_STRIPPING 0
#define GENERIC_HEADER_SIZE 6

//Determine if the environment is 32 or 64 bit
// Check windows
#if _WIN32 || _WIN64
#if _WIN64
#define ENVIRONMENT64
#else
#define ENVIRONMENT32
#endif
#endif

// Check GCC
#if __GNUC__
#if __x86_64__ || __ppc64__
#define ENVIRONMENT64
#else
#define ENVIRONMENT32
#endif
#endif


//enum  	dir_t { DIR_DOWN = -1, DIR_NONE = 0, DIR_UP = 1 }

/** Definition of supported rate */
enum Ofdm_mod_rate {
  OFDM_BPSK_1_2,   /* Efficiency is 1 bps/Hz */
  OFDM_QPSK_1_2,   /* Efficiency is 2 bps/Hz */
  OFDM_QPSK_3_4,   /* Efficiency is 2 bps/Hz */
  OFDM_16QAM_1_2,  /* Efficiency is 4 bps/Hz */
  OFDM_16QAM_3_4,  /* Efficiency is 4 bps/Hz */
  OFDM_64QAM_2_3,  /* Efficiency is 6 bps/Hz */
  OFDM_64QAM_3_4   /* Efficiency is 6 bps/Hz */
};

/** Define fragmentation encoding */
enum fragment_status {
  FRAG_NOFRAG = 0,
  FRAG_LAST,
  FRAG_FIRST,
  FRAG_CONT
};

/** Define Mac management type */
enum mngmt_type {
  MAC_UCD = 0,			//0
  MAC_DCD,			//1
  MAC_DL_MAP,			//2
  MAC_UL_MAP,			//3
  MAC_RNG_REQ,			//4
  MAC_RNG_RSP,			//5
  MAC_REG_REQ,			//6
  MAC_REG_RSP,			//7
  // 8: reserved
  // 9-10: Privacy key, not used 
  MAC_DSA_REQ=11,		//11
  MAC_DSA_RSP,			//12
  MAC_DSA_ACK,			//13
  
  MAC_MOB_NBR_ADV=53,	//53
  MAC_MOB_SCN_REQ,		//54
  MAC_MOB_SCN_RSP,		//55
  MAC_MOB_BSHO_REQ,		//56
  MAC_MOB_MSHO_REQ,		//57
  MAC_MOB_BSHO_RSP,		//58
  MAC_MOB_HO_IND,		//59
  MAC_MOB_SCN_REP,		//60
  MAC_MOB_ASC_REP=66	//66
};

/** Direction of the flow */
enum dir_t {
    NONE,  // flow not used
    DL,    // downlink flow
    UL     // uplink flow
};

/** Enumeration of scanning types */
enum wimax_scanning_type {
  SCAN_WITHOUT_ASSOC,
  SCAN_ASSOC_LVL0,
  SCAN_ASSOC_LVL1,
  SCAN_ASSOC_LVL2
};

/** Ranging status */
enum ranging_status {
  RNG_CONTINUE = 1,
  RNG_ABORT,
  RNG_SUCCESS,
  RNG_RERANGE
};

/** Define the ID for each timer **/
enum timer_id {
  WimaxFrameTimerID,
  WimaxRxTimerID,
  WimaxDCDTimerID,
  WimaxUCDTimerID,
  WimaxRngIntTimerID,
  WimaxLostDLMAPTimerID,
  WimaxLostULMAPTimerID,
  WimaxT1TimerID,
  WimaxT2TimerID,
  WimaxT3TimerID,
  WimaxT6TimerID,
  WimaxT9TimerID,
  WimaxT12TimerID,
  WimaxT16TimerID,
  WimaxT17TimerID,
  WimaxT21TimerID,
  WimaxT44TimerID,
  //mobility extension
  WimaxMobNbrAdvTimerID,

  WimaxScanIntervalTimerID,
  WimaxRdvTimerID
};


class HelperFunctions {
public:
	static std::string getMacAddrString(Mac48Address adr);
	static const char* getPacketTypeString(int gm_sm);

};

}//namespace ns3

#endif
