#include "ns3/wimax2-helper.h"
#include <iostream>
#include "ns3/simulator.h"
#include "ns3/packet.h"
#include "ns3/log.h"
#include "ns3/config.h"
#include "wimax2-net-device.h"
#include "wimax2-channel.h"
#include "wimax2-phy.h"
#include "ns3/node.h"
#include "ns3/packet.h"
#include "ns3/log.h"
#include "ns3/pointer.h"
#include "ns3/error-model.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/wimax2-scheduler.h"
#include "ns3/wimax2-bs-net-device.h"
#include "ns3/wimax2-ss-net-device.h"
#include "framemap.h"



namespace ns3 {

NS_LOG_COMPONENT_DEFINE ("Wimax2Helper");


Wimax2Helper Wimax2Helper::Default(void) {
	Wimax2Helper helper;
	Mac802_16MIB *mmib = new Mac802_16MIB();
	Phy802_16MIB *pmib = new Phy802_16MIB();

	pmib->rtg = 20;
	pmib->ttg = 20;
	pmib->fbandwidth = 10e6;

	mmib->frame_duration = 0.005;
	mmib->ITU_PDP_ = 2;
	helper.setDlRatio(0.3);

	helper.setMacMIB(mmib);
	helper.setPhyMIB(pmib);

	return helper;
}



NetDeviceContainer Wimax2Helper::Install (NodeContainer nodes, Ptr<Wimax2Channel> channel)
{
     NetDeviceContainer devices;
     int nodeCount = 0;
     for (NodeContainer::Iterator i = nodes.Begin (); i != nodes.End (); ++i)
       {
         Ptr<Wimax2NetDevice> dev;
		 dev = CreateObject<Wimax2SSNetDevice>();
         dev->SetAddress (Mac48Address::Allocate ());
         debug2 ("Creating a NetDevice with MAC Address: %s and id_index: %d\n",dev->addr_s(), dev->getIdIndex());
         dev->SetChannel (GetPointer(channel));
         dev->setPhyMIB(phymib);
         dev->setMacMIB(macmib);
         dev->SetSchedulerDlRatio(dlratio_init);
         (*i)->AddDevice (dev);
         devices.Add (dev);
         nodeCount++;

       }
     return devices;
}


NetDeviceContainer Wimax2Helper::InstallBS (NodeContainer nodes, Ptr<Wimax2Channel> channel) {
	NetDeviceContainer devices;
	int nodeCount = 0;

     for (NodeContainer::Iterator i = nodes.Begin (); i != nodes.End (); ++i)
       {
         Ptr<Wimax2NetDevice> dev;
         dev = CreateObject<Wimax2BSNetDevice> ();
         dev->SetAddress (Mac48Address::Allocate ());
         debug2 ("Creating a Wimax2BSNetDevice with MAC Address: %s and id_index: %d\n",dev->addr_s(), dev->getIdIndex());
         dev->SetChannel (GetPointer(channel));
         dev->setPhyMIB(phymib);
         dev->setMacMIB(macmib);
         dev->SetSchedulerDlRatio(dlratio_init);
         (*i)->AddDevice (dev);
         devices.Add (dev);
         nodeCount++;         
       }
     return devices;

}
   
//NetDeviceContainer Wimax2Helper::Install (NodeContainer nodes) 
//{
//     return Install (nodes, CreateObject<Wimax2Channel> ());
//}
//   
//NetDeviceContainer Wimax2Helper::InstallBS (NodeContainer nodes) 
//{
//     return InstallBS (nodes, CreateObject<Wimax2Channel> ());
//}

 Phy802_16MIB* Wimax2Helper::getPhyMIB() {
	return phymib;
 }
 Mac802_16MIB* Wimax2Helper::getMacMIB() {
	return macmib;
 }
 void Wimax2Helper::setPhyMIB(Phy802_16MIB *mib) {
	if (phymib) delete phymib;
	phymib = new Phy802_16MIB(*mib);
 }
 void Wimax2Helper::setMacMIB(Mac802_16MIB *mib) {
	if (macmib) delete macmib;
	macmib = new Mac802_16MIB(*mib);
 }



//void
//Wimax2Helper::EnablePcapInternal (std::string prefix, Ptr<NetDevice> nd, bool explicitFilename, bool promiscuous)
//{
//  //
//  // All of the Pcap enable functions vector through here including the ones
//  // that are wandering through all of devices on perhaps all of the nodes in
//  // the system.  We can only deal with devices of type WimaxNetDevice.
//  //
//  Ptr<Wimax2NetDevice> device = nd->GetObject<Wimax2NetDevice> ();
//  if (device == 0)
//    {
//      NS_LOG_INFO ("Wimax2Helper::EnablePcapInternal(): Device " << &device << " not of type ns3::Wimax2NetDevice");
//      return;
//    }
//
//  Ptr<Wimax2Phy> phy = device->GetPhy ();
//  PcapHelper pcapHelper;
//  std::string filename;
//  if (explicitFilename)
//    {
//      filename = prefix;
//    }
//  else
//    {
//      filename = pcapHelper.GetFilenameFromDevice (prefix, device);
//    }
//
//  Ptr<PcapFileWrapper> file = pcapHelper.CreateFile (filename, std::ios::out, PcapHelper::DLT_EN10MB);
//
//  phy->TraceConnectWithoutContext ("Tx", MakeBoundCallback (&PcapSniffTxRxEvent, file));
//  phy->TraceConnectWithoutContext ("Rx", MakeBoundCallback (&PcapSniffTxRxEvent, file));
//}
//
//
//
//
//static void PcapSniffTxRxEvent (Ptr<PcapFileWrapper> file,
//                              Ptr<const PacketBurst> burst)
//{
//  std::list<Ptr<Packet> > packets = burst->GetPackets ();
//  for (std::list<Ptr<Packet> >::iterator iter = packets.begin (); iter != packets.end (); ++iter)
//    {
//	  Ptr<Packet> p = (*iter)->Copy ();
//	  file->Write (Simulator::Now (), p);
//    }
//}

Wimax2Helper::Wimax2Helper() {
	macmib = NULL;
	phymib = NULL;
}

//note, this implementation is based heavily on the ns3::WimaxHelper::EnableAsciiForConnection function
void Wimax2Helper::EnableAsciiForConnection (Ptr<OutputStreamWrapper> os, uint32_t nodeid, uint32_t deviceid, char *netdevice, char*connection) {
  std::ostringstream oss;
  oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::" << netdevice << "/" << connection
      << "/TxQueue/Enqueue";
  Config::Connect (oss.str (), MakeBoundCallback (&AsciiTraceHelper::DefaultEnqueueSinkWithContext, os));

  oss.str ("");
  oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::" << netdevice << "/" << connection
      << "/TxQueue/Dequeue";
  Config::Connect (oss.str (), MakeBoundCallback (&AsciiTraceHelper::DefaultDequeueSinkWithContext, os));

  oss.str ("");
  oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::" << netdevice << "/" << connection
      << "/TxQueue/Drop";
  Config::Connect (oss.str (), MakeBoundCallback (&AsciiTraceHelper::DefaultDropSinkWithContext, os));	
}

void Wimax2Helper::AsciiRxEvent(Ptr<OutputStreamWrapper> stream, std::string path, Ptr<const Packet> packet, const Mac48Address &source) {
	*stream->GetStream() << "r " <<Simulator::Now().GetSeconds() << " from: " <<source << " ";
	*stream->GetStream() << path << std::endl;
}

void Wimax2Helper::AsciiTxEvent(Ptr<OutputStreamWrapper> stream, std::string path, Ptr<const Packet> packet, const Mac48Address &dest) {
	*stream->GetStream() << "t " <<Simulator::Now().GetSeconds() << " to: " <<dest << " ";
	*stream->GetStream() << path << std::endl;

}

void Wimax2Helper::EnablePcapInternal(std::string prefix, Ptr<NetDevice> netdev, bool explicitFilename, bool promisc) {
	Ptr<Wimax2NetDevice> dev = netdev->GetObject<Wimax2NetDevice>();
	if (dev == 0) {
		NS_LOG_INFO ("Wimax2Helper::EnablePcapInternal() : Device " << &netdev << " not of type ns3::Wimax2NetDevice");
		return;
	}
		

	Ptr<Wimax2Phy> phy = dev->getPhy();	

	PcapHelper pcpHlp;
	std::string filename;
	if (explicitFilename) filename = prefix;
	else filename = pcpHlp.GetFilenameFromDevice(prefix, dev);

	Ptr<PcapFileWrapper> file = pcpHlp.CreateFile(filename, std::ios::out, PcapHelper::DLT_EN10MB);

	phy->TraceConnectWithoutContext ("Tx", MakeBoundCallback (&Wimax2Helper::PcapSniffTxRxEvent, file));
	phy->TraceConnectWithoutContext ("Rx", MakeBoundCallback (&Wimax2Helper::PcapSniffTxRxEvent, file));
}

void Wimax2Helper::EnableAsciiInternal(Ptr<OutputStreamWrapper> stream, std::string prefix, Ptr<NetDevice> netdev, bool explicitFilename) {
	Ptr<Wimax2NetDevice> dev = netdev->GetObject<Wimax2NetDevice>();
	std::string BSorSS = "";
	if (dev->GetObject<Wimax2BSNetDevice>() != 0) BSorSS = "BS";
	else BSorSS = "SS";
	if (dev == 0) {
		NS_LOG_INFO ("Wimax2Helper::EnableAsciiInternal() : Device " << &netdev << " not of type ns3::Wimax2NetDevice");
		return;
	}

	Packet::EnablePrinting();

	if (stream == 0) {
		AsciiTraceHelper ath;
		std::string filename;
		if (explicitFilename) filename = prefix;
		else filename = ath.GetFilenameFromDevice(prefix, dev);

		Ptr<OutputStreamWrapper> str = ath.CreateFileStream(filename);

	uint32_t nodeid = dev->GetNode()->GetId();
	uint32_t deviceid = dev->GetIfIndex ();

	std::ostringstream oss;

	      oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::Wimax2" << BSorSS << "NetDevice/Rx";
	      Config::Connect (oss.str (), MakeBoundCallback (&Wimax2Helper::AsciiRxEvent, str));
	      oss.str ("");
	      oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::Wimax2" << BSorSS << "NetDevice/Tx";
	      Config::Connect (oss.str (), MakeBoundCallback (&Wimax2Helper::AsciiTxEvent, str));
	} else {
		
	uint32_t nodeid = dev->GetNode()->GetId();
	uint32_t deviceid = dev->GetIfIndex ();

	std::ostringstream oss;
	
	      oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::Wimax2" << BSorSS << "NetDevice/Rx";
	      Config::Connect (oss.str (), MakeBoundCallback (&Wimax2Helper::AsciiRxEvent, stream));
	      oss.str ("");
	      oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::Wimax2" << BSorSS << "NetDevice/Tx";
	      Config::Connect (oss.str (), MakeBoundCallback (&Wimax2Helper::AsciiTxEvent, stream));
	}
}

void Wimax2Helper::PcapSniffTxRxEvent (Ptr<PcapFileWrapper> file, Ptr<const Packet> pkt) {
	Ptr<Packet> p =  pkt->Copy();
	file->Write (Simulator::Now(), p);
}


 

} //namespace ns3
