/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef SERVICEFLOWHANDLER_H
#define SERVICEFLOWHANDLER_H

#include "ns3/wimax2-serviceflowqos.h"
#include "ns3/wimax2-serviceflow.h"
#include "ns3/packet.h"
#include <sys/queue.h>

namespace ns3 {

class Wimax2NetDevice;

/**
 * Handler for service flows
 */
class ServiceFlowHandler {

 public:

  /* 
   * Create a service flow
   * @param mac The Mac where it is located
   */
  ServiceFlowHandler ();

  /*
   * Set the mac it is located in 
   * @param mac The mac it is located in 
   */
  void setMac (Ptr<Wimax2NetDevice> mac);
  
  /**
   * Process the given packet. Only service related packets must be sent here.
   * @param p The packet to process
   */
  void  process (Ptr<Packet> p);
  
  /**
   * Add a flow
   * @param qos The qos for the new connection
   */
  Wimax2ServiceFlow* addFlow (Wimax2ServiceFlowQoS * qos);
      
  /**
   * Remove a flow
   * @param id The flow ID
   */
  void  removeFlow (int id);
    
  /**
   * Send a flow request to the given node
   * @param index The node address
   * @param incoming The flow direction
   */
  void sendFlowRequest (Mac48Address index, bool incoming);

  /**
   * Add the Static Flows
   */
  int addStaticFlow (int argc, const char*const* argv);
  
  /**
   * Add the static flows using an arg list
   */
  int addStaticFlow (bool UL, int drate, SchedulingType_t ft, int dsize, int prid, uint8_t arq_en, double arq_ret_time, uint32_t arq_maxwin, uint8_t ack_per, int traffic_pri, int peak_traffic_rate, int min_res_traffic_rate, int req_xmit_pol, int jitter, int sdu_indicator, int min_tol_traffic_rate, int sdusize, int max_burst_size, int said);
  

  /**
   * Initialize the Static Flows
   */
  void init_static_flows (Mac48Address index);

 protected:

  /**
   * process a flow request
   * @param p The received request
   */
  void processDSA_req (Ptr<Packet> p);

  /**
   * process a flow response
   * @param p The received response
   */
  void processDSA_rsp (Ptr<Packet> p);

  /**
   * process a flow request
   * @param p The received response
   */
  void processDSA_ack (Ptr<Packet> p);
  
 private:

  /**
   * The Mac where this handler is located
   */
   Ptr<Wimax2NetDevice> mac_;

  /**
   * The list of current flows
   */
   struct serviceflow_list flow_head_;

   /**
    * List of pending flows
    */
   struct serviceflow_list pendingflow_head_;

   /**
    * The list of current flows
    */
   struct serviceflow_list static_flow_head_; 
};


}//namespace
#endif //SERVICEFLOWHANDLER_H

