/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */
#include <cstdlib> 
#include "wimax2-ss-net-device.h"
#include "ns3/wimax2-scheduler.h"
//#include "scheduling/ssscheduler.h" 
#include "ns3/wimax2-ss-scheduler.h"
#include "ns3/destclassifier.h"
//#include "cmu-trace.h"
#include "ns3/random-variable.h"
#include "globalparams_wimax.h" 
#include "ns3/ulsubframetimer.h"
#include "ns3/dlsubframetimer.h"
#include "ns3/pointer.h"
#include <math.h>

#include "ns3/trace-source-accessor.h"

NS_LOG_COMPONENT_DEFINE("Wimax2SSNetDevice");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED(Wimax2SSNetDevice);

TypeId Wimax2SSNetDevice::GetTypeId (void) {
	  static TypeId tid = TypeId ("ns3::Wimax2SSNetDevice")
    .SetParent<NetDevice> ()
    .AddConstructor<Wimax2SSNetDevice> ()
    .AddAttribute("TxQueue", "A queue used to hit the PyViz hooks.", PointerValue(), MakePointerAccessor(&Wimax2SSNetDevice::m_queue), MakePointerChecker<Queue>())
    .AddTraceSource("Rx", "Receive trace", MakeTraceSourceAccessor(&Wimax2SSNetDevice::m_traceRx))
    .AddTraceSource("Tx", "Transmit trace", MakeTraceSourceAccessor(&Wimax2SSNetDevice::m_traceTx))
//    .AddTraceSource("SSTxDrop", 	"A packet from upper layers dropped before transmission.",  	MakeTraceSourceAccessor (&Wimax2SSNetDevice::m_ssTxDropTrace))
//    .AddTraceSource("SSRx", 		"A packet recieved from the PHY layer (non-promiscuous).", 	MakeTraceSourceAccessor (&Wimax2SSNetDevice::m_ssRxTrace))
//    .AddTraceSource("SSRxDrop", 	"A packet from the PHY layer dropped by the MAC.", 		MakeTraceSourceAccessor (&Wimax2SSNetDevice::m_ssRxDropTrace))
    ;
    return tid;
}

/**
 * Creates a SS Mac Layer (NetDevice) 
 */

 Wimax2SSNetDevice::Wimax2SSNetDevice() : Wimax2NetDevice(Seconds(0.1)) {
//this really shouldn't ever be called, it's just included so the compiler is happier
  NS_LOG_FUNCTION_NOARGS ();
  type_ = STA_MN; //type of MAC. In this case it is for SS
  
  //Create default configuration
  addClassifier (new DestClassifier ());
  scheduler_ = new Wimax2SSScheduler();
  scheduler_->setMac (this); //register the mac
  
  init_default_connections ();

  //We can configure each SS with a different modulation
  //default transmission profile (64 QAM 3_4. Can be changed by TCL)
  //The DIUC to use will be determined via an algorithm considering 
  //the channel performance
  default_diuc_ = DIUC_PROFILE_7;

  //initialize state
  state_ = MAC802_16_DISCONNECTED; //At start up, we are not connected.
  //scanning status
  scan_info_ = (struct scanning_structure *) malloc (sizeof (struct scanning_structure));
  memset (scan_info_, 0, sizeof (struct scanning_structure));
  scan_info_->nbr = NULL;
  scan_info_->substate = NORMAL; //no scanning
  scan_flag_ = false;

  map_ = new FrameMap (this);

  //create timers for state machine
  t1timer_ = new WimaxT1Timer (this);
  t2timer_ = new WimaxT2Timer (this);
  t6timer_ = NULL;
  t12timer_ = new WimaxT12Timer (this);
  t21timer_ = new WimaxT21Timer (this);
  lostDLMAPtimer_ = new WimaxLostDLMAPTimer (this);
  lostULMAPtimer_ = new WimaxLostULMAPTimer (this);
  t44timer_ = NULL;

  //create timers for DL/UL boundaries
  dl_timer_ = new DlTimer (this);
  ul_timer_ = new UlTimer (this);

  //initialize some other variables
  nb_reg_retry_ = 0;
  nb_scan_req_ = 0;
  new_dlmap_processed = false;
  this_frame_ulmap_processed = false;
  this_frame_dlmap_processed = false;
  //note that this is needed instead of sticking the string strait in setBSAddr b/c of the python binding scanner not supporting the implicet copy constructor used there
  Mac48Address blank("00:00:00:00:00:00"); 
  setBSAddr(blank);

  //initialize the queue used to hit PyViz hooks
  m_queue = CreateObject<DropTailQueue>();
 }

Wimax2SSNetDevice::Wimax2SSNetDevice(Time boot) : Wimax2NetDevice (boot)
{
  NS_LOG_FUNCTION_NOARGS ();
  type_ = STA_MN; //type of MAC. In this case it is for SS
  
  //Create default configuration
  addClassifier (new DestClassifier ());
  scheduler_ = new Wimax2SSScheduler();
  scheduler_->setMac (this); //register the mac
  
  init_default_connections ();

  //We can configure each SS with a different modulation
  //default transmission profile (64 QAM 3_4. Can be changed by TCL)
  //The DIUC to use will be determined via an algorithm considering 
  //the channel performance
  default_diuc_ = DIUC_PROFILE_7;

  //initialize state
  state_ = MAC802_16_DISCONNECTED; //At start up, we are not connected.
  //scanning status
  scan_info_ = (struct scanning_structure *) malloc (sizeof (struct scanning_structure));
  memset (scan_info_, 0, sizeof (struct scanning_structure));
  scan_info_->nbr = NULL;
  scan_info_->substate = NORMAL; //no scanning
  scan_flag_ = false;

  map_ = new FrameMap (this);

  //create timers for state machine
  t1timer_ = new WimaxT1Timer (this);
  t2timer_ = new WimaxT2Timer (this);
  t6timer_ = NULL;
  t12timer_ = new WimaxT12Timer (this);
  t21timer_ = new WimaxT21Timer (this);
  lostDLMAPtimer_ = new WimaxLostDLMAPTimer (this);
  lostULMAPtimer_ = new WimaxLostULMAPTimer (this);
  t44timer_ = NULL;

  //create timers for DL/UL boundaries
  dl_timer_ = new DlTimer (this);
  ul_timer_ = new UlTimer (this);

  //initialize some other variables
  nb_reg_retry_ = 0;
  nb_scan_req_ = 0;
  new_dlmap_processed = false;
  this_frame_ulmap_processed = false;
  this_frame_dlmap_processed = false;
  //note that this is needed instead of sticking the string strait in setBSAddr b/c of the python binding scanner not supporting the implicet copy constructor used there
  Mac48Address blank("00:00:00:00:00:00"); 
  setBSAddr(blank);

  //initialize the queue used to hit PyViz hooks
  m_queue = CreateObject<DropTailQueue>();
}


/**
 * Initialize default connections
 * These connections are not linked to a peer node
 */
void Wimax2SSNetDevice::init_default_connections ()
{
  NS_LOG_FUNCTION_NOARGS ();
  Connection * con;

  //create initial ranging and padding connection
  con = new Connection (CONN_INIT_RANGING);
  connectionManager_->add_connection (con, OUT_CONNECTION); 
  con = new Connection (CONN_INIT_RANGING);
  connectionManager_->add_connection (con, IN_CONNECTION); 
  con = new Connection (CONN_PADDING);
  connectionManager_->add_connection (con, OUT_CONNECTION);
  con = new Connection (CONN_PADDING);
  connectionManager_->add_connection (con, IN_CONNECTION);

  //create connection to receive broadcast packets from BS
  con = new Connection (CONN_BROADCAST);
  connectionManager_->add_connection (con, IN_CONNECTION);
}

/**
 * Initialize the MAC
 */
void Wimax2SSNetDevice::init ()
{
  NS_LOG_FUNCTION_NOARGS ();
  //Set physical layer to receiving mode
  getPhy()->setMode (OFDM_RECV);

  //SS is looking for synchronization
  setMacState(MAC802_16_WAIT_DL_SYNCH); 

  //start timer for expiration
  t21timer_->start (macmib_->t21_timeout);

  int nbPS = (int) floor((getFrameDuration()/getPhy()->getPS()));
  int nbPS_left = nbPS - phymib_.rtg - phymib_.ttg;
  int nbSymbols = (int) floor((getPhy()->getPS()*nbPS_left)/getPhy()->getSymbolTime());  // max num of OFDM symbols available per frame. 
  int nbSubcarrier = getPhy()->getFFT();
  int nbSubchannel = getPhy()->getNumsubchannels (DL_); //we receive downlink


  intpower_ = (double**) malloc (nbSymbols*sizeof (double*));
  for (int i = 0 ; i<nbSymbols ; i++) {
    intpower_[i] = (double*) malloc (nbSubcarrier*sizeof (double));
    for (int j = 0 ; j< nbSubcarrier ; j++)
      intpower_[i][j] = 0.0 ; 
  }
  basepower_ = (double**) malloc (nbSymbols*sizeof (double*));
  for (int i = 0 ; i<nbSymbols; i++) {
    basepower_[i] = (double*) malloc (nbSubchannel*sizeof (double));
    for (int j = 0 ; j<nbSubchannel ; j++)
      basepower_[i][j] = 0.0 ; 
  }
     
  //init the scheduler
  scheduler_->init();

  //setup the data structure that stores incoming packet fragments
  initialize_grid();
}


/**
 * Set the mac state
 * @param state The new mac state
 */  
void Wimax2SSNetDevice::setMacState (Mac802_16State state)
{
  NS_LOG_FUNCTION_NOARGS ();
  debug10("At %f, setting SS %s state to %d\n", NOW, addr_s(), state);
 	 state_ = state;
  if(state == MAC802_16_CONNECTED) std::cerr << "At " << NOW << " SS " << addr_s() << " is connected\n";
}

/**
 * Return the mac state
 * @return The new mac state
 */  
Mac802_16State Wimax2SSNetDevice::getMacState ()
{
  NS_LOG_FUNCTION_NOARGS ();
  return state_;
}

/**
 * Backup the state of the Mac
 * @return A structure containing a copy of current MAC state
 */
state_info* Wimax2SSNetDevice::backup_state ()
{
  NS_LOG_FUNCTION_NOARGS ();
  state_info *backup_state = (state_info*) malloc (sizeof (state_info));
  backup_state->state = state_;
  backup_state->frameduration = getFrameDuration();
  backup_state->frame_number = frame_number_;
  backup_state->channel = getChannel();
  backup_state->connectionManager = connectionManager_;
  connectionManager_ = new ConnectionManager (this);
  init_default_connections ();
  backup_state->serviceFlowHandler = serviceFlowHandler_;
  serviceFlowHandler_ = new ServiceFlowHandler();
  backup_state->peer_list = peer_list_;
  backup_state->nb_peer = nb_peer_;
  peer_list_ = (struct peerNode_list *) malloc (sizeof(struct peerNode_list));
  LIST_INIT(peer_list_);
  return backup_state;
}

/**
 * Restore the state of the MAC
 * @param backup_state The MAC state to restore
 */
void Wimax2SSNetDevice::restore_state (state_info *backup_state)
{
  NS_LOG_FUNCTION_NOARGS ();
  state_ = backup_state->state;
  setFrameDuration(backup_state->frameduration);
  frame_number_ = backup_state->frame_number;
  setChannel (backup_state->channel);
  delete (connectionManager_);
  connectionManager_ = backup_state->connectionManager;
  delete (serviceFlowHandler_);
  serviceFlowHandler_ = backup_state->serviceFlowHandler;
  while (getPeerNode_head()!=NULL) {
    removePeerNode (getPeerNode_head());
  }
  peer_list_ = backup_state->peer_list;
  nb_peer_ = backup_state->nb_peer;
}

/**
 * Called when a timer expires
 * @param The timer ID
 */
void Wimax2SSNetDevice::expire (timer_id id)
{
  NS_LOG_FUNCTION_NOARGS ();
		debug10("entering (SS) %s expire\n", addr_s());

  switch (id) {
  case WimaxT21TimerID:
    debug10 ("At %f in Mac %s, synchronization failed\n", NOW, addr_s());
    //go to next channel
    nextChannel();
    t21timer_->start (macmib_->t21_timeout);
    break;
  case WimaxLostDLMAPTimerID:
    debug10 ("At %f in Mac %s, lost synchronization (DL_MAP)\n", NOW, addr_s());
    lost_synch ();
    break;
  case WimaxT1TimerID:
    debug10 ("At %f in Mac %s, lost synchronization (DCD)\n", NOW, addr_s());
    lost_synch ();
    break;
  case WimaxLostULMAPTimerID:
    debug10 ("At %f in Mac %s, lost synchronization (UL_MAP)\n", NOW, addr_s());
    lost_synch ();
    break;
  case WimaxT12TimerID:
    debug10 ("At %f in Mac %s, lost uplink param (UCD)\n", NOW, addr_s());
    lost_synch ();
    break;
  case WimaxT2TimerID:
    debug10 ("At %f in Mac %s, lost synchronization (RNG) so remove cdma initial ranging request\n", NOW, addr_s());
    //getMap()->getUlSubframe()->getRanging()->removeRequest ();
    getMap()->getUlSubframe()->getRanging()->removeRequest_mac (addr());
    Connection *c_tmp;
    c_tmp  = getCManager ()->get_connection (0, true);
    c_tmp->setINIT_REQ_QUEUE(id_index, 0);

    lost_synch ();
    break;
  case WimaxT3TimerID:
    debug10 ("At %f in Mac %s, no response from BS so remove cdma initial ranging request\n", NOW, addr_s());
    //we reach the maximum number of retries
    //mark DL channel usuable (i.e we go to next)
    //getMap()->getUlSubframe()->getRanging()->removeRequest ();
    getMap()->getUlSubframe()->getRanging()->removeRequest_mac (addr());
    nextChannel();
    lost_synch ();
    break;
  case WimaxT6TimerID:
    debug10 ("At %f in Mac %s, registration timeout (nbretry=%d)\n", NOW, addr_s(),
	   nb_reg_retry_);
    if (nb_reg_retry_ == macmib_->reg_req_retry) {
      debug10 ("\tmax retry excedeed\n");
      lost_synch ();
    } else {
      send_registration();
    }
    break;
  case WimaxT44TimerID:
    debug10 ("At %f in Mac %s, did not receive MOB_SCN-RSP (nb_retry=%d/%d)\n", NOW, addr_s(), nb_scan_req_, macmib_->scan_req_retry);
    if (nb_scan_req_ <= macmib_->scan_req_retry) {
      send_scan_request ();
    } else { //reset for next time
      nb_scan_req_ = 0;
    }
    break;
  case WimaxScanIntervalTimerID:
    pause_scanning ();
    break;    
  case WimaxRdvTimerID:
    //we need to meet at another station. We cancel the current scanning
    //lost_synch ();
    debug10 ("At %f in Mac %s Rdv timer expired\n", NOW, addr_s());
    break;
  default:
    debug10 ("Trigger unkown\n");
  }
}


/**** Packet processing methods ****/

/*
 * Process packets going out from the upper layers
 * Note: Assumes packets from upper layers
 * 	 enqueue directly to connection queues if sending from L2
 * @param p The packet to send out
 */
void Wimax2SSNetDevice::sendDown(Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
	debug10("entering (SS) %s sendDown\n", addr_s());

  //We first send it through the CS
  int cid = -1;
  //since we only support higher layer communication in this function
  //and the classification system doesn't really support multiple data
  //connections on a single node, we just grab the outgoing data
  //connection's information for this
  cid = getCManager()->getConOfType(CONN_DATA, OUT_CONNECTION);
//  cid = classify (p);

  if (cid == -1) {
    debug10 ("At %f in Mac %s drop packet because no classification was found\n", NOW, addr_s());
    drop(p, "CID");
    //Packet::free (p);
  } else {
    //enqueue the packet 
    Connection *connection = connectionManager_->get_connection (cid, OUT_CONNECTION);
    if (connection == NULL) {
      if (getMacState() == MAC802_16_CONNECTED && cid == BROADCAST_CID) {
      	  cid = connectionManager_->getConOfType(CONN_DATA, true);
      	  All_Headers ah;
      	  p->RemoveHeader(ah);
	
	  //to hit PyViz's hooks, we need to enqueue and dequeue the packet.  This could be prevented by modifying src/tools/visualizer/model/pyviz.cc to directly
	  //register the Wimax2 classes, but this allows us to use a plugin rather than modifying core ns3 code and is much easier to distrubute
//	  m_queue->Enqueue(p);
//	  m_queue->Dequeue(); //this actually triggers the hook
//	  m_traceTx(p);//, ah.dest);


      	  ah.GMH.GMH.cid = cid;
      	  p->AddHeader(ah);
      	  connection = connectionManager_->get_connection(cid, OUT_CONNECTION);
      } else {
		  debug2 ("Warning: At %f in Mac %s connection with cid = %d does not exist. Please check classifiers\n", NOW, addr_s(), cid);
		  //Packet::free (p);
		  //update_watch (&loss_watch_, 1); //cat - uncomment
		  drop(p, "CID");
		  return;
      }
    }
     //if we have a valid cid with a connection mapped to it
    
    		bool need_drop = false;
 			if(!connection->getArqStatus())
			{
				if(connection->queueLength() >= macmib_->queue_length)
				{
					//update_watch (&loss_watch_, 1); //cat- uncomment later
					need_drop = true;
					drop (p, "QWI");
				}
			}
			else if(connection->getArqStatus() && (connection->getArqStatus()->isArqEnabled() == 1))
			{
				All_Headers ah;
				p->PeekHeader(ah);
				int data_packet_size = ah.GMH.GMH.len;
				int num_arq_blocks = (int) ceil ((double)data_packet_size/arq_block_size_);

				debug2("data_packet_size=%d\t num_arq_blocks=%d\t qlength=%d max_qlength=%d\n",
							data_packet_size,num_arq_blocks,connection->queueLength(),macmib_->queue_length);
				if((connection->queueLength() + num_arq_blocks) >= macmib_->queue_length)
				{
					//update_watch (&loss_watch_, 1); //CAT- uncomment
					need_drop = true;
					drop (p, "QWI");
				}
			}
			
			if( !need_drop)
			{
				//Begin RPI
				if (connection->getArqStatus () == NULL) 
				{
				  //to hit PyViz's hooks, we need to enqueue and dequeue the packet.  This could be prevented by modifying src/tools/visualizer/model/pyviz.cc to directly
				  //register the Wimax2 classes, but this allows us to use a plugin rather than modifying core ns3 code and is much easier to distrubute
				  m_queue->Enqueue(p);
				  m_queue->Dequeue(); //this actually triggers the hook
				  m_traceTx(p, declassify(cid));//, ah.dest);
					//End RPI
					//update mac header information
					//set header information
					All_Headers ah;
					p->RemoveHeader(ah);
					gen_mac_header_t *wimaxHdr = &ah.GMH.GMH;
					wimaxHdr->ht = 0;
					wimaxHdr->ec = 1;
					wimaxHdr->type_mesh = 0;
					wimaxHdr->type_arqfb = 0;
					wimaxHdr->type_ext = 0;
					wimaxHdr->type_frag = 0;
					wimaxHdr->type_pck = 0;
					wimaxHdr->type_fbgm = 0;
					wimaxHdr->ci = 0;
					wimaxHdr->eks = 0;
					wimaxHdr->cid = cid; //default
					wimaxHdr->hcs = 0;
					wimaxHdr->len += HDR_MAC802_16_SIZE;
					p->AddHeader(ah);
					connection ->enqueue (p);
					//Begin RPI
					debug2 ("At %f (SS) in Mac %s, SENDDOWN, Enqueue packet to cid :%d queue size :%d (max :%d)\n", NOW, addr_s(), cid, connection->queueLength (), macmib_->queue_length);
				}
				else if((connection->getArqStatus () != NULL) && (connection->getArqStatus ()->isArqEnabled () == 1)) 
				{
				  //to hit PyViz's hooks, we need to enqueue and dequeue the packet.  This could be prevented by modifying src/tools/visualizer/model/pyviz.cc to directly
				  //register the Wimax2 classes, but this allows us to use a plugin rather than modifying core ns3 code and is much easier to distrubute
				  m_queue->Enqueue(p);
				  m_queue->Dequeue(); //this actually triggers the hook
				  m_traceTx(p, declassify(cid));//, ah.dest);
					//We will have to divide the packet to ARQ Blocks
					//In NS2 it will be implemented by duplicating the packet but changing the size of the packet
					//here we'll actually be creating packets from sub-ranges of the data
					All_Headers ah;
					p->RemoveHeader(ah);
					int packet_size = ah.GMH.GMH.len;
					int number_of_packs= (packet_size/arq_block_size_); 
					gen_mac_header_t *wimaxHdr = &ah.GMH.GMH;
					wimaxHdr->ht = 0;
					wimaxHdr->ec = 1;
					wimaxHdr->type_mesh = 0;
					wimaxHdr->type_arqfb = 0;
					wimaxHdr->type_ext = 0;
					wimaxHdr->type_frag = 0;
					wimaxHdr->type_pck = 0;
					wimaxHdr->type_fbgm = 0;
					wimaxHdr->ci = 0;
					wimaxHdr->eks = 0;
					wimaxHdr->cid = cid; //default
					wimaxHdr->hcs = 0;
					wimaxHdr->len = arq_block_size_;
					p->AddHeader(ah);

					for(int i=1; i<= number_of_packs; i++)
					{
						if(i == 1)
							connection->getArqStatus ()->arqSend(p, connection, FRAG_FIRST);
						else
							connection->getArqStatus ()->arqSend(p, connection, FRAG_CONT);	
						connection ->enqueue (p->Copy());  //?????
					}
					p->RemoveHeader(ah);
					ah.GMH.GMH.len = (packet_size - (number_of_packs * arq_block_size_));
					p->AddHeader(ah);
					
					if(number_of_packs)
					{
						connection->getArqStatus ()->arqSend(p, connection, FRAG_LAST);
					}
					else
					{
						connection->getArqStatus ()->arqSend(p, connection, FRAG_NOFRAG);
					}
					connection ->enqueue (p);  //?????
					debug2 ("ARQ SS Dividing Packet %d into Blocks Packet Size=%d, Number of packets = %d, ARQ_BLOCK_SIZE = %d,Last Packet Size= %d\n",(int)p->GetUid(),packet_size,number_of_packs+1,arq_block_size_, ah.GMH.GMH.len) ;
				}
				//End RPI
			}
		
	}

	//inform upper layer that it can send another packet
	accepting_ = true;
//	resume (NULL);  
}

/*
 * Transmit a packet to the physical layer
 * @param p The packet to send out
 */
void Wimax2SSNetDevice::transmit(Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
  PHY_INFO_Header pih;
  p->PeekHeader(pih);
  debug10 ("At %f (SS) in Mac %s, TRANSMIT, sending packet (size=%d)\n", NOW, addr_s(), p->GetSize());

  //update stats for delay and jitter
  //double delay = NOW - ah.timestamp; 
  //if (flag_cdma_packet == 0) update_watch (&delay_watch_, delay); //CAT- uncomment
  //double jitter = fabs (delay - last_tx_delay_);
  //if (flag_cdma_packet == 0) update_watch (&jitter_watch_, jitter);//CAT- uncomment
  //last_tx_delay_ = delay;
  
  //if (flag_cdma_packet == 0) update_throughput (&tx_traffic_watch_, 8*ch->size());//CAT- uncomment
  
 

  //pass it down
  phy_->Send(p);

}

/*
 * Process incoming packets one allocation slot at a time
 * receive function is called when all slots are received
 * @param p The incoming packet
 * @param subchannel The subchannel on which the slot was transmitted
 * @param time_in The time at which the slot was received
 */

//void Wimax2SSNetDevice::sendUp(Ptr<Packet> p, uint8_t subchannel, Time r_time) {
//	NS_LOG_FUNCTION_NOARGS();
//	debug10("entering (SS) %s sendUp\t", addr_s());
//
//	//the number of symbol-times into the frame we are
//	//note that the preamble should cover the first three
//	int32_t symb = floor( (r_time.GetSeconds()-start_current_frame.GetSeconds()) / getPhy()->getSymbolTime());
//	int32_t slot = symb <= 3 ? 0 : (symb - DL_PREAMBLE ) / getPhy()->getSlotLength(DL_);
//
//	debug10(" received a block to be placed in (symbol-time = %d, subchannel = %d)\n", symb, subchannel);
//
//	if (slot == 0) {
//		//the FCH will be in the top two subchannels
//	}
//
//	
//
//	
//	//we need to specifically catch the FCH
//	//if (subchannel == 0 && r_time.GetSeconds() >= start_current_frame.GetSeconds() + DL_PREAMBLE * getPhy()->getSymbolTime() && r_time.GetSeconds() < start_current_frame.GetSeconds() + (DL_PREAMBLE + 1) * getPhy()->getSymbolTime()) {
//	//}
//
//
//
//
//}

/*
 * Process incoming packets, which will then be passed through the recieve function
 * @param p The incoming packet
 */
//void Wimax2SSNetDevice::sendUp (Ptr<Packet> p, uint8_t subchannel, Time recieved)
//{
//  NS_LOG_FUNCTION_NOARGS ();
//	debug10("entering (SS) %s sendUp\n", addr_s());
//
//	//we see incoming packets during this period, but nothing will actually be bound for us
//	//so we just ignore anything that comes in until the DL SUBFRAME starts
//	if (getPhy()->getMode() != OFDM_RECV) {
////		std::cerr << "Currently in UL Subfame.  SS will not receive data during this period.  Dropping incoming packet.\n";
////		std::cerr<<"ignoring packet at " << NOW << std::endl;
//		p = 0;
//		return;
//	}// else {
////		std::cerr << "Currently in DL Subfame.  SS will receive data during this period.\n";
////	}
//
//  // bool is_MAC_packet = isMACPacket(p);
////   char packet_type[10];
////   if (isMACPacket(p)) packet_type = "MAC";
////   else packet_type = "non-MAC";
//
//  All_Headers ah;
//  p->RemoveHeader(ah);
//  
//  debug10("packet size %d\n", p->GetSize());
//    
//  //fill both a GMH, SMH, and a CRH from the same 6 bytes, decide which to use below
//  p->PeekHeader(ah.GMH);
//  p->PeekHeader(ah.SMH);
//  p->RemoveHeader(ah.CRH);
//
//  
// // struct hdr_cmn *ch = HDR_CMN(p);
//  //hdr_mac802_16 *wimaxHdr;
//  
//  //wimaxHdr = HDR_MAC802_16(p);
//  int flag_cdma_packet = 0;
//
///*
////Chakchai; if the direction is not for SS; return => probably fixed by Richard
//  hdr_cmn* ch_tmp = HDR_CMN(p);
//  hdr_mac802_16 *wimaxHdr_tmp;
//  wimaxHdr_tmp = HDR_MAC802_16(p);
////  debug10 (" SS direction :%d\n", wimaxHdr_tmp->phy_info.direction);
//  if (wimaxHdr_tmp->phy_info.direction != 0) {
//     Packet::free(p);
//     return;
//  }
//*/
//	  
////#ifdef DEBUG_WIMAX
////  debug10 ("At %f in Mac %s receive first bit..over at %f(txtime=%f) (type=%s)\n", NOW, index_, NOW+ch->txtime(),ch->txtime(), packet_info.name(ch->ptype()));
//  debug10 ("At %f (SS) in Mac %s, SENDUP, receive first bit\n", NOW, addr_s());
//
//  //if (is_MAC_packet) {
//    if (ah.GMH.GMH.ht == 0)
//      {
//      	ah.gm_sm = 1;
//      	ah.mac_packet = ah.GMH.GMH.rsv1;
//      	if (p->GetSize() > 0) {
//			Type_Check tp;
//      		p->PeekHeader(tp);
//      		if (! ah.GMH.GMH.type_frag)
//				debug10("\t generic mngt %d\n", tp.type);
//      	}
//      }
//    else { //either a bw_req, cmda_req, or cdma_init... which should not be sent to a SS
//
//        cdma_req_header_t *req_tmp = &ah.CRH.data;
//
//        if (req_tmp->type == 0x2) {
//          	debug10 ("\t Impossible (PANIC) cdma_bw_req code :%d, top :%d\n", req_tmp->code, req_tmp->top);
//  	  		flag_cdma_packet = 1;
//  	  		ah.gm_sm = 2;
//        } else if (req_tmp->type == 0x3) {
//          	debug10 ("\t Impossible (PANIC) cdma_init_req code :%d, top :%d\n", req_tmp->code, req_tmp->top);
//  	  	  	flag_cdma_packet = 1;
//  	  	  	ah.gm_sm = 2;
//        } else {
//          	debug10 ("\t Impossible (PANIC) bwreq\n");
//          	ah.gm_sm = 0;
//        }
//
//    }
//
//
////#endif
//  //sam
//
//  //wimaxHdr = HDR_MAC802_16(p);
//  int cid = -1;
//  //gen_mac_header_t header = ah.GMH.GMH;
//
//  if (flag_cdma_packet == 0) {
//	  cid = ah.GMH.GMH.cid;
//  }
//  else {
//  	cid = ah.CRH.data.cid;
//  }
//  Connection *con = connectionManager_->get_connection (cid, IN_CONNECTION);
//
//  if( (ah.PIH.OFDMSymbol_offset == 0 && ah.PIH.num_OFDMSymbol == 0) 
//  		|| cid == BROADCAST_CID)   // this kind of packets are treated diff(OFDM types) basically it is a bw req packet.
//    {
//      if (con == NULL) {
//		//This packet is not for us
//		//    //    //debug2 ("At %f in Mac %s Connection null\n", NOW, index_);
//		//update_watch (&loss_watch_, 1); //CAT - uncomment
//	 
//		p = 0;
//		debug2 (" packet getting dropped header.ht = %d",ah.GMH.GMH.ht);
//		//pktRx_=NULL;
//		return;
//      } 
//      
///*      if(contPktRxing_ == true) //CAT - uncomment
//	{
//	  // if there is a collision and the power of the currently recvd packet is less by factor 10 then discard it. 
//           
//	  if (head_pkt_ !=NULL) {
//            
//	    if(head_pkt_->p->txinfo_.RxPr / p->txinfo_.RxPr >= p->txinfo_.CPThresh)
//	      p = 0;
//        else if(head_pkt_->p->txinfo_.RxPr / p->txinfo_.RxPr <= p->txinfo_.CPThresh)
//	      {
//			head_pkt_->timer.stop();
//			removePacket(head_pkt_);
//			p->AddHeader(ah);
//			addPacket(p);
//	      }
//            else 
//	      {
//		if(txtime(p) > head_pkt_->timer.expire())
//		  {
//		    head_pkt_->timer.stop();
//		    removePacket(head_pkt_);
//		    ch->error() = 1;
//		    //update_watch (&loss_watch_, 1); //CAT- uncomment
//		    collision_ = true;
//		    p->AddHeader(ah);
//		    addPacket(p);	
//
//		  }
//
//	      }
//            
//       
//	  }
//	}
//
//      else 
//
//        {*/
//          //contPktRxing_ = true;
//          p->AddHeader(ah);
//          addPacket(p);
//		// }  //CAT - uncomment
//    }
//  
//  else {
//
//    if (con == NULL) {
//      //This packet is not for us
//      debug2 ("At %f in Mac %s Connection null -- packet getting dropped at SS not for us \n", NOW, addr_s());
//      //update_watch (&loss_watch_, 1); //CAT- uncomment
//    
//      // remeber to take care of collision, tht is if a packet destined to this is already being recvd and another packets destined to this comes thn thr will be collision    
//
//	  drop(p, "CON");
//      return;
//
//
//    }  // free and end if con == NULL 
//    
//
//    // collision check .  rpi
//
//    bool collision = false; 
//
//    //collision = IsCollision ( wimaxHdr, 0.0);   // removed replace it later
//
//    if (collision == true)
//
//      {
//		collision = false; 
//		//hdr_cmn *hdr = HDR_CMN(p);
//		ah.error = 1;   // we do not discard this packet but just mark it and r=transmit coz, this packet and the other collided packet might not have symbols and carrriers exactly same, but they can collide, we shld also keep track of this packets transmission, because other packets overlapping with this can collide. 
//	 
//		// set to the symbol and corresponding subcarriers over which this packet was transmitted to -1, so tht for future packets we can check for collision and also after the other collided packet is received we can chk for collision.    
//		//collision_ = true; 
//		addPowerinfo(&ah, 1.0,true);  // enable , diabled just to compile // will be in the base power 
//	
//		// set end
//
//      }
//    else
//      addPowerinfo(&ah, BASE_POWER,true ); // enable , disabled just to compile 
//  
//    addPowerinfo(&ah, 0.0,false);   // enable, disabled just to compile  // separated the variabled for interference and collision detection. 
//
//    // collision check rpi 
//
//
//    //remove temporary to chk without collision and interference 
//	p->AddHeader(ah);
//	debug10 ("So now (type=%s) at bottom of sendUp\n",  HelperFunctions::getPacketTypeString(ah.gm_sm));
//
//    addPacket(p);    // adding the collided packet also , which will be dropped in the receive function, this will help for collision detection of packets which might over lap with this.
//
//  } // ends else for bwreq  replace this } later when testing over 
//
//}

/**
 * Process the fully received packet (called after Wimax2NetDevice::assemble_packet finishes it's work)
 */
void Wimax2SSNetDevice::receive (Ptr<Packet> pktRx_)
{
  NS_LOG_FUNCTION_NOARGS ();
  debug10("entering (SS) %s receive\n", addr_s());

  assert (pktRx_);

  All_Headers ah;
  pktRx_->RemoveHeader(ah);

//#ifdef DEBUG_WIMAX
  debug10 ("At %f (SS) in Mac %s, RECEIVE, packet received (type=%s)\n", NOW, addr_s(), HelperFunctions::getPacketTypeString(ah.gm_sm));
  
  //update rx time of last packet received
  PeerNode *peer_ = getPeerNode_head();
  if (peer_) {
    peer_->setRxTime (NOW);
  }
  
  if (! isMACPacket(pktRx_)) {
  	if ( state_ == MAC802_16_CONNECTED) {
	      //we need to get rid of any excess packing zeros added to fill transmission allocations
	      //as the higher layers won't always ignore them correctly
	      ah.GMH.GMH.len -= ah.GMH.GetSerializedSize(); //since the GMH was already removed
	      if (pktRx_->GetSize() > ah.GMH.GMH.len) pktRx_->RemoveAtEnd(pktRx_->GetSize() - ah.GMH.GMH.len);

	  
	      UL_Protocol_Header uph;
	      pktRx_->RemoveHeader(uph);
	      uint16_t proto_num = uph.protocol;
	      debug2("Sending Recieved packet up to IP Layer in SS\n");

	       //note a classification system using more than the first entry of the classifier list might not work with the following
	       Mac48Address origin = declassify(ah.GMH.GMH.cid);
	       m_traceRx(pktRx_, origin);
	       Simulator::ScheduleWithContext(this->GetNode()->GetId(), Seconds(0), &Wimax2NetDevice::Receive, this, pktRx_, proto_num, addr(), origin);
	       return;
        }
	else {
		debug2("At %f in %s, received packet bound for upper layers when not in correct state, dropping packet\n", NOW, addr_s());
		drop(pktRx_, "SS-STATE");
		return;
	}
    }

    //If we make it this far, we know the packet is bound for layer two
  
    if (ah.GMH.GMH.ht == 1) {
    	debug10("At SS %s, a packet was recieved that is labeled as having a signaling MAC header, which is incorect.  Dropping packet\n", addr_s());
	//There are currently no valid situations in which a SS recieves a packet that does not employ a Generic MAC Header
	//otehr than the FCH/DL_MAP which do not pass through this function
	drop(pktRx_, "BAD HEADER TYPE");
	return;
    }


    pktRx_->AddHeader(ah);
    process_mac_packet (pktRx_);
    pktRx_ = 0;
    
}

//Begin RPI
/**
 * The function is used to process the MAC PDU when ARQ,Fragmentation and Packing are enabled
 * @param con The connection by which it arrived
 * @param p The packet to process
 */
void Wimax2SSNetDevice::process_mac_pdu_witharqfragpack (Connection *con, Ptr<Packet>  pkt)
{
  NS_LOG_FUNCTION_NOARGS ();
  Ptr<Packet> mac_pdu, p_current, p, mac_sdu, p_previous, p_temp;
  All_Headers ah_pdu, ah_current, ah, ah_sdu, ah_temp;
  //hdr_cmn *ch_pdu, *ch_current, *ch, *ch_sdu;
  //hdr_mac802_16 *wimaxHdr_pdu, *wimaxHdr_current, *wimaxHdr, *wimaxHdr_sdu;
  double tmp;
  uint8_t isInOrder;
  uint32_t seqno;
  bool mac_sdu_gen = false;
  int pdu_size = 0;	if (pdu_size);//quiet, compiler
  
  debug2("ARQ SS : Entering the process_mac_pdu_witharqfragpack function \n");	 
  //The packet received is always MAC PDU
  mac_pdu = pkt;
  mac_pdu->RemoveHeader(ah_pdu);

  // We shall first handle the acknowledgements received
  if(ah_pdu.GMH.GMH.type_arqfb == 1)
    {
      for(uint16_t i=0; i < ah_pdu.num_of_acks; i++)
	{
	  Connection *connection = this->getCManager ()->get_connection(ah_pdu.arq_ie[i].cid, true);
	  if(connection)
	    { 
	      if(connection->getArqStatus () && connection->getArqStatus ()->isArqEnabled () == 1){
		debug2("ARQ : SS Feedback in Data Payload Received: Has a feedback: Value of i:%d , Value of number of acks:%d \n" , i, ah_pdu.num_of_acks);
		connection->getArqStatus ()->arqRecvFeedback(mac_pdu, i, connection);
	      }
	    }  
	}
      ah_pdu.GMH.GMH.len -= (ah_pdu.num_of_acks * HDR_MAC802_16_ARQFEEDBK_SIZE);	
    }

  //Add loss on data connection
  UniformVariable unfv(0,1);
  tmp = unfv.GetValue(0,1);
  if (tmp < data_loss_) {
    //update_watch(&loss_watch_,1);  //CAT- uncomment
    //drop (pktRx_, DROP_MAC_PACKET_ERROR);
	drop (pkt, "DROP_MAC_PACKET_ERROR");
    //pktRx_ = NULL;
    debug2 ("ARQ: SS process_mac_pdu_witharqfragpack %f Drop data loss %f %f\n", NOW, tmp, data_loss_);
    return;
  } else {
    debug2 ("ARQ: SS process_mac_pdu_witharqfragpack %f No drop data loss %f %f\n", NOW, tmp, data_loss_);
  }
  
  // We need to change the size of the first packet to its original size as in transmission logic 
  ah_pdu.GMH.GMH.len -= HDR_MAC802_16_SIZE+HDR_MAC802_16_FRAGSUB_SIZE; /*Size of generic Mac Header was set (Note one packet header and first ARQ Block is present)*/
  if(ah_pdu.GMH.GMH.ci)
   ah_pdu.GMH.GMH.len -= HDR_MAC802_16_CRC_SIZE;	
   
// //Initial Condition
//   p_previous = mac_pdu->Copy();	
//   p_temp = mac_pdu->Copy();
//   PACKING_SUBHEADER psh_temp, psh_previous;
//   p_temp->RemoveHeader(psh_tmp);
//  
//   while(p_temp)
//     {
//       // Update the Mac PDU Size
//       debug2("ARQ SS: p-previous :%p p-temp : %p \n", PeekPointer(p_previous), PeekPointer(p_temp));	
//       debug2 ("ARQ SS: type of ARQ to update size:  %d, %d\n ",psh_temp.data.fc, ((psh_temp.data.sn - (HDR_MAC802_16(p_previous)->pack_subheader.sn)) &0x7FF));	
      // if((HDR_MAC802_16(p_temp)->pack_subheader.fc == FRAG_FIRST) || ((((HDR_MAC802_16(p_temp)->pack_subheader.sn) - (HDR_MAC802_16(p_previous)->pack_subheader.sn)) &0x7FF) != 1) || (HDR_MAC802_16(p_temp)->pack_subheader.fc == FRAG_NOFRAG))
// 	ch_pdu->size() -= HDR_CMN(p_temp)->size()+HDR_MAC802_16_FRAGSUB_SIZE ;  	
//       else
// 	ch_pdu->size() -= HDR_CMN(p_temp)->size();
// 
//       //Initial conditions
//       p_previous = p_temp;
//       p_temp = (Ptr<Packet> ) p_temp->accessdata();
//     }	
//   debug2("ARQ SS: Finally the ch_pdu size: %d \n",ch_pdu->size()); 
//   
  
  	// Initial Condition
  	int pdu_len = ah.GMH.GMH.len;
	p_previous = mac_pdu;	
	PACKING_SUBHEADER psh_temp, psh_prev;
	p_temp = mac_pdu->Copy();
	p_temp->RemoveHeader(psh_temp);

 	/*Maybe there are a lot of ARQ blocks in this incoming packet, but they belong to different MAC SDU.
 	   We first update the size of first ARQ block. and use a loop to remove all the overheads.*/
 	debug2("ARQ BS: before update, the ch_pdu size: %d \n",ah_pdu.GMH.GMH.len); 
	while(psh_temp.data.fc > 1)
	{
		// Update the Mac PDU Size
		debug2("ARQ:BS p-previous :%p p-temp : %p \n", PeekPointer(p_previous), PeekPointer(p_temp));	
		debug2 ("ARQ : BS type of ARQ to update size:  %d, %d\n ",psh_temp.data.fc, (((psh_temp.data.sn) - (psh_prev.data.sn)) &0x7FF));	
		
		if ((psh_temp.data.fc == FRAG_FIRST) ||
		   (((psh_temp.data.sn - psh_prev.data.sn ) & (0x7FF)) != 1 )|| 
		   psh_temp.data.fc == FRAG_NOFRAG){
			 pdu_len -= (psh_temp.data.length + psh_temp.GetSerializedSize());
		}
		else pdu_len -= psh_temp.GetSerializedSize();

		//Initial conditions
		psh_prev = psh_temp;
		p_temp->RemoveAtStart(psh_temp.data.length);
		p_temp->RemoveHeader(psh_temp);
	}	
	debug2("ARQ BS: Finally the pdu size: %d \n",pdu_len); 
  
	//CAT - the packets being enqueued for ARQ processing need headers
  // Time to perform ARQ for The MAC_PDU since it is the first ns packet
  if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1) {
    isInOrder = 1;
    con->getArqStatus ()->arqReceive(mac_pdu,con,&isInOrder);
    debug2("Receive: ARQ SS :ORDER is:%d\n",isInOrder);
    if (isInOrder == 0) {
      // packet is out of order, so queue it in retransmission queue at receiver end since it is not being used
      debug2("Receive: ARQ SS : Out of order packet, buffering in retransmission queue\n");
      
      con->getArqStatus ()->arq_retrans_queue_->Enqueue(mac_pdu->Copy());
    }
    else if (isInOrder == 1){
      // packet is in order, so queue it in transmission queue at receiver end since it is not being used
      debug2("Receive: ARQ SS : In order packet, buffering in arq transmission queue\n");
      con->getArqStatus ()->arq_trans_queue_->Enqueue(mac_pdu->Copy());
    }
    else {
      debug2("Receive: ARQ SS : Not buffering in arq trans/retrans queue\n");
    }	 
  }
  // The the initial conditions
  p_temp = mac_pdu->Copy();
  p_temp->RemoveHeader(psh_temp);
  p_temp = p_temp->CreateFragment(0,psh_temp.data.length); 
  Ptr<Packet> p_que;
  All_Headers ah_que;
  
  while(p_temp->GetSize() > 0)
    {
      p_temp->PeekHeader(psh_temp);
     // wimaxHdr_current= HDR_MAC802_16(p_current);
     // ch_current = HDR_CMN(p_current);

      //Time to perform ARQ for the new packet
	  if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1) {
		isInOrder = 1;
		con->getArqStatus ()->arqReceive(p_temp,con,&isInOrder);
		debug2("Receive: ARQ SS :ORDER is:%d\n",isInOrder);
		p_que = p_temp->CreateFragment(0,psh_temp.data.length);
		ah_que.GMH.GMH.len = psh_temp.data.length;
		ah_que.direction = DL;
		ah_que.PSH.data = psh_temp.data;
		p_que->AddHeader(ah_que);
		if (isInOrder == 0) {
			  // packet is out of order, so queue it in retransmission queue at receiver end since it is not being used
			  debug2("Receive: ARQ SS : Out of order packet, buffering in retransmission queue\n");
			  con->getArqStatus () -> arq_retrans_queue_->Enqueue(p_que);
		}
		else if(isInOrder == 1) {
		  // packet is in order, so queue it in transmission queue at receiver end since it is not being used
		  debug2("Receive: ARQ SS : In order packet, buffering in arq transmission queue\n");
		  con->getArqStatus ()->arq_trans_queue_->Enqueue(p_que);
		}
		else {
		  debug2("Receive: ARQ SS :  Not buffering in arq trans/retrans queue\n"); 	
		} 
	  }
      // The the initial conditions
      p_temp->RemoveAtStart(psh_temp.data.length+psh_temp.GetSerializedSize());
    }
 
  // Here we will check if packets in retransmission queue can be fit into transmission queue now that a lost packet may have been retransmitted
 primary:
  con->getArqStatus ()->arq_retrans_queue_->resetIterator();
   
  while(con->getArqStatus ()->arq_retrans_queue_)
    {
    secondary:
      p = con->getArqStatus ()->arq_retrans_queue_->getNext ();
     
      if(!p){
		debug2("ARQ SS: Retransmission queue is empty- so nothing to fit or packets are there but of no use\n");
		break;
      }  		 

      //Time to perform ARQ for the new packet
      if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1) {
		isInOrder = 1;
		con->getArqStatus ()->arqReceiveBufferTransfer(p,con,&isInOrder);
		if(isInOrder == 2)
		  {
			con->getArqStatus ()->arq_retrans_queue_->Remove(p);
			p = 0;
			goto secondary;		
		  }
			if (isInOrder == 0) {
			  goto secondary;
			}
			else {
			  con->getArqStatus ()->arq_trans_queue_->Enqueue(p->Copy());
			  con->getArqStatus ()->arq_retrans_queue_->Remove(p);
			  p = 0;
			  
			  while(con->getArqStatus ()->arq_retrans_queue_)
			  {
				  p = con->getArqStatus ()->arq_retrans_queue_->getNext ();
				
				  if(!p){
					debug2("ARQ SS: Retransmission queue is empty- so nothing to fit or packets are there but of no use\n");
					break;
				  }   		
				  //Time to perform ARQ for the new packet
				  if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1) {
					isInOrder = 1;
					con->getArqStatus ()->arqReceiveBufferTransfer(p,con,&isInOrder);
					if(isInOrder == 2)
					  {
						con->getArqStatus ()->arq_retrans_queue_->Remove(p);
						p = 0;
						goto primary; 		
					  }  	
					if (isInOrder == 0) {
					  goto primary;
					}
					else {
					  con->getArqStatus ()->arq_trans_queue_->Enqueue(p->Copy());
					  con->getArqStatus ()->arq_retrans_queue_->Remove(p);
					  p = 0;
					}
				  }
			}
		  //break;
		  goto primary;	  					 
			}
		 
		  }		    	
    }
   
  //The advantage of seperating the packets in transmission and retransmission queue is the all the packets in transmission queue will be in order.
  while(1) {
    con->getArqStatus ()->arq_trans_queue_->resetIterator();
    p = con->getArqStatus ()->arq_trans_queue_->getNext ();
  
    if(!p) {
      debug2("ARQ SS: Transmission Queue is Empty \n ");	
      break;
    }	
    p->PeekHeader(ah);
//    wimaxHdr= HDR_MAC802_16(p);
//    ch = HDR_CMN(p);

    // This will be first part of the MAC SDU
    if(ah.PSH.data.fc == FRAG_FIRST) {
      mac_sdu = p->Copy();
      mac_sdu->RemoveHeader(ah_sdu);
      //wimaxHdr_sdu= HDR_MAC802_16(mac_sdu);
      //ch_sdu = HDR_CMN(mac_sdu);
      seqno = ah_sdu.PSH.data.sn; 
      debug2("ARQ SS: This is the first packet in transmission queue and FRAG_FIRST Fsn: %d Size :%d\n",ah_sdu.PSH.data.sn, ah.GMH.GMH.len);
      while(1) {
        p =  con->getArqStatus ()->arq_trans_queue_->getNext ();
        if(!p){
	  	  debug2("ARQ SS: All ARQ Blocks for the MAC SDU are not present \n");
          break;
		}
		 p->PeekHeader(ah);

		if((ah.PSH.data.fc == FRAG_CONT) && (((ah.PSH.data.sn - seqno) &0x7FF)==1)) {
		  debug2("ARQ SS: Generating the MAC SDU Current SN: %d Previous SN: %d Frag Type: %d Size: %d, %d\n ",ah.PSH.data.sn, seqno, ah.PSH.data.fc, ah.GMH.GMH.len, ah_sdu.GMH.GMH.len); 	
		  ah_sdu.GMH.GMH.len += ah.GMH.GMH.len;
		  seqno++;	
		}
		else if(ah.PSH.data.fc == FRAG_LAST && (((ah.PSH.data.sn - seqno) &0x7FF)==1)) {
			  debug2("ARQ SS: Generating the MAC SDU Current SN: %d Previous SN: %d Frag Type: %d Size: %d, %d\n ",ah.PSH.data.sn, seqno, ah.PSH.data.fc, ah.GMH.GMH.len, ah_sdu.GMH.GMH.len);
		  ah_sdu.GMH.GMH.len += ah.GMH.GMH.len;
		  mac_sdu_gen = true; 
		  break;		
			}
		else {
			  debug2("ARQ SS:This is a unsual case FSN: %d  and Frag_Type: %d \n",ah.PSH.data.sn, ah.PSH.data.fc);		 
		  break;
		}		
      }
      if(mac_sdu_gen == true){
	debug2("ARQ SS: Created the Mac SDU SIZE: %d\n ", ah_sdu.GMH.GMH.len);
        //update_throughput (&rx_data_watch_, 8*pdu_size);    
        //update_throughput (&rx_traffic_watch_, 8*pdu_size);
        ah_sdu.direction = UL;
      //  mac_sdu->AddHeader(ah_sdu);
        
      UL_Protocol_Header uph;
      mac_sdu->RemoveHeader(uph);
      uint16_t proto_num = uph.protocol;
		
	//m_rxCallback (this, mac_sdu, proto_num, classifier_list_.lh_first->declassify(ah.GMH.GMH.cid));        
        //note a classification system using more than the first entry of the classifier list might not work with the following
        Mac48Address origin = declassify(ah.GMH.GMH.cid);
        m_traceRx(mac_sdu, origin);//, origin);
	Simulator::ScheduleWithContext(this->GetNode()->GetId(), Seconds(0), &Wimax2NetDevice::Receive, this, mac_sdu, proto_num, addr(), origin);
//	Receive(mac_sdu, proto_num, addr(), origin);

        //uptarget_->recv(mac_sdu, (Handler*) 0);
		mac_sdu_gen = false;
        // Go ahead and delete all the packets pertaining to this MAC SDU from the transmission queue
	do {
	  p_temp = con->getArqStatus ()->arq_trans_queue_->Dequeue();
	  p_temp = 0;
	  debug2("ARQ SS: To keep a count on how many times Dequeue function is called \n");
	  p = con->getArqStatus ()->arq_trans_queue_->head();
	  if(!p){
	    goto end;
	  }
	  p->RemoveHeader(ah);
	  //wimaxHdr= HDR_MAC802_16(p);
	  //ch = HDR_CMN(p);
	  debug2("ARQ SS: Next Packet to Dequeue FSN: %d FRAG_STATUS: %d \n",ah.PSH.data.sn, ah.PSH.data.fc);
	} while(ah.PSH.data.fc != FRAG_FIRST);
		
      }
      // Either no packets in trans queue or all the mac sdu that could be generated are sent above
      else {
		goto end; 
		debug2("ARQ SS: Either no packets in trans queue or all the mac sdu that could be generated are sent above \n");
      }
    }
    // This will be the only part of the MAC SDU
    else if(ah.PSH.data.fc == FRAG_NOFRAG) {
      mac_sdu = p->Copy();
      mac_sdu->RemoveHeader(ah_sdu);
//      wimaxHdr_sdu= HDR_MAC802_16(mac_sdu);
//      ch_sdu = HDR_CMN(mac_sdu);
      seqno = ah_sdu.PSH.data.sn; 
      debug2("ARQ SS: This is the only fragment  of MAC SDU FRAG_NOFRAG Fsn: %d Size :%d\n",ah_sdu.PSH.data.sn, ah_sdu.GMH.GMH.len);
      debug2("ARQ SS: Created the Mac SDU SIZE: %d\n ", ah_sdu.GMH.GMH.len);
      //update_throughput (&rx_data_watch_, 8*pdu_size);    
      //update_throughput (&rx_traffic_watch_, 8*pdu_size);
      UL_Protocol_Header uph;
      mac_sdu ->RemoveHeader(uph);
      uint16_t proto_num = uph.protocol;
      
      //m_rxCallback (this, mac_sdu, proto_num, classifier_list_.lh_first->declassify(ah.GMH.GMH.cid));
       //note a classification system using more than the first entry of the classifier list might not work with the following
       Mac48Address origin = declassify(ah.GMH.GMH.cid);
       m_traceRx(mac_sdu, origin);//, origin);
	Simulator::ScheduleWithContext(this->GetNode()->GetId(), Seconds(0), &Wimax2NetDevice::Receive, this, mac_sdu, proto_num, addr(), origin);
//      Receive(mac_sdu, proto_num, addr(), origin);
      //uptarget_->recv(mac_sdu, (Handler*) 0);
      mac_sdu_gen = false;
      // Go ahead and delete all the packets pertaining to this MAC SDU from the transmission queue
      p_temp =con->getArqStatus ()->arq_trans_queue_->Dequeue();
      p_temp = 0;
      debug2("ARQ SS: To keep a count on how many times Dequeue function is called \n");
    } 
    else {
      debug2("ARQ SS: Something really wrong with the implementation\n");	 
      abort();
    }			
  }
 end:
  pkt = 0;
  //update_watch(&loss_watch_,0); //CAT- uncomment
  //pktRx_=NULL; 
}
//End RPI

/**
 * Process a MAC packet
 * @param p The MAC packet received
 */
void Wimax2SSNetDevice::process_mac_packet (Ptr<Packet> p) {
  NS_LOG_FUNCTION_NOARGS ();
  assert ( isMACPacket(p) );


  debug10 ("SS %s received MAC packet to process\n", addr_s());
  
  All_Headers ah;
  p->RemoveHeader(ah);
  debug10 ("Packetsize: %d\n", p->GetSize());

  
  if (ah.GMH.GMH.ht == 1) {
    debug10 ("SS %s received bandwitdh request packet..don't process\n", addr_s());
    return;
  }
  //Begin RPI
  
  if (p->GetSize() == 0) {
  	debug10("SS %s recieved empty packet (just GMH).  Not processing\n", addr_s());
  	return;
  }
   
  if(ah.GMH.GMH.type_arqfb == 1)
	{
	  for(uint16_t i=0; i < ah.num_of_acks; i++)
	    {
	      Connection *connection = this->getCManager ()->get_connection(ah.arq_ie[i].cid, true);
	      if(connection)
		{ 
		  if(connection->getArqStatus () && connection->getArqStatus ()->isArqEnabled () == 1){
		    debug2("ARQ SS: Feedback Payload Received: Has a feedback: Value of i:%d , Value of number of acks:%d \n" , i, ah.num_of_acks);
		    connection->getArqStatus ()->arqRecvFeedback(p, i, connection);
		  }
		}  
	    }
	  p = 0;
	  
	  return ;
	}
    
    
  Type_Check th;
  p->PeekHeader(th);
  
  //mac802_16_dl_map_frame *frame = (mac802_16_dl_map_frame*) p->accessdata();

	//note, {} in case statements to define scope of declared frame header objects
  switch (th.type) {
  case MAC_DL_MAP: 
  	{	
	//this really shouldn't get called from here anymore, as it gets directly called after FCH processing
  	getMap()->setStarttime (ah.timestamp - (DL_PREAMBLE * phy_->getSymbolTime()) );
    	debug2 ("SS %s, MAC_DL_MAP of length %d At %f frame start at %f\n", addr_s(), p->GetSize(), NOW, getMap()->getStarttime());
    	DL_MAP_FRAME_HEADER frame;
    	p->RemoveHeader(frame);
    	process_dl_map (&frame, ah.timestamp);
    	}
    break;

    //collect receive signal strength stats
    //START CAT - UNCOMMENT
//     PeerNode *peer = getPeerNode_head(); 
//     if (peer) { //it is possible that during a scan the peer is removed if switching channel after processing the dl_map
//       //peer->getStatWatch()->update(10*log10(p->txinfo_.RxPr*1e3)); //CAT - uncomment
//       //debug10 ("At %f in Mac %s weighted RXThresh: %e rxp average %e lgd %d\n", NOW, index_, macmib_->lgd_factor_*macmib_->RXThreshold_, pow(10,peer->getStatWatch()->average()/10)/1e3); //CAT - uncomment
//       //double avg_w = pow(10,(peer->getStatWatch()->average()/10))/1e3; //CAT - uncomment
//       
//       if ( avg_w < (macmib_->lgd_factor_*macmib_->RXThreshold_)) {
// 	if (state_==MAC802_16_CONNECTED) {
// #ifdef USE_802_21
// 	  if(mih_){
// 	    double probability = ((macmib_->lgd_factor_*macmib_->RXThreshold_)-avg_w)/((macmib_->lgd_factor_*macmib_->RXThreshold_)-macmib_->RXThreshold_);
// 	    Mac::send_link_going_down (addr(), peer->getAddr(), -1, (int)(100*probability), LGD_RC_LINK_PARAM_DEGRADING, eventId_++);
// 	  }else{
// #endif
// 	    if (!peer->isGoingDown ()) { //when we don't use 802.21, we only want to send the scan request once	    
// 	      link_scan (NULL); //replaced send_scan_request (); so it checks for pending requests
// 	    }
// #ifdef USE_802_21
// 	  }
// #endif
// 	  peer->setGoingDown (true);
// 	}
//       }
//       else {
// 	if (peer->isGoingDown()) {
// #ifdef USE_802_21
// 	  Mac::send_link_rollback (addr(), peer->getAddr(), eventId_-1);
// #endif
// 	  peer->setGoingDown (false);
// 	}
//       }
//     }
//END CAT - UNCOMMENT
  //break;
  case MAC_DCD: 
  	{
  	debug2 ("At SS MAC_DCD \n");
  	DCD_FRAME_HEADER dcd_frame;
  	p->RemoveHeader(dcd_frame);
    process_dcd (&dcd_frame);
    }
    break;
  case MAC_UL_MAP: 
  	{
    debug2 ("At SS MAC_UL_MAP \n");
  	UL_MAP_FRAME_HEADER ul_map_frame;
  	p->RemoveHeader(ul_map_frame);
    process_ul_map (&ul_map_frame, ah.txtime + (NOW-ah.timestamp) + 2 * phy_->getSymbolTime());
    }
    break;
  case MAC_UCD: 
  	{
    debug2 ("At SS MAC_UCD \n");
  	UCD_FRAME_DATA ucd_frame;
  	p->RemoveHeader(ucd_frame);
    process_ucd (&ucd_frame);
    }
    break;
  case MAC_RNG_RSP:
  	{
    debug2 ("At SS MAC_RNG_RSP \n");
  	RANGING_RESPONSE_FRAME rng_rsp_frame;
  	p->RemoveHeader(rng_rsp_frame);
    process_ranging_rsp (&rng_rsp_frame);
    }
    break;
  case MAC_REG_RSP:
  	{
    debug2 ("At SS MAC_REG_RSP \n");
  	REGISTRATION_RESPONSE_FRAME reg_rsp_frame;
  	p->RemoveHeader(reg_rsp_frame);
    process_reg_rsp (&reg_rsp_frame);
    }
    break;    
  case MAC_MOB_SCN_RSP:
  	{
  	debug2 ("At MAC_MOB_SCN_RSP \n");
  	mac802_16_mob_scn_rsp_frame scn_rsp_frame;
  	p->RemoveHeader(scn_rsp_frame);
    process_scan_rsp (&scn_rsp_frame);
    }
    break;
  case MAC_MOB_BSHO_RSP:
  	{
  	debug2 ("At MAC_MOB_BSHO_RSP \n");
  	BSHO_RESPONCE_FRAME_HEADER bsho_rsp_frame;
  	p->RemoveHeader(bsho_rsp_frame);
    process_bsho_rsp (&bsho_rsp_frame);
    }
    break;
  case MAC_MOB_NBR_ADV:
  	{
  	debug2 ("At MAC_MOB_NBR_ADV \n");
  	NEIGHBOR_ADVERTISEMENT_FRAME_HEADER nbr_adv_frame;
  	p->RemoveHeader(nbr_adv_frame);
    process_nbr_adv (&nbr_adv_frame);
    }
    break;
  case MAC_DSA_REQ: 
  case MAC_DSA_RSP: 
  case MAC_DSA_ACK: 
  	debug2("At MAC_DSA_REQ, DSA_RSP, or DSA_ACK \n");
  	p->AddHeader(ah);
	serviceFlowHandler_->process (p);
    break;
  default:
    debug10 ("unknown packet in SS %s\n", addr_s());
    //exit (0);
  }
}


/**
 * Process a DL_MAP message
 * @param frame The dl_map information
 */
void Wimax2SSNetDevice::process_dl_map (DL_MAP_FRAME_HEADER *frame_h, double startTime)
{
  NS_LOG_FUNCTION_NOARGS ();
  if (this_frame_dlmap_processed) 
  	return; //this is a retransmission
  else Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(.005 - phy_->getSymbolTime()), &Wimax2SSNetDevice::reset_repetition_detectors, this);
  assert (frame_h);
  mac802_16_dl_map_frame *frame = &frame_h->mf;
  getMap()->getDlSubframe()->setAbsStarttime(startTime);

  //schedule incoming packet assembly
  for (uint32_t i = 0; i < frame->nb_ies; i++) {
  	if (getCManager()->get_connection(frame->ies[i].ie.cid, IN_CONNECTION)) {
		PHY_INFO_Header *pih = new PHY_INFO_Header();
		pih->direction = DL;
		pih->subchannel_offset = frame->ies[i].ie.subchannel_offset;
		pih->OFDMSymbol_offset = frame->ies[i].ie.OFDMASymbol_offset;
		pih->num_subchannels = frame->ies[i].ie.num_subchannel;
		pih->num_OFDMSymbol = frame->ies[i].ie.num_OFDMASymbol;
		Time reception = start_of_frame + Seconds((pih->OFDMSymbol_offset + pih->num_OFDMSymbol) * getPhy()->getSymbolTime() + getPhy()->getPS()) - Simulator::Now();
		//we add a single PS to the time to process the reciept to avoid rounding errors in the time
		debug10("At %f, in SS %s, scheduling assembly of a packet for %f, just after reception should end according to DL_MAP\n", NOW, addr_s(), reception.GetSeconds());
		debug10("\t mapped block is (symbols %d to %d, subchannels %d to %d)\n", pih->OFDMSymbol_offset, pih->OFDMSymbol_offset + pih->num_OFDMSymbol - 1, pih->subchannel_offset, pih->subchannel_offset + pih->num_subchannels);
		Simulator::ScheduleWithContext(Simulator::GetContext(), reception, &Wimax2NetDevice::assemble_incoming_packet, this, pih);
	}
  }
  
  //create an entry for the BS
  if (getPeerNode_head ()==NULL)
    addPeerNode (new PeerNode (frame->bsid));

  getMap()->parseDLMAPframe (frame_h);

  if (getMacState()==MAC802_16_WAIT_DL_SYNCH) {
    debug10 ("At %f in %s, received DL_MAP for synch from %s (substate=%d)\n", NOW, addr_s(), HelperFunctions::getMacAddrString(frame->bsid).c_str(),scan_info_->substate);

    setBSAddr(frame->bsid);
    assert (t21timer_->busy()!=0);
    //synchronization is done
    t21timer_->stop();
    //start lost_dl_map
    lostDLMAPtimer_->start (macmib_->lost_dlmap_interval);
    //start T1: DCD
    t1timer_->start (macmib_->t1_timeout);
    //start T12: UCD
    t12timer_->start (macmib_->t12_timeout);

#ifdef USE_802_21
    if (scan_info_->substate != SCANNING) {
      debug10 ("At %f in Mac %s, send link detected\n", NOW, addr_s());
      send_link_detected (addr(), frame->bsid, 1);
    }
#endif
    
    setMacState(MAC802_16_WAIT_DL_SYNCH_DCD);

    //if I am doing handoff and we have dcd/ucd information 
    //from scanning, use it
    if (scan_info_->substate == HANDOVER || scan_info_->substate == SCANNING) {
      if (scan_info_->substate == SCANNING) {
	if (scan_info_->nbr == NULL || scan_info_->nbr->getID()!=frame->bsid) {
	  //check if an entry already exist in the database
	  scan_info_->nbr = nbr_db_->getNeighbor (frame->bsid);
	  if (scan_info_->nbr == NULL) {
	    //create entry
	    debug2 ("Creating nbr info for node %s\n", HelperFunctions::getMacAddrString(frame->bsid).c_str());
	    scan_info_->nbr = new WimaxNeighborEntry (frame->bsid);
	    nbr_db_->addNeighbor (scan_info_->nbr);
	  } else {
	    debug2 ("loaded nbr info\n");
	    if (scan_info_->nbr->isDetected ()) {
 	      //we already synchronized with this AP...skip channel
 	      nextChannel();
 	      lost_synch ();
 	      return;
 	    } 
	  }
	}
      }//if HANDOVER, scan_info_->nbr is already set

      bool error = false;
      //we check if we can read the DL_MAP
      DCD_FRAME_HEADER *dcd = scan_info_->nbr->getDCD();
      if (dcd!=NULL) {
	debug2 ("Check if we can decode stored dcd\n");
	//check if we can decode dl_map with previously acquired dcd      
	bool found;
	for (int i = 0 ; !error && i < getMap()->getDlSubframe()->getPdu()->getNbBurst() ; i++) {
	  int diuc = getMap()->getDlSubframe()->getPdu()->getBurst(i)->getIUC();
	  if (diuc == DIUC_END_OF_MAP)
	    continue;
	  found = false;
	  for (uint32_t j = 0 ; !found && j < dcd->mf.nb_prof; j++) {
	    found = dcd->mf.profiles[j].diuc==diuc;	    
	  }
	  error = !found;
	}
	if (!error)
	  process_dcd (dcd);
      } else {
	debug2 ("No DCD information found\n");
      }
    }
  } else {
    //maintain synchronization
    assert (lostDLMAPtimer_->busy());
    lostDLMAPtimer_->stop();
    //printf ("update dlmap timer\n");
    lostDLMAPtimer_->start (macmib_->lost_dlmap_interval);

    if (getMacState()!= MAC802_16_WAIT_DL_SYNCH_DCD
	&& getMacState()!=MAC802_16_UL_PARAM) {

      //since the map may have changed, we need to adjust the timer 
      //for the DLSubframe
 //      double stime = getMap()->getStarttime() + getMap()->getDuration() - NOW;
// //      stime += getMap()->getDlSubframe()->getPdu()->getBurst(1)->getStarttime()*getPhy()->getSymbolTime();
//       
//       debug10("rescheding dlsubframetimer from SSNetDevice::process_dl_map\n");
//       debug10("\t\t\t\tstime = %f\n",stime);
//       getMap()->getDlSubframe()->getTimer()->resched (stime);
	dl_timer_->resched (getMap()->getStarttime()+getFrameDuration());
    }
  }
  debug10("new_dlmap_processed now equals TRUE\n");
  new_dlmap_processed = true;
  this_frame_dlmap_processed = true;
}

/**
 * Process a DCD message
 * @param frame The dcd information
 */
void Wimax2SSNetDevice::process_dcd (DCD_FRAME_HEADER *frame_h)
{
  NS_LOG_FUNCTION_NOARGS ();
  mac802_16_dcd_frame *frame = &frame_h->mf;
  if (getMacState()==MAC802_16_WAIT_DL_SYNCH) {
    //we are waiting for DL_MAP, ignore this message
    return;
  }

  getMap()->parseDCDframe (frame);
  if (getMacState()==MAC802_16_WAIT_DL_SYNCH_DCD) {
    debug10 ("At %f in %s, received DCD for synch\n", NOW, addr_s());
    //now I have all information such as frame duration
    //adjust timing in case the frame we received the DL_MAP
    //and the DCD is different
    while (NOW - getMap()->getStarttime () > getFrameDuration()) {
      getMap()->setStarttime (getMap()->getStarttime()+getFrameDuration());
    }
    
    //store information to be used during potential handoff
    if (scan_info_->substate == SCANNING) {
      DCD_FRAME_HEADER *tmp_h = new DCD_FRAME_HEADER();
      DCD_FRAME_HEADER *old_h = scan_info_->nbr->getDCD();
      mac802_16_dcd_frame *tmp = &tmp_h->mf;
      mac802_16_dcd_frame *old = &old_h->mf;
      memcpy (tmp, frame, sizeof (mac802_16_dcd_frame));
      //mac802_16_dcd_frame *old = scan_info_->nbr->getDCD(); 
      if (frame == old)
		frame = tmp;
      if (old)
		free (old); //free previous entry
      scan_info_->nbr->setDCD(tmp_h);    //set new one
    }

    setMacState(MAC802_16_UL_PARAM);
    //we can schedule next frame
    //printf ("SS schedule next frame at %f\n", getMap()->getStarttime()+getFrameDuration());
    dl_timer_->resched(getMap()->getStarttime()+getFrameDuration()-NOW);
  }

  if (t1timer_->busy()!=0) {
    //we were waiting for this packet
    t1timer_->stop();
    t1timer_->start (macmib_->t1_timeout);
  }
}

/**
 * Process a UCD message
 * @param frame The ucd information
 */
void Wimax2SSNetDevice::process_ucd (UCD_FRAME_DATA *frame_h)
{
  NS_LOG_FUNCTION_NOARGS ();
  mac802_16_ucd_frame *frame = &frame_h->ucdf;
  
  if (getMacState()==MAC802_16_WAIT_DL_SYNCH
      ||getMacState()==MAC802_16_WAIT_DL_SYNCH_DCD) {
    //discard the packet
    return;
  }
  assert (t12timer_->busy()!=0); //we are waiting for this packet

  if (getMacState()==MAC802_16_UL_PARAM) {
    //check if uplink channel usable
    debug10 ("At %f in %s, received UL(UCD) parameters\n", NOW, addr_s());
    //start T2: ranging
    t2timer_->start (macmib_->t2_timeout);
    //start Lost UL-MAP
    lostULMAPtimer_->start (macmib_->lost_ulmap_interval);

    //store information to be used during potential handoff
    if (scan_info_->substate == SCANNING) {
      UCD_FRAME_DATA *tmp_h = new UCD_FRAME_DATA();
      UCD_FRAME_DATA *old_h = scan_info_->nbr->getUCD();
      mac802_16_ucd_frame *tmp = &tmp_h->ucdf;
	  mac802_16_ucd_frame *old = &old_h->ucdf;
      memcpy (tmp, frame, sizeof (mac802_16_ucd_frame));
      if (frame == old)
		frame = tmp;
      if (old) 
		free (old); //free previous entry
      scan_info_->nbr->setUCD(tmp_h);    //set new one            
      
    }

    //change state
    debug10 (" received ucd and prepare to do init_ranging set state to MAC802_16_RANGING\n");
    setMacState (MAC802_16_RANGING);
  }

  //reset T12
  t12timer_->stop();
  t12timer_->start (macmib_->t12_timeout);

  getMap()->parseUCDframe (frame);
}

/**
 * Process a UL_MAP message
 * @param frame The ul_map information
 */
void Wimax2SSNetDevice::process_ul_map (UL_MAP_FRAME_HEADER *frame_h, double txtime)
{
  NS_LOG_FUNCTION_NOARGS ();
  if (this_frame_ulmap_processed) return; //this is a retransmission
  
	//the DL frame may come in at the same time, and the scheduler may process this first if they come in
	//at identical times, so we put this off VERY SLIGHTLY if the DL_MAP hasn't been processed yet
  if (!new_dlmap_processed) {
		debug10("pushing the processing of UL_MAP very slightly ahead so DL_MAP is processed first\n");
		cout << "1 PS: " << Seconds(getPhy()->getPS()) << " to total time of " << NOW << std::endl;
		Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(getPhy()->getPS()), &Wimax2SSNetDevice::process_ul_map, this, frame_h, txtime);
		return;
  }
  else {
	  mac802_16_ul_map_frame *frame = &frame_h->mf;
	  if (frame); //quiet, compiler
	  if (getMacState()==MAC802_16_WAIT_DL_SYNCH || getMacState()==MAC802_16_WAIT_DL_SYNCH_DCD) {
		//discard the packet
		debug10("At %f in %s, drop UL_MAP because not in right state (state=%d)\n", NOW, addr_s(), getMacState());
		return;
	  }
	
	  debug10 ("At %f in %s, received UL_MAP for synch (substate=%d) (macstate=%d)\n", NOW, addr_s(),scan_info_->substate, getMacState());
	
	  if (getMacState()==MAC802_16_UL_PARAM) {
		if (scan_info_->substate == HANDOVER || scan_info_->substate==SCANNING) {
		  FrameMap *tmpMap = new FrameMap (this);
		  tmpMap->parseULMAPframe (frame_h, txtime); 
		  //printf ("Checking if we can read UL_MAP\n");
		  bool error = false;
		  //we check if we can read the UL_MAP
		  UCD_FRAME_DATA *ucd = scan_info_->nbr->getUCD();
		  if (ucd!=NULL) {
			//check if we can decode ul_map with previously acquired ucd      
			bool found;
			for (uint32_t i = 0 ; !error && i < tmpMap->getUlSubframe()->getNbPdu() ; i++) {
			  UlBurst *b = (UlBurst*)tmpMap->getUlSubframe()->getPhyPdu(i)->getBurst(0);
			  int uiuc = b->getIUC();
			  if (uiuc == UIUC_EXT_UIUC && b->getExtendedUIUC ()== UIUC_FAST_RANGING)
				uiuc = b->getFastRangingUIUC();	  
			  found = false;
			  for (uint32_t j = 0 ; !found && j < ucd->ucdf.nb_prof; j++) {
				//printf ("\t prof=%d, search=%d\n", ucd->profiles[j].uiuc, uiuc);
				found = ucd->ucdf.profiles[j].uiuc==uiuc;	    
			  }
			  error = !found;
			}
			if (!error)
			  process_ucd (ucd);
	    }
		delete (tmpMap);
		if (error) {
			//we cannot read message
			return;
		}
	  } 
	  else return;
    }
	
	  if (scan_info_->substate == SCANNING) {
		//TBD: add checking scanning type for the given station
		uint8_t scanning_type = 0;
		for (int i = 0 ; i < scan_info_->rsp->n_recommended_bs_full ; i++) {
		  if (scan_info_->rsp->rec_bs_full[i].recommended_bs_id == scan_info_->nbr->getID()) {
				scanning_type = scan_info_->rsp->rec_bs_full[i].scanning_type;
				break;
		  }
		}
		if (scanning_type == 0) {
		  //store information about possible base station and keep scanning
		  scan_info_->nbr->getState()->state_info= backup_state();
		  debug10 ("At %f in Mac %s bs %s detected during scanning\n", NOW, addr_s(), HelperFunctions::getMacAddrString(scan_info_->nbr->getID()).c_str());
		  scan_info_->nbr->setDetected (true);
		  nextChannel();
		  lost_synch ();
		  return;
		}
	  }
	
	  getMap()->parseULMAPframe (frame_h, txtime);  
	  if (getMacState()==MAC802_16_RANGING) 
	  {
		debug10("About to do ranging\n");
		//execute ranging
		assert (t2timer_->busy()!=0); //we are waiting for this packet
		init_ranging ();
	  }

	  //note that the following is duplicated in the process_rng_rsp function, to avoid not seeing allocations because we were in the wrong frame when the UL_MAP was processed
	  else if (getMacState() == MAC802_16_WAIT_RNG_ALLOC) {
		UlSubFrame *ulsubframe = getMap()->getUlSubframe();
		DlSubFrame *dlsubframe = getMap()->getDlSubframe();
	  	for (uint32_t i = 0; i < ulsubframe->getNbPdu(); i++) {
	  		if (ulsubframe->getPhyPdu(i)->getBurst(0)->getB_CDMA_TOP() == ulsubframe->getRanging()->getTOP(addr()) && ulsubframe->getPhyPdu(i)->getBurst(0)->getB_CDMA_CODE() == ulsubframe->getRanging()->getCODE(addr())) {
				UlBurst *b = (UlBurst*) ulsubframe->getPhyPdu(i)->getBurst(0);
				b->setB_CDMA_CODE(ulsubframe->getRanging()->getCODE(addr()));
				b->setB_CDMA_TOP(ulsubframe->getRanging()->getTOP(addr()));
				
				PeerNode *peer = getPeerNode_head();

				Ptr<Packet> p = getPacket();
				All_Headers ah;
				p->RemoveHeader(ah);
				ah.GMH.GMH.cid = INITIAL_RANGING_CID; 
				ah.SMH.BRH.cid = INITIAL_RANGING_CID;
				ah.CRH.data.cid = INITIAL_RANGING_CID;
				RANGING_REQUEST_FRAME frame_h;
				mac802_16_rng_req_frame *frame = &frame_h.data;
				frame->type = MAC_RNG_REQ;
				frame->dc_id = dlsubframe->getChannelID();
				frame->ss_mac_address = addr();
				frame->channel_num = GetInitialChannel();
				peer->setchannel(frame->channel_num);
				//other elements??      
				frame->req_dl_burst_profile = default_diuc_ & 0xF; //we use lower bits only
				ah.GMH.GMH.len += frame_h.GetSerializedSize();
		
				//fill phyinfo header
				//hdr_mac802_16 *wimaxHdr;
				ah.PIH.num_subchannels = b->getnumSubchannels();
				ah.PIH.subchannel_offset = b->getSubchannelOffset();
				ah.PIH.num_OFDMSymbol = b->getDuration();
				ah.PIH.OFDMSymbol_offset = b->getStarttime();
				ah.PIH.channel_index = frame->channel_num;
				ah.PIH.direction = 1;
				ah.PIH.modulation_ = getMap()->getUlSubframe()->getProfile(UIUC_INITIAL_RANGING)->getEncoding();
			
			
				//compute how long it will take to send message
				ah.txtime = getPhy()->getTrxTime (ah.GMH.GMH.len, ulsubframe->getProfile (UIUC_INITIAL_RANGING)->getEncoding());
				//starttime+backoff
				ah.timestamp = NOW; //add timestamp since it bypasses the queue
				p->AddHeader(frame_h);
				p->AddHeader(ah.GMH);
				p->AddHeader(ah.PIH);
				debug10("NOW: %f\n",NOW);
				//b->enqueue(p);
				Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(getMap()->getStarttime()+b->getStarttime()*getPhy()->getSymbolTime() - NOW), &Wimax2NetDevice::transmit, this, p);
				setMacState(MAC802_16_WAIT_RNG_RSP);
	  		}
	  		
	  	}
	  }	
	  //schedule when to take care of outgoing packets
	  debug10("startTime for map: %f\n",getMap()->getDlSubframe()->getAbsStarttime());//getStarttime());
	  debug10("Current Time: %f\n", NOW);
	  debug10("offset for UL Subframe (PS) %d \n",frame->allocation_start);
	  debug10("offset for UL Subframe (s) %f \n", frame->allocation_start * getPhy()->getPS());
	  debug10("startTime for Ul Subframe:%f\n", frame->allocation_start * getPhy()->getPS() + getMap()->getDlSubframe()->getAbsStarttime());
	  double start = getMap()->getDlSubframe()->getAbsStarttime() + frame->allocation_start * getPhy()->getPS() - NOW; //offset for ul subframe
	  getMap()->getUlSubframe()->setStarttime(frame->allocation_start);
	  //start -= NOW; //works with relative time not absolute
	 // debug10("testing offset: %f\n",frame->allocation_start * getPhy()->getPS());
	  debug2 ("Uplink starts in %f (DL-start = %f, framestate=%f) %f %f\n", 
		  start, 
		  getMap()->getDlSubframe()->getAbsStarttime(),
		  getMap()->getStarttime(),
		  getFrameDuration()/getPhy()->getPS(), 
		  getFrameDuration()/getPhy()->getSymbolTime());
		  ul_timer_->resched (start);
	
	//scheduling setting the phy mode for the ul subframe
	//we set it to half a symbol time before the transition to make sure it's set before anything arrives, but nothing arrives after it's changed
	Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(start - 0.5*getPhy()->getSymbolTime()), &Wimax2Phy::setMode, getPhy(), OFDM_SEND);
	//scheduling setting the phy mode back for the dl subframe
	//we set it to half a symbol time before the transition to make sure it's set before anything arrives, but nothing arrives after it's changed
	Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(start+((double) MaxUlduration_ - 0.5) * getPhy()->getSymbolTime()), &Wimax2Phy::setMode, getPhy(), OFDM_RECV);
	
	  //reset Lost UL-Map
	  //lostULMAPtimer_->stop();
	  
	  lostULMAPtimer_->resched (macmib_->lost_ulmap_interval);
	  debug10("new_dlmap_processed now equals FALSE\n");
	  new_dlmap_processed = false;
	}
	this_frame_ulmap_processed = true;
}

/**
 * Process a ranging response message 
 * @param frame The ranging response frame
 */
void Wimax2SSNetDevice::process_ranging_rsp (RANGING_RESPONSE_FRAME *frame_h)
{
  NS_LOG_FUNCTION_NOARGS ();
  if (getMacState() != MAC802_16_WAIT_ANON_RNG_RSP && getMacState() != MAC802_16_WAIT_RNG_RSP) {
  	debug2("Received RNG_RSP at SS %s when not in correct state\n", addr_s());
  	return;
  }
  mac802_16_rng_rsp_frame *frame = &frame_h->data;
  //check the destination
  if (getMacState() == MAC802_16_WAIT_ANON_RNG_RSP && frame->ss_mac_address == Mac48Address::GetBroadcast()) {
  	switch (frame->rng_status) {
  	  case RNG_SUCCESS:
	  	  //we need to check the top and code here to know if we should keep it or not
		  if (getMap()->getUlSubframe()->getRanging()->getCODE(addr()) != frame->code || getMap()->getUlSubframe()->getRanging()->getTOP(addr()) != frame->top) {
				debug10("Incorrect Code or Top for incoming RNG_RSP. Stored code and top are %d and %d vs incomming %d and %d\n", getMap()->getUlSubframe()->getRanging()->getCODE (addr()), getMap()->getUlSubframe()->getRanging()->getTOP (addr()), frame_h->data.code, frame_h->data.top); 
				return;
		  }
  	  	  getMap()->getUlSubframe()->getRanging()->removeRequest_mac (Mac48Address::GetBroadcast());
  	  	  setMacState(MAC802_16_WAIT_RNG_ALLOC);

		  //the following would normally be handled in the UL_MAP processing frame, but since we will probably get this packet after we process the UL_MAP
		  //we're just going to duplicate that code.  Note that there is still a copy in process_ul_map

	  	  if (this_frame_ulmap_processed && getMacState() == MAC802_16_WAIT_RNG_ALLOC) {
				UlSubFrame *ulsubframe = getMap()->getUlSubframe();
				DlSubFrame *dlsubframe = getMap()->getDlSubframe();
				for (uint32_t i = 0; i < ulsubframe->getNbPdu(); i++) {
					if (ulsubframe->getPhyPdu(i)->getBurst(0)->getB_CDMA_TOP() == ulsubframe->getRanging()->getTOP(addr()) && ulsubframe->getPhyPdu(i)->getBurst(0)->getB_CDMA_CODE() == ulsubframe->getRanging()->getCODE(addr())) {
						UlBurst *b = (UlBurst*) ulsubframe->getPhyPdu(i)->getBurst(0);
						
						PeerNode *peer = getPeerNode_head();

						Ptr<Packet> p = getPacket();
						All_Headers ah;
						p->RemoveHeader(ah);
						ah.GMH.GMH.cid = INITIAL_RANGING_CID; 
						ah.SMH.BRH.cid = INITIAL_RANGING_CID;
						ah.CRH.data.cid = INITIAL_RANGING_CID;
						RANGING_REQUEST_FRAME frame_h;
						mac802_16_rng_req_frame *frame = &frame_h.data;
						frame->type = MAC_RNG_REQ;
						frame->dc_id = dlsubframe->getChannelID();
						frame->ss_mac_address = addr();
						frame->channel_num = GetInitialChannel();
						peer->setchannel(frame->channel_num);
						//other elements??      
						frame->req_dl_burst_profile = default_diuc_ & 0xF; //we use lower bits only
						ah.GMH.GMH.len += frame_h.GetSerializedSize();

						//fill phyinfo header
						//hdr_mac802_16 *wimaxHdr;

						//note that we don't use the connection/queue/ulsubframetimer system to transmit here becuase
						//it won't take the code/top into account when identifying bursts, which causes issues here, but that
						//functionality is unnecessary except in this specific case
					
						ah.PIH.num_subchannels = b->getnumSubchannels();
						ah.PIH.subchannel_offset = b->getSubchannelOffset(); //richard: changed 1->0
						ah.PIH.num_OFDMSymbol = b->getDuration();
						ah.PIH.OFDMSymbol_offset = b->getStarttime();
						ah.PIH.channel_index = frame->channel_num;
						ah.PIH.direction = 1;
		  				ah.PIH.modulation_ = getMap()->getUlSubframe()->getProfile(UIUC_INITIAL_RANGING)->getEncoding();
					 
						//compute how long it will take to send message
						ah.txtime = getPhy()->getTrxTime (ah.GMH.GMH.len, ulsubframe->getProfile (UIUC_INITIAL_RANGING)->getEncoding());
						//starttime+backoff
						ah.timestamp = NOW; //add timestamp since it bypasses the queue
						p->AddHeader(frame_h);
						p->AddHeader(ah.GMH);
						p->AddHeader(ah.PIH);
						debug10("NOW: %f\n",NOW);
						//b->enqueue(p);
						Simulator::ScheduleWithContext(Simulator::GetContext(), start_of_frame + Seconds(getMap()->getUlSubframe()->getStarttime() * getPhy()->getPS() - NOW), &Wimax2NetDevice::transmit, this, p);
						setMacState(MAC802_16_WAIT_RNG_RSP);
						  
					}
				}	
	 		 }	
  	  	break;
	  case RNG_ABORT:
		setMacState(MAC802_16_DISCONNECTED);
		break;
	  case RNG_CONTINUE:
	  case RNG_RERANGE:
		setMacState(MAC802_16_RANGING);
		break;
	  default:
		fprintf (stderr, "Unknown status reply\n");
		exit (-1);
  	}
  	return;
  } else if (frame->ss_mac_address != addr()) {
    debug10("SS %s recieved RNG_RSP bound for SS %s\n", addr_s(), HelperFunctions::getMacAddrString(frame->ss_mac_address).c_str());
    return;
  }
  
  Connection *basic, *primary;
  PeerNode *peer;

  //TBD: add processing for periodic ranging

  //check status 
  switch (frame->rng_status) {
  case RNG_SUCCESS:
    debug10 ("Ranging response (remove cdma intial ranging request) : status = Success.Basic :%d, Primary :%d\n",
	   frame->basic_cid, frame->primary_cid);

    peer = getPeerNode_head();
    assert (peer);
//getMap()->getUlSubframe()->getRanging()->removeRequest ();
    getMap()->getUlSubframe()->getRanging()->removeRequest_mac (addr());
    Connection *c_tmp;
    c_tmp  = getCManager ()->get_connection (0, true);
    c_tmp->setINIT_REQ_QUEUE(id_index, 0);

    if (scan_info_->substate == SCANNING) {
      //store information about possible base station and keep scanning
      scan_info_->nbr->getState()->state_info= backup_state();
      scan_info_->nbr->setDetected (true);
      //keep the information for later
      RANGING_RESPONSE_FRAME *tmp = new RANGING_RESPONSE_FRAME();
//      mac802_16_rng_rsp_frame *tmp = (mac802_16_rng_rsp_frame *) malloc (sizeof (mac802_16_rng_rsp_frame));
      memcpy (&tmp->data, frame, sizeof (mac802_16_rng_rsp_frame));
      scan_info_->nbr->setRangingRsp (tmp);
      nextChannel();
      lost_synch ();
      return;
    }

    //ranging worked, now we must register
    basic = peer->getBasic(IN_CONNECTION);
    primary = peer->getPrimary(IN_CONNECTION);
    if (basic!=NULL && basic->get_cid ()==frame->basic_cid) {
      //duplicate response
      assert (primary->get_cid () == frame->primary_cid);
    } else {
      if (basic !=NULL) {
	//we have been allocated new cids..clear old ones
	getCManager ()->remove_connection (basic->get_cid());
	getCManager ()->remove_connection (primary->get_cid());
	if (peer->getSecondary(IN_CONNECTION)!=NULL) {
	  getCManager ()->remove_connection (peer->getSecondary(IN_CONNECTION));
	  getCManager ()->remove_connection (peer->getSecondary(OUT_CONNECTION));
	}
	if (peer->getOutData()!=NULL)
	  getCManager ()->remove_connection (peer->getOutData());
	if (peer->getInData()!=NULL)
	  getCManager ()->remove_connection (peer->getInData());
      } 

      basic = new Connection (CONN_BASIC, frame->basic_cid);
      Connection *upbasic = new Connection (CONN_BASIC, frame->basic_cid);
//      upbasic->setBW_REQ_QUEUE(0);
//      upbasic->initCDMA_SSID();
      primary = new Connection (CONN_PRIMARY, frame->primary_cid);
      Connection *upprimary = new Connection (CONN_PRIMARY, frame->primary_cid);
//      upprimary->setBW_REQ_QUEUE(0);
//      upprimary->initCDMA_SSID();

      //a SS should only have one peer, the BS
      peer->setBasic (basic, upbasic); //set outgoing
      peer->setPrimary (primary, upprimary); //set outgoing
      getCManager()->add_connection (upbasic, OUT_CONNECTION);
      getCManager()->add_connection (basic, IN_CONNECTION);
      getCManager()->add_connection (upprimary, OUT_CONNECTION);
      getCManager()->add_connection (primary, IN_CONNECTION);
    }

    //registration must be sent using Primary Management CID
    setMacState (MAC802_16_REGISTER);
    //stop timeout timer
    t2timer_->stop ();
    nb_reg_retry_ = 0; //first time sending
    send_registration();
    break;
  case RNG_ABORT:
  	setMacState(MAC802_16_DISCONNECTED);
  	break;
  case RNG_CONTINUE:
  case RNG_RERANGE:
	setMacState(MAC802_16_RANGING);
    break;
  default:
    fprintf (stderr, "Unknown status reply\n");
    exit (-1);
  }
}

/**
 * Schedule a ranging
 */
void Wimax2SSNetDevice::init_ranging ()
{
  NS_LOG_FUNCTION_NOARGS ();
  //check if there is a ranging opportunity
  UlSubFrame *ulsubframe = getMap()->getUlSubframe();
  DlSubFrame *dlsubframe = getMap()->getDlSubframe();
  PeerNode *peer = getPeerNode_head();

  // If I am doing a Handoff, check if I already associated 
  // with the target BS
  //printf(" inside init ranging" );  

  if (scan_info_->substate == HANDOVER && scan_info_->nbr->getRangingRsp()!=NULL) {
    debug10 ("At %f in Mac %s MN already executed ranging during scanning\n", NOW, addr_s());
    process_ranging_rsp (scan_info_->nbr->getRangingRsp());
    return;
  }

  //check if there is Fast Ranging IE
  //for (PhyPdu *p = getMap()->getUlSubframe ()->getFirstPdu(); p ; p = p->next_entry()) {
  for (std::list<PhyPdu*>::iterator p = getMap()->getUlSubframe()->getPduList()->begin(); p != getMap()->getUlSubframe()->getPduList()->end(); p++) {
    UlBurst *b = (UlBurst*) (*p)->getBurst(0);
    if (b->getIUC() == UIUC_EXT_UIUC   &&   b->getExtendedUIUC ()== UIUC_FAST_RANGING   &&   b->getFastRangingMacAddr ()==addr()) {
		  debug2 ("Found fast ranging\n");
		  //we should put the ranging request in that burst
		  Ptr<Packet> p = getPacket();
		  All_Headers ah;
		  p->RemoveHeader(ah);
		  ah.GMH.GMH.cid = INITIAL_RANGING_CID; 
		  ah.SMH.BRH.cid = INITIAL_RANGING_CID;
		  ah.CRH.data.cid = INITIAL_RANGING_CID;
	
		  RANGING_REQUEST_FRAME frame_h;
		  mac802_16_rng_req_frame *frame = &frame_h.data;
		  frame->type = MAC_RNG_REQ;
		  frame->dc_id = dlsubframe->getChannelID();
		  frame->ss_mac_address = addr();
		  frame->channel_num = GetInitialChannel();
		  peer->setchannel(frame->channel_num);
		  //other elements??      
		  frame->req_dl_burst_profile = default_diuc_ & 0xF; //we use lower bits only
		  ah.GMH.GMH.len += frame_h.GetSerializedSize();
	
		  //fill phyinfo header
		  //hdr_mac802_16 *wimaxHdr;
	
		  ah.PIH.num_subchannels = getPhy()->getNumsubchannels(UL_);
		  ah.PIH.subchannel_offset = 0; //richard: changed 1->0
		  ah.PIH.num_OFDMSymbol = 0;
		  ah.PIH.OFDMSymbol_offset = 0; //initial_offset;
		  ah.PIH.channel_index = frame->channel_num;
		  ah.PIH.direction = 1;
		  ah.PIH.modulation_ = getMap()->getUlSubframe()->getProfile(UIUC_INITIAL_RANGING)->getEncoding();
	 
	
		  //compute when to send message
		  double txtime = getPhy()->getTrxTime (ah.GMH.GMH.len, ulsubframe->getProfile (b->getFastRangingUIUC ())->getEncoding());
		  ah.txtime = txtime;
		  //starttime+backoff
		  ah.timestamp = NOW; //add timestamp since it bypasses the queue
		  p->AddHeader(frame_h);
		  p->AddHeader(ah);
		  b->enqueue(p);
		  setMacState(MAC802_16_WAIT_RNG_RSP);
		  return;
		}
  }


  //for (PhyPdu *pdu = ulsubframe->getFirstPdu(); pdu ; pdu = pdu->next_entry()) 
  for (std::list<PhyPdu*>::iterator p = getMap()->getUlSubframe()->getPduList()->begin(); p != getMap()->getUlSubframe()->getPduList()->end(); p++) 
  {
    PhyPdu *pdu = *p;
    debug10("IUC: %d init_ranging: %d\n", pdu->getBurst(0)->getIUC(),UIUC_INITIAL_RANGING);
    if (pdu->getBurst(0)->getIUC()==UIUC_INITIAL_RANGING) 
    {
      debug10 ("At %f (SS) Mac %s found ranging opportunity => prepare to create MAC_RNG_REQ (cdma_init_ranging proccess)\n", NOW, addr_s());

      if (getMap()->getUlSubframe()->getRanging()->getRequest_mac (addr())!=NULL) {
        debug2 ("Create CDMA-INIT-REQ, SS addr :%s -- already pending requests => return\n", addr_s());
        continue;  
      } else {
        debug2("Create CDMA-INIT-REQ, SS addr :%s\n", addr_s());
      }
      

      ContentionSlot *slot1 = getMap()->getUlSubframe()->getRanging ();

      int s_init_contention_size = slot1->getSize(); //0 to 4
      int s_nbretry = 0;
      int s_backoff_start = slot1->getBackoff_start();
      int s_backoff_stop = slot1->getBackoff_stop(); if (s_backoff_stop);//quiet, compiler
      int s_window = s_backoff_start;
      int fix_six_subchannel = 6;
      int result = rand() % ((int)(pow (2, s_window)+1));
	  int s_backoff = 0;
      if (s_init_contention_size != 0) s_backoff = floor(result / s_init_contention_size);
      int sub_channel_off = 0;
      if(s_init_contention_size != 0) sub_channel_off = result % s_init_contention_size;
      int s_flagtransmit = 0;
      int s_flagnowtransmit = 0;
      

      Ptr<Packet> p = getPacket();
      All_Headers ah;
      p->RemoveHeader(ah);
      
      //setting all three of these in case some later code looks at the wrong one
      ah.GMH.GMH.cid = INITIAL_RANGING_CID;
      ah.SMH.BRH.cid = INITIAL_RANGING_CID;
      ah.CRH.data.cid = INITIAL_RANGING_CID;
      
      ah.gm_sm = 2;
      ah.PIH.num_subchannels = fix_six_subchannel;
      ah.PIH.subchannel_offset = sub_channel_off*fix_six_subchannel;
      ah.PIH.num_OFDMSymbol = 2;
      ah.PIH.OFDMSymbol_offset = 0; //initial_offset;
      ah.PIH.direction = 1;
      ah.PIH.modulation_ = getMap()->getUlSubframe()->getProfile(UIUC_INITIAL_RANGING)->getEncoding();

      /*Xingting randomly assign channel index for MS.*/
	  ah.PIH.channel_index = GetInitialChannel();
      
      uint8_t header_top = sub_channel_off;

      int code_range = macmib_->cdma_code_init_start - macmib_->cdma_code_init_stop + 1;
      UniformVariable unfv(0,code_range);
      int c_rand = unfv.GetInteger(0,code_range);
      uint8_t header_code = (uint8_t)macmib_->cdma_code_init_start + (uint8_t)c_rand;
//      debug10 (" CODE init start :%d, stop :%d, rand :%d, final code :%d\n", macmib_->cdma_code_init_start, macmib_->cdma_code_init_stop, c_rand, header_code);

//      uint8_t header_code = rand() % (MAXCODE-1);
      cdma_req_header_t *header = &ah.CRH.data;
      header->ht=1;
      header->ec=1;
      header->type = 1; //CDMA_INIT
      header->top = header_top;
      header->br = id_index;
      header->code = header_code;
      header->cid = INITIAL_RANGING_CID;
      double txtime = getPhy()->getSymbolTime();
      ah.txtime = txtime;

      debug10 (" SSscheduler is about to enqueue cdma_initial ranging request for cid :%d, q-len :%d, (nbPacket :%d), code :%d, top :%d, ss_id :%d, backoff :%d, window :%d, nbretry :%d, flagtransmit :%d, flagnowtransmit :%d\n", INITIAL_RANGING_CID, 0, 0, header_code, header_top, header->br, result, s_window, s_nbretry, s_flagtransmit, s_flagnowtransmit);

      debug10 (" symbol_offset :%d, symbol :%d, subchannel_offset :%d, subchannel :%d, channel_index :%d, direction :%d, txtime :%f\n", ah.PIH.OFDMSymbol_offset, ah.PIH.num_OFDMSymbol, ah.PIH.subchannel_offset, ah.PIH.num_subchannels, ah.PIH.channel_index , ah.PIH.direction, ah.txtime);
      
	  p->AddHeader(ah);
      getMap()->getUlSubframe()->getRanging()->addRequest (p, INITIAL_RANGING_CID, ah.CRH.GetSerializedSize(), result, CDMA_TIMEOUT, s_nbretry, s_window, (int)header_code, (int)header_top, s_flagtransmit, s_flagnowtransmit, addr());
      
      int init_code = getMap()->getUlSubframe()->getRanging()->getCODE (addr());
      int init_top = getMap()->getUlSubframe()->getRanging()->getTOP (addr()); 
      	if (init_code || init_top); //quiet, compiler
      
      debug10("init_code: %d, init_top: %d\n",init_code, init_top);
      return;
    }
  }
}

/**
 * Prepare to send a registration message
 */
void Wimax2SSNetDevice::send_registration ()
{
  NS_LOG_FUNCTION_NOARGS ();
  Ptr<Packet> p;
  All_Headers ah;
  REGISTRATION_REQUEST_FRAME rrf;
  mac802_16_reg_req_frame *reg_frame = &rrf.data;;
  PeerNode *peer;

  //create packet for request
  p = getPacket ();
  p->RemoveHeader(ah);
  reg_frame->type = MAC_REG_REQ;
  ah.GMH.GMH.len += rrf.GetSerializedSize();

  peer = getPeerNode_head();  
  ah.GMH.GMH.cid = peer->getPrimary(OUT_CONNECTION)->get_cid();
  p->AddHeader(rrf);
  p->AddHeader(ah);
  peer->getPrimary(OUT_CONNECTION)->enqueue (p);
  debug2(" sending registration\n");

  //start reg timeout
  if (t6timer_==NULL) {
    t6timer_ = new WimaxT6Timer (this);
  }
  t6timer_->start (macmib_->t6_timeout);
  nb_reg_retry_++;
}

/**
 * Process a registration response message 
 * @param frame The registration response frame
 */
void Wimax2SSNetDevice::process_reg_rsp (REGISTRATION_RESPONSE_FRAME *frame_h)
{
  NS_LOG_FUNCTION_NOARGS ();
  if (getMacState() != MAC802_16_REGISTER) {
  	debug10("Received REG_RSP at SS %s when not in correct state\n", addr_s());
  	return;
  }
  mac802_16_reg_rsp_frame *frame = &frame_h->data;
  //check the destination
  PeerNode *peer = getPeerNode_head();

  debug2(" registration response recvd \n " ); 

  if (frame->response == 0) {
    //status OK
    debug2 ("At %f (SS) in Mac %s, registration sucessful (nbretry=%d)\n", NOW, addr_s(),
	    nb_reg_retry_);
    Connection *secondary = peer->getSecondary(IN_CONNECTION);
    if (!secondary) {
      Connection *secondary = new Connection (CONN_SECONDARY, frame->sec_mngmt_cid);
//      secondary->setBW_REQ_QUEUE(0);
//      secondary->initCDMA_SSID();
      Connection *upsecondary = new Connection (CONN_SECONDARY, frame->sec_mngmt_cid);
//      upsecondary->setBW_REQ_QUEUE(0);
//      upsecondary->initCDMA_SSID();
      getCManager()->add_connection (upsecondary, OUT_CONNECTION);
      getCManager()->add_connection (secondary, IN_CONNECTION);
      peer->setSecondary (secondary, upsecondary);
    }
    //cancel timeout (could be idle if receiving retransmissions)
    if (t6timer_->busy()!=0) {
      t6timer_->stop ();
    }
    //update status
    setMacState(MAC802_16_CONNECTED);
    
    //we need to setup a data connection - implementation is nonfunctional 
    //(use static flows from script instead)
    //getServiceHandler ()->sendFlowRequest (peer->getAddr(), OUT_CONNECTION);
    //getServiceHandler ()->sendFlowRequest (peer->getAddr(), IN_CONNECTION);
    
    // we will setup the static flows (which should be created by the script being run)
    getServiceHandler ()->init_static_flows (peer->getAddr());
    

#ifdef USE_802_21
    if (scan_info_->substate==HANDOVER) {
      debug10 ("At %f in Mac %s link handoff complete\n", NOW, addr_s());      
      send_link_handover_complete (addr(), scan_info_->serving_bsid, peer->getAddr());
      scan_info_->handoff_timeout = -1;
    }
    debug10 ("At %f in Mac %s, send link up\n", NOW, addr_s());
    send_link_up (addr(), peer->getAddr(), -1);
#endif
    
  } else {
    //status failure
    debug10 ("At %f in Mac %s, registration failed (nbretry=%d)\n", NOW, addr_s(),
	   nb_reg_retry_);
    if (nb_reg_retry_ == macmib_->reg_req_retry) {
#ifdef USE_802_21
      if (scan_info_ && scan_info_->handoff_timeout == -2) {
	debug10 ("At %f in Mac %s link handoff failure\n", NOW, addr_s());      
	//send_link_handoff_failure (addr(), scan_info_->serving_bsid, peer->getAddr());
	scan_info_->handoff_timeout = -1;
      }
#endif
      lost_synch ();
    } else {
      send_registration();
    }
  }
}

/**
 * Send a scanning message to the serving BS
 */
void Wimax2SSNetDevice::send_scan_request ()
{
  NS_LOG_FUNCTION_NOARGS ();
  Ptr<Packet> p;
  All_Headers ah;
  
  mac802_16_mob_scn_req_frame *req_frame = new mac802_16_mob_scn_req_frame();
  PeerNode *peer;

  //if the mac is not connected, cannot send the request
  if (getMacState() != MAC802_16_CONNECTED) {
    debug10 ("At %f in Mac %s scan request invalid because MAC is disconnected\n", NOW, addr_s());
    return;
  }


  debug10 ("At %f in Mac %s enqueue scan request\n", NOW, addr_s());

  //create packet for request
  p = getPacket ();
  p->RemoveHeader(ah);
  req_frame->type = MAC_MOB_SCN_REQ;

  req_frame->scan_duration = macmib_->scan_duration;
  req_frame->interleaving_interval = macmib_->interleaving;
  req_frame->scan_iteration = macmib_->scan_iteration;
  req_frame->n_recommended_bs_index = 0;
  req_frame->n_recommended_bs_full = 0;

  ah.GMH.GMH.len +=req_frame->GetSerializedSize();
  peer = getPeerNode_head();  
  ah.GMH.GMH.cid = peer->getPrimary(OUT_CONNECTION)->get_cid();
  p->AddHeader(*req_frame);
  p->AddHeader(ah);
  peer->getPrimary(OUT_CONNECTION)->enqueue (p);

  //start reg timeout
  if (t44timer_==NULL) {
    t44timer_ = new WimaxT44Timer (this);
  }
  t44timer_->start (macmib_->t44_timeout);
  nb_scan_req_++;
}

/**
 * Process a scanning response message 
 * @param frame The scanning response frame
 */
void Wimax2SSNetDevice::process_scan_rsp (mac802_16_mob_scn_rsp_frame *frame)
{
  NS_LOG_FUNCTION_NOARGS ();
  //PeerNode *peer = getPeerNode_head();
  if (!t44timer_->busy()) {
    //we are receiving the response too late..ignore
    debug10 ("At %f in Mac %s, scan response arrives too late\n", NOW, addr_s());
    return;
  }


  if (frame->scan_duration != 0) {
    //scanning accepted
    debug10 ("At %f in Mac %s, scanning accepted (dur=%d it=%d)\n", NOW, addr_s(), frame->scan_duration,frame->scan_iteration );
    //allocate data for scanning
    //scan_info_ = (struct scanning_structure *) malloc (sizeof (struct scanning_structure));
    //store copy of frame
    
    scan_info_->rsp = (struct mac802_16_mob_scn_rsp_frame *) malloc (sizeof (struct mac802_16_mob_scn_rsp_frame));
    memcpy (scan_info_->rsp, frame, sizeof (struct mac802_16_mob_scn_rsp_frame));
    scan_info_->iteration = 0;
    scan_info_->count = frame->start_frame;
    scan_info_->substate = SCAN_PENDING;
    scan_info_->handoff_timeout = 0; 
    scan_info_->serving_bsid = getPeerNode_head()->getAddr();
    scan_info_->nb_rdv_timers = 0;

    //mark all neighbors as not detected
    for (int i = 0 ; i < nbr_db_->getNbNeighbor() ; i++) {
      nbr_db_->getNeighbors()[i]->setDetected(false);
    }

    //schedule timer for rdv time (for now just use full)
    //TBD: add rec_bs_index
    debug10 ("\tstart scan in %d frames (%f)\n",frame->start_frame,NOW+frame->start_frame*getFrameDuration());
    for (int i = 0 ; i < scan_info_->rsp->n_recommended_bs_full ; i++) {
      if (scan_info_->rsp->rec_bs_full[i].scanning_type ==SCAN_ASSOC_LVL1 
	  || scan_info_->rsp->rec_bs_full[i].scanning_type==SCAN_ASSOC_LVL2) {
	debug2 ("Creating timer for bs=%s at time %f\n",  HelperFunctions::getMacAddrString(scan_info_->rsp->rec_bs_full[i].recommended_bs_id).c_str(), NOW+getFrameDuration()*scan_info_->rsp->rec_bs_full[i].rdv_time);
	assert (nbr_db_->getNeighbor (scan_info_->rsp->rec_bs_full[i].recommended_bs_id));
	//get the channel
	int ch = getChannel (nbr_db_->getNeighbor (scan_info_->rsp->rec_bs_full[i].recommended_bs_id)->getDCD ()->mf.frequency*1000);
	assert (ch!=-1);
	WimaxRdvTimer *timer = new WimaxRdvTimer (this, ch);
	scan_info_->rdv_timers[scan_info_->nb_rdv_timers++] = timer;
	timer->start(getFrameDuration()*scan_info_->rsp->rec_bs_full[i].rdv_time);
      }
    }

  } else {
    debug10 ("At %f in Mac %s, scanning denied\n", NOW, addr_s());
    //what do I do???
  }

  t44timer_->stop();
  nb_scan_req_ = 0;
}

/**
 * Send a MSHO-REQ message to the BS
 */
void Wimax2SSNetDevice::send_msho_req ()
{
  NS_LOG_FUNCTION_NOARGS ();
	debug2("SS %s send a handover request..\n", addr_s());
  Ptr<Packet> p;
  All_Headers ah;
  MSHO_REQUEST_FRAME_HEADER mrfh;
  mac802_16_mob_msho_req_frame *req_frame = &mrfh.data;
  double rssi = 0;

  PeerNode *peer = getPeerNode_head();

  int nbPref = 0;
  for (int i = 0 ; i < nbr_db_->getNbNeighbor() ; i++) {
    WimaxNeighborEntry *entry = nbr_db_->getNeighbors()[i];
    if (entry->isDetected()) {
      debug10 ("At %f in Mac %s Found new AP %s..need to send HO message\n",NOW, addr_s(), HelperFunctions::getMacAddrString(entry->getID()).c_str());
      nbPref++;
    }  
  }

  if (nbPref==0)
    return; //no other BS found

  //create packet for request
  p = getPacket ();
  p->RemoveHeader(ah);
  //p->allocdata (sizeof (struct mac802_16_mob_msho_req_frame)+nbPref*sizeof (mac802_16_mob_msho_req_bs_index));
  //req_frame = (mac802_16_mob_msho_req_frame*) p->accessdata();
  memset (req_frame, 0, sizeof (mac802_16_mob_msho_req_bs_index));
  req_frame->type = MAC_MOB_MSHO_REQ;
  
  req_frame->report_metric = 0x2; //include RSSI
  req_frame->n_new_bs_index = 0;
  req_frame->n_new_bs_full = nbPref;
  req_frame->n_current_bs = 1;
  //rssi = getPeerNode_head()->getStatWatch()->average(); //CAT - uncomment
  debug2 ("RSSI=%e, %d, peer bs=%s\n", rssi, (uint8_t)((rssi+103.75)/0.25), HelperFunctions::getMacAddrString(getPeerNode_head()->getAddr()).c_str());
  req_frame->bs_current[0].temp_bsid = 1;//getPeerNode_head()->getAddr();
  req_frame->bs_current[0].bs_rssi_mean = (uint8_t)((rssi+103.75)/0.25);
  for (int i = 0, j=0; i < nbr_db_->getNbNeighbor() ; i++) {
    WimaxNeighborEntry *entry = nbr_db_->getNeighbors()[i];
    //TBD: there is an error measuring RSSI for current BS during scanning
    //anyway, we don't put it in the least, so it's ok for now
    if (entry->isDetected() && entry->getID()!= getPeerNode_head()->getAddr()) {
      	//old model used ints for mac addrs, and just used those as index instead of maintaining list based on BSHO_RSP and RNG_RSP frames, we're using id_index, but it should eventually be changed to monitor those frames and keep a lsit
      req_frame->bs_full[j].neighbor_bs_index = getIdIndex();
      //rssi = entry->getState()->state_info->peer_list->lh_first->getStatWatch()->average(); //CAT - uncomment
      debug2 ("RSSI=%e, %d, neighbor bs=%d\n", rssi, (uint8_t)((rssi+103.75)/0.25), req_frame->bs_full[j].neighbor_bs_index);
      req_frame->bs_full[j].bs_rssi_mean = (uint8_t)((rssi+103.75)/0.25);
      //the rest of req_frame->bs_full is unused for now..
      req_frame->bs_full[j].arrival_time_diff_ind = 0;
      j++;
    }
  }
  
  ah.GMH.GMH.len += mrfh.GetSerializedSize();
  ah.GMH.GMH.cid = peer->getPrimary(OUT_CONNECTION)->get_cid();
  p->AddHeader(mrfh);
  p->AddHeader(ah);
  peer->getPrimary(OUT_CONNECTION)->enqueue (p);
}

/**
 * Process a BSHO-RSP message 
 * @param frame The handover response frame
 */
void Wimax2SSNetDevice::process_bsho_rsp (BSHO_RESPONCE_FRAME_HEADER *frame_h)
{
  NS_LOG_FUNCTION_NOARGS ();
  mac802_16_mob_bsho_rsp_frame *frame = &frame_h->data;
  debug10 ("At %f in Mac %s, received handover response\n", NOW, addr_s());
 
  //go and switch to the channel recommended by the BS
  Mac48Address targetBS = frame->n_rec[0].neighbor_bsid;
  PeerNode *peer = getPeerNode_head();      

  if (peer->getAddr ()==targetBS) {
    debug10 ("\tDecision to stay in current BS\n");
    return;
  }
  scan_info_->nbr = nbr_db_->getNeighbor (targetBS);

  Ptr<Packet> p;
  All_Headers ah;
  
  
  p = getPacket ();
  p->RemoveHeader(ah);
  MOB_HO_IND_FRAME_HEADER mhif;
  mac802_16_mob_ho_ind_frame *ind_frame = &mhif.data;
  ind_frame->type = MAC_MOB_HO_IND;
  
  ind_frame->mode = 0; //HO
  ind_frame->ho_ind_type = 0; //Serving BS release
  ind_frame->rng_param_valid_ind = 0;
  ind_frame->target_bsid = targetBS;
  
  ah.GMH.GMH.len += mhif.GetSerializedSize();
  ah.GMH.GMH.cid = peer->getPrimary(OUT_CONNECTION)->get_cid();
  p->AddHeader(mhif);
  p->AddHeader(ah);
  peer->getPrimary(OUT_CONNECTION)->enqueue (p);
  
#ifdef USE_802_21
  send_link_handover_imminent (addr(), peer->getAddr(), targetBS);
  debug10 ("At %f in Mac %s link handover imminent\n", NOW, addr_s());
#endif 
  
  debug10 ("\tHandover to BS %s\n", HelperFunctions::getMacAddrString(targetBS).c_str());
  scan_info_->handoff_timeout = 20;
  scan_info_->substate = HANDOVER_PENDING;

  Connection *head  = getCManager()->get_out_connection();
  while(head != NULL)
  {
  		debug2("SS ARQ handling out cid %d\n", head->get_cid());
		if((head->getArqStatus() != NULL))
		{
			debug2("SS delete ARQ timer CID is [%d]\n",head->get_cid());
			head->getArqStatus()->cancelTimer();
			//delete head->getArqStatus();
			//getCManager()->remove_connection (head);
			//delete head;
		}
		head = head->next_entry();
  }
  //setChannel (scan_info_->bs_infos[i].channel);
  //lost_synch ();  
}


/**
 * Process a NBR_ADV message 
 * @param frame The handover response frame
 */
void Wimax2SSNetDevice::process_nbr_adv (NEIGHBOR_ADVERTISEMENT_FRAME_HEADER *frame_h)
{
  NS_LOG_FUNCTION_NOARGS ();
  mac802_16_mob_nbr_adv_frame *frame = &frame_h->data;
  debug10 ("At %f in Mac %s, received neighbor advertisement\n", NOW, addr_s());

  //mac802_16_mob_nbr_adv_frame *copy;
  //copy  = (mac802_16_mob_nbr_adv_frame *) malloc (sizeof (mac802_16_mob_nbr_adv_frame));
  //memcpy (copy, frame, sizeof (mac802_16_mob_nbr_adv_frame));
  
  //all we need is to store the information. We will process that only
  //when we will look for another station
  for (int i = 0 ; i < frame->n_neighbors ; i++) {
    Mac48Address nbrid = frame->nbr_info[i].nbr_bsid;
    mac802_16_nbr_adv_info *info = (mac802_16_nbr_adv_info *) malloc (sizeof(mac802_16_nbr_adv_info));
    WimaxNeighborEntry *entry = nbr_db_->getNeighbor (nbrid);
    if (entry==NULL){
      entry = new WimaxNeighborEntry (nbrid);
      nbr_db_->addNeighbor (entry);
    }
    memcpy(info, &(frame->nbr_info[i]), sizeof(mac802_16_nbr_adv_info));
    if (entry->getNbrAdvMessage ())
      free (entry->getNbrAdvMessage());
    entry->setNbrAdvMessage(info);
    if (info->dcd_included) {
      //set DCD 
      DCD_FRAME_HEADER *tmp_h = new DCD_FRAME_HEADER();
      mac802_16_dcd_frame *tmp = &tmp_h->mf;
      memcpy(tmp, &(info->dcd_settings), sizeof(mac802_16_dcd_frame));
      entry->setDCD(tmp_h);
    }
    else 
      entry->setDCD(NULL);
    if (info->ucd_included) {
      //set DCD 
      UCD_FRAME_DATA *tmp_h = new UCD_FRAME_DATA();
      mac802_16_ucd_frame *tmp = &tmp_h->ucdf;
      memcpy(tmp, &(info->ucd_settings), sizeof(mac802_16_ucd_frame));
      entry->setUCD(tmp_h);
#ifdef DEBUG_WIMAX
      mac802_16_ucd_profile *profiles = tmp->profiles;
      int nb_prof = tmp->nb_prof;
      debug2 ("Dump information nbr in Mac %s for nbr %s %lx\n", addr_s(), HelperFunctions::getMacAddrString(nbrid).c_str(), (long)tmp);
      for (int i = 0 ; i < nb_prof ; i++) {
		debug2 ("\t Reading ul profile %i: f=%d, rate=%d, iuc=%d\n", i, 0, profiles[i].fec, profiles[i].uiuc);
      }
#endif
    }
    else
      entry->setUCD(NULL);
  }  

}

/**** Internal methods ****/

#ifdef USE_802_21

/*
 * Connect to the PoA
 */
void Wimax2SSNetDevice::link_connect(int poa)
{
  NS_LOG_FUNCTION_NOARGS ();
  debug10 ("At %f in Mac %s, received link connect to BS %d\n", NOW, addr_s(), poa);
 
  set_mode (NORMAL_MODE);

  //go and switch to the channel recommended by the BS
  int targetBS = poa;
  PeerNode *peer = getPeerNode_head();      

  if (peer->getAddr ()==targetBS) {
    debug10 ("\tDecision to stay in current BS\n");
    return;
  }
  scan_info_->nbr = nbr_db_->getNeighbor (targetBS);

  Ptr<Packet> p;
  struct hdr_cmn *ch;
  hdr_mac802_16 *wimaxHdr;
  mac802_16_mob_ho_ind_frame *ind_frame;
  
  
  p = getPacket ();
  ch = HDR_CMN(p);
  wimaxHdr = HDR_MAC802_16(p);
  p->allocdata (sizeof (struct mac802_16_mob_ho_ind_frame));
  ind_frame = (mac802_16_mob_ho_ind_frame*) p->accessdata();
  ind_frame->type = MAC_MOB_HO_IND;
  
  ind_frame->mode = 0; //HO
  ind_frame->ho_ind_type = 0; //Serving BS release
  ind_frame->rng_param_valid_ind = 0;
  ind_frame->target_bsid = targetBS;
  
  ch->size() += Mac802_16pkt::getMOB_HO_IND_size(ind_frame);
  wimaxHdr->header.cid = peer->getPrimary(OUT_CONNECTION)->get_cid();
  peer->getPrimary(OUT_CONNECTION)->enqueue (p);
  
#ifdef USE_802_21
  send_link_handover_imminent (addr(), peer->getAddr(), targetBS);
  debug10 ("At %f in Mac %s link handover imminent\n", NOW, addr_s());
  
#endif 
  
  debug10 ("\tHandover to BS %d\n", targetBS);
  scan_info_->handoff_timeout = 20;
  scan_info_->substate = HANDOVER_PENDING;
  //setChannel (scan_info_->bs_infos[i].channel);
  //lost_synch ();    
}


/*
 * Disconnect from the PoA
 */
void Wimax2SSNetDevice::link_disconnect ()
{
  NS_LOG_FUNCTION_NOARGS ();
  //force losing synchronization
  lost_synch ();
  set_mode (POWER_DOWN); //not sure if we should turn it off
}


/*
 * Set the operation mode
 * @param mode The new operation mode
 * @return true if transaction succeded
 */
bool Wimax2SSNetDevice::set_mode (mih_operation_mode_t mode)
{
  NS_LOG_FUNCTION_NOARGS ();
  switch (mode) {
  case NORMAL_MODE:
    if (op_mode_ != NORMAL_MODE) {
      getPhy()->node_on(); //turn on phy
      debug10 ("Turning on mac\n");
    }
    op_mode_ = mode;
    return true;
    break;
  case POWER_SAVING:
    //not yet supported 
    return false;
    break;
  case POWER_DOWN:
    if (op_mode_ != POWER_DOWN) {
      getPhy()->node_off(); //turn off phy
      debug10 ("Turning off mac\n");
    }
    op_mode_ = mode;
    return true;
    break;
  default:
    return false;
  }
}

#endif

/*
 * Scan chanels
 */
void Wimax2SSNetDevice::link_scan (void *req)
{
  NS_LOG_FUNCTION_NOARGS ();
  if(!isScanRunning()){
    setScanFlag(true);  
    send_scan_request ();
  }
}

/**
 * Update the given timer and check if thresholds are crossed
 * @param watch the stat watch to update
 * @param value the stat value
 */
 //CAT - uncomment
// void Wimax2SSNetDevice::update_watch (StatWatch *watch, double value)
// {
//   char *name;
// 
// #ifdef USE_802_21 //Switch to activate when using 802.21 modules (external package)
//   threshold_action_t action = watch->update (value);
// 
//   if (action != NO_ACTION_TH) {
//     link_parameter_type_s param;
//     union param_value old_value, new_value;
// 
//     if (watch == &loss_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_FRAME_LOSS;
//     } else if (watch == &delay_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_PACKET_DELAY;
//     } else if (watch == &jitter_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_PACKET_JITTER;
//     }
//     old_value.data_d = watch->old_average();
//     new_value.data_d = watch->average();
// 
//     send_link_parameters_report (addr(), getPeerNode_head()->getAddr(), param, old_value, new_value);      
//   }
// #else
//   watch->update (value);
// #endif
// 
//   if (watch == &loss_watch_) {
//     name = "loss";
//   } else if (watch == &delay_watch_) {
//     name = "delay";
//   } else if (watch == &jitter_watch_) {
//     name = "jitter";
//   } else {
//     name = "other";
//   }
//   if (print_stats_)
//     debug2 ("At %f in Mac %s, updating stats %s: %f\n", NOW, addr_s(), name, watch->average());
// }

/**
 * Update the given timer and check if thresholds are crossed
 * @param watch the stat watch to update
 * @param value the stat value
 */
 //CAT - uncomment
// void Wimax2SSNetDevice::update_throughput (ThroughputWatch *watch, double size)
// {
//   char *name;
// 
// #ifdef USE_802_21 //Switch to activate when using 802.21 modules (external package)
//   threshold_action_t action = watch->update (size, NOW);
//   if (action != NO_ACTION_TH) {
//     link_parameter_type_s param;
//     union param_value old_value, new_value;
//     if (watch == &rx_data_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_RX_DATA_THROUGHPUT;
//     } else if (watch == &rx_traffic_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_RX_TRAFFIC_THROUGHPUT;
//     } else if (watch == &tx_data_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_TX_DATA_THROUGHPUT;
//     } else if (watch == &tx_traffic_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_TX_TRAFFIC_THROUGHPUT;
//     }
//     old_value.data_d = watch->old_average();
//     new_value.data_d = watch->average();
// 
//     send_link_parameters_report (addr(), getPeerNode_head()->getAddr(), param, old_value, new_value);      
//   }
// #else
//   watch->update (size, NOW);
// #endif 
// 
//   if (watch == &rx_data_watch_) {
//     name = "rx_data";
//     rx_data_timer_->resched (watch->get_timer_interval());
//   } else if (watch == &rx_traffic_watch_) {
//     rx_traffic_timer_->resched (watch->get_timer_interval());
//     name = "rx_traffic";
//   } else if (watch == &tx_data_watch_) {
//     tx_data_timer_->resched (watch->get_timer_interval());
//     name = "tx_data";
//   } else if (watch == &tx_traffic_watch_) {
//     tx_traffic_timer_->resched (watch->get_timer_interval());
//     name = "tx_traffic";
//   }
// 
//   if (print_stats_)
//     debug2 ("At %f in Mac %s, updating stats %s: %f\n", NOW, addr_s(), name, watch->average());
// }

/**
 * Start a new frame
 */
void Wimax2SSNetDevice::start_dlsubframe ()
{
  StartSubframe(true);
  NS_LOG_FUNCTION_NOARGS ();
  debug2 ("At %f in Mac %s SS scheduler dlsubframe expires %d\n", NOW, addr_s(), scan_info_->substate);
  r_delta_ss = 0;


  switch (scan_info_->substate) {
  case SCAN_PENDING: 
    if (scan_info_->count == 0) {
      resume_scanning();
      return;
    } 
    scan_info_->count--;
    break;
  case HANDOVER_PENDING:
    if (scan_info_->handoff_timeout == 0) {
      assert (scan_info_->nbr);
#ifdef USE_802_21
      //debug10 ("At %f in Mac %s link handoff proceeding\n", NOW, addr());
      //send_link_handoff_proceeding (addr(), getPeerNode_head()->getAddr(), scan_info_->nbr->getID());
#endif 
      scan_info_->substate = HANDOVER;
      //restore previous state 
      //restore_state (scan_info_->nbr->getState()->state_info);
      setChannel (scan_info_->nbr->getState()->state_info->channel);
      lost_synch ();
      //add target as peer
      addPeerNode (new PeerNode(scan_info_->nbr->getID()));
      return;
    }
    scan_info_->handoff_timeout--;
    break;
  default:
    break;
  }
    
  //this is the begining of new frame
  map_->setStarttime (NOW);

  //start handler of dlsubframe
  debug10("starting dlsubframetimer(0) from SSNetDevice::start_dlsubframe\n");
  if (map_->getDlSubframe()->getTimer()->busy() != 0) map_->getDlSubframe()->getTimer()->resched(0);
  else map_->getDlSubframe()->getTimer()->start (0);

  //reschedule for next frame
  dl_timer_->resched (getFrameDuration());

}

/**
 * Start a new frame
 */
void Wimax2SSNetDevice::start_ulsubframe ()
{
  NS_LOG_FUNCTION_NOARGS ();
  frame_number_++;

  debug10 ("At %f in Mac %s SS scheduler ulsubframe expires\n", NOW, addr_s());

  
  //change state of PHY: even though it should have been done before
  //there are some cases where it does not (during scanning)
  getPhy()->setMode (OFDM_SEND);

	debug2("About to schedule SS\n");
  scheduler_->schedule();
  	debug2("Done scheduling SS\n");

  //start handler for ulsubframe
  if (getMap()->getUlSubframe()->getNbPdu ()>0) {
    Burst *b = getMap()->getUlSubframe()->getPhyPdu (0)->getBurst (0);
    getMap()->getUlSubframe()->getTimer()->resched (b->getStarttime()*getPhy()->getSymbolTime());
  }//else there is no uplink phy pdu defined

  //reschedule for next frame
  ul_timer_->resched (getFrameDuration());     
 
}

/**
 * Called when lost synchronization
 */
void Wimax2SSNetDevice::lost_synch ()
{
  NS_LOG_FUNCTION_NOARGS ();
#ifdef USE_802_21
  int poa = -1;
  bool disconnect = false;
#endif

  //reset timers
  if (t1timer_->busy()!=0)
    t1timer_->stop();
  if (t12timer_->busy()!=0)
    t12timer_->stop();
  if (t21timer_->busy()!=0)
    t21timer_->stop();
  if (lostDLMAPtimer_->busy()!=0)
    lostDLMAPtimer_->stop(); 
  if (lostULMAPtimer_->busy()!=0)
    lostULMAPtimer_->stop(); 
  if (t2timer_->busy()!=0)
    t2timer_->stop(); 
  if (t44timer_ && t44timer_->busy()!=0)
    t44timer_->stop();

  //we need to go to receiving mode
  //printf ("Set phy to recv %x\n", getPhy());
  getPhy()->setMode (OFDM_RECV);
  if (getMacState()==MAC802_16_CONNECTED) {
    //remove possible pending requests
    map_->getUlSubframe()->getBw_req()->removeRequests(); 

#ifdef USE_802_21
    poa = getPeerNode_head()->getAddr ();
    disconnect = true;
#endif
  }

  //remove information about peer node
  if (getPeerNode_head())
    removePeerNode (getPeerNode_head());

  //start waiting for DL synch
  setMacState (MAC802_16_WAIT_DL_SYNCH);
  t21timer_->start (macmib_->t21_timeout);
  if (dl_timer_->IsRunning()==TIMER_PENDING)
    dl_timer_->Cancel();
  map_->getDlSubframe()->getTimer()->reset();
  if (ul_timer_->IsRunning()==TIMER_PENDING)
    ul_timer_->Cancel();
  map_->getUlSubframe()->getTimer()->reset();

#ifdef USE_802_21
  if (disconnect) {
    debug10 ("At %f in Mac %s, send link down\n", NOW, addr_s());
    send_link_down (addr(), poa, LD_RC_FAIL_NORESOURCE);
  }
#endif

  if (scan_info_->substate == HANDOVER_PENDING || scan_info_->substate == SCAN_PENDING) {
    //we have lost synch before scanning/handover is complete
    for (int i=0 ; i < scan_info_->nb_rdv_timers ; i++) {
      //debug10 ("canceling rdv timer\n");
      if (scan_info_->rdv_timers[i]->busy()) {
	scan_info_->rdv_timers[i]->stop();
      }
      delete (scan_info_->rdv_timers[i]);
    }
    scan_info_->nb_rdv_timers = 0;
  }

  if (scan_info_->substate == HANDOVER_PENDING) {
    debug10 ("Lost synch during pending handover (%d)\n", scan_info_->handoff_timeout);
    //since we lost connection, let's execute handover immediately 
    scan_info_->substate = HANDOVER;
    setChannel (scan_info_->nbr->getState()->state_info->channel);
    //add target as peer
    addPeerNode (new PeerNode(scan_info_->nbr->getID()));
    return;
  } 
  if (scan_info_->substate == SCAN_PENDING) {
    debug10 ("Lost synch during pending scan (%d)\n", scan_info_->count);
    //we must cancel the scanning
    scan_info_->substate = NORMAL;
  }
}

/**
 * Start/Continue scanning
 */
void Wimax2SSNetDevice::resume_scanning ()
{
  NS_LOG_FUNCTION_NOARGS ();
  if (scan_info_->iteration == 0) 
    debug10 ("At %f in Mac %s, starts scanning\n", NOW, addr_s());
  else 
    debug10 ("At %f in Mac %s, resume scanning\n", NOW, addr_s());
  
  scan_info_->substate = SCANNING;

  //backup current state
  scan_info_->normal_state.state_info = backup_state();
  if (t1timer_->busy())
    t1timer_->pause();
  scan_info_->normal_state.t1timer = t1timer_;
  if (t2timer_->busy())
    t2timer_->pause();
  scan_info_->normal_state.t2timer = t2timer_;
  if (t6timer_->busy())
    t6timer_->pause();
  scan_info_->normal_state.t6timer = t6timer_;
  if (t12timer_->busy())
    t12timer_->pause();
  scan_info_->normal_state.t12timer = t12timer_;
  if (t21timer_->busy())
    t21timer_->pause();
  scan_info_->normal_state.t21timer = t21timer_;
  if (lostDLMAPtimer_->busy())
    lostDLMAPtimer_->pause();
  scan_info_->normal_state.lostDLMAPtimer = lostDLMAPtimer_;
  if (lostULMAPtimer_->busy())
    lostULMAPtimer_->pause();
  scan_info_->normal_state.lostULMAPtimer = lostULMAPtimer_;
  scan_info_->normal_state.map = map_;

  if (scan_info_->iteration == 0) {
    //reset state
    t1timer_ = new WimaxT1Timer (this);
    t2timer_ = new WimaxT2Timer (this);
    t6timer_ = new WimaxT6Timer (this);
    t12timer_ = new WimaxT12Timer (this);
    t21timer_ = new WimaxT21Timer (this);
    lostDLMAPtimer_ = new WimaxLostDLMAPTimer (this);
    lostULMAPtimer_ = new WimaxLostULMAPTimer (this);
    
    map_ = new FrameMap (this);
    
    nextChannel();

    scan_info_->scn_timer_ = new WimaxScanIntervalTimer (this);

    //start waiting for DL synch
    setMacState (MAC802_16_WAIT_DL_SYNCH);
    t21timer_->start (macmib_->t21_timeout);
    if (dl_timer_->IsRunning()==TIMER_PENDING)
      dl_timer_->Cancel();
    map_->getDlSubframe()->getTimer()->reset();
    if (ul_timer_->IsRunning()==TIMER_PENDING)
      ul_timer_->Cancel();
    map_->getUlSubframe()->getTimer()->reset();
    

  }else{
    //restore where we left
    //restore previous timers
    restore_state(scan_info_->scan_state.state_info);
    t1timer_ = scan_info_->scan_state.t1timer;
    if (t1timer_->paused())
      t1timer_->resume();
    t2timer_ = scan_info_->scan_state.t2timer;
    if (t2timer_->paused())
      t2timer_->resume();
    t6timer_ = scan_info_->scan_state.t6timer;
    if (t6timer_->paused())
      t6timer_->resume();
    t12timer_ = scan_info_->scan_state.t12timer;
    if (t12timer_->paused())
      t12timer_->resume();
    t21timer_ = scan_info_->scan_state.t21timer;
    if (t21timer_->paused())
      t21timer_->resume();
    lostDLMAPtimer_ = scan_info_->scan_state.lostDLMAPtimer;
    if (lostDLMAPtimer_->paused())
      lostDLMAPtimer_->resume();
    lostULMAPtimer_ = scan_info_->scan_state.lostULMAPtimer;
    if (lostULMAPtimer_->paused())
      lostULMAPtimer_->resume();
    map_ = scan_info_->scan_state.map;
    
    getPhy()->setMode (OFDM_RECV);

    if (ul_timer_->IsRunning()==TIMER_PENDING)
      ul_timer_->Cancel();
  }
  setNotify_upper (false);
  //printf ("Scan duration=%d, frameduration=%f\n", scan_info_->rsp->scan_duration, getFrameDuration());
  scan_info_->scn_timer_->start (scan_info_->rsp->scan_duration*getFrameDuration());
  scan_info_->iteration++;
  
}

/**
 * Pause scanning
 */
void Wimax2SSNetDevice::pause_scanning ()
{
  NS_LOG_FUNCTION_NOARGS ();
  if (scan_info_->iteration < scan_info_->rsp->scan_iteration)
    debug10 ("At %f in Mac %s, pause scanning\n", NOW, addr_s());
  else 
    debug10 ("At %f in Mac %s, stop scanning\n", NOW, addr_s());

  //return to normal mode
  if (scan_info_->iteration < scan_info_->rsp->scan_iteration) {
    //backup current state
    scan_info_->scan_state.state_info = backup_state();
    if (t1timer_->busy())
      t1timer_->pause();
    scan_info_->scan_state.t1timer = t1timer_;
    if (t2timer_->busy())
      t2timer_->pause();
    scan_info_->scan_state.t2timer = t2timer_;
    if (t6timer_->busy())
      t6timer_->pause();
    scan_info_->scan_state.t6timer = t6timer_;
    if (t12timer_->busy())
      t12timer_->pause();
    scan_info_->scan_state.t12timer = t12timer_;
    if (t21timer_->busy())
      t21timer_->pause();
    scan_info_->scan_state.t21timer = t21timer_;
    if (lostDLMAPtimer_->busy())
      lostDLMAPtimer_->pause();
    scan_info_->scan_state.lostDLMAPtimer = lostDLMAPtimer_;
    if (lostULMAPtimer_->busy())
      lostULMAPtimer_->pause();
    scan_info_->scan_state.lostULMAPtimer = lostULMAPtimer_;
    scan_info_->scan_state.map = map_;

    scan_info_->count = scan_info_->rsp->interleaving_interval;

  } else {
    //else scanning is over, no need to save data
    //reset timers
    if (t1timer_->busy()!=0)
      t1timer_->stop();
    delete (t1timer_);
    if (t12timer_->busy()!=0)
      t12timer_->stop();
    delete (t12timer_);
    if (t21timer_->busy()!=0)
      t21timer_->stop();
    delete (t21timer_);
    if (lostDLMAPtimer_->busy()!=0)
      lostDLMAPtimer_->stop(); 
    delete (lostDLMAPtimer_);
    if (lostULMAPtimer_->busy()!=0)
      lostULMAPtimer_->stop(); 
    delete (lostULMAPtimer_);
    if (t2timer_->busy()!=0)
      t2timer_->stop(); 
    delete (t2timer_);
  }
  //restore previous timers
  restore_state(scan_info_->normal_state.state_info);
  t1timer_ = scan_info_->normal_state.t1timer;
  if (t1timer_->paused())
    t1timer_->resume();
  t2timer_ = scan_info_->normal_state.t2timer;
  if (t2timer_->paused())
    t2timer_->resume();
  t6timer_ = scan_info_->normal_state.t6timer;
  if (t6timer_->paused())
    t6timer_->resume();
  t12timer_ = scan_info_->normal_state.t12timer;
  if (t12timer_->paused())
    t12timer_->resume();
  t21timer_ = scan_info_->normal_state.t21timer;
  if (t21timer_->paused())
    t21timer_->resume();
  lostDLMAPtimer_ = scan_info_->normal_state.lostDLMAPtimer;
  if (lostDLMAPtimer_->paused())
    lostDLMAPtimer_->resume();
  lostULMAPtimer_ = scan_info_->normal_state.lostULMAPtimer;
  if (lostULMAPtimer_->paused())
    lostULMAPtimer_->resume();
  map_ = scan_info_->normal_state.map;

  setNotify_upper (true);
  dl_timer_->resched (0);

  if (scan_info_->iteration == scan_info_->rsp->scan_iteration) {
    scan_info_->substate = NORMAL;
    
    /** here we check if there is a better BS **/
#ifdef USE_802_21
    if(mih_){
      int nbDetected = 0;
      for (int i = 0 ; i < nbr_db_->getNbNeighbor() ; i++) {
	if (nbr_db_->getNeighbors()[i]->isDetected()) {
	  nbDetected++;
	}
      }
      int *listOfPoa = new int[nbDetected];
      int itr = 0;
      for (int i = 0 ; i < nbr_db_->getNbNeighbor() ; i++) {
	WimaxNeighborEntry *entry = nbr_db_->getNeighbors()[i];
	if (entry->isDetected()) {
	  listOfPoa[itr] = entry->getID();
	  itr++;
	}  
      }  
      send_scan_result (listOfPoa, itr*sizeof(int));	
    }else
#endif
      {
	send_msho_req();
      }

    setScanFlag(false);
    scan_info_->count--; //to avoid restarting scanning
  } else {
    scan_info_->substate = SCAN_PENDING;
  }
}

/**
 * Set the scan flag to true/false
 * param flag The value for the scan flag
 */
void Wimax2SSNetDevice::setScanFlag(bool flag)
{
  NS_LOG_FUNCTION_NOARGS ();
  scan_flag_ = flag;
}

/**
 * return scan flag
 * @return the scan flag
 */
bool Wimax2SSNetDevice::isScanRunning()
{
  NS_LOG_FUNCTION_NOARGS ();
  return scan_flag_;
}


/**
 * rpi - get initial random channel for propagation model
 */
int  Wimax2SSNetDevice::GetInitialChannel()
{
  NS_LOG_FUNCTION_NOARGS ();
  int Rand_Num=0;
  Rand_Num = ((int)(rand() % 500)/*NUM_REALIZATIONS*/+ 1);
  return Rand_Num; 
}



void Wimax2SSNetDevice::addPowerinfo(All_Headers *wimaxHdr,double power_per_subchannel, bool collision )
{
  NS_LOG_FUNCTION_NOARGS ();

  uint32_t num_data_subcarrier = getPhy()->getNumSubcarrier (DL_);
  uint32_t num_subchannel = getPhy()->getNumsubchannels (DL_);
  uint32_t num_symbol_per_slot = getPhy()->getSlotLength (DL_); 

  if (collision == true)

    if(wimaxHdr->PIH.num_OFDMSymbol % num_symbol_per_slot == 0)   // chk this condition , chk whether broacastcid reqd ir not. 
      // if((wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels -1) > 30)
      //if(wimaxHdr->PIH.num_OFDMSymbol >= num_symbol_per_slot)
      {
	if(wimaxHdr->PIH.num_OFDMSymbol > num_symbol_per_slot) 
	  {
            //numsymbols = (int)  ceil((((wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels -1) - 30) % 30));
            //numsymbols *= num_symbol_per_slot;  
	    // for the first num_symbol_per_slot symbols 
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	      for(uint32_t j = (wimaxHdr->PIH.subchannel_offset) ; j<=num_subchannel ; j++ )
                (basepower_[i][j] = power_per_subchannel);
	    // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
          
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset+num_symbol_per_slot ; i<  (wimaxHdr->PIH.OFDMSymbol_offset) + (wimaxHdr->PIH.num_OFDMSymbol-num_symbol_per_slot) ; i++)
	      for(uint32_t j = 1 ; j<=num_subchannel ; j++ )
                (basepower_[i][j] = power_per_subchannel);  

	    // last num_symbol_per_slot 
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset +wimaxHdr->PIH.num_OFDMSymbol-num_symbol_per_slot ; i< wimaxHdr->PIH.OFDMSymbol_offset + wimaxHdr->PIH.num_OFDMSymbol/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	      for(uint32_t j = 1 ; j<=((wimaxHdr->PIH.num_subchannels - (num_subchannel - (wimaxHdr->PIH.subchannel_offset)) )%num_subchannel) ; j++ )
		(basepower_[i][j] = power_per_subchannel); 
          }
	else 
          {

	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
              for(uint32_t j = (wimaxHdr->PIH.subchannel_offset) ; j< (wimaxHdr->PIH.subchannel_offset) + wimaxHdr->PIH.num_subchannels ; j++ )
                (basepower_[i][j] = power_per_subchannel); 	    

          }

      }

    else

      {
	//numsymbols = (int)  ceil((((wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels -1) - 30) % 30));
	//numsymbols *= num_symbol_per_slot;  
        // for the first num_symbol_per_slot symbols 
	if(wimaxHdr->PIH.num_OFDMSymbol > 1) 
	  {
            for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + 1)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
              for(uint32_t j = (wimaxHdr->PIH.subchannel_offset) ; j<=num_subchannel ; j++ )
		(basepower_[i][j] = power_per_subchannel); 
	    // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
          
            for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset+1 ; i<  ((wimaxHdr->PIH.OFDMSymbol_offset) + wimaxHdr->PIH.num_OFDMSymbol-1) ; i++)
	      for(uint32_t j = 1 ; j<=num_subchannel ; j++ )
		basepower_[i][j] = power_per_subchannel;

	    // last num_symbol_per_slot 
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset +wimaxHdr->PIH.num_OFDMSymbol-1 ; i<  (wimaxHdr->PIH.OFDMSymbol_offset) + wimaxHdr->PIH.num_OFDMSymbol/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	      for(uint32_t j = 1 ; j<=((wimaxHdr->PIH.num_subchannels - (num_subchannel - (wimaxHdr->PIH.subchannel_offset)) )%num_subchannel) ; j++ )
		(basepower_[i][j] = power_per_subchannel); 

          }
	else
          {
          
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + 1)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	      for(uint32_t j = (wimaxHdr->PIH.subchannel_offset) ; j< wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels ; j++ )
		(basepower_[i][j] = power_per_subchannel); 	

          }
             
      }

  else
       
    if(wimaxHdr->PIH.num_OFDMSymbol % num_symbol_per_slot == 0)   // chk this condition , chk whether broacastcid reqd ir not. 
      // if((wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels -1) > num_subchannel)
      //if(wimaxHdr->PIH.num_OFDMSymbol >= num_symbol_per_slot)
      {
	if(wimaxHdr->PIH.num_OFDMSymbol > num_symbol_per_slot) 
	  {
            //numsymbols = (int)  ceil((((wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels -1) - num_subchannel) % num_subchannel));
            //numsymbols *= num_symbol_per_slot;  
	    // for the first num_symbol_per_slot symbols 
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	      for(uint32_t j = (wimaxHdr->PIH.subchannel_offset)*num_data_subcarrier ; j<num_subchannel*num_data_subcarrier ; j++ )
                (intpower_[i][j] = power_per_subchannel);
	    // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
          
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset+num_symbol_per_slot ; i<  (wimaxHdr->PIH.OFDMSymbol_offset) + (wimaxHdr->PIH.num_OFDMSymbol-num_symbol_per_slot) ; i++)
	      for(uint32_t j = 0 ; j<num_subchannel*num_data_subcarrier ; j++ )
                (intpower_[i][j] = power_per_subchannel);  

	    // last num_symbol_per_slot 
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset +wimaxHdr->PIH.num_OFDMSymbol-num_symbol_per_slot ; i< wimaxHdr->PIH.OFDMSymbol_offset + wimaxHdr->PIH.num_OFDMSymbol/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	      for(uint32_t j = 0 ; j<(((wimaxHdr->PIH.num_subchannels - (num_subchannel - (wimaxHdr->PIH.subchannel_offset)) )%num_subchannel))*num_data_subcarrier ; j++ )
		(intpower_[i][j] = power_per_subchannel); 
          }
	else 
          {

	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
              for(uint32_t j = (wimaxHdr->PIH.subchannel_offset)*num_data_subcarrier ; j<(((wimaxHdr->PIH.subchannel_offset)*num_data_subcarrier) + wimaxHdr->PIH.num_subchannels*num_data_subcarrier) ; j++ )
                (intpower_[i][j] = power_per_subchannel); 	    

          }

      }

    else

      {
	//numsymbols = (int)  ceil((((wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels -1) - num_subchannel) % num_subchannel));
	//numsymbols *= num_symbol_per_slot;  
        // for the first num_symbol_per_slot symbols 
	if(wimaxHdr->PIH.num_OFDMSymbol > 1) 
	  {
            for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + 1)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
              for(uint32_t j = (wimaxHdr->PIH.subchannel_offset)*num_data_subcarrier ; j<num_subchannel*num_data_subcarrier ; j++ )
		(intpower_[i][j] = power_per_subchannel); 
	    // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
          
            for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset+1 ; i<  ((wimaxHdr->PIH.OFDMSymbol_offset) + wimaxHdr->PIH.num_OFDMSymbol-1) ; i++)
	      for(uint32_t j = 0 ; j<num_subchannel*num_data_subcarrier ; j++ )
		intpower_[i][j] = power_per_subchannel;

	    // last num_symbol_per_slot 
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset +wimaxHdr->PIH.num_OFDMSymbol-1 ; i<  (wimaxHdr->PIH.OFDMSymbol_offset) + wimaxHdr->PIH.num_OFDMSymbol/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	      for(uint32_t j = 0 ; j<(((wimaxHdr->PIH.num_subchannels - (num_subchannel - (wimaxHdr->PIH.subchannel_offset)) )%num_subchannel))*num_data_subcarrier ; j++ )
		(intpower_[i][j] = power_per_subchannel); 

          }
	else
          {
          
	    for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + 1)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	      for(uint32_t j = (wimaxHdr->PIH.subchannel_offset)*num_data_subcarrier ; j<(((wimaxHdr->PIH.subchannel_offset)*num_data_subcarrier) + wimaxHdr->PIH.num_subchannels*num_data_subcarrier) ; j++ )
		(intpower_[i][j] = power_per_subchannel); 	

          }
             
      }

}



bool Wimax2SSNetDevice::IsCollision (const All_Headers *wimaxHdr,double power_subchannel)
{
  NS_LOG_FUNCTION_NOARGS ();
  bool collision = false; 
  uint32_t num_subchannel = getPhy()->getNumsubchannels (DL_); 
  uint32_t num_symbol_per_slot = getPhy()->getSlotLength (DL_); 

  if(wimaxHdr->PIH.num_OFDMSymbol % num_symbol_per_slot == 0 )   // chk this condition , chk whether broacastcid reqd ir not. 
    // if((wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels -1) > num_subchannel)
    //if(wimaxHdr->PIH.num_OFDMSymbol >= num_symbol_per_slot)
    {
      if(wimaxHdr->PIH.num_OFDMSymbol > num_symbol_per_slot) 
	{
	  //numsymbols = (int)  ceil((((wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels -1) - num_subchannel) % num_subchannel));
	  //numsymbols *= num_symbol_per_slot;  
	  // for the first num_symbol_per_slot symbols 
          for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
            {
              for(uint32_t j = (wimaxHdr->PIH.subchannel_offset) ; j<=num_subchannel ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);			
		    return collision;
		  }
              //if(collision == true) return collision;  
	    }
	  // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
          
          for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset+num_symbol_per_slot ; i<  (wimaxHdr->PIH.OFDMSymbol_offset) + (wimaxHdr->PIH.num_OFDMSymbol-num_symbol_per_slot) ; i++)
            {
              for(uint32_t j = 1 ; j<=num_subchannel ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  } 
	      //              if(collision == true) return collision; 
            } 
	  // last num_symbol_per_slot 
          for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset +wimaxHdr->PIH.num_OFDMSymbol-num_symbol_per_slot ; i<  wimaxHdr->PIH.OFDMSymbol_offset + wimaxHdr->PIH.num_OFDMSymbol/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)        
            {
	      for(uint32_t j = 1 ; j<=((wimaxHdr->PIH.num_subchannels - (num_subchannel - (wimaxHdr->PIH.subchannel_offset)) )%num_subchannel) ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision ;
		  }
	      //              if(collision == true) return collision;    
            }
	}
      else 
	{

	  for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
            {
	      for(uint32_t j = (wimaxHdr->PIH.subchannel_offset) ; j< (wimaxHdr->PIH.subchannel_offset) + wimaxHdr->PIH.num_subchannels ; j++ )
                if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  }    
	      //            if(collision == true) return collision;  
	    }  

	}

    }

  else

    {
      //numsymbols = (int)  ceil((((wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels -1) - num_subchannel) % num_subchannel));
      //numsymbols *= num_symbol_per_slot;  
      // for the first num_symbol_per_slot symbols 
      if(wimaxHdr->PIH.num_OFDMSymbol > 1) 
	{
	  for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + 1)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	    {
	      for(uint32_t j = (wimaxHdr->PIH.subchannel_offset) ; j<=num_subchannel ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  } 
	      //	      if(collision == true) return collision; 
	    }	
	  // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
          
	  for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset+1 ; i<  ((wimaxHdr->PIH.OFDMSymbol_offset) + wimaxHdr->PIH.num_OFDMSymbol-1) ; i++)
	    {
	      for(uint32_t j = 1 ; j<=num_subchannel ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  }  
	      //		 if(collision == true) return collision; 
	    }
	  // last num_symbol_per_slot 
          for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset +wimaxHdr->PIH.num_OFDMSymbol-1 ; i<  (wimaxHdr->PIH.OFDMSymbol_offset) + wimaxHdr->PIH.num_OFDMSymbol/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	    {
	      for(uint32_t j = 1 ; j<=((wimaxHdr->PIH.num_subchannels - (num_subchannel - (wimaxHdr->PIH.subchannel_offset)) )%num_subchannel) ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  }
	      //               if(collision == true) return collision; 
            }
	}
      else
	{
          
	  for(uint32_t i = wimaxHdr->PIH.OFDMSymbol_offset ; i<  (wimaxHdr->PIH.OFDMSymbol_offset + 1)/*wimaxHdr->PIH.num_OFDMSymbol)*/ ; i++)
	    {
	      for(uint32_t j = (wimaxHdr->PIH.subchannel_offset) ; j< wimaxHdr->PIH.subchannel_offset + wimaxHdr->PIH.num_subchannels ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  } 	
	      //                if(collision == true) return collision;  
	    }

	}
             
    }

  return collision; 

}

//resets these values for the start of a new frame
void Wimax2SSNetDevice::reset_repetition_detectors() {
  NS_LOG_FUNCTION_NOARGS ();
	this_frame_dlmap_processed = false;
	this_frame_ulmap_processed = false;
}

void Wimax2SSNetDevice::DoDispose() {
	Wimax2NetDevice::DoDispose();
}

void Wimax2SSNetDevice::process_FCH() {
	debug10("At %f, entering Wimax2SSNetDevice::process_FCH in SS %s\n", NOW, addr_s());
	//note that symbol time 0 in grid_ is the preamble, so symbol 1 is the first that's actually used
	Ptr<Packet> first  = get_grid_element(1,0).data;
		if (first == 0) {
			first = get_grid_element(1,1).data;
			if (first == 0) return; //we didn't get a viable copy of the FCH, so we can't decode DL_MAP, and therefore can't really do anything useful this frame
		}
	Ptr<Packet> second = get_grid_element(2,0).data;
		if (second == 0) {
			second = get_grid_element(2,1).data;
			if (second == 0) return;//we didn't get a viable copy of the FCH, so we can't decode DL_MAP, and therefore can't really do anything useful this frame
		}
	first->AddAtEnd(second);	
	Frame_Prefix_Header fph;
	first->RemoveHeader(fph);

	debug10("At %f in %s, assembled FCH, re-assembling DL_MAP\n", NOW, addr_s());
	
	Ptr<Packet> assembled = Create<Packet>();
	All_Headers ah;
	DL_MAP_FRAME_HEADER dmfh;
	
	//the repetition_coding field in the fph is only 2 bits, so it uses the following encoding
	// 00 - only one transmission
	// 01 - two total transmissions  (one retransmission)
	// 10 - four total transmissions (three retransmissions)
	// 11 - six total transmissions  (five retransmissions)
	int repetitions = 1;
	if (fph.repetition_coding > 0) repetitions *= fph.repetition_coding*2;

	for (int i = 0; i < repetitions; i++) {
		PHY_INFO_Header *dl_map_pih = new PHY_INFO_Header();
		dl_map_pih->num_subchannels = fph.DL_MAP_length / repetitions;
		dl_map_pih->num_OFDMSymbol = getPhy()->getSlotLength(DL_);
		dl_map_pih->subchannel_offset = 2 + i * (fph.DL_MAP_length / repetitions); //from the FCH
		dl_map_pih->OFDMSymbol_offset = DL_PREAMBLE;
		dl_map_pih->direction = DL;

        	std::vector<Ptr<Packet> > slot_chunks;

                uint8_t last_subchannel = dl_map_pih->subchannel_offset + dl_map_pih->num_subchannels;

                for (uint8_t i = dl_map_pih->subchannel_offset; i < last_subchannel; i++) 
                        for (uint8_t j = dl_map_pih->OFDMSymbol_offset; j < DL_PREAMBLE + getPhy()->getSlotLength(DL_); j++) {
                                if (get_grid_element(j,i).collision || ! get_grid_element(j,i).received) {
                                        debug10("Error assembling packet from DL Subframe.  It should have spanned subchannels %u to %u and OFDM-Symbol-Times %u to %u, but %s\n", dl_map_pih->subchannel_offset, dl_map_pih->num_subchannels, dl_map_pih->OFDMSymbol_offset, dl_map_pih->num_OFDMSymbol, get_grid_element(j,i).collision ? "there was a collision" : "at least one block was not received");
                                        delete dl_map_pih; 
                                        goto next_rep;
                                }    
                                else slot_chunks.push_back(get_grid_element(j,i).data);
                        }
		ah.PIH = *dl_map_pih;
		

		debug10("At %f in %s, about to join blocks of DL_MAP\n", NOW, addr_s());
		for (uint32_t i = 0; i < slot_chunks.size(); i++) 
			assembled->AddAtEnd(slot_chunks[i]);
	     
		if (assembled->GetSize() < 6) { 
			debug10("ERROR Assembling packet.  Too short.  dropping\n");
			delete dl_map_pih; 
			goto next_rep;
		}
		assembled->RemoveHeader(dmfh);
		debug10("At %f in %s, DL_MAP assembled, sending for processing\n", NOW, addr_s());
		process_dl_map(&dmfh, start_of_frame.GetSeconds());
		return;  
next_rep:
	debug10("bad copy of DL_MAP, going to next repetition\n");
	}


}

}//namespace 
