import sys
import ns3

DISTANCE = 100 # (m)
NUM_NODES_SIDE = 3

def main(argv):
	
	cmd = ns3.CommandLine()
	
#	cmd = ns3.CommandLine()
	cmd.NumNodesSide = None
#	cmd.AddValue("NumNodesSide", "Grid side number of nodes (total number of nodes will be this number squared)")

	cmd.Plot = None
	cmd.AddValue("Plot", "Plot the results using the matplotlib python module")

	cmd.Parse(argv)

	wimax = ns3.Wimax2Helper.Default()

	
	internet = ns3.InternetStackHelper()
	list_routing = ns3.Ipv4ListRoutingHelper()
	list_routing.Add(static_routing, 0)
	internet.SetRoutingHelper(list_routing)

	ipv4Addresses = ns3.Ipv4AddressHelper()
	ipv4Addresses.setBase(ns3.Ipv4Address("10.0.0.0"), ns3.Ipv4Mask("255.255.255.0"))

	port = 125
	onOffHelper = ns3.OnOffHelper("ns3::UdpSocketFactory", ns3.Address(ns3.InetSocketAddress(ns3.Ipv4Address("10.0.0.1"), port)))
	onOffHelper.SetAttribut("DataRate", ns3.DataRateValue(ns3.DataRate("100kbps")))
	onOffHelper.SetAttribut("OnTime", ns3.RandomVariableValue(ns3.ConstantVariable(1)))
	onOffHelper.SetAttribut("OffTime", ns3.RandomVariableValue(ns3.ConstantVariable(0)))

	addresses = []
	nodes = []

	if cmd.NumNodesSide is None:
		num_nodes_side = NUM_NODES_SIDE
	else:
		num_nodes_side = int(cmd.NumNodesSide)
	
	for xi in range(num_nodes_side):
#		for yi in range(num_nodes_side):
			
		nodes = ns3.Node()
		nodes.append(node)

		static_routing = list_routing.GetRoutingProtocol(0)
		static_routing.SetDefaultRoute(addresses[0], 0, 0)

		internet.Install(ns3.NodeContainer(node))

		mobility = ns3.ConstantPositionMobilityModel()
		mobility.SetPosition(ns3.Vector(xi*DISTANCE, xi%2 ,0))
		node.AggregateObject(mobility)

		devices = wimax.Install(node)
		ipv4_interfaces = ipv4Addresses.Assign(devices)
		addresses.append(ipv4_interfaces.GetAddress(0))

	for i, node in enumerate(nodes, start=1):
		destaddr = addresses[0]
		onOffHelper.SetAttribut("Remote", ns3.AddressValue(ns3.InetSocketAddress(destaddr, port)))

		app.Start(ns3.Seconds(ns3.UniformVariable(20,30).GetValue()))

		flowmon_helper = ns3.FlowMonitorHelper()
		monitor = flowmon_helper.InstallAll()
		monitor.SetAttribut("DelayBinWidth", ns3.DoubleValue(0.001))
		monitor.SetAttribut("JitterBinWidth", ns3.DoubleValue(0.001))
		monitor.SetAttribut("PacketSizeBinWidth", ns3.DoubleValue(20))

		ns3.Simulator.Stop(ns3.Seconds(44.0))
		ns3.Simulator.Run()

	def print_stats(os, st):
		print >> os, "  Tx Bytes: ", st.txBytes
		print >> os, "  Rx Bytes: ", st.rxBytes
		print >> os, "  Tx Packets: ", st.txPackets
		print >> os, "  Rx Packets: ", st.rxPackets

	monitor.CheckForLostPackets()
	classifier = flowmon_helper.GetClassifier()

	if cmd.Plot is not None:
		import pylab
		delays = []
		for flow_id, flow_stats in monitor.GetFlowStats():
			tupl = classifier.FindFlow(flow_id)
			if tupl.protocol == 17 and tupl.sourcePort == 698:
				continue
			delays.append(flow_stats.delaySum.GetSeconds()/ flow_stats.rxPackets)
		pylab.hist(delays, 20)

		pylab.xlabel("Delay (s)")
		pylab.ylabel("Number of Flows")
		pylab.show()
	
	return 0

if __name__ == '__main__':
	sys.exit(main(sys.argv))
		
