/* This class contains the control agent located in IEEE 802.16 BS responsible
 * for synchronization between BSs.
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#include "ns3/wimax2-control-agent.h"
#include "ns3/wimax2-net-device.h" 
#include "ns3/wimax2-bs-net-device.h"
#include "ns3/wimax2-bs-scheduler.h"
#include "ns3/wimaxneighborentry.h"
#include "ns3/random-variable.h"
#include "ns3/wimax2-common.h"

#define MYNUM	HelperFunctions::getMacAddrString(Mac48Address::ConvertFrom(mac_->GetAddress())).c_str()

 namespace ns3 {

/*
 * Handler for timer expiration
 */
void UpdateTimer::handle ()
{
	busy_ = 0;
	paused_ = 0;
	stime  = 0.0;
	rtime = 0.0;
  a_->sendUpdate();
}

/*
 * Handler for response timer expiration
 */
void ScanRspTimer::handle ()
{
	busy_ = 0;
	paused_ = 0;
	stime  = 0.0;
	rtime = 0.0;
  a_->agent()->send_scan_response(a_->cid());
}

UpdateTimer::UpdateTimer(Wimax2ControlAgent *a)  : Wimax2Timer ( a->mac_->GetObject<Wimax2NetDevice>()) 
	  { a_ = a;}

/*
 * Creates a Wimax controler agent
 * Initializes the agent and bind variable to be accessible in TCL
 */
Wimax2ControlAgent::Wimax2ControlAgent() : mac_(0), updatetimer_ (this)
{
  nbmapentry_=0;
  LIST_INIT (&scan_req_head_);
  
 // debug_ = 0;
  adv_interval_ = 1.0;
//  default_association_level_ = 0;
  synch_frame_delay_ = 50;


  //schedule first update
  UniformVariable ru(0,UPDATE_JITTER);
  updatetimer_.start (ru.GetValue(0,UPDATE_JITTER));
}

/* 
 * Interface with TCL interpreter
 * @param argc The number of elements in argv
 * @param argv The list of arguments
 * @return TCL_OK if everything went well else TCL_ERROR
 */
// int Wimax2ControlAgent::command(int argc, const char*const* argv)
// {
//   //Tcl& tcl= Tcl::instance();
// 
//   if (argc == 3) {
//     // set the Minimum interval between two RAs
//     if (strcmp(argv[1], "set-mac") == 0) {
//       mac_ = (Mac802_16BS*) TclObject::lookup(argv[2]);
//       mac_->setCtrlAgent (this);
//       return TCL_OK;
//     }
//   } else if (argc == 4) {
//     if (strcmp(argv[1], "add-neighbor") == 0) {
//       //the parameter is the mac, and we also extract the node
//       Mac802_16 *tmp = (Mac802_16 *) TclObject::lookup(argv[2]);
//       if (nbmapentry_ == MAX_MAP_ENTRY) {
// 	fprintf (stderr, "Table size exceeding. Increase MAX_MAP_ENTRY\n");
//       }
//       // The following is not supported outside of our mobility
//       // package so we provide a workaround
//       //tcl.evalf ("%s get-node", argv[2]);
//       //Node *tmpNode = (Node *) TclObject::lookup(tcl.result());
//       Node *tmpNode = (Node *) TclObject::lookup(argv[3]);
//       
//       //add entry
//       maptable_[nbmapentry_][0] = tmp->addr();
//       maptable_[nbmapentry_][1] = tmpNode->address();
//       debug2 ("Adding neighbor %s (mac %d) in %s\n", Address::instance().print_nodeaddr(tmpNode->address()),
// 	      tmp->addr(), MYNUM);
//       nbmapentry_++;
//       return TCL_OK;
//     }
//   }
// 
//   return (Agent::command(argc, argv));
// }

/*
 * Send an update (DCD/UCD) to all neighboring BSs
 */
void Wimax2ControlAgent::sendUpdate ()
{
  //get the DCD/UCD message to include in the update
  Ptr<Packet> dcd = mac_->getMap()->getDCD();
  Ptr<Packet> ucd = mac_->getMap()->getUCD();
  
  All_Headers ah_dcd, ah_ucd;

  //allocate data to store information
  DCD_FRAME_HEADER dcdframe;
  dcd->RemoveHeader(ah_dcd);
  dcd->RemoveHeader(dcdframe);
  UCD_FRAME_DATA ucdframe;
  ucd->RemoveHeader(ah_ucd);
  ucd->RemoveHeader(ucdframe);
  //mac802_16_dcd_frame *dcdframe = (mac802_16_dcd_frame*) dcd->accessdata();
  //mac802_16_ucd_frame *ucdframe = (mac802_16_ucd_frame*) ucd->accessdata();
  
  Ptr<Packet> p = mac_->getPacket();
  All_Headers ah;
  p->RemoveHeader(ah);
  // hdr_ip *iph = HDR_IP(p);
//   hdr_wimaxbs *rh = HDR_WIMAXBS(p);
//   hdr_cmn *hdrc = HDR_CMN(p);
  
  //rh->getType() = WIMAX_BS_ADV; 
  ah.GMH.GMH.len = ah_dcd.GMH.GMH.len+ah_ucd.GMH.GMH.len; //TBD: remove double header
  //set content
  ah.dest = mac_->addr();
  //p->allocdata (sizeof (mac802_16_dcd_frame)+sizeof (mac802_16_ucd_frame)); 
  //unsigned char *data = p->accessdata();
  //memcpy (data, dcdframe, sizeof (mac802_16_dcd_frame));
  //memcpy (data+sizeof (mac802_16_dcd_frame), ucdframe, sizeof (mac802_16_ucd_frame));
  p->AddHeader(ucdframe);
  p->AddHeader(dcdframe);
  p->AddHeader(ah);
  
  Ptr<Packet> tmpPkt; 
  Ipv4Header iph;
  for (int i = 0; i < nbmapentry_ ; i++) {
    tmpPkt = p->Copy();
    tmpPkt->RemoveHeader(ah);
    //iph = HDR_IP(tmpPkt);
    //set packet destination
    ah.dest = maptable2_[i];
    //iph->daddr() = maptable_[i][1];
    //iph->dport() = port();
    debug2 ("At %f in node %s, send update to node %s\n", NOW, MYNUM,HelperFunctions::getMacAddrString(ah.dest).c_str());
    debug2 ("frame number=%d\n", dcdframe.mf.frame_number);
    mac_->Send(tmpPkt, ah.dest, 0);
  }

  //reschedule timer
  updatetimer_.resched (adv_interval_);
}

/* 
 * Process received packet
 * @param p The packet received
 */
void Wimax2ControlAgent::recv(Ptr<Packet> p)//, Handler *h)
{
  assert (p);

  hdr_wimaxbs rh;
  p->PeekHeader(rh);
  switch (rh.subtype_) {
  case WIMAX_BS_ADV:
    processUpdate (p);
    break;
  case WIMAX_BS_SYNCH_REQ:
    process_synch_request (p);
    break;
  case WIMAX_BS_SYNCH_RSP:
    process_synch_response (p);
    break;
  default:
    fprintf (stderr, "Unknown message type in Wimax2ControlAgent\n");
  }
  p=0;
}

/* 
 * Process received packet
 * @param p The update received
 * @param h The handler that sent the packet
 */
void Wimax2ControlAgent::processUpdate(Ptr<Packet> p)
{
  Ipv4Header iph;
  p->RemoveHeader(iph);
  debug2 ("At %f in node %s, Wimax2ControlAgent received update message from %d\n", NOW, MYNUM, iph.GetDestination().Get());

  hdr_wimaxbs rh;
  p->RemoveHeader(rh);
  WimaxNeighborEntry *entry = mac_->nbr_db_->getNeighbor(rh.macAddr());

  //check if we know about this neighbor
  bool found = false;
  for (int i = 0; i < nbmapentry_ ; i++) {
    if (maptable1_[i]==iph.GetDestination())
      found = true;
  }
  assert (found);

  if (entry==NULL) {
    debug2 ("\tNew neighbor detected...add entry for mac %s\n", HelperFunctions::getMacAddrString(rh.macAddr()).c_str());
    entry = new WimaxNeighborEntry (rh.macAddr());
    mac_->nbr_db_->addNeighbor(entry);
  }
  //update entry
  //unsigned char *data = p->accessdata();
  DCD_FRAME_HEADER dfh;
  UCD_FRAME_DATA ufh;
  DCD_FRAME_HEADER *dcdframe = new DCD_FRAME_HEADER();
  UCD_FRAME_DATA *ucdframe = new UCD_FRAME_DATA();

  
  debug2 ("\tframe number=%d ccc=%d\n", dcdframe->mf.frame_number,ucdframe->ucdf.config_change_count);
  DCD_FRAME_HEADER *dcdtmp = entry->getDCD();
  if (dcdtmp)
    free(dcdtmp);
  UCD_FRAME_DATA *ucdtmp = entry->getUCD();
  if (ucdtmp)
    free(ucdtmp);
  entry->setDCD (dcdframe);
  entry->setUCD (ucdframe);
  //free (p);
}

/**
 * Process scanning request
 * @param p The request
 */
void Wimax2ControlAgent::process_scan_request (Ptr<Packet> req)
{
  All_Headers wimaxHdr_req;
  req->RemoveHeader(wimaxHdr_req);
  gen_mac_header_t *header_req = &wimaxHdr_req.GMH.GMH;
  mac802_16_mob_scn_req_frame req_frame;
  req->RemoveHeader(req_frame);
  
  debug2 ("At %f in Mac %s received scanning request from %d\n", NOW, HelperFunctions::getMacAddrString(mac_->addr()).c_str(), header_req->cid);
  
  //for first implementation we disregard the information 
  //sent by MN. Just use default association mechanisms

  //should check if there is already pending request: TBD
  Scan_req *entry = NULL;
  
  for (entry = scan_req_head_.lh_first ; entry && (entry->cid() != header_req->cid); entry=entry->next_entry());

  if (entry) {
    debug2 ("\tDuplicate requests. Skip...\n");
    return;
  } else {
    entry = new Scan_req (this, synch_frame_delay_*mac_->getFrameDuration(), header_req->cid, &req_frame);
    entry->insert_entry (&scan_req_head_);
    entry->start_frame()=mac_->getFrameNumber();
  }

  switch (defaultlevel_){
  case 0: //Scan without association
    entry->response()->scan_duration = req_frame.scan_duration;
    entry->response()->start_frame = 2;
    entry->response()->report_mode = 0; //no report for now
    entry->response()->interleaving_interval = req_frame.interleaving_interval;
    entry->response()->scan_iteration = req_frame.scan_iteration;
    entry->response()->n_recommended_bs_index = 0;
    entry->response()->n_recommended_bs_full = nbmapentry_;
    for (int i = 0; i < nbmapentry_ ; i++) {
      entry->response()->rec_bs_full[i].recommended_bs_id = maptable2_[i];
      entry->response()->rec_bs_full[i].scanning_type = SCAN_WITHOUT_ASSOC;
    }
    //send response
    mac_->send_scan_response (entry->response(), header_req->cid);
    //clean data
    entry->remove_entry();
    delete entry;
    break;

  case 1: //Association without coordination
    entry->response()->scan_duration = req_frame.scan_duration;
    entry->response()->start_frame = 2;
    entry->response()->report_mode = 0; //no report for now
    entry->response()->interleaving_interval = req_frame.interleaving_interval;
    entry->response()->scan_iteration = req_frame.scan_iteration;
    entry->response()->n_recommended_bs_index = 0;
    entry->response()->n_recommended_bs_full = nbmapentry_;
    for (int i = 0; i < nbmapentry_ ; i++) {
      entry->response()->rec_bs_full[i].recommended_bs_id = maptable2_[i];
      entry->response()->rec_bs_full[i].scanning_type = SCAN_ASSOC_LVL0;
    }
    //send response
    mac_->send_scan_response (entry->response(), header_req->cid);
    //clean data
    entry->remove_entry();
    delete entry;
    break;

  case 2: //Association with coordination
    //init data
    entry->response()->n_recommended_bs_index = 0;
    entry->response()->n_recommended_bs_full = 0;
    entry->pending_rsp () = 0;
 
    //send request to neighbors
    for (int i = 0; i < nbmapentry_ ; i++) {
      Ptr<Packet> p = mac_->getPacket();
      Ipv4Header iph;
      hdr_wimaxbs rh;
	  All_Headers ah;
	  p->RemoveHeader(ah);
		
      rh.subtype_ = WIMAX_BS_SYNCH_REQ; 
      ah.GMH.GMH.len = 30; //We need to define proper size
      //set content
      iph.SetDestination(maptable1_[i]);
      //iph->dport() = port();
      
      rh.macAddr() = mac_->getCManager()->get_connection(header_req->cid, true)->getPeerNode()->getAddr();
      rh.cid = header_req->cid;
      rh.scanning_type = (wimax_scanning_type) defaultlevel_;
      //we suggest a rendez-vous time at the beginning of each 
      //scan iteration
      rh.current_frame = mac_->getFrameNumber();
      //if we want to start scanning 2 frames after, then add one (i.e 3) because 
      //the message will be sent on the next frame
      printf ("scan_duration=%d, scan_interval=%d\n", req_frame.scan_duration,
	      req_frame.interleaving_interval);
      rh.rdvt = (i+1)*(req_frame.scan_duration+req_frame.interleaving_interval)+synch_frame_delay_+START_FRAME_OFFSET+1;
      rh.rendezvous_time = NOW+rh.rdvt * mac_->getFrameDuration();     
      printf ("Request: current frame=%d, rdv frame=%d, rdv time=%f\n",
	      rh.current_frame, rh.rdvt, rh.rendezvous_time); 

      entry->pending_rsp ()++;
      
      p->AddHeader(iph);
      p->AddHeader(rh);
      p->AddHeader(ah);
      mac_->Send(p, rh.macAddr(), 0);

      
      entry->response()->n_recommended_bs_full = i + 1;
      entry->response()->rec_bs_full[i].recommended_bs_id = maptable2_[i];
      entry->response()->rec_bs_full[i].scanning_type = rh.scanning_type;
      entry->response()->rec_bs_full[i].rdv_time = rh.rdvt-synch_frame_delay_;  
    }

    //continue initializing response
    entry->response()->scan_duration = req_frame.scan_duration;
    entry->response()->start_frame = START_FRAME_OFFSET;
    entry->response()->report_mode = 0; //no report for now
    entry->response()->interleaving_interval = entry->request()->interleaving_interval;
    entry->response()->scan_iteration = entry->request()->scan_iteration;
    entry->response()->n_recommended_bs_index = 0;
    
    //printf ("Response: current frame=%d (now=%f), start in %d frame (t=%f)\n",
    //	    mac_->getFrameNumber (), NOW, entry->response()->start_frame, NOW+(entry->response()->start_frame*mac_->getFrameDuration()));
    //send response
    //((BSScheduler*)mac_->getScheduler())->send_scan_response (entry->response(), entry->cid());

    break;
  case 3: //Network Assisted Association reporting

    break;
  default:
    break;
  }
}

/**
 * Process synchronization request
 * @param req The request
 */
void Wimax2ControlAgent::process_synch_request (Ptr<Packet> req)
{
  Ipv4Header iph_req;
  hdr_wimaxbs rh_req;
  req->RemoveHeader(iph_req);
  req->PeekHeader(rh_req);
  debug2 ("At %f in node %s, Wimax2ControlAgent received synch request from %s\n", NOW, MYNUM,
	 HelperFunctions::getMacAddrString(rh_req.macaddr_).c_str());

  //schedule rendez-vous time
  //schedule sending of Fast-ranging-IE
  mac_->addNewFastRanging (rh_req.rendezvous_time, rh_req.macAddr());
  

  //send response
  Ptr<Packet> p = mac_->getPacket();
  Ipv4Header iph;
  hdr_wimaxbs rh;

  
  rh.subtype_ = WIMAX_BS_SYNCH_RSP; 
  //hdrc->size() = 30; //We need to define proper size
  //set content
  //iph = HDR_IP(p);
  iph.SetDestination(iph_req.GetSource());
  
  //we accept what the serving BS sent
  rh.cid = rh_req.cid;
  rh.scanning_type = rh_req.scanning_type;
  rh.current_frame = rh_req.current_frame;
  rh.rdvt =rh_req.rdvt;
  rh.rendezvous_time = rh_req.rendezvous_time; 

  //send (p,0);
}

/**
 * Process synchronization response
 * @param p The response
 */
void Wimax2ControlAgent::process_synch_response (Ptr<Packet> p)
{
  Ipv4Header iph;
  p->RemoveHeader(iph);
  hdr_wimaxbs rh;
  p->RemoveHeader(rh);
  debug2 ("At %f in node %s, Wimax2ControlAgent received synch response from %d\n", NOW, MYNUM,
	 iph.GetSource().Get());

  
  int i;

  //update information
  Scan_req *entry;
  for (entry = scan_req_head_.lh_first ; entry && (entry->cid() != rh.cid); entry=entry->next_entry());

  assert (entry);

  i = entry->response()->n_recommended_bs_full;
  entry->response()->n_recommended_bs_full = i + 1;
  entry->response()->rec_bs_full[i].recommended_bs_id = maptable2_[i];
  entry->response()->rec_bs_full[i].scanning_type = rh.scanning_type;
  entry->response()->rec_bs_full[i].rdv_time = rh.rdvt;
}

/**
 * Send a scan response to the MN that has the given CID
 * @param cid The CID of the MN
 */
void Wimax2ControlAgent::send_scan_response (int cid)
{
  Scan_req *entry;
  for (entry = scan_req_head_.lh_first ; entry && (entry->cid() != cid); entry=entry->next_entry());
  
  assert (entry);
  
  entry->response()->scan_duration = entry->request()->scan_duration;
  printf ("Response: current frame=%d, start frame=%d diff=%d\n", 
	  mac_->getFrameNumber(), entry->start_frame(), 100-(mac_->getFrameNumber()-entry->start_frame()));
  printf ("Response: current frame=%d (now=%f), start in %d frame (t=%f)\n",
	  mac_->getFrameNumber (), NOW, entry->response()->start_frame, NOW+(entry->response()->start_frame*mac_->getFrameDuration()));
  //send response
  mac_->send_scan_response (entry->response(), entry->cid());
  //clean data
  entry->remove_entry();
  delete entry;
}


hdr_wimaxbs::hdr_wimaxbs() {}
hdr_wimaxbs::~hdr_wimaxbs() {}
TypeId hdr_wimaxbs::GetTypeId() {
	static TypeId tid = TypeId("ns3::hdr_wimaxbs")
	.SetParent<Header> ()
	.AddConstructor<hdr_wimaxbs>()
	;
	return tid;
}
TypeId hdr_wimaxbs::GetInstanceTypeId() const{
	return GetTypeId();
}

void hdr_wimaxbs::Print(std::ostream &os) const{
	os << "PRINTING SAMPLE hdr_wimaxbs Header\n";	
}

uint32_t hdr_wimaxbs::GetSerializedSize() const {
	return 28;
}

void hdr_wimaxbs::Serialize (Buffer::Iterator start) const{
	Buffer::Iterator i = start;
	uint8_t temp = subtype_;
	i.WriteU8(temp);
	uint8_t adr[6];
	macaddr_.CopyTo(adr);
	for (int j = 0; j < 6; j++)
		i.WriteU8(adr[j]);
		
	i.WriteHtonU16(cid);
	temp = scanning_type;
	i.WriteU8(scanning_type);
	i.WriteHtonU32(current_frame);
	i.WriteHtonU32(rdvt);
	const double *conversion_holder = &rendezvous_time;
	i.WriteHtonU64(*reinterpret_cast<const uint64_t *>(conversion_holder));
}

uint32_t hdr_wimaxbs::Deserialize(Buffer::Iterator start) {
	debug10("Deseriailizing from wimax2-ctrl-agent\n");
	Buffer::Iterator i = start;
	uint8_t temp;
	temp = i.ReadU8();
	subtype_ = static_cast<wimaxbs_type>(temp);
	uint8_t adr[6];
	for (int j = 0; j < 6; j++)
		adr[j] = i.ReadU8();
	macaddr_.CopyFrom(adr);
	
	cid = i.ReadNtohU16();
	temp = i.ReadU8();
	scanning_type = static_cast<wimax_scanning_type>(temp);
	current_frame = i.ReadNtohU32();
	rdvt = i.ReadNtohU32();
	
	uint64_t temp2 = i.ReadNtohU64();
	uint64_t *conversion_holder = &temp2;
	rendezvous_time = *reinterpret_cast<double *>(conversion_holder);
	return GetSerializedSize();
}

ScanRspTimer::ScanRspTimer(Scan_req *a)  : Wimax2Timer(a->agent()->mac_) 
	  { a_ = a;}


}//namespace ns3
