/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2008 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
 
#include <iostream>
#include "wimax2-net-device.h"
#include "wimax2-channel.h"
#include "wimax2-phy.h"
#include "ns3/node.h"
#include "ns3/packet.h"
#include "ns3/log.h"
#include "ns3/pointer.h"
#include "ns3/error-model.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/wimax2-scheduler.h"
#include "ns3/wimax2-bs-net-device.h"
#include "ns3/wimax2-ss-net-device.h"
#include "framemap.h"

#include "ns3/arp-header.h"


NS_LOG_COMPONENT_DEFINE ("Wimax2NetDevice");

namespace ns3 {


/* Defines frequencies for 3.5 GHz band and 7 Mhz freqency bandwidth */
/* Will be removed when a dynamic way is added */
static const int nbFreq = 5;
static const double frequencies[] = { 3.486e+9, 3.493e+9, 3.5e+9, 3.507e+9, 3.514e+9};


TypeId 
Wimax2NetDevice::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::Wimax2NetDevice")
    .SetParent<NetDevice> ()
    .AddConstructor<Wimax2NetDevice> ()
    //
    // Trace sources at the "top" of the net device, where packets transition
    // to/from higher layers.
    //
//     .AddTraceSource ("MacTx", 
//                      "Trace source indicating a packet has arrived for transmission by this device",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_macTxTrace))
//     .AddTraceSource ("MacTxDrop", 
//                      "Trace source indicating a packet has been dropped by the device before transmission",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_macTxDropTrace))
//     .AddTraceSource ("MacPromiscRx", 
//                      "A packet has been received by this device, has been passed up from the physical layer "
//                      "and is being forwarded up the local protocol stack.  This is a promiscuous trace,",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_macPromiscRxTrace))
//     .AddTraceSource ("MacRx", 
//                      "A packet has been received by this device, has been passed up from the physical layer "
//                      "and is being forwarded up the local protocol stack.  This is a non-promiscuous trace,",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_macRxTrace))
    .AddTraceSource ("MacRxDrop", 
                     "Trace source indicating a packet was dropped before being forwarded up the stack",
                     MakeTraceSourceAccessor (&Wimax2NetDevice::m_macRxDropTrace))
//     //
//     // Trace souces at the "bottom" of the net device, where packets transition
//     // to/from the channel.
//     //
//     .AddTraceSource ("PhyTxBegin", 
//                      "Trace source indicating a packet has begun transmitting over the channel",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_phyTxBeginTrace))
//     .AddTraceSource ("PhyTxEnd", 
//                      "Trace source indicating a packet has been completely transmitted over the channel",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_phyTxEndTrace))
//     .AddTraceSource ("PhyTxDrop", 
//                      "Trace source indicating a packet has been dropped by the device during transmission",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_phyTxDropTrace))
// 
//     .AddTraceSource ("PhyRxBegin", 
//                      "Trace source indicating a packet has begun being received by the device",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_phyRxBeginTrace))
//     .AddTraceSource ("PhyRxEnd", 
//                      "Trace source indicating a packet has been completely received by the device",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_phyRxEndTrace))
//     .AddTraceSource ("PhyRxDrop", 
//                      "Trace source indicating a packet has been dropped by the device during reception",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_phyRxDropTrace))
// 
//     // Trace sources designed to simulate a packet sniffer facility (tcpdump).
//     .AddTraceSource ("Sniffer", 
//                      "Trace source simulating a non-promiscuous packet sniffer attached to the device",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_snifferTrace))
//     .AddTraceSource ("PromiscSniffer", 
//                      "Trace source simulating a promiscuous packet sniffer attached to the device",
//                      MakeTraceSourceAccessor (&SmNetDevice::m_promiscSnifferTrace))
//   .AddTraceSource("Rx", "Receive trace", MakeTraceSourceAccessor (&Wimax2NetDevice::m_traceRx))
//   .AddTraceSource("Tx", "Transmit trace", MakeTraceSourceAccessor (&Wimax2NetDevice::m_traceTx))
//     .AddAttribute ("ReceiveErrorModel",
//                    "The receiver error model used to simulate packet loss",
//                    PointerValue (),
//                    MakePointerAccessor (&Wimax2NetDevice::m_receiveErrorModel),
//                    MakePointerChecker<ErrorModel> ())
//     .AddTraceSource ("PhyRxDrop",
//                      "Trace source indicating a packet has been dropped by the device during reception",
//                      MakeTraceSourceAccessor (&Wimax2NetDevice::m_phyRxDropTrace))
    ;
  return tid;
}


//no boot time, used for BS
Wimax2NetDevice::Wimax2NetDevice() 
  : phy_ (0),
    m_node (0),
    m_mtu (0xffff),
    m_ifIndex (0)
{
  NS_LOG_FUNCTION(this);
  static uint32_t last_id_index = 1;
  id_index = last_id_index++;
  phy_ = CreateObject<Wimax2Phy> (this);
  macmib_ = new Mac802_16MIB();
  scheduler_ = new Wimax2Scheduler();
  scheduler_->setMac(this);
  //init variable
  LIST_INIT(&classifier_list_);
  peer_list_ = (struct peerNode_list *) malloc (sizeof(struct peerNode_list));
  classifier_list_ = *((struct sduClassifier_list *) malloc(sizeof(struct sduClassifier_list)));
  LIST_INIT(peer_list_);
  LIST_INIT(&classifier_list_);
  DestClassifier *clas = new DestClassifier(this);
  addClassifier(clas);
  nb_peer_ = 0;
  head_pkt_ = NULL;
  tail_pkt_ = NULL;
  trash_pkt_ = NULL;
  frame_number_ = 0;
  nbr_db_ = new NeighborDB();
  pktBuf_ = Ptr<Packet>(NULL);
  notify_upper_ = false; 
  serviceFlowHandler_ = new ServiceFlowHandler();
  serviceFlowHandler_->setMac(this);

  
  initTimer_ = new InitTimer (this);
  
  initTimer_->Schedule(Seconds(0));
  debug10("past initTimer schedule\n");
 // Ptr<Wimax2NetDevice> smartThis(this);
  //connectionManager_ = new ConnectionManager (smartThis);

  
  connectionManager_ = new ConnectionManager (this);
  data_loss_ = 0;
}

Wimax2NetDevice::Wimax2NetDevice (Time boot)
  : phy_ (0),
    m_node (0),
    m_mtu (0xffff),
    m_ifIndex (0)
{
  debug2("Creating new net device, boot time = %f\n", boot.GetSeconds());
  NS_LOG_FUNCTION(this);
  static uint32_t last_id_index = 1;
  id_index = last_id_index++;
  phy_ = CreateObject<Wimax2Phy> (this);
  macmib_ = new Mac802_16MIB();
  scheduler_ = new Wimax2Scheduler();
  scheduler_->setMac(this);
  //init variable
  LIST_INIT(&classifier_list_);
  peer_list_ = (struct peerNode_list *) malloc (sizeof(struct peerNode_list));
  classifier_list_ = *((struct sduClassifier_list *) malloc(sizeof(struct sduClassifier_list)));
  LIST_INIT(peer_list_);
  LIST_INIT(&classifier_list_);
  DestClassifier *clas = new DestClassifier(this);
  addClassifier(clas);
  nb_peer_ = 0;
  head_pkt_ = NULL;
  tail_pkt_ = NULL;
  trash_pkt_ = NULL;
  frame_number_ = 0;
  nbr_db_ = new NeighborDB();
  pktBuf_ = Ptr<Packet>(NULL);
  notify_upper_ = false; 
  serviceFlowHandler_ = new ServiceFlowHandler();
  serviceFlowHandler_->setMac(this);

  
  initTimer_ = new InitTimer (this);
  
  initTimer_->Schedule(boot);
  debug10("past initTimer schedule\n");
 // Ptr<Wimax2NetDevice> smartThis(this);
  //connectionManager_ = new ConnectionManager (smartThis);

  
  connectionManager_ = new ConnectionManager (this);
  data_loss_ = 0;
}

Wimax2NetDevice::~Wimax2NetDevice() {
}

void Wimax2NetDevice::SetSchedulerDlRatio(double rat) {
  NS_LOG_FUNCTION_NOARGS ();
	scheduler_->dlratio_ = rat;
}

void 
Wimax2NetDevice::Receive (Ptr<Packet> packet, uint16_t protocol, Mac48Address to, Mac48Address from)
{
  NS_LOG_FUNCTION_NOARGS ();
  std::cout<< "RECEIVED PACKET (SIZE= " << packet->GetSize() << ", PROTOCOL= " << protocol << ") FROM: " << from << " TO: " << to << " AT TIME " << NOW <<"\n";
  NS_LOG_FUNCTION (packet << protocol << to << from);
  NetDevice::PacketType packetType;

  if (packet->GetSize() == 28) {
	debug10("Received ARP Packet\n");
  }
  
  if (m_receiveErrorModel && m_receiveErrorModel->IsCorrupt (packet) )
    {
      m_phyRxDropTrace (packet);
      return;
    }

  if (to == m_address)
    {
      //m_traceRx(packet, from);
//       m_snifferTrace(packet);
//       m_phyRxEndTrace
      packetType = NetDevice::PACKET_HOST;
    }
  else if (to.IsBroadcast ())
    {
      //m_traceRx(packet, from);
      cout<<"Recieved broadcast packet\n";
      packetType = NetDevice::PACKET_HOST;
    }
  else if (to.IsGroup ())
    {
      //m_traceRx(packet, from);
      cout<<"Recieved multicast packet\n";
      packetType = NetDevice::PACKET_MULTICAST;
    }
  else 
    {
      m_macRxDropTrace(packet);
      cout<<"Recieved packet bound for another host\n";
      packetType = NetDevice::PACKET_OTHERHOST;
      return;
    }
  m_rxCallback (this, packet->Copy(), protocol, from);

   if (!m_promiscCallback.IsNull ())
     {
       m_promiscCallback (this, packet, protocol, from, to, packetType);
     }
}

void 
Wimax2NetDevice::SetChannel (Ptr<Wimax2Channel> channel)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_channel = channel;
  m_channel->Add (getPhy());
  phy_->setChannel(m_channel);
}

void
Wimax2NetDevice::SetReceiveErrorModel (Ptr<ErrorModel> em)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_receiveErrorModel = em;
}

void 
Wimax2NetDevice::SetIfIndex(const uint32_t index)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_ifIndex = index;
}

uint32_t 
Wimax2NetDevice::GetIfIndex(void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_ifIndex;
}

Ptr<Channel> 
Wimax2NetDevice::GetChannel (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  Ptr<Channel> ns3ptr (m_channel);
  return ns3ptr;
}

void
Wimax2NetDevice::SetAddress (Address address)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_address = Mac48Address::ConvertFrom(address);
}

Address 
Wimax2NetDevice::GetAddress (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  //
  // Implicit conversion from Mac48Address to Address
  //
  return m_address;
}

bool 
Wimax2NetDevice::SetMtu (const uint16_t mtu)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_mtu = mtu;
  return true;
}

uint16_t 
Wimax2NetDevice::GetMtu (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_mtu;
}

bool 
Wimax2NetDevice::IsLinkUp (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return true;
}

void 
Wimax2NetDevice::AddLinkChangeCallback (Callback<void> callback)
{
  NS_LOG_FUNCTION_NOARGS ();
}

bool 
Wimax2NetDevice::IsBroadcast (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return true;
}

Address
Wimax2NetDevice::GetBroadcast (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return Mac48Address ("ff:ff:ff:ff:ff:ff");
}

bool 
Wimax2NetDevice::IsMulticast (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return false;
}

Address 
Wimax2NetDevice::GetMulticast (Ipv4Address multicastGroup) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return Mac48Address::GetMulticast (multicastGroup);
}

Address Wimax2NetDevice::GetMulticast (Ipv6Address addr) const
{
  NS_LOG_FUNCTION_NOARGS ();
	return Mac48Address::GetMulticast (addr);
}

bool 
Wimax2NetDevice::IsPointToPoint (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return false;
}

bool 
Wimax2NetDevice::IsBridge (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return false;
}


//rpi linked list of packets to be received 

void Wimax2NetDevice::addPacket( Ptr<Packet> p ) {
	 NS_LOG_FUNCTION_NOARGS ();
  PacketTimerRecord *ptr = new PacketTimerRecord(this);
  //HDR_CMN *ch = HDR_CMN(p);
  // Create a new packet record and set the timer and power for it
  ptr = new PacketTimerRecord(this);
  ptr->p = p;

  double del_inc = 0.000000001;
  All_Headers hd;
  p->PeekHeader(hd);
  double txtime_delta = hd.txtime;
  debug10 ("\t before scheduled event type :%d; at time %lf, r_delta_bs :%lf, r_delta_ss :%lf, recv_delta :%lf\n", type_, hd.txtime, r_delta_bs, r_delta_ss, recv_delta);
  if (type_ == STA_BS) {
        txtime_delta = txtime_delta + r_delta_bs;
        r_delta_bs = r_delta_bs + del_inc;
  } else {
        txtime_delta = txtime_delta + r_delta_ss;
        r_delta_ss = r_delta_ss + del_inc;
  }
  recv_delta = recv_delta + del_inc;
  debug10 ("\t after scheduled event type :%d; at time %lf, sum :%lf, r_delta_bs :%lf, r_delta_ss :%lf, recv_delta :%lf\n", type_, hd.txtime, txtime_delta, r_delta_bs, r_delta_ss, recv_delta);


 ptr->timer.start(txtime_delta); 
 
 
 
//  ptr->timer.sched( duration );
//  ptr->timer.start(HDR_CMN(p)->txtime()-0.000000001);

  //debug("\t scheduled interference event\n");
  // Add the new object to the tail of the list (ensures that packet powers
  // will always be added in the same order (probably doesn't matter but...))
  ptr->prev = tail_pkt_;
  ptr->next = NULL;
  if (head_pkt_ == NULL) {
    head_pkt_ = ptr;
  }
  if (tail_pkt_ != NULL) {
    tail_pkt_->next = ptr;
  }
  tail_pkt_ = ptr;
  
}


bool 
Wimax2NetDevice::Send(Ptr<Packet> p, const Address& dest, uint16_t protocolNumber)
{
	 NS_LOG_FUNCTION_NOARGS ();
	std::cout<< "SENDING PACKET (SIZE = " << p->GetSize() << ", PROTOCOL = " << protocolNumber << ") FROM " << addr_s() << " TO " << Mac48Address::ConvertFrom(dest) << " AT TIME: " << NOW << std::endl;


	//m_traceTx(p, Mac48Address::ConvertFrom(dest));
	UL_Protocol_Header uph(protocolNumber);
	p->AddHeader(uph);
	
	All_Headers ah;
	ah.mac_packet = false;
	ah.dest = Mac48Address::ConvertFrom(dest);
	ah.GMH.GMH.len = p->GetSize(); //6 for GMH will be added later, when that header is actually attached
	ah.GMH.GMH.rsv1 = 0; //used to track whether a packet will be processed by the MAC layer or passed up
						 //but only in the event that CID == BROADCAST_CID (ARP, really) where we can't use that
	ah.PIH.g_ = getPhy()->getCPRatio();
	p->AddHeader(ah);
	MAC_TAG mt(false);
	MAC_TAG clearer(false);
	p->RemovePacketTag(clearer); //in case there's a stray tag on there for some reason
	p->AddPacketTag(mt);
	sendDown(p);
// 	
// 	
// 	cout << "DEST: " << dest << std::endl;
// 	Connection *conn;
// 	Mac48Address d = Mac48Address::ConvertFrom (dest);
// 	//classify flow
// 	int cid = -1;
// 	cout<<"searching existing connections...\n";
// 	cid = classify ( p, d );
// 	
// 	if (cid == -1) {
// 		PeerNode *nNode = new PeerNode(d);
// 		conn = new Connection(CONN_BASIC);
// 		nNode->setOutData(conn);
// 		nNode->insert_entry (peer_list_);
// 		conn->setPeerNode(nNode);
// 		connectionManager_->add_connection(conn, true);
// 		cid = conn->get_cid();
// 		
// 		cout<<"No connection found, creating new connection to send to MAC " << conn->getPeerNode()->getAddr() << std::endl;
// 		cout<<"the new connection's cid is: " << cid << "\n";
// 	}
// 	else {
// 		cout<<"connection found.  cid is " <<cid << std::endl;
// 		conn = connectionManager_->get_connection(cid, OUT_CONNECTION);
// 		
// 		}
// 	
// 	
// 	bool need_drop = false;
// //  			if(!conn->getArqStatus())
// // 			{
// // 				if(conn->queueLength() >= macmib_.queue_length)
// // 				{
// // 					update_watch (&loss_watch_, 1);
// // 					need_drop = true;
// // 					drop (p, "QWI");
// // 				}
// // 			}
// 	if (!need_drop && conn->getArqStatus () == NULL) {
// 		//End RPI
// 		//update mac header information
// 		//set header information
// 		Generic_MAC_Header *wimaxHdr = new Generic_MAC_Header();
// 		wimaxHdr->setHeaderType(false);
// 		wimaxHdr->setEncryptionControl(false);
// 		wimaxHdr->setARQFB(false);
// 		wimaxHdr->setExt(false);
// 		wimaxHdr->setFrag(false);
// 		wimaxHdr->setPck(false);
// 		wimaxHdr->setFFBGM(false);
// 		wimaxHdr->setCI(false);
// 		wimaxHdr->setEks(0);
// 		wimaxHdr->setCID(cid);
// 		wimaxHdr->setHCS(0);
// 		wimaxHdr->setPduLen(1024+wimaxHdr->GetSerializedSize() );
// 		cout<<"Wimax Generic Mac Header:\n";
// 		wimaxHdr->Print(cout);
// 		p->AddHeader(*wimaxHdr);
// 		//Begin RPI
// 	}
// 	//DL scheduler
// 	
// 	//frame assembler
// 	
// 	//PHY
// 	cout<<"SENDING PACKET\n";
//   NS_LOG_FUNCTION (p << dest << protocolNumber);
//   Mac48Address to = Mac48Address::ConvertFrom (dest);
//   conn->enqueue(p);
//   
//   Burst *b;
//   PhyPdu *ppu;
//   map_ = new FrameMap(this);
//   ppu = new PhyPdu(map_, 0);
//   b = new Burst(ppu);
//   scheduler_->transfer_packets1 (conn, b,  1030);
//   phy_->Send (conn->dequeue(), protocolNumber, to, m_address, this);
  return true;
}

/**
 * Change the channel
 * @param channel The new channel
 */
void Wimax2NetDevice::setChannel (int channel)
{
  NS_LOG_FUNCTION_NOARGS ();
  assert (channel < nbFreq);
  phymib_.channel = channel;
  double tmp = frequencies[phymib_.channel];
  getPhy()->setFrequency (tmp);
}

uint8_t Wimax2NetDevice:: get_diuc() {
  NS_LOG_FUNCTION_NOARGS ();
	return 0;
	//handle by subclass
}


/**
 * Return the channel number for the given frequency
 * @param freq The frequency
 * @return The channel number of -1 if the frequency does not match
 */
int Wimax2NetDevice::getChannel (double freq)
{
  NS_LOG_FUNCTION_NOARGS ();
  for (int i = 0 ; i < nbFreq ; i++) {
    if (frequencies[i]==freq)
      return i;
  }
  return -1;
}

/**
 * Return the channel index
 * @return The channel
 */
int Wimax2NetDevice::getChannel ()
{
  NS_LOG_FUNCTION_NOARGS ();
  return phymib_.channel;
}


/**
 * Set the channel to the next from the list
 * Used at initialisation and when loosing signal
 */
void Wimax2NetDevice::nextChannel ()
{
  NS_LOG_FUNCTION_NOARGS ();
  debug2 ("At %f in Mac %d Going to channel %d\n", NOW, id_index, (phymib_.channel+1)%nbFreq);
  setChannel ((phymib_.channel+1)%nbFreq);
}


const char* Wimax2NetDevice::addr_s() {
	if (mac_48_address_string) return mac_48_address_string; //so a new char* isn't allocated each time, causing a memory leak

	//the first time...
	mac_48_address_string = new char[17];
	strcpy(mac_48_address_string,  HelperFunctions::getMacAddrString(m_address).c_str());
	return mac_48_address_string;
}


Ptr<Node> 
Wimax2NetDevice::GetNode (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_node;
}

void 
Wimax2NetDevice::SetNode (Ptr<Node> node)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_node = node;
}

bool 
Wimax2NetDevice::NeedsArp (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return true;
}

void 
Wimax2NetDevice::SetReceiveCallback (NetDevice::ReceiveCallback cb)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_rxCallback = cb;
}

void
Wimax2NetDevice::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  free(peer_list_);
  if (phy_) phy_->Dispose();
  m_node = 0;
  m_receiveErrorModel = 0;
  delete [] mac_48_address_string;
  NetDevice::DoDispose ();
}


void
Wimax2NetDevice::SetPromiscReceiveCallback (PromiscReceiveCallback cb)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_promiscCallback = cb;
}

bool
Wimax2NetDevice::SupportsSendFrom (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return false;
}


/**
 * Return the peer node that has the given address
 * @param index The address of the peer
 * @return The peer node that has the given address
 */
PeerNode* Wimax2NetDevice::getPeerNode (Mac48Address addr)
{
  NS_LOG_FUNCTION_NOARGS ();
  for (PeerNode *p=peer_list_->lh_first;p;p=p->next_entry()) {
    if (p->getAddr() == addr)
      return p;
  }
  return NULL;
}

/**
 * Add the peer node
 * @param The peer node to add
 */
void Wimax2NetDevice::addPeerNode (PeerNode *node)
{
  NS_LOG_FUNCTION_NOARGS ();
  node->insert_entry (peer_list_);
  nb_peer_++;
  //update Rx time so for default value
  node->setRxTime(NOW);
  //node->getStatWatch()->set_alpha(macmib_.rxp_avg_alpha);
}

/**
 * Remove the peer node
 * @param The peer node to remove
 */
void Wimax2NetDevice::removePeerNode (PeerNode *peer)
{
  NS_LOG_FUNCTION_NOARGS ();
  //when removing, we give the CID and it removes IN and OUT connections
  if (peer->getBasic(IN_CONNECTION)) {
    getCManager()->remove_connection (peer->getBasic(IN_CONNECTION)->get_cid());
    delete (peer->getBasic(IN_CONNECTION));
    delete (peer->getBasic(OUT_CONNECTION));
  }
  if (peer->getPrimary(IN_CONNECTION)) {
    getCManager()->remove_connection (peer->getPrimary(IN_CONNECTION)->get_cid());
    delete (peer->getPrimary(IN_CONNECTION));
    delete (peer->getPrimary(OUT_CONNECTION));
  }
  if (peer->getSecondary(IN_CONNECTION)) {
    getCManager()->remove_connection (peer->getSecondary(IN_CONNECTION)->get_cid());
    delete (peer->getSecondary(IN_CONNECTION));
    delete (peer->getSecondary(OUT_CONNECTION));
  }
  if (peer->getInData()) {
    getCManager()->remove_connection (peer->getInData()->get_cid());
    delete (peer->getInData());
  }
  if (peer->getOutData()) {
    getCManager()->remove_connection (peer->getOutData()->get_cid());
    delete (peer->getOutData());
  }
  peer->remove_entry ();
  nb_peer_--;
  delete (peer);
}

/**
 * Return the number of peer nodes
 */
int Wimax2NetDevice::getNbPeerNodes ()
{
  NS_LOG_FUNCTION_NOARGS ();
  return nb_peer_;
}


int Wimax2NetDevice::classify(Ptr<Packet> p, const Address& dest) {
  NS_LOG_FUNCTION_NOARGS ();
//	Mac48Address to = Mac48Address::ConvertFrom(dest);
	int cid = -1;
	int i = 0;
	for (SDUClassifier *n = classifier_list_.lh_first; n && cid==-1; n = n->next_entry()) {
		i++;
		debug10("\t Attempting to classify address using classifier %d\n", i);
		//cid = reinterpret_cast<DestClassifier*>(n)->classify(to);
		cid = n->classify(p, dest);
	}
	return cid;
}

int Wimax2NetDevice::classify (Ptr<Packet> p) {
  NS_LOG_FUNCTION_NOARGS ();
	All_Headers ah;
	p->PeekHeader(ah);
	return classify (p, ah.dest);
}

Mac48Address Wimax2NetDevice::declassify(int cid) {
  NS_LOG_FUNCTION_NOARGS ();
//	Mac48Address to = Mac48Address::ConvertFrom(dest);
	Mac48Address dest = Mac48Address::GetBroadcast();
	int i = 0;
	for (SDUClassifier *n = classifier_list_.lh_first; n && dest == Mac48Address::GetBroadcast(); n = n->next_entry()) {
		i++;
		debug10("\t Attempting to declassify address using classifier %d\n", i);
		//cid = reinterpret_cast<DestClassifier*>(n)->declassify(to);
		dest = n->declassify(cid);
	}
	return dest;
}



/*
 * Add a classifier
 * @param clas The classifier to add
 */
void Wimax2NetDevice::addClassifier (SDUClassifier *clas) 
{
  NS_LOG_FUNCTION_NOARGS ();
  SDUClassifier *n=classifier_list_.lh_first;
  SDUClassifier *prev=NULL;
  int i = 0;
  if (!n || (n->getPriority () >= clas->getPriority ())) {
    //the first element
    debug2 ("Add first classifier\n");
    clas->insert_entry_head (&classifier_list_);
  } else {
    while ( n && (n->getPriority () < clas->getPriority ()) ) {
      prev=n;
      n=n->next_entry();
      i++;
    }
    debug2 ("insert entry at position %d\n", i);
    clas->insert_entry (prev);
  }
  //Register this mac with the classifier
  clas->setMac (this);
}

void Wimax2NetDevice::init() {
  NS_LOG_FUNCTION_NOARGS ();
  initialize_grid();
  //this is actually going to be done by the subclass, which probably won't bother to call up, but the reset_grid call is included here
  //as a reminder incase more subclasses are added later
}

Ptr<Packet> Wimax2NetDevice::getPacket() {
	Ptr<Packet> p = Create<Packet>();
	All_Headers ah;
	ah.mac_packet = true;
	ah.PIH.g_ = getPhy()->getCPRatio();
 	Generic_MAC_Header *GMH = &ah.GMH;
 	GMH->setHeaderType(0);		
 	GMH->setEncryptionControl(0); 
 	GMH->setARQFB(0);	
 	GMH->setExt(0);		
 	GMH->setFrag(0);		
 	GMH->setPck(0);		
 	GMH->setFFBGM(0);	
 	GMH->setCI(0); 		
 	GMH->setEks(0);
 	GMH->setPduLen(6); 
 	GMH->setCID(BROADCAST_CID);	  
 	GMH->setHCS(0);	 
 	//GMH->GMH.len = 0;
 	GMH->GMH.rsv1 = 1; //used to label mac vs non-mac packets if cid == BROADCAST_CID
 	p->AddHeader(ah);
 	MAC_TAG mt(true);
 	p->AddPacketTag(mt);
	return p;
}

/**
 * Return the type of MAC
 * @return the type of node
 */
station_type_t Wimax2NetDevice::getNodeType()
{
  NS_LOG_FUNCTION_NOARGS ();
  return type_;
}

   
   /**
 * Called when a timer expires
 * @param The timer ID
 */
void Wimax2NetDevice::expire (timer_id id)
{
  NS_LOG_FUNCTION_NOARGS ();
  //handle by subclass
}

/**
 * Process the fully received packet - changed for OFDMA 
 * @param p The packet to receive  -rpi
 */
void Wimax2NetDevice::receive (Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
  //handle by subclass
}


/**
 * Start a new DL subframe
 */
void Wimax2NetDevice::start_dlsubframe ()
{
  NS_LOG_FUNCTION_NOARGS ();
  //handle by subclass
}

/**
 * Start a new UL subframe
 */
void Wimax2NetDevice::start_ulsubframe ()
{
  NS_LOG_FUNCTION_NOARGS ();
  //handle by subclass
}

void
Wimax2NetDevice::removePacket( PacketTimerRecord *expired ) {

  NS_LOG_FUNCTION_NOARGS ();
  assert( head_pkt_ != NULL );
  assert( tail_pkt_ != NULL );
  // Remove the expired packet 
  if (expired->prev == NULL) {
    head_pkt_ = expired->next;
  } else {
    expired->prev->next = expired->next;
  }
  if (expired->next == NULL){
    tail_pkt_ = expired->prev;
  } else {
    expired->next->prev = expired->prev;
  }

  // Cannot delete the packet record right away because the TimerHandler class
  // has to perform some more housekeeping.  So we move the packet to the
  // "trash" so it can be deleted the next time a packet expires.

  // If there is a packet in the trash, delete it.
  if (trash_pkt_ != NULL) {
    delete trash_pkt_;
  }
  // Move the expired packet to the trash.
  trash_pkt_ = expired;

}

/*
 * Transmit a packet to the physical layer
 * @param p The packet to send out
 */
void Wimax2NetDevice::transmit(Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
  //handle by subclass
}


/*
 * Return the code for the frame duration
 * @return the code for the frame duration
 */
int Wimax2NetDevice::getFrameDurationCode () {
  NS_LOG_FUNCTION_NOARGS ();
  if (macmib_->frame_duration == 0.0025)
    return 0;
  else if (macmib_->frame_duration == 0.004)
    return 1;
  else if (macmib_->frame_duration == 0.005)
    return 2;
  else if (macmib_->frame_duration == 0.008)
    return 3;
  else if (macmib_->frame_duration == 0.01)
    return 4;
  else if (macmib_->frame_duration == 0.0125)
    return 5;
  else if (macmib_->frame_duration == 0.02)
    return 6;
  else {
    fprintf (stderr, "Invalid frame duration %f\n", macmib_->frame_duration);
    exit (1);
  }
}

/*
 * Set the frame duration using code
 * @param code The frame duration code
 */
void Wimax2NetDevice::setFrameDurationCode (int code) 
{
  NS_LOG_FUNCTION_NOARGS ();
  switch (code) {
  case 0:
    macmib_->frame_duration = 0.0025;
    break;
  case 1:
    macmib_->frame_duration = 0.004;
    break;
  case 2:
    macmib_->frame_duration = 0.005;
    break;
  case 3:
    macmib_->frame_duration = 0.008;
    break;
  case 4:
    macmib_->frame_duration = 0.01;
    break;
  case 5:
    macmib_->frame_duration = 0.0125;
    break;
  case 6:
    macmib_->frame_duration = 0.02;
    break;
  default:
    fprintf (stderr, "Invalid frame duration code %d\n", code);
    exit (1);
  }
}

  
/**
 * Set the variable used to find out if upper layers
 * must be notified to send packets. During scanning we
 * do not want upper layers to send packet to the mac.
 */
void Wimax2NetDevice::setNotify_upper (bool notify) { 
  NS_LOG_FUNCTION_NOARGS ();
  debug2 ("At %f in Mac %d notify %d pktBuf_=%p\n", NOW, id_index, notify, PeekPointer(pktBuf_));
  notify_upper_ = notify; 
  if (notify_upper_ && PeekPointer(pktBuf_)) {
    sendDown (pktBuf_);
    pktBuf_ = Ptr<Packet>(NULL);
  }
}

/*
 * Process packets going out
 * @param p The packet to send out
 */
void Wimax2NetDevice::sendDown(Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
  //handle by subclass
}

void Wimax2NetDevice::drop(Ptr<Packet> p, std::string reason) {
  NS_LOG_FUNCTION_NOARGS ();
	debug2 ("dropping packet %d because of a problem relating to %s\n", (int)p->GetUid(), reason.c_str());
//if we were tracking drops, this is where the trace would fire
}

 
 Mac802_16MIB::Mac802_16MIB() {
	queue_length  = 50;
	frame_duration  = 0.005;
	dcd_interval  = 5;
	ucd_interval  = 5;
	init_rng_interval  = 1;
	lost_dlmap_interval  = 0.6;
	lost_ulmap_interval  = 0.6;
	t1_timeout  = 5* dcd_interval;
	t2_timeout  = 5* init_rng_interval;
	t3_timeout  = 0.2;
	t6_timeout  = 3;
	t12_timeout  = 5*ucd_interval;
	t16_timeout  = 0.01;
	t17_timeout  = 5 * 60; //5 MINUTES - see 802.16d Table 342
	t21_timeout  = 0.02;
	contention_rng_retry  = 16;
	invited_rng_retry  = 16;
	request_retry  = 2;
	reg_req_retry  = 3;
	tproc  = 0.001;
	dsx_req_retry  = 3;
	dsx_rsp_retry  = 3;
	
	rng_backoff_start = 2;
	rng_backoff_stop = 6;
	bw_backoff_start = 2;
	bw_backoff_stop = 6;
	bw_req_contention_size = 4;
	init_contention_size = 4;
	cdma_code_bw_start = 1;
	cdma_code_bw_stop = 63;
	cdma_code_init_start = 64;
	cdma_code_init_stop = 127;
	cdma_code_cqich_start = 128;
	cdma_code_cqich_stop = 195;
	cdma_code_handover_start = 196;
	cdma_code_handover_stop = 255;
	 
	  //mobility extension
	scan_duration  = 50;
	interleaving  = 50;
	scan_iteration  = 2;
	t44_timeout  = 0.1;
	max_dir_scan_time  = 0.2;
	nbr_adv_interval  = 0.5;
	scan_req_retry  = 3;
	
	client_timeout  = 0.5;
	rxp_avg_alpha = 1;
	lgd_factor_ = 1;
	//ITU_PDP_  = 
  
  
   	dcd_interval = 0.05;
 	ucd_interval = 0.05;
 }
 
 Phy802_16MIB::Phy802_16MIB() {
	channel  = 0;
	fbandwidth  = 10e6;
	ttg  = 10;
	rtg  = 10;
	dl_perm = 0;
	ul_perm = 0;
//	disableInterference = 
 }

bool Wimax2NetDevice::isMACPacket(Ptr<Packet> p) {
  NS_LOG_FUNCTION_NOARGS ();
	
	MAC_TAG mt(false);
	p->RemovePacketTag(mt);
	p->AddPacketTag(mt);
	return mt.isMAC;
}

bool Wimax2NetDevice::TagIncomingMACPacket(Ptr<Packet> p) {
  NS_LOG_FUNCTION_NOARGS ();
	All_Headers ah;
	p->RemoveHeader(ah);
	bool isMAC = true;
	int cid = ah.GMH.GMH.cid;
	if (cid > TRANSPORT_SEC_CID_START && cid < TRANSPORT_SEC_CID_STOP) isMAC = false;
	
	if (cid == BROADCAST_CID) {
		//could be either, since broadcast will be selected for things like ARP that
		//are broadcast but should still be sent up
		//note if we were using DHCP like we should be, SS wouldn't need to ARP and
		//wouldn't need to send anything broadcast...
		if (ah.GMH.GMH.rsv1 == 1) isMAC = true;
		else isMAC = false;
	}
	
	if (isMAC) ah.mac_packet = 1;
	else ah.mac_packet = 0;
	
	p->AddHeader(ah);
	
	MAC_TAG old;
	p->RemovePacketTag(old);
	MAC_TAG mt(isMAC);
	p->AddPacketTag(mt);
	return isMAC;
}

int Wimax2NetDevice::GetActualLength(Ptr<Packet> p, bool uplink) {
	NS_LOG_FUNCTION_NOARGS();
	p = p->Copy();
	
	All_Headers ah;
	p->RemoveHeader(ah);

	debug10("GMH.len = %d, actual packet len = %d\n", ah.GMH.GMH.len, p->GetSize());


	if (ah.GMH.GMH.ht != 0) return 6;
	int len = ah.GMH.GetSerializedSize() + p->GetSize();
	if (ah.GMH.GMH.type_arqfb) len += ah.num_of_acks * HDR_MAC802_16_ARQFEEDBK_SIZE;
	if (ah.GMH.GMH.type_frag) len += ah.FSH.GetSerializedSize();
	if (ah.GMH.GMH.type_pck) len += ah.PSH.GetSerializedSize();
	if (ah.GMH.GMH.type_fbgm) {
		if (uplink)	len+= ah.GMSH.GetSerializedSize();
		else len += ah.FFSH.GetSerializedSize();
	}
	return len;
}

Ptr<Wimax2Channel> Wimax2NetDevice::GetWimaxChannel(void ) const {
	return m_channel;
}

//this is a useful function for getting things guarenteed at the start of a frame
void Wimax2NetDevice::SynchronizeFrameStart(Time start) {
	debug10("At %f, synchronising station %s.  Next frame starts at total simulation time %f\n", NOW, addr_s(), start.GetSeconds());
	start_of_frame = start;

	if (this->GetObject<Wimax2SSNetDevice>() != 0) {
		if (this->GetObject<Wimax2SSNetDevice>()->getMacState() == MAC802_16_DISCONNECTED) return;
		Simulator::ScheduleWithContext(Simulator::GetContext(), start + Seconds (getPhy()->getSymbolTime() * (DL_PREAMBLE + getPhy()->getSlotLength(DL_))) - Simulator::Now(), &Wimax2SSNetDevice::process_FCH, this->GetObject<Wimax2SSNetDevice>());
	}

	Simulator::ScheduleWithContext(Simulator::GetContext(), start - Simulator::Now(), &Wimax2NetDevice::reset_grid, this);
	getMap()->setStarttime(start.GetSeconds());
}

incoming_packet Wimax2NetDevice::get_grid_element(uint8_t x, uint8_t y) {
	NS_LOG_FUNCTION_NOARGS();
	//note that the casting trick works here becuase direction is just 0 (false) = DL_, 1 (true) = UL_
	if (x < getMaxDlduration() + getMaxUlduration() && y < getPhy()->getNumsubchannels(static_cast<direction>(x < getMaxDlduration())))
		return grid_[x][y];
	
	debug10("error, invalid grid position access attempted\n");
	incoming_packet dead;
	dead.data = 0;
	return dead;
}

bool Wimax2NetDevice::set_grid_element(uint8_t x, uint8_t y, incoming_packet inc_p) {
	NS_LOG_FUNCTION_NOARGS();
	//note that the casting trick works here becuase direction is just 0 (false) = DL_, 1 (true) = UL_
	if (x < getMaxDlduration() + getMaxUlduration() && y < getPhy()->getNumsubchannels (static_cast<direction>(x < getMaxDlduration()))) {
		grid_[x][y] = inc_p;
		return true;
	}
	return false;
}

void Wimax2NetDevice::reset_grid() {
	NS_LOG_FUNCTION_NOARGS();
	//for this function to work, we need to know the dimensions of a frame
	//to know that,we need to have obtained a MAC_MIB already
	debug10("At %f in station %s, resetting the grid used to store incoming data for the frame\n", NOW, addr_s());
	int max_subch = getPhy()->getNumsubchannels(UL_) > getPhy()->getNumsubchannels(DL_) ? getPhy()->getNumsubchannels(UL_) : getPhy()->getNumsubchannels(DL_);
	int max_symb = getMaxUlduration() + getMaxDlduration() + 1;

	if (grid_) {
		for (uint8_t i = 0; i < max_symb; i++) delete [] grid_[i];

		delete[] grid_;
	}

	grid_ = new incoming_packet*[max_symb];

	for (uint8_t i = 0; i < max_symb; i++) {
		grid_[i] = new incoming_packet[max_subch];
		for (uint8_t j = 0; j < max_subch; j++) {
			grid_[i][j].data = 0;
			grid_[i][j].collision = false;
			grid_[i][j].received = false;
		}
	}
//	Simulator::ScheduleWithContext(15, Seconds(getFrameDuration()), &Wimax2NetDevice::reset_grid, this);
}

void Wimax2NetDevice::initialize_grid() {
	//for this function to work, we need to know the dimensions of a frame
	//to know that,we need to have obtained a MAC_MIB already
	debug10("At %f in station %s, resetting the grid used to store incoming data for the frame\n", NOW, addr_s());
	int max_subch = getPhy()->getNumsubchannels(UL_) > getPhy()->getNumsubchannels(DL_) ? getPhy()->getNumsubchannels(UL_) : getPhy()->getNumsubchannels(DL_);
	int max_symb = getMaxUlduration() + getMaxDlduration() + 1;

	if (grid_) {
		for (uint8_t i = 0; i < max_symb; i++) delete [] grid_[i];

		delete[] grid_;
	}

	grid_ = new incoming_packet*[max_symb];

	for (uint8_t i = 0; i < max_symb; i++) {
		grid_[i] = new incoming_packet[max_subch];
		for (uint8_t j = 0; j < max_subch; j++) {
			grid_[i][j].data = 0;
			grid_[i][j].collision = false;
			grid_[i][j].received = false;
		}
	}


}

/*
 * Process incoming packet fragments on a single allocation block (1xSymbol Time * 1xSubchannel)
 * This adds data to grid_, which is decoded by scheduled assemble_packet calls
 * @param p The packet object storing the data chunk transmitted in this specific block
 * @param subchannel The logical subchannel over which the block was transmitted
 * @param r_time The (global total) simulation time at which the packet was recieved
 */


void Wimax2NetDevice::sendUp(Ptr<Packet> p, uint8_t subchannel, Time r_time) {
	//determine which symbol-duration column this is in the allocation grid
	uint8_t symbol = floor((r_time - start_of_frame).GetSeconds() / getPhy()->getSymbolTime());
	
	if (symbol < getMaxDlduration() + getMaxUlduration() && subchannel < getPhy()->getNumsubchannels(static_cast<direction>(symbol > getMaxDlduration()))) {
		grid_[symbol][subchannel].received = true;
		if (grid_[symbol][subchannel].data != 0 || grid_[symbol][subchannel].collision) {
			debug10("Colission detected in station %s in allocation grid position (symbol_time = %u, subchannel = %u)\n", addr_s(), symbol, subchannel);
			grid_[symbol][subchannel].collision = true;
		}
		grid_[symbol][subchannel].data = p;
		grid_[symbol][subchannel].received = true;
		debug15("Packet fragment recieved and stored at allocation block (symbol = %d, subchannel = %d)\n", symbol, subchannel);
	} else {
		std::cerr << "ERROR: At " << NOW << " in " << addr_s() << ", tried to sendUp data (PHY->MAC) for an invalid position on the allocation grid (symbol = " << (int)symbol << ", subchannel = " << (int)subchannel << ")\n";
	}
}

/*
	We're going to do the following here:

	1. Ensure that the current time is high enough that the packet should have been completely received, and reschedule for when it will be if necessary
	2. Examine each block in the region specified by the combination of the parameters in the PIH
		2.1 If any of these blocks has not arrived or a collision was detected in that space on the allocation table, give up... we don't have a valid packet reciept
		2.2 For all validly received blocks, add them to a vector of Ptr<Packet>s in the correct order
	3. Assemble these packet fragments into one packet
	4. call process_new_packet on what we get to deal with packing and fragmentation (if it exists)
*/
 // -#- @pih(transfer_ownership=true) -#-
void Wimax2NetDevice::assemble_incoming_packet(PHY_INFO_Header *pih) {
	debug10("At %f, in %s, attempting to assemble a packet that should have been fully received by now\n", NOW, addr_s());
	//determine which symbol-duration column this is in the allocation grid
	uint8_t current_symbol = floor ((Simulator::Now() - start_of_frame).GetSeconds() / getPhy()->getSymbolTime());

	std::vector<Ptr<Packet> > slot_chunks;

	if (pih->direction == DL) {
		debug10("\t packet is in DL Subframe and covers symbols %d to %d, and subchannels %d to %d\n", pih->OFDMSymbol_offset, pih->OFDMSymbol_offset + pih->num_OFDMSymbol - 1, pih->subchannel_offset, pih->subchannel_offset + pih->num_subchannels - 1);
		//first we need to make sure the packet SHOULD all be here and this didn't get called early for some reason
		uint8_t last_symbol = pih->OFDMSymbol_offset + pih->num_OFDMSymbol; //since we start with symbol 0, this is actually one after it, but that makes for faster comparisons below anyway
		if (last_symbol > current_symbol) {
			Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds ( getPhy()->getSymbolTime() * (last_symbol-current_symbol) + getPhy()->getPS()), &Wimax2NetDevice::assemble_incoming_packet, this, pih); 
			return;
		}
		uint8_t last_subchannel = pih->subchannel_offset + pih->num_subchannels;

		//now that we know it should all be here, we're going to iterate through the blocks and ensure that it all arrived (and there weren't any collision issues)
		for (uint8_t i = pih->subchannel_offset; i < last_subchannel; i++)
			for (uint8_t j = pih->OFDMSymbol_offset; j < last_symbol; j++) {
				if (grid_[j][i].collision || ! grid_[j][i].received) {
					debug10("Error assembling packet from DL Subframe.  It should have spanned subchannels %u to %u and OFDM-Symbol-Times %u to %u, but %s\n", pih->subchannel_offset, pih->subchannel_offset + pih->num_subchannels, pih->OFDMSymbol_offset, pih->OFDMSymbol_offset + pih->num_OFDMSymbol, grid_[j][i].collision ? "there was a collision" : "at least one block was not received");
					delete pih;
					return;
				}
				else slot_chunks.push_back(grid_[j][i].data);
			}
	} else { //if we're assembling a packet in the UL Subframe
		debug10("\t packet is in UL Subframe.  It starts at (OFDM Symbol = %d, subchannel = %d) and lasts for %d symbols, wrapping at the end of rows in if necessary\n", pih->OFDMSymbol_offset, pih->subchannel_offset, pih->num_OFDMSymbol);
		//first we need to make sure the packet SHOULD all be here and this didn't get called early for some reason
		bool is_contention = pih->OFDMSymbol_offset == 0; //the first slot is reserved for contention packets
		//this is the last symbol that's actually used
		uint8_t last_symbol = pih->OFDMSymbol_offset + pih->num_OFDMSymbol > static_cast<uint32_t>(getMaxUlduration()) ? getMaxUlduration() : pih->OFDMSymbol_offset + pih->num_OFDMSymbol;
		last_symbol += getMaxDlduration();
		if (last_symbol > current_symbol) {
//			if (is_contention && last_symbol == current_symbol) {} //since contention packets start from 0, these being equal means we're fine
//			else {
				Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds ( getPhy()->getSymbolTime() * (last_symbol - current_symbol + 1) + getPhy()->getPS()), &Wimax2NetDevice::assemble_incoming_packet, this, pih);
				return;
//			}
		}
		
		//now that we know it should all be here, we're going to iterate through the blocks and ensure that it all arrived (and there weren't any collision issues)
		uint32_t on_first_row = pih->OFDMSymbol_offset + getMaxDlduration();;
		uint32_t current_block = 0;
		uint32_t total_symbol_duration = getMaxUlduration() + getMaxDlduration();
		for (uint8_t i = pih->subchannel_offset; i < getPhy()->getNumsubchannels(UL_); i++)
			for (uint8_t j = on_first_row; j < total_symbol_duration; j++) {
				if (is_contention) {
					if (current_block++ == pih->num_OFDMSymbol) goto do_assembly;
					if (grid_[j][i].collision) {
						//collision on contention slots is bad, but non-reception is fine
						//we don't want to print an error message if it's just that nobody transmitted during the slot
						debug10("Collision on contention slot number %d\n", i);
						delete pih;
						return;
					}
					if (grid_[j][i].received) slot_chunks.push_back(grid_[j][i].data);
					else return; //it's not generally an error if we don't have a reception in a contention slot - most of the time they're empty
				} else { //if it's not a contention slot
					on_first_row = getPhy()->getSlotLength(UL_) + getMaxDlduration();
					if (current_block++ == pih->num_OFDMSymbol) goto do_assembly;
					if ( grid_[j][i].collision || ! grid_[j][i].received) {
						debug10("Error assembling packet from UL Subframe.  It should have start at (subchannel = %u, OFDM-symbol-time = %u) and run for %u allocation blocks, but at (symbol = %d, subchannel = %d), %s\n", pih->subchannel_offset, pih->OFDMSymbol_offset, pih->num_OFDMSymbol, j - getMaxDlduration(), i, grid_[j][i].collision ? "there was a collision" : "a block was not received");
						delete pih;
						return;
					}
					else slot_chunks.push_back(grid_[j][i].data);
				}
			}
	}
do_assembly:
	Ptr<Packet> assembled = Create<Packet>();
	All_Headers ah;
	ah.PIH = *pih;

	debug10("At %f in %s, gathered %d blocks to assemble\n", NOW, addr_s(), (int)slot_chunks.size());
	for (uint32_t i = 0; i < slot_chunks.size(); i++) {
		//The follow is a workaround for a bug in NS-3 v. 3.10 with Packet::AddAtEnd
		//it should be removed in favor of that function when it works correctly, probably in ns3.11 or later
		assembled->AddAtEnd(slot_chunks[i]);
//		Copy_Trailer ch(slot_chunks[i]->GetSize());
//		slot_chunks[i]->CopyData(ch.d, ch.length);
//		uint8_t *cp = new uint8_t[assembled->GetSize() + slot_chunks[i]->GetSize()];
//		assembled->CopyData(cp, assembled->GetSize());
//		for (int j = 0; j < ch.length; j++) {
//			cp[assembled->GetSize() + j] = ch.d[j];
//		}
//		assembled = Create<Packet>(cp, assembled->GetSize() + slot_chunks[i]->GetSize());
//		delete [] cp;
//		assembled->AddAtEnd(slot_chunks[i]);
	}
	
	if (assembled->GetSize() < 6) {
		debug10("ERROR Assembling packet.  Too short.  dropping\n");
		delete pih;
		return;
	}

	debug10("At %f in %s, assembled a packet of size %d\n", NOW, addr_s(), assembled->GetSize());
	
	assembled->PeekHeader(ah.GMH);
	assembled->PeekHeader(ah.SMH);
	assembled->RemoveHeader(ah.CRH);
	assembled->AddHeader(ah);
	process_new_packet(assembled);

	delete pih;
}

/*
	1. If that packet utilizes fragmentation or packing, seperate it / cache the fragment appropriately
	2. Send each individual, complete packet into the receive function
*/

void Wimax2NetDevice::process_new_packet(Ptr<Packet> p) {
	All_Headers ah;
	p->RemoveHeader(ah);

	debug10("At %f in %s, processing just-assembled packet of size %u\n", NOW, addr_s(), p->GetSize());

	//first we make sure that the packet is supposed to be coming into this node by checking the CID
	
	int cid = -1;
	if (ah.GMH.GMH.ht == 0) cid = ah.GMH.GMH.cid;
	else {
		if (ah.SMH.BRH.type == 0) cid = ah.SMH.BRH.cid;
		else cid = ah.CRH.data.cid;
	}

	if (getCManager()->get_connection(cid, IN_CONNECTION) == NULL) {
			debug10("At %f, in %s, recieved packet that does not have a valid cid for this node, dropping\n", NOW, addr_s());
			drop(p, "WRONG DEST");
			return;
	}



	//note that when passing up from this function, we can safely assume that any packet pulled out of one 
	//employing packing, or assembled from several using fragmentation, uses a GMH (not a BW-Req or CDMA MAC 
	//Header), as those are all either contention based (so they wouldn't be coming in an allocated, assembled 
	//packet, or can be replaced with the bandwidth request header, so they would not be used.


	//if it's just a signaling header, send it along
	if (ah.GMH.GMH.ht == 1) {
		debug10("At %f in %s, sending a signaling header to subclass processing funtction\n", NOW, addr_s());
		p->AddHeader(ah);
		TagIncomingMACPacket(p);
		receive(p);
		return;
	}


	//piggybacked BW-requests always directly follow the GMH, so if it's on the packet, grab it before worrying about anything else
	if (this->GetObject<Wimax2BSNetDevice>() != 0 && (ah.GMH.GMH.type_fbgm == 1)) {
		p->RemoveHeader(ah.GMSH);
		ah.GMH.GMH.len -= ah.GMSH.GetSerializedSize();
		debug10("\tReceived piggybacked bandwidth request of size %d\n", ah.GMSH.data.piggyback_req);
		getCManager()->get_connection(ah.GMH.GMH.cid, IN_CONNECTION)->setBw(ah.GMSH.data.piggyback_req);
	}

	//if it's a GMH-headed packet, check for fragmentation and packing, and deal with them appropriately
	//send along any complete packets
	//it's just one simple packet, send it along
	if (ah.GMH.GMH.type_frag == 0 && ah.GMH.GMH.type_pck == 0) {
		debug10("At %f in %s, sending a packet without packing or fragmentation to subclass processing function\n", NOW, addr_s());
		p->AddHeader(ah);
		TagIncomingMACPacket(p);
		receive(p);
		return;
	}
	
	//it's just several sdu's packed into one pdu, seperate them and send each along
	else if (ah.GMH.GMH.type_frag == 0 && ah.GMH.GMH.type_pck == 1) {
		debug10("At %f in %s, sending packets with packing but no fragmentation to subclass processing function\n", NOW, addr_s());
		PACKING_SUBHEADER psh;
		Ptr<Packet> extract;
		while (p->GetSize() > psh.GetSerializedSize()) {
			p->RemoveHeader(psh);
			if (psh.data.length == 0) {
				//this isn't another packed in SDU, it's just some extra padded-on data to fill out the allocation
				return;
			}
			if (p->GetSize() < psh.data.length) {
				drop(p, "PACK MISSING");
				return;
			}
			extract = p->Copy();
			extract->RemoveAtEnd(extract->GetSize()- (psh.data.length - psh.GetSerializedSize()));
			p->RemoveAtStart(psh.data.length - psh.GetSerializedSize()); //since we already removed the PSH

			All_Headers ah_e;
			ah_e.GMH.GMH.len = extract->GetSize() + ah.GMH.GetSerializedSize();
			ah_e.GMH.GMH.cid = ah.GMH.GMH.cid;
			ah_e.GMH.GMH.type_frag = 0;
			ah_e.GMH.GMH.type_pck = 0;
			ah_e.GMH.GMH.type_fbgm = 0;
			ah_e.GMH.GMH.ci = ah.GMH.GMH.ci;
			ah_e.GMH.GMH.eks = ah.GMH.GMH.eks;
			ah_e.GMH.GMH.ht = 0;
			ah_e.GMH.GMH.ec = 0;
			extract->AddHeader(ah_e);
			TagIncomingMACPacket(extract);
			receive (extract);
		}
	}
	//it's one fragment, add it to the queue (or assemble a fully recieved fragmented packet
	else if (ah.GMH.GMH.type_frag == 1 && ah.GMH.GMH.type_pck == 0) {
		debug10("At %f in %s, processing recieved packet fragment, sending to subclass processing functions if it's the last one \n", NOW, addr_s());
		Connection * incoming = getCManager()->get_connection(ah.GMH.GMH.cid, false); //get an incoming connection with the given CID
		if (! incoming->isFragEnable()) {
			drop(p, "NO FRAG");
			return;
		}

		p->AddHeader(ah);
		TagIncomingMACPacket(p);
		p->RemoveHeader(ah);

		FRAG_SUBHEADER fsh;
		Ptr<Packet> extract;
		p->RemoveHeader(fsh);
		//note that cases are enclosed in braces to prevent scope errors with variables declared within them
		switch (fsh.data.fc) {

		case FRAG_NOFRAG:
		{
			//strictly speaking this shouldn't happen... if it was going to be FRAG_NOFRAG, it shouldn't have fragmentation enabled
			//TagIncomingMACPacket(extract);
			ah.GMH.GMH.len -= ah.FSH.GetSerializedSize();
			p->AddHeader(ah);
			receive (p);
			return;
		}
		case FRAG_FIRST:
		{
			if (incoming->getFragmentationStatus() != FRAG_NOFRAG) {
				incoming->flush_queue();
			}

			if (incoming->incrementPackNumber() != fsh.data.fsn) {
				//drop(p, "MISSING FRAG");
				//don't actually do anything, since we're starting a new fragmented packet, but ARQ would care about this
			}

			
			All_Headers ah_f;
			ah_f.GMH.GMH.cid = ah.GMH.GMH.cid;
			ah_f.GMH.GMH.type_frag = 1;
			//GMH.len is expected to include the GMH that would be included on an independent packet
			ah_f.GMH.GMH.len = p->GetSize() + ah_f.GMH.GetSerializedSize();
			ah_f.FSH = fsh;
			p->AddHeader(ah_f);
			incoming->enqueue(p);
			incoming->incrementFragNumber();
			incoming->setFragStatus(FRAG_CONT);
			return;

		}
		case FRAG_CONT:
		{
			//if we're either missing a fragment or aren't expecting a middle-fragment, drop it all...
			if (incoming->getFragmentationStatus() != FRAG_CONT || incoming->incrementPackNumber() != fsh.data.fsn) {
				incoming->flush_queue();
				drop (p, "MISSING START FRAG");
				incoming->updateFragmentation(FRAG_NOFRAG, fsh.data.fsn, 0);
				//ARQ would care about this, were it implemented
				return;
			}


			All_Headers ah_f;
			ah_f.GMH.GMH.cid = ah.GMH.GMH.cid;
			ah_f.GMH.GMH.type_frag = 1;
			//GMH.len is expected to include the GMH that would be included on an independent packet
			ah_f.GMH.GMH.len = p->GetSize() + ah_f.GMH.GetSerializedSize();
			ah_f.FSH = fsh;
			p->AddHeader(ah_f);

			incoming->enqueue(p);
			incoming->updateFragmentation(FRAG_CONT, fsh.data.sn, p->GetSize() + incoming->getFragmentBytes());
			return;
		}
		case FRAG_LAST:
		{
			if (incoming->getFragmentationStatus() != FRAG_CONT || incoming->incrementPackNumber() != fsh.data.fsn) {
				incoming->flush_queue();
				drop (p, "MISSING START FRAG");
				incoming->updateFragmentation(FRAG_NOFRAG, fsh.data.fsn, 0);
				//ARQ would care about this, were it implemented
				return;
			}
			
			incoming->enqueue(p);

			Ptr<Packet> q_frag = incoming->dequeue();
			Ptr<Packet> assemble = getPacket();
			All_Headers ah_asmb, ah_part;
			assemble->RemoveHeader(ah_asmb);
			//we're done getting fragments now, so put them together and send it to receive
			while (q_frag != 0 && incoming->queueLength() > 0) {
				q_frag->RemoveHeader(ah_part);
				assemble->AddAtEnd(q_frag);
				q_frag = incoming->dequeue();
			}
			
			//GMH.len is expected to include the GMH that would be included on an independent packet
			ah_asmb.GMH.GMH.len = assemble->GetSerializedSize() + ah_asmb.GetSerializedSize();
			ah_asmb.GMH.GMH.cid = ah.GMH.GMH.cid;
			assemble->AddHeader(ah_asmb);
			TagIncomingMACPacket(assemble);
			incoming->flush_queue();
			incoming->updateFragmentation(FRAG_NOFRAG, incoming->getFragmentNumber(), 0);
			receive (assemble);
			return;

		}
		default:
			debug10("impossible value stored to fc field of Fragmentation subheader... panic\n");
			drop(p, "MAGIC STORAGE");
			return;
		}
	}
	//packed and fragmented.  check each packed sdu and send it to receive or the fragmentation reception queue as appropriate
	else if (ah.GMH.GMH.type_frag == 1 && ah.GMH.GMH.type_pck == 1) {
		debug10("At %f in %s, processing recieved packet that employs both packing and fragmentation\n", NOW, addr_s());
		Connection * incoming = getCManager()->get_connection(ah.GMH.GMH.cid, IN_CONNECTION); //get an incoming connection with the given CID
		if (! (incoming->isFragEnable() && incoming->isPackingEnable())) {
			drop(p, "NO FRAG/PACK");
			return;
		}

		PACKING_SUBHEADER psh;
		Ptr<Packet> extract;
		bool first = true;
		while (p->GetSize() > psh.GetSerializedSize()) {
			p->RemoveHeader(psh);
			if (psh.data.length == 0) {
				//this isn't another packed in SDU, it's just some extra padded-on data to fill out the allocation
				return;
			}
			if (psh.data.fc == FRAG_NOFRAG && p->GetSize() < psh.data.length - psh.GetSerializedSize()) {
				drop(p, "PACK MISSING");
				return;
			}

			extract = p->Copy();
			extract->RemoveAtEnd(extract->GetSize()- (psh.data.length - psh.GetSerializedSize()));
			p->RemoveAtStart(psh.data.length - psh.GetSerializedSize());

			if (psh.data.fc == FRAG_NOFRAG) {
				All_Headers ah_e;
				//GMH.len is expected to include the GMH that would be included on an independent packet
				ah_e.GMH.GMH.len = extract->GetSize() + ah_e.GMH.GetSerializedSize();
				ah_e.GMH.GMH.cid = ah.GMH.GMH.cid;
				ah_e.GMH.GMH.type_frag = 0;
				ah_e.GMH.GMH.type_pck = 0;
				ah_e.GMH.GMH.type_fbgm = 0;
				ah_e.GMH.GMH.ci = ah.GMH.GMH.ci;
				ah_e.GMH.GMH.eks = ah.GMH.GMH.eks;
				ah_e.GMH.GMH.ht = 0;
				ah_e.GMH.GMH.ec = 0;
				extract->AddHeader(ah_e);
				TagIncomingMACPacket(extract);
				debug10("At %f in %s, sending un-packed packet to subclass recieve function\n", NOW ,addr_s());
				receive (extract);
				first = false;
				continue;
			} else if (psh.data.fc == FRAG_FIRST) {
				//this must be the last fragment in the received burst, or it would be merged with the preceeding fragmetns until it is

				Connection * incoming = getCManager()->get_connection(ah.GMH.GMH.cid, false); //get an incoming connection with the given CID
				if (incoming->getFragmentationStatus() != FRAG_NOFRAG) {
					incoming->flush_queue();
				}

				if (incoming->incrementPackNumber() != psh.data.sn) {
					//drop(p, "MISSING FRAG");
					//don't actually do anything, since we're starting a new fragmented packet, but ARQ would care about this
				}


				All_Headers ah_e;
				//GMH.len is expected to include the GMH that would be included on an independent packet
				ah_e.GMH.GMH.len = extract->GetSize() + ah_e.GMH.GetSerializedSize();
				ah_e.GMH.GMH.cid = ah.GMH.GMH.cid;
				ah_e.GMH.GMH.type_frag = 0;
				ah_e.GMH.GMH.type_pck = 0;
				ah_e.GMH.GMH.type_fbgm = 0;
				ah_e.GMH.GMH.ci = ah.GMH.GMH.ci;
				ah_e.GMH.GMH.eks = ah.GMH.GMH.eks;
				ah_e.GMH.GMH.ht = 0;
				ah_e.GMH.GMH.ec = 0;
				debug10("At %f in %s, enqueueing packet fragment of size %u for future re-assembly\n", NOW ,addr_s(), extract->GetSize());
				extract->AddHeader(ah_e);
				TagIncomingMACPacket(extract);
				incoming->enqueue(extract);
				incoming->incrementFragNumber();
				incoming->setFragStatus(FRAG_CONT);
				return;

			} else if (psh.data.fc == FRAG_CONT) {
				//this is invalid.  There is no reason to pack a middle-fragment, as it could be merged with the adjacent fragments
				//until either the entire frame is used by this fragment (so no packing, and a FSH instead of a PSH), or the adjacent
				//packet is not a fragment, so this is either the first or last fragment (depeneding on which side the non-fragment lies)
				drop (extract, "BAD FRAG-PACK");
				first = false;
				continue;

			} else if (psh.data.fc == FRAG_LAST) {
				//this must be the first fragment in the recieved burst, or it would be merged with the subsequent fragment until it is
				if (!first) {
					drop(extract, "BAD FRAG-PACK");
					first = false;
					continue;
				} else {
					Connection * incoming = getCManager()->get_connection(ah.GMH.GMH.cid, false); //get an incoming connection with the given CID

					if (incoming->getFragmentationStatus() != FRAG_CONT || incoming->incrementPackNumber() != psh.data.sn) {
						incoming->flush_queue();
						drop (extract, "MISSING START FRAG");
						incoming->updateFragmentation(FRAG_NOFRAG, psh.data.sn, 0);
						//ARQ would care about this, were it implemented
						return;
					}

					All_Headers ah_e;
					//GMH.len is expected to include the GMH that would be included on an independent packet
					ah_e.GMH.GMH.len = extract->GetSize() + ah_e.GMH.GetSerializedSize();
					ah_e.GMH.GMH.cid = ah.GMH.GMH.cid;
					ah_e.GMH.GMH.type_frag = 0;
					ah_e.GMH.GMH.type_pck = 0;
					ah_e.GMH.GMH.type_fbgm = 0;
					ah_e.GMH.GMH.ci = ah.GMH.GMH.ci;
					ah_e.GMH.GMH.eks = ah.GMH.GMH.eks;
					ah_e.GMH.GMH.ht = 0;
					ah_e.GMH.GMH.ec = 0;
					debug10("At %f in %s, enqueueing packet fragment of size %u for future re-assembly\n", NOW ,addr_s(), extract->GetSize());
					extract->AddHeader(ah_e);
					TagIncomingMACPacket(extract);
					incoming->enqueue(extract);
					incoming->incrementFragNumber();

					Ptr<Packet> q_frag = incoming->dequeue();
					Ptr<Packet> assemble = getPacket();
					All_Headers ah_asmb;
					assemble->RemoveHeader(ah_asmb);
					//we're done getting fragments now, so put them together and send it to receive
					debug10("At %f in %s, assembling packet fragments\n", NOW ,addr_s());
					while (q_frag != 0 && incoming->queueLength() > 0) {
						assemble->AddAtEnd(q_frag);
						q_frag = incoming->dequeue();
					}
					 
					//GMH.len is expected to include the GMH that would be included on an independent packet
					ah_asmb.GMH.GMH.len = assemble->GetSize() + ah_asmb.GMH.GetSerializedSize();
					ah_asmb.GMH.GMH.cid = ah.GMH.GMH.cid;
					ah_asmb.GMH.GMH.type_frag = 0;
					ah_asmb.GMH.GMH.type_pck = 0;
					ah_asmb.GMH.GMH.type_fbgm = 0;
					ah_asmb.GMH.GMH.ci = ah.GMH.GMH.ci;
					ah_asmb.GMH.GMH.eks = ah.GMH.GMH.eks;
					ah_asmb.GMH.GMH.ht = 0;
					ah_asmb.GMH.GMH.ec = 0;
					debug10("At %f in %s, sending reassembled packet of size %u up to subclass revieve function\n", NOW, addr_s(), assemble->GetSize());
					assemble->AddHeader(ah_asmb);
					TagIncomingMACPacket(assemble);
					incoming->flush_queue();
					incoming->updateFragmentation(FRAG_NOFRAG, incoming->getFragmentNumber(), 0);
					receive (assemble);
				}

				first = false;
			}
	}
		//we pull out a packed packet, and if it has a fragmentation type set deal with it
		//note that only the first and last packed packets can legally have a fragmentation type




	}

}


} // namspace ns3
