/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2008 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 */
#ifndef WIMAX2_NET_DEVICE_H
#define WIMAX2_NET_DEVICE_H


#include "connectionmanager.h"
#include "ns3/net-device.h"
#include "ns3/mac48-address.h"
#include "ns3/net-device-container.h"
#include "ns3/node-container.h"
#include <stdint.h>
#include <string>
#include "ns3/traced-callback.h"
#include "ns3/simulator.h"
#include "ns3/ptr.h"
#include "ns3/error-model.h"
#include "ns3/trace-helper.h"
#include "ns3/llc-snap-header.h"

//#include "scheduling/Wimax2Scheduler.h"

//#include "arqstatus.h"
#include "ns3/connection.h"
//#include "connectionmanager.h"
#include "ns3/sduclassifier.h"
#include "ns3/destclassifier.h"
#include "ns3/peernode.h"
//#include "ns3/wimax2-timer.h"
#include "ns3/wimax2-common.h"
#include "ns3/neighbordb.h"
#include "wimax2-phy.h"
#include "ns3/wimax2-serviceflowhandler.h"
//#include "serviceflowqos.h"

#include "802_16headers.h"

namespace  ns3 {

//class Wimax2Phy;


//class Mac802_16MIB;
class Wimax2Channel;
class Wimax2Scheduler;
class Node;
class ErrorModel;
class FrameMap;
class InitTimer;
class DlTimer;
class UlTimer;
class WimaxRxTimer;
struct PacketTimerRecord;

/** Defines different types of nodes */
enum station_type_t {
  STA_UNKNOWN,
  STA_MN,
  STA_BS
};

/** MAC MIB */
class Mac802_16MIB {
 public: 
  Mac802_16MIB ();
 
  int queue_length;
  double frame_duration;

  double dcd_interval;
  double ucd_interval;
  double init_rng_interval;
  double lost_dlmap_interval;
  double lost_ulmap_interval;
  
  double t1_timeout;
  double t2_timeout;
  double t3_timeout;
  double t6_timeout;
  double t12_timeout;
  double t16_timeout;
  double t17_timeout;
  double t21_timeout;
  double t44_timeout;

  uint32_t contention_rng_retry;
  uint32_t invited_rng_retry;
  uint32_t request_retry;
  uint32_t reg_req_retry;
  double    tproc;
  uint32_t dsx_req_retry;
  uint32_t dsx_rsp_retry;

  uint32_t rng_backoff_start;
  uint32_t rng_backoff_stop;
  uint32_t bw_backoff_start;
  uint32_t bw_backoff_stop;

  uint32_t init_contention_size;
  uint32_t bw_req_contention_size;

  uint32_t cdma_code_bw_start;
  uint32_t cdma_code_bw_stop;
  uint32_t cdma_code_init_start;
  uint32_t cdma_code_init_stop;
  uint32_t cdma_code_cqich_start;
  uint32_t cdma_code_cqich_stop;
  uint32_t cdma_code_handover_start;
  uint32_t cdma_code_handover_stop;

  //mobility extension
  uint32_t scan_duration;
  uint32_t interleaving;
  uint32_t scan_iteration;
  uint32_t max_dir_scan_time;
  double    nbr_adv_interval;
  uint32_t scan_req_retry;

  //miscalleous
  double rxp_avg_alpha;  //for measurements
  double lgd_factor_; 
  double RXThreshold_;
  double client_timeout; //used to clear information on BS side
  int ITU_PDP_; // To get value of ITU-PDP model being used

};

/** PHY MIB */
class Phy802_16MIB
 {
 public: 
  Phy802_16MIB ();

//RPI MIMO 
  unsigned char MIMO_enable:1;
  unsigned char num_antenna;
//RPI MIMO
  int channel; //current channel
  double fbandwidth;
  uint32_t ttg;  //gap between downlink burst and subsequent uplink burst
  uint32_t rtg;  //gap between uplink burst and subsequent downlink burst
  uint32_t dl_perm;
  uint32_t ul_perm;
  int disableInterference;
};

//stores the data recieved in a single subchannel * ODFMSymbolDuration
struct incoming_packet {
	Ptr<Packet> data;
	bool collision;
	bool received;
};


/**
 * \ingroup netdevice
 * This device and assumes 48-bit mac addressing;
 */
class Wimax2NetDevice : public NetDevice
{
  friend class Wimax2BSScheduler;
  friend class Wimax2SSScheduler;
  friend class Wimax2Timer;
  friend class StatTimer;
  friend class InitTimer;
  friend class FrameMap;

public:
  Mac802_16MIB* getMacMIB() {return macmib_;}
   // -#- @mib(transfer_ownership=false) -#-
  void setMacMIB(Mac802_16MIB *mib) {macmib_ = mib;}
  Phy802_16MIB getPhyMIB() {return phymib_;}
   // -#- @mib(transfer_ownership=false) -#-
  void setPhyMIB(Phy802_16MIB *mib) {phymib_ = *mib;}
  static TypeId GetTypeId (void);
  Wimax2NetDevice (Time boot);
  Wimax2NetDevice ();
  ~Wimax2NetDevice();

  virtual void Receive (Ptr<Packet> packet, uint16_t protocol, Mac48Address to, Mac48Address from);
  virtual void SetChannel (Ptr<Wimax2Channel> channel);
  virtual void SetReceiveCallback (NetDevice::ReceiveCallback cb);


  /**
   * Attach a receive ErrorModel to the Wimax2NetDevice.
   *
   * The Wimax2NetDevice may optionally include an ErrorModel in
   * the packet receive chain.
   *
   * \see ErrorModel
   * \param em Ptr to the ErrorModel.
   */
  virtual void SetReceiveErrorModel(Ptr<ErrorModel> em);

  // inherited from NetDevice base class.
  virtual void SetIfIndex(const uint32_t index);
  virtual uint32_t GetIfIndex(void) const;
  virtual Ptr<Channel> GetChannel (void) const;
  virtual void SetAddress (Address address);
  virtual Address GetAddress (void) const;
  virtual bool SetMtu (const uint16_t mtu);
  virtual uint16_t GetMtu (void) const;
  virtual bool IsLinkUp (void) const;
  virtual void AddLinkChangeCallback (Callback<void> callback);
  virtual bool IsBroadcast (void) const;
  virtual Address GetBroadcast (void) const;
  virtual bool IsMulticast (void) const;
  virtual Address GetMulticast (Ipv4Address multicastGroup) const;
  virtual bool IsPointToPoint (void) const;
  virtual bool IsBridge (void) const;
  
  /**
   * Enqueues the packet on the correct Connection
   */
  virtual bool Send(Ptr<Packet> packet, const Address& dest, uint16_t protocolNumber);
  virtual bool SendFrom(Ptr<Packet>,const Address &source, const Address &dest, uint16_t protocolNumber) {return false;}
  virtual Ptr<Node> GetNode (void) const;
  virtual void SetNode (Ptr<Node> node);
  virtual bool NeedsArp (void) const;
//  virtual void SetReceiveCallback (NetDevice::ReceiveCallback cb);

  virtual Address GetMulticast (Ipv6Address addr) const;

  virtual void SetPromiscReceiveCallback (PromiscReceiveCallback cb);
  virtual bool SupportsSendFrom (void) const;
  
  
  /**
   * Return a pointer to the Wimax2Channel associated with this net device
   */
  Ptr<Wimax2Channel> GetWimaxChannel(void) const;

  
  /**
   * Return the head of the peer nodes list
   * @return the head of the peer nodes list
   */
  virtual PeerNode * getPeerNode_head () { return peer_list_->lh_first; }

  /**
   * Return the peer node that has the given address
   * @param index The address of the peer
   * @return The peer node that has the given address
   */
  virtual PeerNode *getPeerNode (Mac48Address addr);
  //PeerNode *getPeerNode (int index);

  /**
   * Add the peer node
   * @param The peer node to add
   */
   // -#- @node(transfer_ownership=true) -#-
  virtual void addPeerNode (PeerNode *node);

  /**
   * Remove a peer node
   * @param The peer node to remove
   */
   // -#- @node(transfer_ownership=true) -#-
  virtual void removePeerNode (PeerNode *node);

  /**
   * Return the number of peer nodes
   */
  virtual int getNbPeerNodes ();
  
     /**
    * Add a classifier
    */
   // -#- @SDUClassifier(transfer_ownership=true) -#-
   virtual void addClassifier (SDUClassifier *);

   /**
    * Run the packet through the classifiers
    * to find the proper connection
    * @param p the packet to classify
    */
   virtual int classify (Ptr<Packet> p, const Address& dest);
  
   /**
    * Just pulls the destination address out of an All_Headers
    * assumed to be serialized to the beginning of the packet
    * then calls classify (p, dest)
    */
   virtual int classify (Ptr<Packet> p);

   /**
    * Runs the CID through the declassifiers
    * to find the proper Mac48Address
    */
    virtual Mac48Address declassify(int cid);
   
   
  /**
   * Return the connection manager
   * @return the connection manager
   */
  inline ConnectionManager *  getCManager () { return connectionManager_; }
  
    /**
   * Return the frame duration (in s)
   * @return the frame duration (in s)
   */
  double  getFrameDuration () { return macmib_->frame_duration; }
  
  /**
   * Set the frame duration
   * @param duration The frame duration (in s)
   */
  void  setFrameDuration (double duration) { macmib_->frame_duration = duration; }
    
  /**
   * Return the Scheduler
   * @return the Scheduler
   */
  inline Wimax2Scheduler * getScheduler () { return scheduler_; }

  /**
   * Return the current frame number
   * @return the current frame number
   */
  inline int getFrameNumber () {return frame_number_; }

  /**
   * Change the channel
   * @param channel The new channel
   */
  void setChannel (int channel);

  /**
   * Return the channel index
   * @return The channel
   */
  int getChannel ();

  /**
   * Return the channel number for the given frequency
   * @param freq The frequency
   * @return The channel number of -1 if the frequency does not match
   */
  int getChannel (double freq);

  /**
   * Set the channel to the next from the list
   * Used at initialisation and when loosing signal
   */
  void nextChannel ();
  
  inline void SetPhy(Ptr<Wimax2Phy> phy) {if (phy) phy_ = phy;}
  inline Ptr<Wimax2Phy> getPhy() {return phy_;}
  
  inline Mac48Address addr() {return m_address;}
  //note that if the net devices mac address is changed after the first time this is called, it's not going to be updated
  //if this ever becomes an issue, maybe add a static member that stores the Mac48Address it's representing, and check that
  const char* addr_s();// {return HelperFunctions::getMacAddrString(m_address).c_str();}
  
  
     /**
    * The MAC MIB
    */
   Mac802_16MIB * macmib_;
   
   
   /**
    * The Physical layer MIB
    */
   Phy802_16MIB phymib_;
   
   
  /**
   *  Set the total dl duration left after preamble for dl timer 
   */
   inline void setMaxDlduration (int maxdlduration){ MaxDlduration_ = maxdlduration; }

  /*
   *  Get the dl duration in OFDM symbols
   */
   inline int getMaxDlduration() { return MaxDlduration_;}

  /**
   *  Set the total dl duration left after preamble for dl timer 
   */
   inline void setMaxUlduration (int maxulduration){ MaxUlduration_ = maxulduration; }

  /*
   *  Get the ul duration in OFDM symbols
   */
    inline int getMaxUlduration() { return MaxUlduration_;} 

  /**
   *  Set the total dl duration left after preamble for dl timer 
   */
//    inline void setStartUlduration (int ulduration){ Ulduration_ = ulduration; }

  /*
   *  Get the dl duration 
   */
 //   inline int getStartUlduration() { return Ulduration_;}
    
  /**
   * Return the type of MAC
   * @return the type of node
   */
  station_type_t getNodeType();
  
    /**
   * Return the MAP of the current frame
   * @return the MAP of the current frame
   */
  FrameMap *getMap () { return map_;}
  
  inline uint32_t getIdIndex() {return id_index;}
  
  
   /**
    * Return a new allocated packet
    * @return A newly allocated packet 
    */
   Ptr<Packet> getPacket();

   virtual uint8_t get_diuc();
   
   /**
   * Called when a timer expires
   * @param The timer ID
   */
  virtual void expire (timer_id id);
  
 /**
   * Process the packet after receiving last bit
   * @param p the packet to be received
   */
  virtual void receive(Ptr<Packet> p);
  
 /**
   * Start a new DL subframe
   */
  virtual void start_dlsubframe ();
  
  /**
   * Start a new UL subframe
   */
  virtual void start_ulsubframe ();


  /**
   * rpi - remove a packet from the record. 
   */
  void removePacket(PacketTimerRecord*);
  
    /**
   * Process packets going out
   * @param p The packet to transmit
   */
  virtual void transmit(Ptr<Packet> p);
  
  /**
  * rpi- record a packet 
  */
  void addPacket( Ptr<Packet> p); 
  
    /**
   * Set the variable used to find out if upper layers
   * must be notified to send packets. During scanning we
   * do not want upper layers to send packet to the mac.
   * @param notify Value indicating if we want to receive packets 
   * from upper layers
   */
  void setNotify_upper (bool notify);
  
    /**
   * Process packets going out
   * @param p The packet to transmit
   */
  virtual void sendDown(Ptr<Packet> p);
  
  virtual void drop(Ptr<Packet> p, std::string reason);
  
  inline bool isLinkUp() {return accepting_;}
  
  
 /* 
  *   get the ITU_PDP model 
  */ 
    inline int getITU_PDP () { return macmib_->ITU_PDP_; } // 1 - PEDA, 2 - PEDB, 3 -VEHIC
  /**
   * Return The Service Flow handler
   * @return The Service Flow handler
   */
  inline ServiceFlowHandler *  getServiceHandler () { return serviceFlowHandler_; }
  
  /**
   * Set the dlratio for the scheduler
   * @param rat The new ratio
   */
  void SetSchedulerDlRatio(double rat);
  
   /**
   * Determines if the packet payload should be processed by the MAC layer or sent up
   * @param p - the packet in question
   * Use this function to check a tagged packet
   */
  bool isMACPacket(Ptr<Packet> p);
  
   /**
   * Determines if the packet payload should be processed by the MAC layer or sent up
   * @param p - the packet in question
   * Use this function to add a tag to an untagged packet
   */
  bool TagIncomingMACPacket(Ptr<Packet> p);
  
    /**
   * Process incoming packets 
   * @param p The received packet
   */
  virtual void sendUp(Ptr<Packet> p, uint8_t subchannel, Time r_time);
  
  inline DestClassifier* d_classifier() {return reinterpret_cast<DestClassifier*>(classifier_list_.lh_first);}

/*
 * Used to calculate the actual transmitted length of a packet
 * @param p The packet to be transmitted
 * @param uplink true if the packet is being transmitted in the UL Subframe, false otherwise
 */
  int GetActualLength(Ptr<Packet> p, bool uplink);

  //called between frames (during the preamble) to sync all stations for the start of a new frame
  void SynchronizeFrameStart(Time start);


  //examine the receieved OFDMA allocation grid blocks to assemble a packet based on the region of the grid
  //defined by this PHY_INFO_Header
   virtual void assemble_incoming_packet(PHY_INFO_Header *pih);

//protected:	note from here down should be protected instead of public, but 
//				the python bindings generator cannot process protected member variables
//				this should be changed eventually, and until then 
  uint32_t id_index;
  Ptr<Wimax2Phy> phy_;
  Ptr<Wimax2Channel> m_channel;
  NetDevice::ReceiveCallback m_rxCallback;
  NetDevice::PromiscReceiveCallback m_promiscCallback;
  Ptr<Node> m_node;
  uint16_t m_mtu;
  uint32_t m_ifIndex;
  Mac48Address m_address;
  bool accepting_; //true if mac is accepting packets, false otherwise, accesed through isLinkUp
  Ptr<ErrorModel> m_receiveErrorModel;
  uint8_t nohead_; //used to count bursts recieved at the start of a DL Subframe
  Time start_of_frame;
  

    /**
   * Timer to init the MAC
   */
  InitTimer *initTimer_;
  
     /**
    * Database of neighboring BS
    */
   NeighborDB *nbr_db_;
  
  
  /**
   * The trace source fired when the phy layer drops a packet it has received
   * due to the error model being active.
   * \see class CallBackTraceSource
   * - now declared in each subclass because inheritance didn't agree with the ns3 tracing system -
   */
  //TracedCallback<Ptr<const Packet>, Mac48Address& > m_traceRx;
  //TracedCallback<Ptr<const Packet>, Mac48Address& > m_traceTx;


  /**
   * The following are used to generate trace files
   */
   TracedCallback<Ptr<const Packet> > m_macTxTrace;
   TracedCallback<Ptr<const Packet> > m_macTxDropTrace;
   TracedCallback<Ptr<const Packet> > m_macPromixcRxTrace;
   TracedCallback<Ptr<const Packet> > m_macRxTrace;
   TracedCallback<Ptr<const Packet> > m_macRxDropTrace;
   TracedCallback<Ptr<const Packet> > m_phyTxBeginTrace;
   TracedCallback<Ptr<const Packet> > m_phyTxEndTrace;
   TracedCallback<Ptr<const Packet> > m_phyTxDropTrace;
   TracedCallback<Ptr<const Packet> > m_phyRxBeginTrace;
   TracedCallback<Ptr<const Packet> > m_phyRxEndTrace;
   TracedCallback<Ptr<const Packet> > m_phyRxDropTrace;
   TracedCallback<Ptr<const Packet> > m_snifferTrace;
   TracedCallback<Ptr<const Packet> > m_promixcSnifferTrace;

  
     /**
    * The module that handles flow requests
    */
   ServiceFlowHandler * serviceFlowHandler_;
  
     /**
    * Loss rate for data connections
    * Allows to test loss on data and not management
    */
   double data_loss_;
  
  //Begin RPI
   /**
    * Allows ARQ_BLOCK_SIZE in data connection  
    */
   int arq_block_size_ ;
//End RPI


   /**
    * Last time a packet was sent
    */
   double last_tx_time_;

   /**
    * Last transmission duration
    */
   double last_tx_duration_;

   /*
    * For collision  detection of contention based packet
    * as they are treated differently here. (OFDM based cont slots) 
    */
   bool contPktRxing_;   
   
   
   /**
    * Indicates if a collision occured
    */
   bool collision_;

  
     /**
    * A packet buffer used to temporary store a packet 
    * received by upper layer. Used during scanning
    */
   Ptr<Packet> pktBuf_;

  
  
   /**
     * Record the number of SS.
     */
     int reg_SS_number;
     
        
   /**
    * Timer used to mark the begining of downlink subframe (i.e new frame)
    */
   DlTimer *dl_timer_;
   
   /**
    * Timer used to mark the begining of uplink subframe
    */
   UlTimer *ul_timer_;

   /**
    * Delay for last packet
    */
   double last_tx_delay_;


   /**
    * Array to store power per subcarrier, to compute interference
    */
   double **intpower_;
   double **basepower_;
   
  /**
    * Indicate if upper layer must be notified to send more packets
    */
   bool notify_upper_;
  
    /**
    * List of connected peer nodes. Only one for SSs.
    */
   struct peerNode_list *peer_list_;

   /**
    * Number of peer in the list
    */
   int nb_peer_;

   /**
    * The list of classifiers
    */
   struct sduClassifier_list classifier_list_;
   
   /**
    * The class to handle connections
    */
   ConnectionManager * connectionManager_;
   
 
   /**
    * Current frame number
    */
   int frame_number_;

   /**
    * The packet scheduler
    */
   Wimax2Scheduler * scheduler_;
   
   
   
   /**rpi
    * dlduration left after preamble in num OFDM  symbols
    */
   int MaxDlduration_; 

  
   /**rpi
    * ulduration left after preamble in num OFDM  symbols
    */
   int MaxUlduration_; 

  
   /**rpi
    * Ulduration left after intial ranging and all in num OFDM  symbols
    */
   //int Ulduration_;

   /**rpi 
   * dlduration left after preamble in num OFDM symbols
   */ 
   int Dlduration_; 
   
   /**
    * Defines BS vs SS
    */
   station_type_t type_;
  
  	/** 
    * The map of the frame
    */
   FrameMap *map_;
   
  /**
  * rpi packet record to keep a record of packets and their timers when multiple packets are recieved in OFDMA
  */
   PacketTimerRecord *head_pkt_;
   PacketTimerRecord *tail_pkt_;
   PacketTimerRecord *trash_pkt_;
   

   
      /**
    * Allows sending ARQFB in data connection  
    */
   int arqfb_in_dl_data_;
 
   /**
    * Allows sending ARQFB in data connection  
    */
   int arqfb_in_ul_data_;
   
   
   /**
     * Rec delta time
     */
     double recv_delta;
     double r_delta_bs;
     double r_delta_ss;
     
     
protected:
	
   /*
    * Return the code for the frame duration
    * @return the code for the frame duration
    */
   int getFrameDurationCode ();
   
   /**
    * Set the frame duration using code
    * @param code The frame duration code
    */
   void setFrameDurationCode (int code);
   
      

 	/**
    * Init the MAC
    */
   virtual void init ();
   
   virtual void DoDispose (void);


   //after assembling, do things like defragmentation and unpacking
   virtual void process_new_packet(Ptr<Packet> p);

   //these functions are used to interact with the grid of incoming packets
   incoming_packet get_grid_element(uint8_t x, uint8_t y);
   bool set_grid_element(uint8_t x, uint8_t y, incoming_packet inc_p);
   void reset_grid();
   void initialize_grid();

private:

//accessed by addr_s() - stores a string representation of the Mac48Address for this NetDevice
//which is needed for printf output (debug2 and debug10 are #defined as this)
char *mac_48_address_string;

//stores received data for this frame
incoming_packet **grid_;

};






// class Wimax2BSNetDevice : public Wimax2NetDevice 
// {
// public:
// 	Wimax2BSNetDevice() {}
// 	
// 	static TypeId GetTypeId(void) {
// 		static TypeId tid = TypeId("ns3::Wimax2BSNetDevice");
// 		return tid;
// 	}
// 
// 
// };

// class Wimax2SSNetDevice : public Wimax2NetDevice 
// {
// public:
// 	Wimax2SSNetDevice() {}
// 	
// 	static TypeId GetTypeId(void) {
// 		static TypeId tid = TypeId("ns3::Wimax2SSNetDevice");
// 		return tid;
// 	}
// 
// 
// };




} // namespace ns3

#endif /* WIMAX2_NET_DEVICE_H */
