/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  Richard Rouil
 * @modified by Chakchai So-In
 * @modified by Xingting Guo
 */

#include "wimax2-scheduler.h"
#include "ns3/ptr.h"
#include <iostream>

#define HDR_PIG 2
#define BW_REQ_TIMEOUT 2
#define INIT_REQ_TIMEOUT 2
#define UL_HORIZONTAL 1
//#define IF_PIG

namespace ns3 {

using std::flush;
/*
 * Create a scheduler
 * @param mac The Mac where it is located
 */
Wimax2Scheduler::Wimax2Scheduler (){
	dlratio_ = 0.3; //should really be getting from macmib
}

/*
 * Set the mac
 * @param mac The Mac where it is located
 */
void Wimax2Scheduler::setMac (Ptr<Wimax2NetDevice> mac)
{
  assert (mac!=NULL);

  mac_ = mac;
}

/**
 * Initialize the scheduler.
 */
void Wimax2Scheduler::init()
{
  // initializing dl and ul duration  --rpi
  Ptr<Wimax2Phy> phy = mac_->getPhy();

  int nbPS = (int) floor((mac_->getFrameDuration()/phy->getPS()));
#ifdef DEBUG_WIMAX
  assert (nbPS*phy->getPS()<=mac_->getFrameDuration()); //check for rounding errors
#endif
  int nbPS_left = nbPS - mac_->phymib_.rtg - mac_->phymib_.ttg;
  int nbSymbols = (int) floor((phy->getPS()*nbPS_left)/phy->getSymbolTime());  // max num of OFDM symbols available per frame. 
  assert (nbSymbols*phy->getSymbolTime()+(mac_->phymib_.rtg + mac_->phymib_.ttg)*phy->getPS() < mac_->getFrameDuration());
  int maxdlduration = (int) (nbSymbols * (1.0 - dlratio_)); //number of symbols for downlink
  int maxulduration = nbSymbols - maxdlduration;            //number of symbols for uplink

  mac_->setMaxDlduration (maxdlduration);
  mac_->setMaxUlduration (maxulduration);
  
  //debug(" in wimax scheduler maxdlduration = %d maxulduration =%d mac =%d \n", maxdlduration, maxulduration, mac_->addr());
}

/**
 * This function is used to schedule bursts/packets
 */
void Wimax2Scheduler::schedule ()
{
  //defined by subclasses
}


/**
 * Transfer the packets from the given connection to the given burst
 * @param con The connection
 * @param b The burst
 * @param b_data Amount of data in the burst
 * @return the new burst occupation
 */
//int Wimax2Scheduler::transfer_packets (Connection *c, Burst *b, int b_data, int subchannel_offset, int symbol_offset)
//{
//  debug10("entering transfer_packets (Connection*, Burst* int, int, int)\n");
//  Ptr<Packet> p;
//  All_Headers ch;
//  Generic_MAC_Header * gmh = &ch.GMH;  //to make some old code functional
//  double txtime, txtime2, txtime3;
//  int txtime_s;
//  bool pkt_transfered = false;
//  Ptr<Wimax2Phy> phy = mac_->getPhy();
//  //int offset = b->getStarttime( );
//  PeerNode *peer;
//
//  debug2 (" entering transfer packets");// , bdata = %d  mac_ = %d \n ",b_data, mac_);
//  debug10 ("\tEntered transfer_packets MacADDR :%s, with bdata  :%d\n ", HelperFunctions::getMacAddrString(mac_->addr()).c_str(), b_data);
//  peer = mac_->getPeerNode_head();
//
//  p = c->get_queue()->head();
//  p->PeekHeader(ch);
//
//  uint32_t max_data;
//  if (mac_->getNodeType()==STA_BS) 
//    { 
//      max_data = phy->getMaxPktSize (b->getDuration(), mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding())-b_data;
//
//      debug10 ("\tBS_transfer_packets, CID :%d with bdata :%d, but MaxData (MaxSize-bdata) :%d, MaxSize :%d\n", c->get_cid(), b_data, max_data, phy->getMaxPktSize (b->getDuration(), mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding()) );
//      debug10 ("\t                     Bduration :%d, DIUC :%d\n",b->getDuration(), b->getIUC());
//
//    } 
//  else 
//    {
//      max_data = phy->getMaxPktSize (b->getDuration(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding())-b_data;
//
//      debug10 ("\tSS_transfer_packets, CID :%d with bdata :%d, but MaxData (MaxSize-bdata) :%d, MaxSize :%d\n", c->get_cid(), b_data, max_data, phy->getMaxPktSize (b->getDuration(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding()) );
//      debug10 ("\t                     Bduration :%d, UIUC :%d\n",b->getDuration(), b->getIUC());
//    }
//
//  if (max_data < gmh->GetSerializedSize() ||
//      (c->getFragmentationStatus()!=FRAG_NOFRAG && max_data < gmh->GetSerializedSize()+ch.FSH.GetSerializedSize()))
//    return b_data; //not even space for header
//
//  while (PeekPointer(p)) {   
//    p->RemoveHeader(ch);
//    gmh = &ch.GMH;
//
//
//    if (mac_->getNodeType()==STA_BS) 
//    	max_data = phy->getMaxPktSize (b->getDuration(), mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding())-b_data;
//    else 
//    	max_data = phy->getMaxPktSize (b->getDuration(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding())-b_data;
//
//    debug2 ("\tIn Loop MacADDR :%s, MaxDATA :%d (burst duration :%d, b_data :%d)\n", HelperFunctions::getMacAddrString(mac_->addr()).c_str(), max_data, b->getDuration(), b_data);
//    if (max_data < gmh->GetSerializedSize() ||
//        (c->getFragmentationStatus()!=FRAG_NOFRAG && max_data < gmh->GetSerializedSize()+ch.FSH.GetSerializedSize()))
//      return b_data; //not even space for header
//
//	//if the packet is already fragmented
//    if (c->getFragmentationStatus()!=FRAG_NOFRAG) {
//      //if we can fit the new data in the extant fragment
//      if (max_data >= ch.GMH.GMH.len-c->getFragmentBytes()+ch.FSH.GetSerializedSize()) {
//        //add fragmentation header
//        gmh->GMH.type_frag = true;
//        //no need to fragment again
//        ch.FSH.data.fc = FRAG_LAST;
//        ch.FSH.data.fsn = c->getFragmentNumber ();
//        c->dequeue();  /*remove packet from queue */
//
//		ch.GMH.GMH.len = ch.GMH.GMH.len-c->getFragmentBytes()+ch.FSH.GetSerializedSize(); //new packet size
//        //update fragmentation
//       debug2 ("\tEnd of fragmentation (FRAG :%x), FSN :%d (max_data :%d, bytes to send :%d\n", FRAG_LAST, ch.FSH.data.fsn, max_data, ch.GMH.GMH.len);
//
//        c->updateFragmentation (FRAG_NOFRAG, 0, 0);
//      } 
//      //we need to fragment the packet again
//      else { 
//        p = p->Copy(); //copy packet to send
//        p->PeekHeader(ch);
//        gmh = &ch.GMH;
//        //add fragmentation header
//        gmh->setFrag(true);
//        ch.FSH.data.fc = FRAG_CONT;
//        ch.FSH.data.fsn = c->getFragmentNumber ();
//        gmh->setPduLen(max_data); //new packet size
//        //update fragmentation
//        c->updateFragmentation (FRAG_CONT, (c->getFragmentNumber ()+1)%8, c->getFragmentBytes()+max_data-(gmh->GetSerializedSize()+ch.FSH.GetSerializedSize()));
//        debug2 ("\tContinue fragmentation (FRAG :%x), FSN :%d\n", FRAG_CONT, ch.FSH.data.fsn);
//      }
//    } 
//    else //if the connection's frag status is FRAG_NOFRAG (the packet is unfragmented at the moment)
//    {
//	if (max_data < ch.GMH.GMH.len && c->isFragEnable()) {
//	  //need to fragment the packet for the first time
//	  p = p->Copy(); //copy packet to send
//	  p->PeekHeader(ch);
//	  gmh = &(ch.GMH);
//	  //add fragmentation header
//	  gmh->setFrag(true);
//	  ch.FSH.data.fc = FRAG_FIRST;
//	  ch.FSH.data.fsn = c->getFragmentNumber();
//	  gmh->setPduLen(max_data); //new packet size
//	  //update fragmentation
//	  c->updateFragmentation (FRAG_FIRST, 1, c->getFragmentBytes()+max_data-(gmh->GetSerializedSize()+ch.FSH.GetSerializedSize()));
//	  debug2 ("\tFirst fragmentation (FRAG :%x), FSN :%d\n", FRAG_FIRST, c->getFragmentNumber());
//
//	 } else if (max_data < ch.GMH.GMH.len && !c->isFragEnable()) {
//	  //we need to fragment, but can't.  return without transfer
//	  debug10("because connection does not support fragmentation, cannot move all requested data into it.  Returning\n");
//	  return b_data;
//	 } else { //we have room in the burst for the whole packet, so dequeue and continue
//	  c->dequeue();
//	 }
//    }
//
//
//	//calculate the transmission time based on the direction, amount of data, and the connection
//    if (mac_->getNodeType()==STA_BS) 
//      {
//		txtime = phy->getTrxTime (ch.GMH.GMH.len, mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding());
//		txtime2 = phy->getTrxSymbolTime (b_data+ch.GMH.GMH.len, mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding());
//		txtime3 = phy->getTrxSymbolTime (ch.GMH.GMH.len, mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding());
//      } else {
//		txtime = phy->getTrxTime (ch.GMH.GMH.len, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding());
//		txtime2 = phy->getTrxSymbolTime (b_data+ch.GMH.GMH.len, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding());
//		txtime3 = phy->getTrxSymbolTime (ch.GMH.GMH.len, mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding());
//      }
//    ch.txtime = txtime3;
//    txtime_s = (int)round (txtime3/phy->getSymbolTime ()); //in units of symbol
//    debug2 ("symbtime=%f\n", phy->getSymbolTime ());
//    debug2 ("\tCheck packet to send size :%d txtime :%f(%d) duration :%d(%f)\n", ch.GMH.GMH.len, txtime, txtime_s, b->getDuration(), b->getDuration()*phy->getSymbolTime ());
//    assert ( txtime2 <= b->getDuration()*phy->getSymbolTime () );
//    debug2 ("\tTransfert to burst (txtime :%f, txtime2 :%f, bduration :%f, txtime3 :%f)\n", txtime,txtime2,b->getDuration()*phy->getSymbolTime (),txtime3);
//
//
//    int total_DL_num_subchannel = phy->getNumsubchannels(DL_);
//    Ofdm_mod_rate dlul_map_mod = mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding();
//    int num_of_slots = (int) ceil(ch.GMH.GMH.len/phy->getSlotCapacity(dlul_map_mod,DL_)); //get the slots	needs.
//    int num_of_subchannel = num_of_slots*2;
//    int num_of_symbol = (int) ceil((double)(subchannel_offset + num_of_subchannel)/total_DL_num_subchannel);		
//
//
//    //wimaxHdr = HDR_MAC802_16(p);
//    //if(ch)
//    //  {
//		ch.PIH.num_subchannels = num_of_subchannel;
//		ch.PIH.subchannel_offset = subchannel_offset;
//		ch.PIH.num_OFDMSymbol = num_of_symbol;
//		ch.PIH.OFDMSymbol_offset = symbol_offset;//(int)round ((phy->getTrxSymbolTime (b_data, mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding()))/phy->getSymbolTime ());//offset;
//		ch.PIH.modulation_ = dlul_map_mod;
//	
//	
//		// if(c->getPeerNode())
//		//wimaxHdr->phy_info.channel_index = c->getPeerNode()->getchannel();
//		ch.PIH.channel_index = 1; 
//		if (mac_->getNodeType()==STA_BS)
//		  ch.PIH.direction = 0;
//		else 
//		  ch.PIH.direction = 1;
//   //   } 
//		
//    symbol_offset += num_of_symbol-1;
//    //symbol_offset+= txtime_s;
//
//    //p = c->dequeue();   //dequeue connection queue
//    debug10("enqueue into the burst for cid [%d].\n",c->get_cid());
//    debug10("New packet in [%d] CID burst ||| sym_#[%d]\t sym_off[%d]\t subch_#[%d]\t subch_off[%d]\n",
//	    c->get_cid(), ch.PIH.num_OFDMSymbol, ch.PIH.OFDMSymbol_offset,ch.PIH.num_subchannels,
//	    ch.PIH.subchannel_offset);
//	p->AddHeader(ch);
//    b->enqueue(p);         //enqueue into burst
//    b_data += ch.GMH.GMH.len; //increment amount of data enqueued
//    if (!pkt_transfered && mac_->getNodeType()!=STA_BS){ //if we transfer at least one packet, remove bw request
//      pkt_transfered = true;
//      mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (c->get_cid());
//    }
//    p = c->get_queue()->head(); //get new head
//  }
//  return b_data;
//
//} 
//
///**
// * Transfer the packets from the given connection to the given burst
// * @param con The connection
// * @param b The burst
// * @param b_data Amount of data in the burst
// * @return the new burst occupation
// */
//int Wimax2Scheduler::transfer_packets1 (Connection *c, Burst *b, int b_data)
//{
//	debug10("entering transfer_packets1(Connection*, Burst*, int)\n");
//	Ptr<Packet> testing_death;
//	if (c->queueLength() > 0) {
//		testing_death = c->dequeue();
//		c->enqueue_head(testing_death);
//	}
//	
//  Ptr<Packet> p = mac_->getPacket();
//  All_Headers ch, th;
//  p->RemoveHeader(ch);
//  Generic_MAC_Header *wimaxHdr;
//  //double txtime;
//  int txtime_s;
//  bool pkt_transfered = false; 
//  Ptr<Wimax2Phy> phy = mac_->getPhy();
//  int offset = b->getStarttime( );
//  PeerNode *peer;
//  int numsubchannels,subchannel_offset; 
//  int initial_offset = b->getStarttime( );
//  int initial_subchannel_offset = b->getSubchannelOffset ();
//  int bw_req_packet = ch.GMH.GetSerializedSize();
//  	if (bw_req_packet); //quiet, compiler
//  int flag_bw_req = 0;
//  int flag_pig = 0;
//  int old_subchannel_offset = 0;
//
//  int num_symbol_per_slot;
//
//  subchannel_offset = initial_subchannel_offset;
//  old_subchannel_offset = subchannel_offset;
//
//  if (mac_->getNodeType()==STA_BS) {
//    num_symbol_per_slot = phy->getSlotLength (DL_); 
//  } else {
//    num_symbol_per_slot = phy->getSlotLength (UL_); 
//  }
//  
//
//  debug10 ("\tEntered transfer_packets1 for MacADDR :%s, for connection :%d with bdata :%d, bytes in connection %d, #subchannels :%d, #symbol :%d, IUC :%d\n", HelperFunctions::getMacAddrString(mac_->addr()).c_str(), c->get_cid(), b_data, c->queueByteLength(), b->getnumSubchannels(), b->getDuration(), b->getIUC());
//
////Start sending registration message
//  printf ("TOTO: check init ranging %d \n", c->getINIT_REQ_QUEUE(mac_->getIdIndex()));
//  if ( (c->get_cid() == 0) && (c->getINIT_REQ_QUEUE(mac_->getIdIndex()) == 0) && (mac_->getNodeType()!=STA_BS) ) {
//  
//
////    Ptr<Packet> p1 = mac_->getPacket();
////    p1->RemoveHeader(ch);
////    
////    wimaxHdr = &(ch.GMH);
////    wimaxHdr->GMH.cid = INITIAL_RANGING_CID;
////
////
////    int max_data = phy->getMaxPktSize (b->getnumSubchannels(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_) - b_data;    
////    
////    RANGING_REQUEST_FRAME rrf;
////    
////
////    //p1->allocdata (sizeof (struct mac802_16_rng_req_frame));
////    mac802_16_rng_req_frame *frame = &(rrf.data);
////    frame->type = MAC_RNG_REQ;
////    frame->ss_mac_address = mac_->addr();
////
////    wimaxHdr->GMH.type_frag = 0;
////    wimaxHdr->GMH.type_fbgm = 0;
////    //frame->req_dl_burst_profile = DIUC_PROFILE_7 & 0xF; //we use lower bits only
////    frame->req_dl_burst_profile = mac_->get_diuc() & 0xF; //we use lower bits only
////
////    ch.GMH.GMH.len = RNG_REQ_SIZE;
////
////    debug10 ("\t   Sending INIT-RNG-MSG-REQ, cid :%d, ssid :%s, MaxSize :%d, msg-size :%d\n", c->get_cid(), HelperFunctions::getMacAddrString(mac_->addr()).c_str(), max_data, ch.GMH.GMH.len);
////    max_data++; max_data--; // to stop compiler from complaining that variable is unused
////
////    int assigned_subchannels;
////    int symbol_offset;
////
////    assigned_subchannels = phy->getNumSubchannels(b_data, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
////    numsubchannels = phy->getNumSubchannels(ch.GMH.GMH.len, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
////
////    //Ofdm_mod_rate dlul_map_mod = mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding();
////    //int num_of_slots = (int) ceil((double)ch.GMH.GMH.len/phy->getSlotCapacity(dlul_map_mod,UL_)); //get the slots needs.
////    symbol_offset = b->getStarttime() + num_symbol_per_slot * ((b->getSubchannelOffset()+assigned_subchannels)/ phy->getNumsubchannels(UL_));
////    txtime_s = (int)(ceil((double)(numsubchannels + subchannel_offset) / (double)phy->getNumsubchannels(UL_)))*num_symbol_per_slot;
////    offset = b->getStarttime( )+ txtime_s-num_symbol_per_slot;
////    subchannel_offset = (int) ceil((numsubchannels + subchannel_offset) % (phy->getNumsubchannels(UL_)));
////
////	
////    ch.txtime = txtime_s*phy->getSymbolTime ();
////
////    debug10 ("\t   Initial_Ranging_Msg, In station %d, Check packet to send: con id in BURST :%d, con id in connection :%d, size :%d, txtime :%f(%d), duration :%d(%f), numsubchnn :%d, subchnoff :%d\n",mac_->getNodeType(), c->get_cid(),b->getCid(), ch.GMH.GMH.len, ch.txtime, txtime_s, b->getDuration(), b->getDuration()*phy->getSymbolTime (),b->getnumSubchannels(),b->getSubchannelOffset ());
////
////    ch.PIH.num_subchannels = numsubchannels;
////    ch.PIH.subchannel_offset = initial_subchannel_offset;//subchannel_offset;
////    ch.PIH.num_OFDMSymbol = txtime_s;
////    ch.PIH.OFDMSymbol_offset = symbol_offset; //initial_offset;
////    ch.PIH.channel_index = (mac_->getPeerNode_head())->getchannel();
////    ch.PIH.direction = 1;
////
////    debug10 ("\t   Initial_Ranging_Msg, In station %d, Packet phy header: numsymbols :%d, symboloffset :%d, subchanneloffset :%d, numsubchannel :%d, Channel_num :%d, direction :%d\n", mac_->getNodeType(), txtime_s,initial_offset, initial_subchannel_offset,numsubchannels,ch.PIH.channel_index,ch.PIH.direction);
////
////    initial_offset = offset;
////    initial_subchannel_offset = subchannel_offset;
////
////    p1->AddHeader(rrf);
////    p1->AddHeader(ch);
////    
////	debug10("Enqueueing Ranging Request Frame (RRF) to burst\n");
////	
////    b->enqueue(p1);      //enqueue into burst
////    b_data = b_data+ch.GMH.GMH.len;
////    c->setINIT_REQ_QUEUE(mac_->addr(), INIT_REQ_TIMEOUT);
//    int t6time = (ceil)((double)(mac_->macmib_->t6_timeout)/(double)(mac_->macmib_->frame_duration));
////    subchannel_offset = old_subchannel_offset;
//
//    if ( mac_->getMap()->getUlSubframe()->getRanging()->getRequest_mac (mac_->addr()) != NULL ) {
//       mac_->getMap()->getUlSubframe()->getRanging()->setTIMEOUT (mac_->addr(), t6time);
//       debug10("\t   Reset CDMA_INIT_REQ timer to #frame :%d (%f sec) since send INIT-RNG-MSGQ out for cid :%d, ssid :%s\n", t6time, mac_->macmib_->t6_timeout, c->get_cid(), HelperFunctions::getMacAddrString(mac_->addr()).c_str());
//       c->setINIT_REQ_QUEUE(mac_->getIdIndex(), t6time);
//    }
//    if (mac_->getNodeType()!=STA_BS) {
//       //if there exists a ranging request to this device's MAC Address, we're going to remove it
//       if ( mac_->getMap()->getUlSubframe()->getRanging()->getRequest_mac (mac_->addr()) != NULL ) {
//          Ptr<Packet> p_tmp(NULL);// = NULL;
//          FrameMap *map1 = mac_->getMap();
//          Burst *b1;
//
//		  //if we have a pointer to a packet containing a ranging request to this device's MAC Address
//          if (PeekPointer((p_tmp = mac_->getMap()->getUlSubframe()->getRanging()->getPacket_P_mac(mac_->addr()))) != NULL) {
//             int flag_transmit = mac_->getMap()->getUlSubframe()->getRanging()->getFLAGNOWTRANSMIT(mac_->addr());
//
//             debug10 ("Search flag_transmit :%d; remove cdma_bw_req if flag == 1, nbpdu :%d\n", flag_transmit, map1->getUlSubframe()->getNbPdu ());
//        	 //if that packet it ready to transmit
//             if (flag_transmit == 1) {
//                int break_frame = 0;
//                //itterate through the subframe's PhyPDUs
//                for (uint32_t index = 0 ; index < map1->getUlSubframe()->getNbPdu (); index++) {
//                    debug10 ("Search index :%d for CDMA enqueue :%s\n", index,HelperFunctions::getMacAddrString(mac_->addr()).c_str());
//                    b1 = map1->getUlSubframe()->getPhyPdu (index)->getBurst (0);
//
//					//if the burst we just pulled out of this PhyPdu is an initial ranging request
//                    if (b1->getIUC()==UIUC_INITIAL_RANGING) {
//                        int $index_p_burst = 0;
//                        int q_len = map1->getUlSubframe()->getPhyPdu(index)->getBurst(0)->getQueueLength_packets();
//
//						//itterate through the packets in this burst's queue
//                        while(q_len>0) {
//                           Ptr<Packet> p_tmp_b = map1->getUlSubframe()->getPhyPdu(index)->getBurst(0)->lookup($index_p_burst);
//                           q_len--;
//                           $index_p_burst++;
//
//                           if (PeekPointer(p_tmp_b)) {
//                                All_Headers* ch_tmp = new All_Headers;
//                                p_tmp_b->PeekHeader(*ch_tmp);
//                                Generic_MAC_Header *wimaxHdr_send;
//                                wimaxHdr_send = &(ch_tmp->GMH);
//                                
//                               // cdma_req_header_t *header_s = (cdma_req_header_t *)&(HDR_MAC802_16(p_tmp_b)->header);
//                                //if (mac_->addr() == header_s->br) {
//                                if (mac_->addr() == ch_tmp->dest) {
//                                   debug10 ("\tFound enqueued cdma_init_req in BURST (remove it) CID :%d, ADDR :%d, #symbol :%d, #symbol_offset :%d, #subchannel :%d, #subchannel_offset :%d, code :%d\n", ch_tmp->GMH.GMH.cid, ch_tmp->CRH.data.br, ch_tmp->PIH.num_OFDMSymbol, ch_tmp->PIH.OFDMSymbol_offset, ch_tmp->PIH.num_subchannels, ch_tmp->PIH.subchannel_offset, ch_tmp->CRH.data.code);
//                                   b1->remove(PeekPointer(p_tmp_b));
//                                   mac_->getMap()->getUlSubframe()->getRanging()->setFLAGNOWTRANSMIT(mac_->addr(), 0);
//                                   break_frame = 1;
//                                   break;
//                                }
//                                delete ch_tmp;
//                            }
//                        }
//                    }
//                    if (break_frame == 1) break;
//                }
//             }
//          }
//
//       }
//
//    }//end SS
//
//
//    return b_data;
//  }
////end sending registration message
//
//  peer = mac_->getPeerNode_head();
//  p = c->get_queue()->head();
//
//  uint32_t max_data;
//
//  if (mac_->getNodeType()==STA_BS) {
//  	//we're going to figure out how much room we have in the burst, then decide on fragmentation
//  	Ofdm_mod_rate mr = mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding();
//    max_data = phy->getMaxPktSize (b->getnumSubchannels(), mr, DL_) - b_data;    
//	
//	//all the packets in the queues still have the All_Headers on them, so the total byte count is inflated
////	int actual_queue_byte = c->queueByteLength() - c->queueLength() * ch.GetSerializedSize();
//	int actual_queue_byte = 0;
//	for (int i = 0; i < c->queueLength(); i++) {
//		actual_queue_byte += mac_->GetActualLength(c->queueLookup(i), false);
//	}
//	if (actual_queue_byte); //quiet, compiler
//	
//    debug10 ("\tBS_transfer1.1, CID :%d with bdata :%d, but MaxData (MaxSize-bdata) :%d, MaxSize :%d, q->bytes :%d\n", c->get_cid(), b_data, max_data, phy->getMaxPktSize (b->getnumSubchannels(), mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding(), DL_), actual_queue_byte );
//    
//    debug10 ("\t   Bduration :%d, Biuc :%d, B#subchannel :%d\n",b->getDuration(), b->getIUC(), b->getnumSubchannels());
//
//  } else {
//
//    max_data = phy->getMaxPktSize (b->getDuration(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_) - b_data;    
//
//    debug10 ("\tSS_transfer1.1, CID :%d with bdata :%d, but MaxData (MaxSize-bdata) :%d, MaxSize :%d, bw-header-size :%d, q->bytes :%d\n", c->get_cid(), b_data, max_data, phy->getMaxPktSize (b->getnumSubchannels(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_), bw_req_packet, c->queueByteLength() );
//    debug10 ("\t   Bduration :%d, Biuc :%d, B#subchannel :%d\n",b->getDuration(), b->getIUC(), b->getnumSubchannels());
////	cout <<"this is a test\n" << std::flush;
//
//	//if we have data to send and can send at least some of it in this burst
//    if ( (max_data >= ch.GMH.GetSerializedSize()) && (mac_->getMap()->getUlSubframe()->getBw_req()->getRequest (c->get_cid())!=NULL) && (c->queueByteLength() > 0) ) {
//	//if ((mac_->getMap()->getUlSubframe()->getBw_req()->getRequest (c->get_cid())!=NULL) && (c->queueByteLength() > 0)) {
//
//
////    int slot_br = 0;
//    Connection *c_tmp;
//    c_tmp = c;
//
//    int i_packet = 0;
//    int already_frag = 0;
//    uint32_t real_bytes = 0;
//
//    Ptr<Packet> np;
//    All_Headers nph;
//    debug10 ("Retrive connection :%d, qlen :%d\n", c_tmp->get_cid(), c_tmp->queueLength());
//    //loop through packets on this connection and update the total number of bytes for this transmission
//    for (int j_p = 0; j_p<c_tmp->queueLength(); j_p++) {
//        if ( (np = c_tmp->queueLookup(i_packet)) != NULL ) {
//        	np->PeekHeader(nph);
//            int p_size = nph.GMH.GMH.len;
//            debug10 ("CON CID :%d, packet-id :%d, q->byte :%d, q->len :%d, packet_size :%d, frag_no :%d, frag_byte :%d, frag_stat :%d, real_bytes :%d\n", c_tmp->get_cid(), i_packet, c_tmp->queueByteLength(), c_tmp->queueLength(), p_size, c_tmp->getFragmentNumber(), c_tmp->getFragmentBytes(), (int)c_tmp->getFragmentationStatus(), real_bytes );            
//	    	i_packet++;
//
//            if ( (c_tmp->getFragmentBytes()>0) && (already_frag == 0) ){
//               p_size = p_size - c_tmp->getFragmentBytes() + 4;
//               already_frag = 1;
//            }
//
//            Ofdm_mod_rate dlul_map_mod = mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding();
//            int num_of_slots = (int) ceil((double)p_size/(double)phy->getSlotCapacity(dlul_map_mod,UL_));
//
//            real_bytes = real_bytes + (int) ceil((double)num_of_slots*(double)(phy->getSlotCapacity(dlul_map_mod,UL_)));
//        }
//   }
//
//      if (c_tmp->getBW_REQ_QUEUE() == 0) flag_bw_req = 0;
//      if (max_data >= real_bytes) {
//         flag_bw_req = 0;
//      } else {
//
//#ifndef IF_PIG
//        flag_bw_req = 6;
//        debug10 ("\t   CheckBW-REQ (cdma_bw_req exist), generate BW-REQ, maxdata :%d, bw-req-size :%d, queue :%d\n",max_data, bw_req_packet, c->queueByteLength());
//#endif
//
//#ifdef IF_PIG
//        if (max_data <= 10) {
//           flag_bw_req = 6;
//           debug10 ("\t   CheckBW-REQ (cdma_bw_req exist), generate BW-REQ, maxdata :%d, queue :%d\n",max_data, c->queueByteLength());
//        } else {
//           debug10 ("\t   CheckBW-REQ (cdma_bw_req exist), generate Piggybacking, maxdata :%d, queue :%d\n",max_data, c->queueByteLength());
//           flag_pig = 2;
//           flag_bw_req = 0;
//        }
//#endif
//
//      }
//
//    } else {
//
//      if (max_data >= c->queueByteLength()) {
//         flag_bw_req = 0;
//         flag_pig = 0;
//      } else {
//
//#ifdef IF_PIG
//      if ( ( max_data < c->queueByteLength() ) && (max_data > 10) && (c->queueByteLength() > 0)) {
//         debug10 ("\t   CheckBW-REQ (cdma_bw_req exist), generate Piggybacking, maxdata :%d, queue :%d\n",max_data, c->queueByteLength());
//         flag_pig = 2;
//      }
//#endif
//      }
//
//    }//end if sending bw-req
//
//
//  }
//
////Start sending BW-REQ packet
//  if (flag_bw_req > 0) {
//  
//    Ptr<Packet> np_tmp = mac_->getPacket();
//    All_Headers np_hdr;
//    np_tmp->RemoveHeader(np_hdr);
//    Signaling_MAC_Header *header = &(np_hdr.SMH);
//    if (header); //quiet, compiler
//    np_hdr.gm_sm = 0;
//    
//    int real_bytes = 0;
//
//    int slot_br = 0;
//    //uint32_t max_tmp_data_t1 = phy->getMaxPktSize (b->getnumSubchannels(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_) - b_data;
//    Ofdm_mod_rate dlul_map_mod = mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding();
//    int num_slot_t1 =ceil((double)ch.GMH.GetSerializedSize()/(double)phy->getSlotCapacity(dlul_map_mod,UL_));
//    uint32_t max_tmp_data = phy->getMaxPktSize (b->getnumSubchannels()-num_slot_t1, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_) - b_data;
//
////    real_bytes = c->queueByteLength()-max_tmp_data;
//
//    Connection *c_tmp;
//    c_tmp = c;
//
//    int i_packet = 0;
//    int  already_frag = 0;
//
//	Ptr<Packet> p_tmp;
//	All_Headers p_hdr_tmp;
//    debug10 ("Retrive connection :%d, qlen :%d\n", c_tmp->get_cid(), c_tmp->queueLength());
//    //loop through the connection and update the total number of bytes that we're requesting bandwidth for
//    for (int j_p = 0; j_p<c_tmp->queueLength(); j_p++) {
//        if ( PeekPointer((p_tmp = c_tmp->queueLookup(i_packet))) != NULL ) {
//        	p_tmp->PeekHeader(p_hdr_tmp);
//            int p_size = p_hdr_tmp.GMH.GMH.len;
//            debug10 ("CON CID :%d, packet-id :%d, q->byte :%d, q->len :%d, packet_size :%d, frag_no :%d, frag_byte :%d, frag_stat :%d, real_bytes :%d\n", c_tmp->get_cid(), i_packet, c_tmp->queueByteLength(), c_tmp->queueLength(), p_size, c_tmp->getFragmentNumber(), c_tmp->getFragmentBytes(), (int)c_tmp->getFragmentationStatus(), real_bytes );
//            i_packet++;
//
//            if ( (c_tmp->getFragmentBytes()>0) && (already_frag == 0) ){
//               p_size = p_size - c_tmp->getFragmentBytes() + 4;
//               already_frag = 1;
//            }
//
//            //Ofdm_mod_rate dlul_map_mod = mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(); //chackhai
//            int num_of_slots = (int) ceil((double)p_size/(double)phy->getSlotCapacity(dlul_map_mod,UL_));
//
//            real_bytes = real_bytes + (int) ceil((double)num_of_slots*(double)(phy->getSlotCapacity(dlul_map_mod,UL_)));
//        }
//     }
//
//    slot_br = real_bytes - max_tmp_data;
//    if (slot_br <0) slot_br = 0;
//    if (max_tmp_data < np_hdr.GMH.GetSerializedSize()+ch.FSH.GetSerializedSize()) {
//		slot_br = real_bytes;
//    }
//    np_hdr.SMH.BRH.br = slot_br;
//    np_hdr.SMH.BRH.cid = c->get_cid();
//    np_hdr.GMH.GMH.cid = c->get_cid();
//	np_hdr.SMH.BRH.ht = 1;
//	np_hdr.GMH.GMH.ht = 1;
//	np_hdr.CRH.data.ht = 1;
//	np_hdr.SMH.BRH.ec = 0;
//	np_hdr.SMH.BRH.type = 0;
//	
//    debug10 ("\t   Sending BW-REQ, cid :%d, qBytes :%d, qLen :%d, bw-req :%d, real_bytes :%d\n", header->BRH.cid, c->queueByteLength(), c->queueLength(), header->BRH.br, real_bytes);
//
//    np_hdr.GMH.GMH.type_frag = 0;
//    np_hdr.GMH.GMH.type_fbgm = 0;
//
//    int assigned_subchannels;
//    int symbol_offset;
//    np_hdr.GMH.GMH.len = np_hdr.SMH.GetSerializedSize();
//
//    assigned_subchannels = phy->getNumSubchannels(b_data, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
//    numsubchannels = phy->getNumSubchannels(ch.GMH.GMH.len, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
//
//    dlul_map_mod = mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(); //chakchai
//    //int num_of_slots = (int) ceil((double)ch.GMH.GMH.len/phy->getSlotCapacity(dlul_map_mod,UL_)); //get the slots needs.
//    symbol_offset = b->getStarttime() + num_symbol_per_slot * ((b->getSubchannelOffset()+assigned_subchannels)/ phy->getNumsubchannels(UL_));
//    txtime_s = (int)(ceil((double)(numsubchannels + subchannel_offset) / (double)phy->getNumsubchannels(UL_)))*num_symbol_per_slot;
//    offset = b->getStarttime( )+ txtime_s-num_symbol_per_slot; 
//    subchannel_offset = (int) ceil((numsubchannels + subchannel_offset) % (phy->getNumsubchannels(UL_)));
//
//    ch.txtime = txtime_s*phy->getSymbolTime ();
//
//    debug10 ("\t   BW-REQ, In station %d, Check packet to send: con id in BURST :%d, con id in connection :%d, size :%d, txtime :%f(%d), duration :%d(%f), numsubchnn :%d, subchnoff :%d\n",mac_->getNodeType(), c->get_cid(),b->getCid(), ch.GMH.GMH.len, ch.txtime, txtime_s, b->getDuration(), b->getDuration()*phy->getSymbolTime (),b->getnumSubchannels(),b->getSubchannelOffset ());
//
//	//now that we've calculated the physical slots over which this burst will be transmitted, fill
//	//out the PHY_INFO_HEADER with that information
//    np_hdr.PIH.num_subchannels = numsubchannels;
//    np_hdr.PIH.subchannel_offset = initial_subchannel_offset;//subchannel_offset;
//    np_hdr.PIH.num_OFDMSymbol = txtime_s;
//    np_hdr.PIH.OFDMSymbol_offset = symbol_offset; //initial_offset;
//    np_hdr.PIH.channel_index = (mac_->getPeerNode_head())->getchannel();
//    np_hdr.PIH.direction = 1;
//    np_hdr.PIH.modulation_ = dlul_map_mod;
//
//    debug10 ("\t   BW-REQ, In station %d, Packet phy header: numsymbols :%d, symboloffset :%d, subchanneloffset :%d, numsubchannel :%d, Channel_num :%d, direction :%d\n", mac_->getNodeType(), txtime_s,initial_offset, initial_subchannel_offset,numsubchannels, ch.PIH.channel_index, ch.PIH.direction);
//	np_tmp->AddHeader(np_hdr);
//    b->enqueue(np_tmp);      //enqueue into burst
////    mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (c->get_cid());
////    b_data = b_data+flag_bw_req;
////    c->setBW_REQ_QUEUE(BW_REQ_TIMEOUT);
////
//    b_data += np_hdr.PIH.num_subchannels*phy->getSlotCapacity(dlul_map_mod,UL_);
//    int t16time = ceil((double)(mac_->macmib_->t16_timeout)/(double)(mac_->macmib_->frame_duration));
//    c->setBW_REQ_QUEUE(t16time);
//
//    initial_offset = offset;
//    initial_subchannel_offset = subchannel_offset;
//
//    if (mac_->getNodeType()!=STA_BS) {
//       if ( mac_->getMap()->getUlSubframe()->getBw_req()->getRequest (c->get_cid()) != NULL ) {
////           mac_->getMap()->getUlSubframe()->getBw_req()->setTIMEOUT (c->get_cid(), BW_REQ_TIMEOUT);
//           mac_->getMap()->getUlSubframe()->getBw_req()->setTIMEOUT (c->get_cid(), t16time);
//           debug10("Reset CDMA_BW_REQ timer since send BW-REQ out for cid [%d].\n",c->get_cid());
//       }
//    }
//
//    if (mac_->getNodeType()!=STA_BS) { 
//       if ( mac_->getMap()->getUlSubframe()->getBw_req()->getRequest (c->get_cid()) != NULL ) {
//          Ptr<Packet> p_tmp = NULL;
//          FrameMap *map1 = mac_->getMap();
//          Burst *b1;
//
//          if (PeekPointer((p_tmp = mac_->getMap()->getUlSubframe()->getBw_req()->getPacket_P(c->get_cid()))) != NULL) {
//	     int flag_transmit = mac_->getMap()->getUlSubframe()->getBw_req()->getFLAGNOWTRANSMIT(c->get_cid());
//
//             debug10 ("Search flag_transmit :%d; remove cdma_bw_req if flag == 1, nbpdu :%d\n", flag_transmit, map1->getUlSubframe()->getNbPdu ());
//	     if (flag_transmit == 1) {
//				int break_frame = 0;
//                for (uint32_t index = 0 ; index < map1->getUlSubframe()->getNbPdu (); index++) {
//					debug10 ("Search index :%d for CDMA enqueue :%d\n", index, c->get_cid());
//					b1 = map1->getUlSubframe()->getPhyPdu (index)->getBurst (0);
//				
//					if (b1->getIUC()==UIUC_REQ_REGION_FULL) {
//					int $index_p_burst = 0;
//					int q_len = map1->getUlSubframe()->getPhyPdu(index)->getBurst(0)->getQueueLength_packets();
//	
//					while(q_len>0) {
//					   Ptr<Packet> p_tmp_b = map1->getUlSubframe()->getPhyPdu(index)->getBurst(0)->lookup($index_p_burst);
//					   q_len--;
//					   $index_p_burst++;
//		
//					   if (PeekPointer(p_tmp_b)) {
//						    All_Headers *ah_send = new All_Headers;
//						    p_tmp_b->PeekHeader(*ah_send);
//							CDMA_REQUEST_HEADER *header_s = &(ah_send->CRH); 
//							if (c->get_cid() == header_s->data.cid) {
//								   debug10 ("\tFound enqueued cdma_bw_req in BURST (remove it) CID :%d, #symbol :%d, #symbol_offset :%d, #subchannel :%d, #subchannel_offset :%d, code :%d\n", header_s->data.cid, ah_send->PIH.num_OFDMSymbol, ah_send->PIH.OFDMSymbol_offset, ah_send->PIH.num_subchannels, ah_send->PIH.subchannel_offset, header_s->data.code);
//								   b1->remove(p_tmp_b);
//								   mac_->getMap()->getUlSubframe()->getBw_req()->setFLAGNOWTRANSMIT(c->get_cid(), 0);
//								   break_frame = 1;
//								   break;
//							} 
//							delete ah_send;
//						}
//					}
//				}
//				if (break_frame == 1) break;
//  	        }
//           }
//         }
//
////        debug10("\tRemove CDMA_BW_REQ (if exist)  since send BW-REQ out for cid [%d].\n",c->get_cid());
////      	mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (c->get_cid());
//
//       }
//
//    }//end SS
//
//    return b_data;
//  }
////end sending BW-REQ packet
//
//  if (max_data < ch.GMH.GetSerializedSize() ||
//      (c->getFragmentationStatus()!=FRAG_NOFRAG && max_data < ch.GMH.GetSerializedSize() + ch.FSH.GetSerializedSize()))
//    {
//      debug2("In station %d returning as no space ( %d < %d )\n", mac_->getNodeType(), max_data, ch.GMH.GetSerializedSize());
//      return b_data; //not even space for header
//    }
//
//  if (flag_bw_req>0) {
//      if ( (max_data) < ch.GMH.GetSerializedSize() ||
//      	 (c->getFragmentationStatus() != FRAG_NOFRAG && (max_data-flag_bw_req) < ch.GMH.GetSerializedSize()+ch.FSH.GetSerializedSize())) {
//
//         debug2("In station %d returning as no space\n", mac_->getNodeType());
//         return b_data; //not even space for header
//      }
//  }
//
//  int gm_flag = 0;
//  //int more_bw = 0;
//#ifdef IF_PIG
//
////Piggybacking 1-May
//  if (flag_pig > 0 ) {
//
//     more_bw = c->queueByteLength() - max_data;
//     if (mac_->getNodeType()!=STA_BS && PeekPointer(p)) {
//        //ch = HDR_CMN(p);
//        p->PeekHeader(ch);
//        //wimaxHdr = HDR_MAC802_16(p);
//
//        debug10("In station %d (SS) ch-size :%d, max_data :%d, queue_bytes :%d, queue_len :%d\n", mac_->getNodeType(), ch.GMH.GMH.len,  max_data, c->queueByteLength(), c->queueLength());
//        if (max_data < c->queueByteLength()) {
//           ch.GMH.GMH.type_fbgm = true;
//           ch.GMSH.data.piggyback_req = more_bw;
//           debug10 ("In station %d (SS) Piggybacking flag :%d, queue_bytes :%d, queue_len :%d, max_data :%d, askmore :%d\n", mac_->getNodeType(), wimaxHdr->header.type_fbgm, c->queueByteLength(), c->queueLength(), max_data, more_bw);
//           gm_flag = 2;
//        } else {
//           ch.GMH.GMH.type_fbgm = 0;
//           debug10 ("In station %d (SS) no gm :%d, maxdata :%d, queue :%d\n",  mac_->getNodeType(), wimaxHdr->header.type_fbgm, max_data, c->queueByteLength());
//        }
//
//    }
// }
//
////Remove cdma-bw-req since we send piggybacking
//    if (mac_->getNodeType()!=STA_BS) {
//       if ( mac_->getMap()->getUlSubframe()->getBw_req()->getRequest (c->get_cid()) != NULL ) {
//          Ptr<Packet> p_tmp = NULL;
//          FrameMap *map1 = mac_->getMap();
//          Burst *b1;
//
//         if (PeekPointer((p_tmp = mac_->getMap()->getUlSubframe()->getBw_req()->getPacket_P(c->get_cid()))) != NULL) {
//			int flag_transmit = mac_->getMap()->getUlSubframe()->getBw_req()->getFLAGNOWTRANSMIT(c->get_cid());
//	
//			debug10 ("Search flag_transmit :%d; remove cdma_bw_req if flag == 1, nbpdu :%d\n", flag_transmit, map1->getUlSubframe()->getNbPdu ());
//		    debug10 ("p_tmp address :%d\n", p_tmp);
//			if (flag_transmit == 1) {
//			int break_frame = 0;
//					for (int index = 0 ; index < map1->getUlSubframe()->getNbPdu (); index++) {
//						debug10 ("Search index :%d for CDMA enqueue :%d\n", index, c->get_cid());
//						b1 = map1->getUlSubframe()->getPhyPdu (index)->getBurst (0);
//					
//						if (b1->getIUC()==UIUC_REQ_REGION_FULL) {
//						int $index_p_burst = 0;
//						int q_len = map1->getUlSubframe()->getPhyPdu(index)->getBurst(0)->getQueueLength_packets();
//		
//						while(q_len>0) {
//							Ptr<Packet>p_tmp_b = map1->getUlSubframe()->getPhyPdu(index)->getBurst(0)->lookup($index_p_burst);
//					  		q_len--;
//					   		$index_p_burst++;
//		
//						   if (PeekPointer(p_tmp_b)) {
//				     			debug10 ("p_tmp address from burst :%d\n", p_tmp_b);
//				     			All_Headers *ah_tmp;
//				     			p_tmp_b->PeekHeader(*ah_tmp);
//								//hdr_cmn* ch_tmp = HDR_CMN(p_tmp_b);
//								//hdr_mac802_16 *wimaxHdr_send;
//								//wimaxHdr_send = HDR_MAC802_16(p_tmp_b);
//								//CDMA_REQUEST_HEADER *header_s = (cdma_req_header_t *)&(HDR_MAC802_16(p_tmp_b)->header);
//								if ( (ah_tmpt->gm_sm && ah_tmp->GMH.GMH.cid == header_s->cid) || (!ah_tmpt->gm_sm && ah_tmp->SMH.BRH.cid == header_s->cid)){
//								   debug10 ("\tFound enqueued cdma_bw_req in BURST (remove it) CID :%d, #symbol :%d, #symbol_offset :%d, #subchannel :%d, #subchannel_offset :%d, code :%d\n", header_s->cid, wimaxHdr_send->phy_info.num_OFDMSymbol, wimaxHdr_send->phy_info.OFDMSymbol_offset, wimaxHdr_send->phy_info.num_subchannels, wimaxHdr_send->phy_info.subchannel_offset, header_s->code);
//								   b1->remove(p_tmp_b);
//								   mac_->getMap()->getUlSubframe()->getBw_req()->setFLAGNOWTRANSMIT(c->get_cid(), 0);
//								break_frame = 1;
//								break;
//							} 
//						}
//							}
//						}
//					if (break_frame == 1) break;
//				}
//				 }
//          }
//
//        debug10("\tRemove CDMA_BW_REQ if enqueued since send BW-REQ (PIGGYBACKING) out for cid [%d].\n",c->get_cid());
//      	mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (c->get_cid());
//
//       }
//
//    }//end SS
//#endif //end ifdef IF_PIG
// 
//  // overloaded function added by rpi 
//  while (PeekPointer(p)) { 
//
//    debug10 ("In station %d Entering while loop, connection type %d, CID :%d\n", mac_->getNodeType(), c->get_category(), c->get_cid());
//    
//	int actual_size = mac_->GetActualLength(p, mac_->getNodeType() == STA_BS ); 
//    p->RemoveHeader(ch);
//	ch.GMH.GMH.len = actual_size;
//	p->AddHeader(ch);
//
//    //ch = HDR_CMN(p);
//    wimaxHdr = &ch.GMH;
//
//    //wimaxHdr = HDR_MAC802_16(p);
//    if(ch.GMH.GMH.len < 0 ) debug2(" packet size negative --- panic!!! "); 		
//
//    if (mac_->getNodeType()==STA_BS)
//      max_data = phy->getMaxPktSize (b->getnumSubchannels()*b->getDuration(), mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding(), DL_) - b_data;
//    else {
//		//int nsch = b->getDuration();
//		//Ofdm_mod_rate mdrt;
//		//if (mac_->getMap()->getDlSubframe()->getProfile(b->getIUC()))//->getEncoding();
//// 		if (mac_->getMap()->getUlSubframe()->getProfile(b->getIUC()))//->getEncoding();
//// 	    	mdrt = mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding();
//// 		max_data = phy->getMaxPktSize (nsch, mdrt, UL_);
//		max_data = phy->getMaxPktSize (b->getDuration(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_) - b_data - gm_flag;
//    }
//
//
//    if (max_data <= wimaxHdr->GetSerializedSize() ||
//	(c->getFragmentationStatus()!=FRAG_NOFRAG && max_data <= wimaxHdr->GetSerializedSize()+ch.FSH.GetSerializedSize())) 
//      return b_data; //not even space for header
//
//
//    if (c->getFragmentationStatus()==FRAG_NOFRAG && max_data <= wimaxHdr->GetSerializedSize()+ch.FSH.GetSerializedSize() && (ch.GMH.GMH.len>wimaxHdr->GetSerializedSize()+ch.FSH.GetSerializedSize()) ) {
//      return b_data; //not even space for header 
//    }
//    
//
//	//if we're already fragmented....
//    if (c->getFragmentationStatus()!=FRAG_NOFRAG) { 
//      //if we have room for all (or all that is remaining of) this packet...
//	  if (max_data >= ch.GMH.GMH.len-c->getFragmentBytes()+ch.FSH.GetSerializedSize()){
//	    //add fragmentation header
//	    wimaxHdr->GMH.type_frag = true;
//	    ch.FSH.data.fc  = FRAG_LAST;
//	    ch.FSH.data.fsn = c->getFragmentNumber ();
//        int more_bw = 0;
//        if (more_bw); //quiet, compiler
//	   
//	    c->dequeue();  /*remove packet from queue */
//
//	    ch.GMH.GMH.len = ch.GMH.GMH.len-c->getFragmentBytes()+ch.FSH.GetSerializedSize(); //new packet size
//	    //update fragmentation
//	    if(ch.GMH.GMH.len < 0 ) debug2(" packet size negative -- panic !!! \n") ; 
//
//        debug2 ("\nEnd of fragmentation %d, CID :%d, (max_data :%d, bytes to send :%d, getFragmentBytes :%d, getFragNumber :%d, updated Frag :%d, update FragBytes :%d, con->qBytes :%d, con->qlen :%d, more_bw :%d)\n", ch.FSH.data.fsn, c->get_cid(), max_data, ch.GMH.GMH.len, c->getFragmentBytes(), c->getFragmentNumber(), 0, 0, c->queueByteLength(), c->queueLength(), more_bw);
//            
//	    c->updateFragmentation (FRAG_NOFRAG, 0, 0);
//	    if (gm_flag>0) { gm_flag = 0; ch.GMH.GMH.len = ch.GMH.GMH.len + HDR_PIG;}
//  	  } else {
//		  //need to fragment the packet again
//		  p = p->Copy(); //copy packet to send
//		  p->PeekHeader(ch);
//		  //add fragmentation header
//		  ch.GMH.GMH.type_frag = true;
//		  ch.FSH.data.fc = FRAG_CONT;
//		  ch.FSH.data.fsn = c->getFragmentNumber();
//		  int more_bw = 0;
//			if (more_bw); //quiet, compiler
//		  ch.GMH.GMH.len = max_data; //new packet size
//	
//		  if (gm_flag>0) { gm_flag = 0; ch.GMH.GMH.len = ch.GMH.GMH.len + HDR_PIG;}
//		  //update fragmentation
//
//          debug2 ("\nContinue fragmentation %d, CID :%d, (max_data :%d, bytes to send :%d, getFragmentBytes :%d, getFragNumber :%d, updated Frag :%d, update FragBytes :%d, con->qBytes :%d, con->qlen :%d, more_bw :%d)\n", ch.FSH.data.fsn, c->get_cid(), max_data, ch.GMH.GMH.len,  c->getFragmentBytes(), c->getFragmentNumber(), (c->getFragmentNumber()+1)%8, c->getFragmentBytes()+max_data-(ch.GMH.GetSerializedSize()+ch.FSH.GetSerializedSize()), c->queueByteLength(), c->queueLength(), more_bw);
//		  c->updateFragmentation (FRAG_CONT, (c->getFragmentNumber ()+1)%8, c->getFragmentBytes()+max_data-(ch.GMH.GetSerializedSize()+ch.FSH.GetSerializedSize()));
//		}
//      } else {//else we're not already fragmented
//      if (max_data < ch.GMH.GMH.len && c->isFragEnable()) {
//		//need to fragment the packet for the first time
//		p = p->Copy(); //copy packet to send so we don't mess up what's still in the queue
//		p->PeekHeader(ch);
//		int ori_ch = ch.GMH.GMH.len;
//			if (ori_ch); //quiet, compiler
//		//add fragmentation header
//		ch.GMH.GMH.type_frag = true;
//		ch.FSH.data.fc = FRAG_FIRST;
//		ch.FSH.data.fsn = c->getFragmentNumber ();
//		int more_bw = 0;
//			if (more_bw); //quiet, compiler
//		ch.GMH.GMH.len = max_data; //new packet size
//	
//		debug2 ("\nFirst fragmentation %d, CID :%d, (max_data :%d, bytes to send :%d, ori_size :%d, getFragmentBytes :%d, FRAGSUB :%d, getFragNumber :%d, updated Frag ;%d, update FragBytes :%d, con->qBytes :%d, con->qlen :%d, more_bw :%d)\n", ch.FSH.data.fsn, c->get_cid(), max_data, ch.GMH.GMH.len, ori_ch, c->getFragmentBytes(), ch.FSH.GetSerializedSize(), c->getFragmentNumber (), 1, c->getFragmentBytes()+max_data-(ch.GMH.GetSerializedSize()+ch.FSH.GetSerializedSize()),c->queueByteLength(), c->queueLength (), more_bw);
//	
//		c->updateFragmentation (FRAG_FIRST, 1, c->getFragmentBytes()+max_data-(ch.GMH.GetSerializedSize()+ch.FSH.GetSerializedSize()));
//		if (gm_flag>0) { gm_flag = 0; ch.GMH.GMH.len = ch.GMH.GMH.len + HDR_PIG;}
//      
//    } else if (max_data < ch.GMH.GMH.len && !c->isFragEnable()) {
//		//the connection does not support fragmentation
//		//can't move packets anymore
//		return b_data;
//      } else {
//	//no fragmentation necessary/possible
//        int more_bw = 0;
//        	if (more_bw); //quiet, compiler
//		if (gm_flag>0) { gm_flag = 0; ch.GMH.GMH.len = ch.GMH.GMH.len + HDR_PIG;}
//	
//		debug2 ("\nNo fragmentation %d, (max_data :%d, bytes to send :%d, con->qBytes :%d, con->qlen :%d, more_bw :%d\n", ch.FSH.data.fsn, max_data, ch.GMH.GMH.len, c->queueByteLength(), c->queueLength(), more_bw);
//		
//		c->dequeue();
//			    	    
//
//      }//end frag
//    }
//
//    debug10 ("B_data1 :%d, ch.size :%d\n", b_data, ch.GMH.GMH.len);
////you did send something; reset BW_REQ_QUEUE
//    c->setBW_REQ_QUEUE(0);
//
//    if (mac_->getNodeType()!=STA_BS) {
//       if ( mac_->getMap()->getUlSubframe()->getBw_req()->getRequest (c->get_cid()) != NULL ) {
//          Ptr<Packet> p_tmp(NULL);
//          FrameMap *map1 = mac_->getMap();
//          Burst *b1;
//
//        if (PeekPointer((p_tmp = mac_->getMap()->getUlSubframe()->getBw_req()->getPacket_P(c->get_cid()))) != NULL) {
//	     	int flag_transmit = mac_->getMap()->getUlSubframe()->getBw_req()->getFLAGNOWTRANSMIT(c->get_cid());
//
//            debug10 ("Search flag_transmit :%d; remove cdma_bw_req if flag == 1, nbpdu :%d\n", flag_transmit, map1->getUlSubframe()->getNbPdu ());
//	     	//debug10 ("p_tmp address :%d\n", p_tmp);
//	     	if (flag_transmit == 1) {
//			int break_frame = 0;
//            for (uint32_t index = 0 ; index < map1->getUlSubframe()->getNbPdu (); index++) {
//		    	debug10 ("Search index :%d for CDMA enqueue :%d\n", index, c->get_cid());
//    		    b1 = map1->getUlSubframe()->getPhyPdu (index)->getBurst (0);
//		    
//    		    if (b1->getIUC()==UIUC_REQ_REGION_FULL) {
//					int $index_p_burst = 0;
//      				int q_len = map1->getUlSubframe()->getPhyPdu(index)->getBurst(0)->getQueueLength_packets();
//
//					while(q_len>0) {
//					   Ptr<Packet> p_tmp_b = map1->getUlSubframe()->getPhyPdu(index)->getBurst(0)->lookup($index_p_burst);
//					   q_len--;
//					   $index_p_burst++;
//
//					   if (PeekPointer(p_tmp_b)) {
//							All_Headers *ah_tmp_b = new All_Headers;
//							p_tmp_b->PeekHeader(*ah_tmp_b);
//							ah_tmp_b->gm_sm = 3;
//							if (c->get_cid() == ah_tmp_b->CRH.data.cid) {
//							   debug10 ("\tFound enqueued cdma_bw_req in BURST (remove it) CID :%d, #symbol :%d, #symbol_offset :%d, #subchannel :%d, #subchannel_offset :%d, code :%d\n", ah_tmp_b->CRH.data.cid, ah_tmp_b->PIH.num_OFDMSymbol, ah_tmp_b->PIH.OFDMSymbol_offset, ah_tmp_b->PIH.num_subchannels, ah_tmp_b->PIH.subchannel_offset, ah_tmp_b->CRH.data.code);
//							   b1->remove(p_tmp_b);
//							   mac_->getMap()->getUlSubframe()->getBw_req()->setFLAGNOWTRANSMIT(c->get_cid(), 0);
//							   break_frame = 1;
//							   break;
//							} 
//							delete ah_tmp_b;
//						}
//					}
//    		    }
//		    if (break_frame == 1) break;
//  	        }
//           }
//          }
//
//        debug10("\tRemove CDMA_BW_REQ (if exist) since send some packet out for cid [%d].\n",c->get_cid());
//      	mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (c->get_cid());
//
//       }
//
//    }//end SS
//
//    debug10 ("B_data2 :%d, ch.size :%d\n", b_data, ch.GMH.GMH.len);
//    
//    int assigned_subchannels;
//    int symbol_offset;
//    //Trying to calculate how to fill up things here? Before this, frag and arq business?
//
///*Richard fixed*/
//	Ofdm_mod_rate dlul_map_mod = OFDM_BPSK_1_2;
//    if (mac_->getNodeType()==STA_BS) 
//     {
//     		//total number we have to work with
//     		assigned_subchannels = b->getnumSubchannels();
//		//total number we need to add this packet in
//		numsubchannels = phy->getNumSubchannels(ch.GMH.GMH.len, mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding(), DL_);
//	
//		 dlul_map_mod = mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding();
//		 //the number of slots we need to hold this packet
//		int num_of_slots = (int) ceil((double)ch.GMH.GMH.len/phy->getSlotCapacity(dlul_map_mod,DL_)); //get the slots needs.
//			if(num_of_slots); //quiet, compiler
//		debug10(" (BS) is going to use [%d] slots.\n",num_of_slots);
//		symbol_offset = b->getStarttime() + num_symbol_per_slot * ((b->getSubchannelOffset()+assigned_subchannels)/ phy->getNumsubchannels(DL_));
//	
//		debug2("number subchannel before calculation is[%d]\t assigned_subchn[%d]\t subch_off[%d]\t PerSymbolSubch#[%d] %f\n", numsubchannels,assigned_subchannels, subchannel_offset, phy->getNumsubchannels(DL_), ceil((double)(numsubchannels + subchannel_offset) / phy->getNumsubchannels(DL_)));
//		
//		txtime_s = (int)(ceil((double)(numsubchannels + subchannel_offset) / (double)phy->getNumsubchannels(DL_))) * num_symbol_per_slot;
//		offset = b->getStarttime( )+ txtime_s-num_symbol_per_slot;
//		subchannel_offset = (int) ceil((numsubchannels + subchannel_offset) % (phy->getNumsubchannels(DL_)));
//		
//      }
//    else { //if its the SS
//
//      assigned_subchannels = phy->getNumSubchannels(b_data, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
//
//      numsubchannels = phy->getNumSubchannels(ch.GMH.GMH.len, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
//       dlul_map_mod = mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding();
//      int num_of_slots = (int) ceil((double)ch.GMH.GMH.len/phy->getSlotCapacity(dlul_map_mod,UL_)); //get the slots needs.
//      int numsubchannel_xingting = num_of_slots;
//      	if (numsubchannel_xingting); //quiet, compiler
//      debug10(" (SS) packet size[%d]\t Richard calculation[%d]\t Vs. xingting calculation[%d]\n", ch.GMH.GMH.len, numsubchannels,numsubchannel_xingting);
//	  symbol_offset = b->getStarttime() + num_symbol_per_slot * ((b->getSubchannelOffset()+assigned_subchannels)/ phy->getNumsubchannels(UL_));
//
//      debug2("number subchannel before calculation is[%d]\t assigned_subchn[%d]\t subch_off[%d]\t PerSYmbolSubch#[%d] %f\n", numsubchannels,assigned_subchannels, subchannel_offset, phy->getNumsubchannels(UL_), ceil((double)(numsubchannels + subchannel_offset) / phy->getNumsubchannels(UL_)));
//      
//      txtime_s = (int)(ceil((double)(numsubchannels + subchannel_offset) / (double)phy->getNumsubchannels(UL_)))*num_symbol_per_slot;
//      offset = b->getStarttime( )+ txtime_s-num_symbol_per_slot; 
//      subchannel_offset = (int) ceil((numsubchannels + subchannel_offset) % (phy->getNumsubchannels(UL_)));
//
//    }
//    debug2 ("txtime_s=%d, symbol time = %f txtime=%f\n", txtime_s, phy->getSymbolTime (), txtime_s*phy->getSymbolTime ());
//
//    ch.txtime = txtime_s*phy->getSymbolTime ();
//
//    debug2 ("In station %d Check packet to send: con id in burst = %d con id in connection = %d size=%d txtime=%f(%d) duration=%d(%f) numsubchnn = %d subchnoff = %d\n", mac_->getNodeType(), c->get_cid(),b->getCid(), ch.GMH.GMH.len,ch.txtime, txtime_s, b->getDuration(), b->getDuration()*phy->getSymbolTime (),b->getnumSubchannels(),b->getSubchannelOffset ());
//
//	//ch still has it's value from way above
//   	ch.PIH.num_subchannels = numsubchannels;
//    ch.PIH.subchannel_offset = initial_subchannel_offset;//subchannel_offset;
//    ch.PIH.num_OFDMSymbol = txtime_s;
//    ch.PIH.OFDMSymbol_offset = symbol_offset; //initial_offset;
//    ch.PIH.modulation_ = dlul_map_mod;
//
//    debug10("In transfer_packets1-- packets info: symbol_off[%d]\t symbol_#[%d]\t subch_Off[%d]\t subch_#[%d]\n", ch.PIH.OFDMSymbol_offset,ch.PIH.num_OFDMSymbol, ch.PIH.subchannel_offset,ch.PIH.num_subchannels);
//
//    //wimaxHdr->phy_info.channel_index = c->getPeerNode()->getchannel();
//    if (mac_->getNodeType()==STA_BS)
//      {
//		ch.PIH.direction = 0;
//		if(c->getPeerNode())
//		{
//			ch.PIH.channel_index = c->getPeerNode()->getchannel();
//		}
//      }
//    else
//      {
//		ch.PIH.channel_index = (mac_->getPeerNode_head())->getchannel();     
//		ch.PIH.direction = 1;
//      }
//
//    debug2("In station %d Packet phy header : numsymbols =%d , symboloffset =%d,  subchanneloffset= %d, numsubchannel = %d, Channel_num = %d, direction =%d \n", mac_->getNodeType(), txtime_s,initial_offset, initial_subchannel_offset,numsubchannels,ch.PIH.channel_index,ch.PIH.direction);
//
//    initial_offset = offset;
//    initial_subchannel_offset = subchannel_offset;
//
//    //  rpi end put phy header info - like num subchannels and ofdmsymbols.... 
//    debug10 ("B_data3 :%d, ch.size :%d\n", b_data, ch.GMH.GMH.len);
//    //replace the All_Headers on p
//    p->RemoveHeader(th);
//	p->AddHeader(ch);
//    b->enqueue(p);      //enqueue into burst
//
//    if (mac_->getNodeType()==STA_BS) {
//      dlul_map_mod = mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding();
//    } else {
//      dlul_map_mod = mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding();
//    }
////    b_data += ch.PIH.num_subchannels*phy->getSlotCapacity(dlul_map_mod,UL_);
//	  b_data += mac_->GetActualLength(p, mac_->getNodeType() == STA_BS);
////    b_data += ch.GMH.GMH.len; //increment amount of data enqueued
//
//    debug2 ("In station %d packet enqueued b_data = %d \n", mac_->getNodeType(), b_data);
//
//    if (!pkt_transfered && mac_->getNodeType()!=STA_BS)
//      { //if we transfert at least one packet, remove bw request
//		pkt_transfered = true;
//		mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (c->get_cid());
//      }
//
//    p = c->get_queue()->head(); //get new head
//
//  }
//  return b_data;
//}
///**
// * Transfer the packets from the given connection to the given burst
// * @param con The connection
// * @param b The burst
// * @param b_data Amount of data in the burst
// * @return the new burst occupation 
// */
//
// int Wimax2Scheduler::transfer_packets_with_fragpackarq(Connection *c, Burst *b, int b_data)
// {
// 	debug10("entering transfer_packets_with_fragpackarq(Connection*, Burst*, int)\n");
//   Ptr<Packet> p, mac_pdu, p_temp, p_current;
//   //hdr_cmn* ch, *ch_pdu, *ch_temp, *ch_current;
//   //hdr_mac802_16 *wimaxHdr, *wimaxHdr_pdu, *wimaxHdr_temp, *wimaxHdr_current;
//   All_Headers wimaxHdr, wimaxHdr_pdu, *wimaxHdr_temp, wimaxHdr_current;
//   wimaxHdr_temp = NULL; //quiet, compiler
//   double txtime = 0; if (txtime);//to stop compiler form complaining about unused variable
//   int txtime_s;
//   bool pkt_transfered = false;
//   Ptr<Wimax2Phy> phy = mac_->getPhy();
//   int offset = b->getStarttime( );
//   PeerNode *peer;
//   int numsubchannels,subchannel_offset; 
//   int initial_offset = b->getStarttime( );
//   int initial_subchannel_offset = b->getSubchannelOffset ();
//   uint32_t bw_req_packet = wimaxHdr.GMH.GetSerializedSize();
//   int flag_bw_req = 0;
// 
//   int num_symbol_per_slot;
//   if (mac_->getNodeType()==STA_BS) 
//   {
//     num_symbol_per_slot = phy->getSlotLength (DL_); 
//   } 
//   else 
//   {
//     num_symbol_per_slot = phy->getSlotLength (UL_); 
//   }
// 
//   //debug2 ("Entered transfer packets with ARQ for connection %d with bdata = %d\n ", c->get_cid(), b_data);
//   peer = mac_->getPeerNode_head();
//   //debug("ARQ Transfer_Packets with Frag pack arq\n");
// 
//     /*Firstly, check the retransmission queue status. */
// 	if (c->getArqStatus () != NULL && c->getArqStatus ()->isArqEnabled() == 1)
// 	{
// 		if((c->getArqStatus ()->arq_retrans_queue_) && (c->getArqStatus ()->arq_retrans_queue_->GetNPackets() > 0))
// 		{
// 			//debug("ARQ Transfer Packets: Transferring a Retransmission packet\n");
// 			mac_pdu = c->getArqStatus ()->arq_retrans_queue_->head();
// 		}
// 		else
// 			mac_pdu = c->get_queue()->head();			
// 	} 
// 	else  /*If ARQ is not supported, directly get the header packet of the queue.*/
// 		mac_pdu = c->get_queue()->head();
//  
// 	if(!mac_pdu)
// 	{
// 		//debug("ARQ: NO Data present to be sent\n");
// 		return b_data;
// 	}
// 
// 	uint32_t max_data;
// 	mac_pdu->PeekHeader(wimaxHdr);
//// 	wimaxHdr_pdu= HDR_MAC802_16(mac_pdu);
// 	//ch_pdu = HDR_CMN(mac_pdu);
// 	//printf("this brust has [%d] of subchannels.\n",b->getnumSubchannels());
// 
// 	int arq_block_size = wimaxHdr_pdu.GMH.GMH.len;
// 		if (arq_block_size);//to stop compiler form complaining about unused variable
// 
// 	/*get the maximum size of the data which could be transmitted using this data burst without changing anything.*/
// 	if (mac_->getNodeType()==STA_BS)
// 	{
// 		max_data = phy->getMaxPktSize (b->getnumSubchannels(), mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding(), DL_) - b_data;    
// 		//debug ("ARQ: In Mac %d max data=%d (b_data=%d) ch_pdu size=%d\n", mac_->addr(), max_data, b_data,ch_pdu->size ());
// 	}
// 	else
// 	{
// 		max_data = phy->getMaxPktSize (b->getnumSubchannels(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_) - b_data;    
// 		//debug ("ARQ: In Mac %d max data=%d (b_data=%d) ch_pdu size=%d\n", mac_->addr(), max_data, b_data,ch_pdu->size ());
// 
// 		/*If this data burst could hold one BW REQ message and could not hold even one ARQ block. We need sponser a BW REQ message.*/
// 		if ( (max_data >= bw_req_packet) && (max_data < wimaxHdr.GMH.GetSerializedSize()+wimaxHdr.FSH.GetSerializedSize()+wimaxHdr_pdu.GMH.GMH.len) ) 
// 		{
// 			debug10 ("\t   CheckBW-REQ, SS Should generate BW-REQ, maxdata :%d, bw-req-size :%d, queue :%d\n", max_data, bw_req_packet, c->queueByteLength());
// 			flag_bw_req = 1;
// 		} 
// 		else /*Otherwise, this data burst could hold the all the ARQ block or even could not hold a BW REQ message, we donot need a 
// 		       BW REQ.*/
// 		{
// 			flag_bw_req = 0;
// 		}
// 	}
// 
// 
// //#ifdef WIMAX_TEST
// 	/*This data burst could not hold even a BW REQ message, then we simply return. Maybe this will be a dead-end loop?*/
// 	if (max_data < wimaxHdr.GMH.GetSerializedSize() ||
//       (c->getFragmentationStatus()!=FRAG_NOFRAG && max_data < wimaxHdr.GMH.GetSerializedSize()+wimaxHdr.FSH.GetSerializedSize()))
//     {
//       //debug("In station %d returning as no space\n", mac_->getNodeType());
//       return b_data; //not even space for header
//     }
// 
//      /*If this data burst could hold a BW REQ, we fill out the necessary info. to ask for more BW.*/
// 	if (flag_bw_req == 1) 
// 	{
// 		Ptr<Packet> p1= mac_->getPacket();
// 		All_Headers wimaxHdr;
// 		p1->RemoveHeader(wimaxHdr);
// 
// 		Signaling_MAC_Header *header = &wimaxHdr.SMH;
// 		wimaxHdr.gm_sm = 0;
// 		//bw_req_header_t *header = (bw_req_header_t *)&(HDR_MAC802_16(p1)->header);
// 		header->BRH.ht=1;
// 		header->BRH.ec=1;
// 		header->BRH.type = 0x1; //aggregate
// 		//header->br = c->queueByteLength()+gmh->GetSerializedSize()+ch->FSH.GetSerializedSize();  /*set the BW we need.*/
// 		header->BRH.br = c->getArqStatus ()->arq_retrans_queue_->GetNBytes() +wimaxHdr.GMH.GetSerializedSize()+wimaxHdr.FSH.GetSerializedSize();  /*set the BW we need.*/
// 		header->BRH.cid = c->get_cid();
// 		debug10 ("\t   Sending BW-REQ, cid :%d, qBytes :%d \n", header->BRH.cid, header->BRH.br);
// 
// 		int assigned_subchannels;
// 		int symbol_offset;
// 		wimaxHdr.GMH.GMH.len = bw_req_packet;
// 
// 		assigned_subchannels = phy->getNumSubchannels(b_data, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
// 		numsubchannels = phy->getNumSubchannels(wimaxHdr.GMH.GMH.len, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
// 
// 		/*old code Chakchai*/
// 		symbol_offset = b->getStarttime() + (b->getSubchannelOffset()+assigned_subchannels)/ phy->getNumsubchannels(UL_);
// 
// 		if ((numsubchannels + assigned_subchannels) > (phy->getNumsubchannels(UL_) - b->getSubchannelOffset () + 1))
// 		{
// 			txtime_s = num_symbol_per_slot + (int) ((ceil((numsubchannels + assigned_subchannels - (phy->getNumsubchannels(UL_) - b->getSubchannelOffset () + 1)) / (double)phy->getNumsubchannels(UL_)))) * num_symbol_per_slot ;
// 			offset = b->getStarttime( )+ txtime_s-num_symbol_per_slot;
// 			subchannel_offset = (int) ceil((numsubchannels + assigned_subchannels - (phy->getNumsubchannels(UL_) - b->getSubchannelOffset () + 1)) % (phy->getNumsubchannels(UL_)));
// 			subchannel_offset++;
// 		} 
// 		else
// 		{
// 			subchannel_offset = numsubchannels + assigned_subchannels;
// 			txtime_s = num_symbol_per_slot;
// 		}
// 
// 		wimaxHdr.txtime = txtime_s*phy->getSymbolTime ();
// 
// 		//debug ("\t   In station %d, Check packet to send: con id in burst :%d, con id in connection :%d, size :%d, txtime :%f(%d), duration :%d(%f), numsubchnn :%d, subchnoff :%d\n", mac_->getNodeType(), c->get_cid(),b->getCid(), ch->size(),txtime, txtime_s, b->getDuration(), b->getDuration()*phy->getSymbolTime (),b->getnumSubchannels(),b->getSubchannelOffset ());
// 
// 		wimaxHdr.PIH.num_subchannels = numsubchannels;
// 		wimaxHdr.PIH.subchannel_offset = initial_subchannel_offset;//subchannel_offset;
// 		wimaxHdr.PIH.num_OFDMSymbol = txtime_s;
// 		wimaxHdr.PIH.OFDMSymbol_offset = symbol_offset; //initial_offset;
// 		wimaxHdr.PIH.channel_index = (mac_->getPeerNode_head())->getchannel();
// 		wimaxHdr.PIH.direction = 1;
//		wimaxHdr.PIH.modulation_ = mac_->getMap()->getUlSubframe()->getProfile(b->getIUC())->getEncoding();
// 
// 		//debug ("\t   In station %d, Packet phy header: numsymbols :%d, symboloffset :%d, subchanneloffset :%d, numsubchannel :%d, Channel_num :%d, direction :%d\n", mac_->getNodeType(), txtime_s,initial_offset, initial_subchannel_offset,numsubchannels,wimaxHdr.PIH.channel_index,wimaxHdr.PIH.direction);
// 
// 		initial_offset = offset;
// 		initial_subchannel_offset = subchannel_offset;
// 
// 		p1->AddHeader(wimaxHdr);
// 		b->enqueue(p1);      //enqueue into burst
// 		// mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (c->get_cid());
// 
// 		return b_data;
// 	}
// //#endif
// 
// 
// 
// 	// Frag/pack subheader size is always equal to 2 bytes
// 	// For each MAC PDU Created, there is definitely MAC Header and Packet Sub Header and at least one ARQ Block
// 	//if (max_data < gmh->GetSerializedSize()+ch->FSH.GetSerializedSize()+ch_pdu->size ()) 
// 	//debug("ARQ retrans :The max_data is %d\n", max_data);
// 	if (max_data < wimaxHdr.GMH.GetSerializedSize()+wimaxHdr.FSH.GetSerializedSize()) 
// 	{
// 		//debug("ARQ: Not enough space even for one block and headers.\n");
// 		return b_data; //not even space for one block and headers
// 	}
// 	
// 	//If the Current window for ARQ is less than one, cannot transmit the packet
// 	//debug("ARQ1.0 Transfer_Packets PDU: Current window : %d , FSN: %d\n", (((c->getArqStatus()->getAckSeq()) + (c->getArqStatus()->getMaxWindow())) &0x7FF),(wimaxHdr_pdu->pack_subheader.sn));
// 	int seqno = wimaxHdr_pdu.PSH.data.sn;
// 	int ack_seq = (c->getArqStatus()->getAckSeq());
// 	int max_window = (c->getArqStatus()->getMaxWindow()); 	
// 	int max_seq =  (c->getArqStatus()->getMaxSeq());	
// 		if (max_seq);//to stop compiler form complaining about unused variable
// 	int curr_window = (((c->getArqStatus()->getAckSeq()) + (c->getArqStatus()->getMaxWindow())) &0x7FF);  // Get the right most of the ARQ slide window point
// 		if (curr_window);//to stop compiler form complaining about unused variable
// 
// 	debug2("xingting== seqno=%d,  ack_seq=%d,  max_window=%d,   max_seq=%d,  curr_window=%d\n", seqno,ack_seq,max_window, max_seq,curr_window);
// 	
// 	/*if the current ARQ block is outside the ARQ slide window, do not transfer it.*/
// 	//if((((seqno - ack_seq) > 0) && ((seqno - ack_seq) > max_window)) 
// 	if((((seqno - ack_seq) > 0) && ((seqno - ack_seq) >= max_window)) 
// 		//|| ((seqno >= curr_window && seqno < max_window) && (ack_seq < max_seq && ack_seq >= (max_seq - max_window)))) 
// 		||((seqno-ack_seq)<0 && ((seqno+max_seq-ack_seq)>=max_window)))
// 	{
// 		debug2("ARQ1.0 Transfer_Packets: Current window Low :curr_window_ : %d\n", (((c->getArqStatus()->getAckSeq()) + (c->getArqStatus()->getMaxWindow())) &0x7FF));
// 		debug2("ARQ1.0 Transfer_Packets: this ARQ block is outside the slide window. will not be transferred.\n");
// 		return b_data;
// 	}
// 
// 
//   // MAC PDU will be the final packet to be transmitted 
//   //mac_pdu = mac_pdu->copy();
// 	//wimaxHdr_pdu= HDR_MAC802_16(mac_pdu);
// 	//ch_pdu = HDR_CMN(mac_pdu);
// 	mac_pdu->RemoveHeader(wimaxHdr_pdu);
// 
// #if 0
// 	int gm_flag = 0;
// 	p->RemoveHeader(wimaxHdr);
// 	//ch = HDR_CMN(p);
// 	//wimaxHdr = HDR_MAC802_16(p);
// 	p->RemoveHeader(wimaxHdr);
// 
// 	int more_bw = c->queueByteLength() - max_data + wimaxHdr.GMH.GetSerializedSize() + wimaxHdr.FSH.GetSerializedSize();
// 	if (mac_->getNodeType()==STA_BS) 
// 	{
// 	//     wimaxHdr->header.type_fbgm = false;
// 	}
// 	else 
// 	{
// 		 printf("SS ch-size :%d, max_data :%d, queue :%d\n", wimaxHdr.GMH.GMH.len,  max_data, c->queueByteLength());
//         if (max_data < c->queueByteLength()+ wimaxHdr.GMH.GetSerializedSize() + wimaxHdr.FSH.GetSerializedSize()) //modify this to account for all the extra junk on the packets from All_Headers
//         {
//              wimaxHdr->type_fbgm = true;
//              wimaxHdr.GMSH.data.piggyback_req = more_bw;
//              wimaxHdr.GMH.GMH.len += wimaxHdr.FSH.GetSerializedSize(); //new packet size
//              debug10 ("In station %d Piggybacking flag :%d, queue_bytes :%d, max_data :%d, askmore :%d\n", mac_->getNodeType(), wimaxHdr->type_fbgm, c->queueByteLength(), max_data, more_bw);
//              gm_flag = 2;
//         } 
//         else 
//         {
//              wimaxHdr->type_fbgm = false;
//              printf("SS no gm :%d, maxdata :%d, queue :%d\n", wimaxHdr->type_fbgm, max_data, c->queueByteLength());
//         }
//     }
//     p->AddHeader(wimaxHdr);
// #endif
// 
//  
// 	//We can remove the packet from the queue now
// 	if (c->getArqStatus () != NULL && c->getArqStatus()->isArqEnabled() == 1)
// 	{
// 		if((c->getArqStatus ()->arq_retrans_queue_) && (c->getArqStatus ()->arq_retrans_queue_->GetNPackets() > 0))
// 		{
// 			//debug("ARQ Transfer Packets: Transferring a Retransmission packet\n");	
// 			p = c->getArqStatus ()->arq_retrans_queue_->getNext();
// 		}	
// 		else
// 			p = c->dequeue();			
// 	} 
// 	else  
// 		p = c->dequeue();
// 
//   //set header information
//   wimaxHdr_pdu.GMH.GMH.ht = 0;
//   wimaxHdr_pdu.GMH.GMH.ec = 1;
//   wimaxHdr_pdu.GMH.GMH.type_mesh = 0;
//   wimaxHdr_pdu.GMH.GMH.type_ext = 1; /*ARQ is enabled */
//   wimaxHdr_pdu.GMH.GMH.type_frag = 0;
//   wimaxHdr_pdu.GMH.GMH.type_pck = 1; /*Packing header is present */ 
//   wimaxHdr_pdu.GMH.GMH.type_fbgm = 0;
//   wimaxHdr_pdu.GMH.GMH.ci = 1;
//   wimaxHdr_pdu.GMH.GMH.eks = 0;
//   wimaxHdr_pdu.GMH.GMH.cid = c->get_cid ();
//   wimaxHdr_pdu.GMH.GMH.hcs = 0;
//   wimaxHdr_pdu.GMH.GMH.len+=wimaxHdr_pdu.GMH.GetSerializedSize()+wimaxHdr_pdu.FSH.GetSerializedSize()+HDR_MAC802_16_CRC_SIZE; /*Size of generic Mac Header and CRC is set (Note one packet header and first ARQ Block is present)*/
//   debug2 ("ARQ: MAC [%s] Generated PDU Size: %d FSN is %d \n",HelperFunctions::getMacAddrString(mac_->addr()).c_str(),wimaxHdr_pdu.GMH.GMH.len,  wimaxHdr_pdu.PSH.data.sn);
//   
//   mac_pdu->AddHeader(wimaxHdr_pdu);
// 
//   // Set the Initial Conditions
//   p_temp = mac_pdu->Copy();
//   p_temp -> PeekHeader(*wimaxHdr_temp);
//   debug2 ("ARQ: MAX_data %d \t MAC %s PDU: ARQ Fragment Added... PDU Size: %d FSN is %d re-trans-Q len is %d  data-trans-Q len is %d\n", max_data, HelperFunctions::getMacAddrString(mac_->addr()).c_str(),wimaxHdr_temp->GMH.GMH.len,  wimaxHdr_temp->PSH.data.sn, c->getArqStatus ()->arq_retrans_queue_->GetNPackets()+1, c->get_queue()->GetNPackets()+1);
// 
//   if (c->getArqStatus () != NULL && c->getArqStatus ()->isArqEnabled() == 1)
//     {
//       if((c->getArqStatus ()->arq_retrans_queue_) && (c->getArqStatus ()->arq_retrans_queue_->GetNPackets() > 0)){
// 	debug2("ARQ Transfer Packets: Transferring a Retransmission packet\n");
// 	p = c->getArqStatus ()->arq_retrans_queue_->head();
//       }
//       else
//         p = c->get_queue()->head();			
//     } 
//   else  
//     p = c->get_queue()->head();
//   
// 	while (p)
// 	{
// 		p->RemoveHeader(wimaxHdr);
// 		
// 		if (mac_->getNodeType()==STA_BS)
// 			max_data = phy->getMaxPktSize (b->getnumSubchannels(), mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding(), DL_) - b_data;    
// 		else
// 			max_data = phy->getMaxPktSize (b->getnumSubchannels(), mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_) - b_data;  
// 
// 		debug2("ARQ : In Mac %s max data=%d (burst duration=%d, b_data=%d) PDU SIZE = %d\n", HelperFunctions::getMacAddrString(mac_->addr()).c_str(), max_data, b->getDuration(), b_data, wimaxHdr_pdu.GMH.GMH.len);
// 
// 		if((wimaxHdr.PSH.data.fc == FRAG_FIRST) 
// 			|| (((wimaxHdr.PSH.data.sn - wimaxHdr_temp->PSH.data.sn) &0x7FF)!= 1) 
// 			|| (wimaxHdr.PSH.data.fc == FRAG_NOFRAG))
// 		{   
// 			if (max_data < wimaxHdr_pdu.GMH.GMH.len + wimaxHdr.GMH.GMH.len +wimaxHdr.FSH.GetSerializedSize())  // why at less need to fit in 2 blocks?
// 			{
// 				debug2("Can not fit in 2 blocks + frag Subheader in this data burst.\n");
// 				break; //cannot fit in more
// 			}
// 		} 
// 		else
// 		{   
// 			if ((int)max_data < wimaxHdr_pdu.GMH.GMH.len+ wimaxHdr.GMH.GMH.len)   // why at less need to fit in 2 blocks??
// 			{
// 				debug2("Can not fit in 2 blocks in this data burst.\n");
// 				break; //cannot fit in more
// 			}
// 		} 
// 
//     // If the Current window for ARQ is less than one, cannot transmit the packet
//     debug2("ARQ2.0 Transfer_Packets: Current window : %d , FSN: %d\n", (((c->getArqStatus()->getAckSeq()) + (c->getArqStatus()->getMaxWindow())) &0x7FF),(wimaxHdr.PSH.data.sn));
//     int seqno = wimaxHdr.PSH.data.sn;
//     int ack_seq = (c->getArqStatus()->getAckSeq());
//     int max_window = (c->getArqStatus()->getMaxWindow());
//     int max_seq =  (c->getArqStatus()->getMaxSeq());	
//     int curr_window = (((c->getArqStatus()->getAckSeq()) + (c->getArqStatus()->getMaxWindow())) &0x7FF);
//     	if (curr_window) ; //to stop compiler from complaining about unused variable
//     if((((seqno - ack_seq) > 0) && ((seqno - ack_seq) > max_window)) 
// 	    //|| ((seqno >= curr_window && seqno < max_window) && (ack_seq < max_seq && ack_seq >= (max_seq - max_window)))) 
// 	    ||((seqno-ack_seq)<0 && ((seqno+max_seq-ack_seq)>max_window)))
//     {	 	
//   		printf("ARQ2.0 Transfer_Packets: seqno=%d, ack_seq=%d, 1st judge=%d,  2nd judge=%d",
// 					seqno,ack_seq,(((seqno - ack_seq) > 0) && ((seqno - ack_seq) > max_window)),((seqno-ack_seq)<0 && ((seqno+max_seq-ack_seq)>max_window)));
//       debug2("ARQ2.0 Transfer_Packets: this ARQ block is outside of ARQ slide window, Will not be transferred.\n ");
//       break;
//     }
// 	
//     //p_temp->allocdata (sizeof (class Packet));
//     //p_current = (Packet*) p_temp->accessdata();
//     
//     wimaxHdr.direction = UL; //to be received properly
//     
//     p->AddHeader(wimaxHdr);
// 
//     // Copies the entire contents of the current packet into data part of the previous packet 
//     p_current->AddAtEnd(p);
//     p_current->RemoveHeader(wimaxHdr_current);
//     
//     // Update the Mac PDU Size
//     if((wimaxHdr.PSH.data.fc == FRAG_FIRST) || (((wimaxHdr.PSH.data.sn - wimaxHdr_temp->PSH.data.sn) &0x7FF)!= 1)|| (wimaxHdr.PSH.data.fc == FRAG_NOFRAG))
//       wimaxHdr_pdu.GMH.GMH.len += wimaxHdr.GMH.GMH.len + wimaxHdr.FSH.GetSerializedSize() ;  	
//     else
//       wimaxHdr_pdu.GMH.GMH.len += wimaxHdr.GMH.GMH.len;
// 
//     //debug ("ARQ: MAC PDU: ARQ Fragment Added... PDU Size: %d FSN is %d \n",ch_pdu->size (),  wimaxHdr_current->pack_subheader.sn);
//     
//     // Set the Conditions
//     p_temp = p_current->Copy();	
//     p_temp->PeekHeader(*wimaxHdr_temp);
//     //ch_temp = ch_current;
// 
//     // Remove the packet from the queue
//     if (c->getArqStatus () != NULL && c->getArqStatus()->isArqEnabled() == 1)
//       {
// 	if((c->getArqStatus ()->arq_retrans_queue_) && (c->getArqStatus ()->arq_retrans_queue_->GetNPackets() > 0)){
// 	  //debug("ARQ Transfer Packets: Transferring a Retransmission packet\n");	
// 	  p = c->getArqStatus ()->arq_retrans_queue_->getNext();
// 	}	
// 	else
//           p = c->dequeue();			
//       } 
//     else  
//       p = c->dequeue();
// 
//     // Fetch the next packet
//     if (c->getArqStatus () != NULL && c->getArqStatus ()->isArqEnabled() == 1)
//       {
// 	if((c->getArqStatus ()->arq_retrans_queue_) && (c->getArqStatus ()->arq_retrans_queue_->GetNPackets() > 0)){
// 	  //debug("ARQ Transfer Packets: Transferring a Retransmission packet\n");
// 	  p = c->getArqStatus ()->arq_retrans_queue_->head();
// 	}
// 	else
// 	  p = c->get_queue()->head();			
//       } 
//     else  
//       p = c->get_queue()->head();   				
//   }
//   //debug ("ARQ: MAC PDU Generated Size : %d \n", ch_pdu->size());
//   int assigned_subchannels;
//   int symbol_offset;
//   //Trying to calculate how to fill up things here? Before this, frag and arq business?
//   if (mac_->getNodeType()==STA_BS) {
//     assigned_subchannels = phy->getNumSubchannels(b_data, mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding(), DL_);
//     numsubchannels = phy->getNumSubchannels(wimaxHdr_pdu.GMH.GMH.len, mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding(), DL_);
//     //RPI not working
//     //      symbol_offset = b->getStarttime() + (b->getSubchannelOffset()+assigned_subchannels)/numsubchannels;
//     //Richard fixed it.
//     symbol_offset = b->getStarttime() + (b->getSubchannelOffset()+assigned_subchannels)/ phy->getNumsubchannels(DL_);
// 
//     if((numsubchannels+assigned_subchannels) > (phy->getNumsubchannels(DL_) - b->getSubchannelOffset () + 1)){ //its crossing over to the next symbol
//       txtime_s = num_symbol_per_slot + (int) ((ceil((numsubchannels + assigned_subchannels - (phy->getNumsubchannels(DL_) - b->getSubchannelOffset () + 1)) / (double)phy->getNumsubchannels(DL_)))) * num_symbol_per_slot;
//       offset = b->getStarttime( )+ txtime_s-num_symbol_per_slot;
//       subchannel_offset = (int) ceil((numsubchannels + assigned_subchannels - (phy->getNumsubchannels(DL_) - b->getSubchannelOffset () + 1)) % (phy->getNumsubchannels(DL_)));
//       subchannel_offset++;
//     }  
//     else{
//       txtime_s=num_symbol_per_slot; 
//       subchannel_offset = numsubchannels + assigned_subchannels;
//     }
//   }
//   else { //if its the SS
//     //RPI not working
//     //      assigned_subchannels = phy->getNumSubchannels(b_data, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), DL_);
//     //Richard fixed it.
//     assigned_subchannels = phy->getNumSubchannels(b_data, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
// 
//     numsubchannels = phy->getNumSubchannels(wimaxHdr_pdu.GMH.GMH.len, mac_->getMap()->getUlSubframe()->getProfile (b->getIUC())->getEncoding(), UL_);
//     //RPI not working
//     //      symbol_offset = b->getStarttime() + (b->getSubchannelOffset()+assigned_subchannels)/numsubchannels;
//     //Richard fixed it.
//     symbol_offset = b->getStarttime() + (b->getSubchannelOffset()+assigned_subchannels)/ phy->getNumsubchannels(UL_);
// 
//     if((numsubchannels + assigned_subchannels) > (phy->getNumsubchannels(UL_) - b->getSubchannelOffset () + 1)){
//       txtime_s = num_symbol_per_slot + (int) ((ceil((numsubchannels + assigned_subchannels - (phy->getNumsubchannels(UL_) - b->getSubchannelOffset () + 1)) / (double)phy->getNumsubchannels(UL_)))) * num_symbol_per_slot ; 
//       offset = b->getStarttime( )+ txtime_s-num_symbol_per_slot; 
//       subchannel_offset = (int) ceil((numsubchannels + assigned_subchannels - (phy->getNumsubchannels(UL_) - b->getSubchannelOffset () + 1)) % (phy->getNumsubchannels(UL_)));
//       subchannel_offset++;
//     }
//     else{
//       subchannel_offset = numsubchannels + assigned_subchannels;
//       txtime_s=num_symbol_per_slot;
//     }
//   }
// 
//   wimaxHdr_pdu.txtime = txtime_s*phy->getSymbolTime ();
// 
//   debug10 ("In station %d Check packet to send: con id in burst = %d con id in connection = %d size=%d txtime=%f(%d) duration=%d(%f) numsubchnn = %d subchnoff = %d\n",mac_->getNodeType(), c->get_cid(),b->getCid(), wimaxHdr_pdu.GMH.GMH.len,txtime, txtime_s, b->getDuration(), b->getDuration()*phy->getSymbolTime (),b->getnumSubchannels(),b->getSubchannelOffset ());
//    
//   // rpi start put phy header info - like num subchannels and ofdmsymbols.... 
//   
//   
//   wimaxHdr_pdu.PIH.num_subchannels = numsubchannels;
//   wimaxHdr_pdu.PIH.subchannel_offset = initial_subchannel_offset;//subchannel_offset;
//   wimaxHdr_pdu.PIH.num_OFDMSymbol = txtime_s;
//   wimaxHdr_pdu.PIH.OFDMSymbol_offset = symbol_offset; //initial_offset;
//   if (mac_->getNodeType()==STA_BS)
//     {
//       wimaxHdr_pdu.PIH.direction = 0;
//       wimaxHdr_pdu.PIH.channel_index = c->getPeerNode()->getchannel();
//       wimaxHdr_pdu.PIH.modulation_ = mac_->getMap()->getDlSubframe()->getProfile(b->getIUC())->getEncoding();
//     }
//   else
//     {
//       wimaxHdr_pdu.PIH.channel_index = (mac_->getPeerNode_head())->getchannel();     
//       wimaxHdr_pdu.PIH.direction = 1;
//       wimaxHdr_pdu.PIH.modulation_ = mac_->getMap()->getUlSubframe()->getProfile(b->getIUC())->getEncoding();
//     }
//   
//   debug10 ("In station %d Packet phy header : numsymbols =%d , symboloffset =%d,  subchanneloffset= %d, numsubchannel = %d, Channel_num = %d, direction =%d \n", mac_->getNodeType(), txtime_s,initial_offset, initial_subchannel_offset,numsubchannels,wimaxHdr_pdu.PIH.channel_index,wimaxHdr_pdu.PIH.direction);
//         
//  	 
//   initial_offset = offset;
//   initial_subchannel_offset = subchannel_offset;
//   
//   mac_pdu->AddHeader(wimaxHdr_pdu);
//  
//   b->enqueue(mac_pdu);      //enqueue into burst
//   b_data += wimaxHdr_pdu.GMH.GMH.len; //increment amount of data enqueued
//   //debug ("In station %d packet enqueued b_data = %d" , mac_->getNodeType(), b_data);
//   if (!pkt_transfered && mac_->getNodeType()!=STA_BS){ //if we transfert at least one packet, remove bw request
//     pkt_transfered = true;
//     mac_->getMap()->getUlSubframe()->getBw_req()->removeRequest (c->get_cid());
//   }
// 
//   return b_data;
// }
//
} // namespace ns3 

