/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef WIMAX_SCHEDULER_H
#define WIMAX_SCHEDULER_H

#include "ns3/packet.h"
#include "ns3/wimax2-net-device.h"
#include "ns3/802_16headers.h"
#include "ns3/wimax2-phy.h"
#include "ns3/framemap.h"
#include "ns3/neighbordb.h"
#include "burst.h"
#include "dlburst.h"
#include "ulburst.h"
#include "contentionslot.h"
#include "phypdu.h"

namespace ns3 {

#define TX_GAP 0.000001 //Time in seconds between sending 2 packets (to avoid collision)

/**
 * Super class for schedulers (BS and MS schedulers)
 */ 
class Wimax2Scheduler	: public Object
{
friend class Wimax2NetDevice;
public:
  /*
   * Create a scheduler
   */
  Wimax2Scheduler ();

  /*
   * Set the mac
   * @param mac The Mac where it is located
   */
  void setMac (Ptr<Wimax2NetDevice> mac);

  /**
   * Initializes the scheduler
   */
  virtual void init ();

  /**
   * Return the Mac layer
   */
  inline Ptr<Wimax2NetDevice>  getMac () { return mac_;}
   

	// NOTE - for transfer packet functions, see subclasses
	//these are using the old nS-2 pack/frag systems and may assume OFDM (not OFDMA)



  /**
   * Transfert the packets from the given connection to the given burst
   * @param con The connection
   * @param b The burst
   * @param b_data The amount of data in burst
   * @return the new burst occupation
   */
//  int transfer_packets (Connection *c, Burst *b, int b_data, int subchannel_offset, int symbol_offset);

//Begin RPI
/**
   * Transfert the packets from the given connection to the given burst
   * @param con The connection
   * @param b The burst
   * @param b_data The amount of data in burst
   * @return the new burst occupation
   */
 // int transfer_packets_with_fragpackarq (Connection *c, Burst *b, int b_data);
//End RPI


  /**
   * Transfert the packets from the given connection to the given burst
   * @param con The connection
   * @param b The burst
   * @param b_data The amount of data in burst
   * @return the new burst occupation
   */
  //int transfer_packets1 (Connection *c, Burst *b, int b_data);



  /**
   * This function is used to schedule bursts/packets
   */
  virtual void schedule ();

//protected: - uncomment if and when python bindings generator can handle protected variables

  /**
   * The Mac layer
   */
  Ptr<Wimax2NetDevice> mac_;

  /**
   * The ratio for downlink subframe
   */
  double dlratio_;
  
private:



};

} //namespace ns3
#endif //WIMAX_SCHEDULER_H

