/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef SERVICEFLOW_H
#define SERVICEFLOW_H

//#include "serviceflowqos.h"
//#include "connection.h"
#include "ns3/packet.h"
#include "ns3/wimax2-serviceflowqos.h"
#include <sys/queue.h>
#include "ns3/wimax2-common.h"


namespace ns3 {

#define UNASSIGNED_FLOW_ID -1


/** Defines the supported scheduling mechanism for the flow */
enum SchedulingType_t {
  SERVICE_UGS,
  SERVICE_ertPS,
  SERVICE_rtPS,
  SERVICE_nrtPS,
  SERVICE_BE
};


class Wimax2ServiceFlow;
LIST_HEAD (serviceflow_list, Wimax2ServiceFlow);
struct service_flow_link {
	Wimax2ServiceFlow *le_next;
	Wimax2ServiceFlow **le_prev;
};

/**
 * Class Wimax2ServiceFlow
 * The service flow identifies the service requirement 
 * for the associated connection 
 */ 
 
 
class Wimax2ServiceFlow {
 public:
  /**
   * Constructor
   */
  Wimax2ServiceFlow (SchedulingType_t, Wimax2ServiceFlowQoS*);

  /**
   * Return the service flow id
   * @return The service flow id. -1 if not yet assigned
   */
  inline int getID () { return id_; }
  
  /**
   * Assign an ID to the service flow
   * @param id The ID to set
   */
  void setID (int id);

  /**
   * Pick the next available ID. Should be called by a BS to assign a unique ID
   */
  void pickID ();

  /**
   * Set the scheduling mechanism for this flow
   * @param scheduling The scheduling type
   */
  inline void setScheduling (SchedulingType_t scheduling) {scheduling_ = scheduling;}
  
  /**
   * Return the scheduling type for this service flow
   */
  inline SchedulingType_t getScheduling () { return scheduling_; }
  
  /**
   * Set the QoS for this flow
   * @param qos The new QoS for this flow
   */
  inline void setQoS (Wimax2ServiceFlowQoS* qos) { qos_ = qos; }

  /**
   * Return the direction for this connection
   */
  inline dir_t getDirection () { return direction_; }

  /**
   * Set the direction for this flow
   * @param dir The new direction for this flow
   */
  inline void setDirection (dir_t dir) { direction_ = dir; }

  /**
   * Return the QoS for this connection
   */
  inline Wimax2ServiceFlowQoS * getQoS () { return qos_; }
  
  // Chain element to the list
  inline void insert_entry_head(struct serviceflow_list *head) {
    LIST_INSERT_HEAD(head, this, link);
  }
  
  // Chain element to the list
  inline void insert_entry(Wimax2ServiceFlow *elem) {
    LIST_INSERT_AFTER(elem, this, link);
  }

  // Return next element in the chained list
  Wimax2ServiceFlow* next_entry(void) const { return link.le_next; }

  // Remove the entry from the list
  inline void remove_entry() { 
    LIST_REMOVE(this, link); 
  }

 //protected: commented to pybindgen can work (no protected member variables allowed)

  /**
   * Pointer to next in the list
   */
   service_flow_link link;

 private:
  /**
   * The service flow id
   */
   int id_;

  /**
   * The scheduling type (UGS, rtPS...)
   */
   SchedulingType_t scheduling_;

   /**
    * Flow direction
    */
   dir_t direction_;

   /**
    * The quality of service for this flow
    */
   Wimax2ServiceFlowQoS * qos_;

};

}//namespace
#endif //SERVICEFLOW_H
