/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#include "wimax2-timer.h"
#include "wimax2-net-device.h"
#include "ns3/wimax2-scheduler.h"

namespace ns3 {

Wimax2Timer::Wimax2Timer(Ptr<Wimax2NetDevice> m) {
	mac = m;
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
	status_ = TIMER_IDLE;
}

/* 
 * Starts a timer for the given duration
 * @param time The timer duration
 */
void Wimax2Timer::start(double time)
{
  //debug10("Starting a timer at time: %f\n", time);
  //Scheduler &s = Scheduler::instance();
  assert(busy_ == 0);
  status_ = TIMER_PENDING;
  busy_ = 1;
  paused_ = 0;
  stime = Simulator::Now ().GetSeconds ();
  rtime = time;
  assert(rtime >= 0.0);
  //intr = Simulator::ScheduleWithContext(mac->GetNode()->GetId(), Seconds(rtime), &Wimax2Timer::handle, this, &intr/*function,args*/);
  debug10("about to schedule the start of a timer %f seconds from now\n", time);
  if (time == 0) intr = Simulator::Schedule(Seconds(0), &Wimax2Timer::handle, this);
  else intr = Simulator::Schedule(Seconds(rtime), &Wimax2Timer::handle, this/*function,args*/);
  //if (time == 0) intr = Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(0), &Wimax2Timer::handle, this);
  //else intr = Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(rtime), &Wimax2Timer::handle, this/*function,args*/);

//  s.schedule(this, &intr, rtime); //schedule the event
}

/*
 * Stop the timer
 */
void Wimax2Timer::stop(void)
{
  //Scheduler &s = Scheduler::instance();
  
  assert(busy_);
  
  if(paused_ == 0) Simulator::Cancel(intr);
    //s.cancel(&intr); //cancel the event
  
  status_ = TIMER_IDLE;
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;
}

void Wimax2Timer::pause(void)
{
	//Scheduler &s = Scheduler::instance();

	//assert(busy_ && ! paused_);
	if (!busy_ || paused_) {
		debug2 ("trying to pause a timer that is not running, doing nothing.\n");
		return;
	}

	paused_ = 1;
	rtime -= Simulator::Now ().GetSeconds ()-stime;

	assert(rtime >= 0.0);

	Simulator::Cancel(intr);
//	s.cancel(&intr);
}


void Wimax2Timer::resume(void)
{
	//Scheduler &s = Scheduler::instance();

	assert(busy_ && paused_);
	status_ = TIMER_PENDING;

	paused_ = 0;
	stime = Simulator::Now ().GetSeconds ();

	assert(rtime >= 0.0);
	debug10("Resuming a timer\n");
    intr = Simulator::Schedule(Seconds(rtime), &Wimax2Timer::handle, this/*function,args*/);

//	Simulator::Schedule(mac->GetNode()->GetId(), rtime, &handle);
    //   	s.schedule(this, &intr, rtime );
}


void Wimax2Timer::resched(double delay) {
	debug10("rescheduling a timer for %f seconds from now\n", delay);
	if (busy_ == 1)
		Simulator::Cancel(intr);
	rtime = delay;
	busy_ = 1;
	paused_ = 0;
	stime = NOW;
    rtime = delay;
    assert(rtime >= 0.0);
//    intr = Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(rtime), &Wimax2Timer::handle, this/*function,args*/);
    intr = Simulator::Schedule(Seconds(rtime), &Wimax2Timer::handle, this/*function,args*/);
	status_ = TIMER_PENDING;	
}

void Wimax2Timer::resched_absolute(double expiration_time) {
	debug10("rescheduling a timer for %f seconds from the start of simulation\n", expiration_time);
	if (busy_ == 1)
		Simulator::Cancel(intr);
	busy_ = 1;
	paused_ = 0;
	stime = Simulator::Now ().GetSeconds ();
    rtime = expiration_time - stime;
    assert(rtime >= 0.0);
    //intr = Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(rtime), &Wimax2Timer::handle, this/*function,args*/);
    intr = Simulator::Schedule( Seconds(rtime), &Wimax2Timer::handle, this/*function,args*/);
	status_ = TIMER_PENDING;	
}



/*
 * Handling function for WimaxFrameTimer
 */
void WimaxRxTimer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;
  mac->receive(record_->p);
  mac->removePacket(record_);
}

WimaxRxTimer::WimaxRxTimer(Ptr<Wimax2NetDevice> m, PacketTimerRecord *ptr) : Wimax2Timer(m) {
	record_ = ptr;
}

/*
 * Handling function for WimaxFrameTimer
 */
void WimaxT1Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;
  mac->expire(WimaxT1TimerID);
}

/*
 * Handling function for WimaxFrameTimer
 */
void WimaxT2Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxT2TimerID);
}

/*
 * Handling function for WimaxT3Timer
 */
void WimaxT3Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxT3TimerID);
}

/*
 * Handling function for WimaxT6Timer
 */
void WimaxT6Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxT6TimerID);
}

/*
 *Handling function for WimaxT9Timer
 */
void WimaxT9Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxT9TimerID);
}

/*
 * Handling function for WimaxT12Timer
 */
void WimaxT12Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxT12TimerID);
}

/*
 * Handling function for WimaxT16Timer
 */
void WimaxT16Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxT16TimerID);
}

/*
 * Handling function for WimaxT17Timer
 */
void WimaxT17Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  /** The node did not send a registration: release and
   *  age out Basic and Primary CIDs
   */
  PeerNode * peer = mac->getPeerNode (peerIndex_);
  debug2 ("At %f in Mac %s did not register on time...release CIDs\n", Simulator::Now ().GetSeconds (), HelperFunctions::getMacAddrString(mac->addr()).c_str());
  mac->removePeerNode (peer);
}


/*
 * Handling function for WimaxT21Timer
 */
void WimaxT21Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxT21TimerID);
}

/*
 * Handling function for WimaxFrameTimer
 */
void WimaxLostDLMAPTimer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxLostDLMAPTimerID);
}

/*
 * Handling function for WimaxFrameTimer
 */
void WimaxLostULMAPTimer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxLostULMAPTimerID);
}

/*
 * Handling function for WimaxDCDTimer
 */
void WimaxDCDTimer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxDCDTimerID);
}

WimaxDCDTimer::WimaxDCDTimer(Ptr<Wimax2NetDevice> m) : Wimax2Timer(m) {}

/*
 * Handling function for WimaxUCDTimer
 */
void WimaxUCDTimer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxUCDTimerID);
}

WimaxUCDTimer::WimaxUCDTimer(Ptr<Wimax2NetDevice> m) : Wimax2Timer(m) {}
/*
 * Handling function for WimaxScanIntervalTimer
 */
void WimaxScanIntervalTimer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxScanIntervalTimerID);
}

/*
 * Handling function for WimaxT44Timer
 */
void WimaxT44Timer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxT44TimerID);
}

/*
 * Handling function for WimaxMobNbrAdvTimer
 */
void WimaxMobNbrAdvTimer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxMobNbrAdvTimerID);
}

/*
 * Handling function for WimaxRdvTimer
 */
void WimaxRdvTimer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxRdvTimerID);
  printf ("Rdv timeout going to channel %d\n", channel_);
  mac->setChannel (channel_);
}




/*
 * Handling function for WimaxT21Timer
 */
void WimaxRngIntTimer::handle()
{
  busy_ = 0;
  paused_ = 0;
  stime = 0.0;
  rtime = 0.0;

  mac->expire(WimaxRngIntTimerID);
}

/**
 Constructors for most of the timer subclasses
 */

 WimaxRngIntTimer::WimaxRngIntTimer(Ptr<Wimax2NetDevice> m) : Wimax2Timer (m) {}
 WimaxLostDLMAPTimer::WimaxLostDLMAPTimer (Ptr<Wimax2NetDevice> m) : Wimax2Timer (m) {}
 WimaxLostULMAPTimer::WimaxLostULMAPTimer(Ptr<Wimax2NetDevice> m ) : Wimax2Timer (m) {}
 WimaxT1Timer::WimaxT1Timer (Ptr<Wimax2NetDevice> m ) : Wimax2Timer(m) {}
 WimaxT2Timer::WimaxT2Timer (Ptr<Wimax2NetDevice> m ) : Wimax2Timer(m) {}
 WimaxT3Timer::WimaxT3Timer (Ptr<Wimax2NetDevice> m ) : Wimax2Timer(m) {}
 WimaxT6Timer::WimaxT6Timer (Ptr<Wimax2NetDevice> m ) : Wimax2Timer(m) {}
 WimaxT9Timer::WimaxT9Timer (Ptr<Wimax2NetDevice> m ) : Wimax2Timer(m) {}
 WimaxT12Timer::WimaxT12Timer (Ptr<Wimax2NetDevice> m ) : Wimax2Timer(m) {}
 WimaxT16Timer::WimaxT16Timer (Ptr<Wimax2NetDevice> m ) : Wimax2Timer(m) {}
 WimaxT17Timer::WimaxT17Timer (Ptr<Wimax2NetDevice> m, Mac48Address peerIndex) : Wimax2Timer(m) {peerIndex_ = peerIndex;}
 WimaxT21Timer::WimaxT21Timer (Ptr<Wimax2NetDevice> m ) : Wimax2Timer(m) {}
 WimaxT44Timer::WimaxT44Timer (Ptr<Wimax2NetDevice> m ) : Wimax2Timer(m) {}
 WimaxScanIntervalTimer::WimaxScanIntervalTimer(Ptr<Wimax2NetDevice> m) : Wimax2Timer (m) {}
 WimaxMobNbrAdvTimer::WimaxMobNbrAdvTimer(Ptr<Wimax2NetDevice> m) : Wimax2Timer (m) {}
 WimaxRdvTimer::WimaxRdvTimer (Ptr<Wimax2NetDevice> m, int channel) : Wimax2Timer (m) {channel_ = channel;}






/*
 * Indicates the beginning of DL subframe
 */
void DlTimer::expire ()
{
  m_->start_dlsubframe();
}

void DlTimer::resched(double d) {
	if (IsRunning()) Cancel();
	SetDelay(Seconds(d));
	Schedule();
}

void UlTimer::resched(double d) {
	if (IsRunning()) Cancel();
	SetDelay(Seconds(d));
	Schedule();
}

/*
 * Indicates the beginning of UL subframe
 */
void UlTimer::expire ()
{
  m_->start_ulsubframe();
}

UlTimer::UlTimer(Ptr<Wimax2NetDevice> m) : Timer(REMOVE_ON_DESTROY) {
	m_=m;
	SetFunction(&ns3::UlTimer::expire, this);
}

DlTimer::DlTimer(Ptr<Wimax2NetDevice> m) : Timer(REMOVE_ON_DESTROY) {
	m_=m;
	SetFunction(&ns3::DlTimer::expire, this);
}

/*
 * Initialize the MAC
 */
void InitTimer::expire (/**/)
{
  m_->init();
}

InitTimer::InitTimer(Ptr<Wimax2NetDevice> m) : Timer(REMOVE_ON_DESTROY) {
  	m_=m;
  	SetFunction(&ns3::InitTimer::expire, this);
}


/*
 * Handling function for ArqTimer
 */
void ARQTimer::expire()
{
//	connection_->getArqStatus ()->arqTimerHandler();
}

PacketTimerRecord::PacketTimerRecord(Ptr<Wimax2NetDevice> m ) : timer(m, this) {
	prev = NULL;
	next = NULL;
	return;
}

} //namespace ns3
