/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef Wimax2NetDeviceTIMER_H
#define Wimax2NetDeviceTIMER_H



#include "ns3/scheduler.h"
//#include "ns3/wimax2-net-device.h"
#include "ns3/ptr.h"
#include "ns3/packet.h"
#include "ns3/timer.h"
#include "ns3/event-id.h"
#include "ns3/mac48-address.h"
#include "ns3/wimax2-common.h"
#include "ns3/wimax2-common.h"
//#include "timer-handler.h"

namespace ns3 {



class Wimax2NetDevice;
struct PacketTimerRecord; /// added rpi. 




/**
 * Super class for timers used in wimax 
 */
class Wimax2Timer {//: public Handler {
public:
	Wimax2Timer(Ptr<Wimax2NetDevice> m);
	virtual void handle() = 0;

	virtual void start(double time);
	virtual void stop(void);
	void pause(void); /*{ assert(0); }*/
	void resume(void);/* { assert(0); }*/

	inline int busy(void) { return busy_; }
	inline int paused(void) { return paused_; }
	inline double expire(void) {
		return ((stime + rtime) - Simulator::Now ().GetSeconds ());
	}
	
	void resched(double delay);
	void resched_absolute(double expiration_time);

	inline int status() {return status_;}

//protected: this would be uncommented but the python binding generator can't handle protected variables
	Ptr<Wimax2NetDevice> mac;
	int		busy_;
	int		paused_;
	EventId		intr;
	double		stime;	// start time
	double		rtime;	// remaining time
	int 		status_;
};

/** Timer for receiving a packet */
class WimaxRxTimer : public Wimax2Timer {
 public:
  WimaxRxTimer(Ptr<Wimax2NetDevice> m, PacketTimerRecord *ptr);// : Wimax2Timer(m) {
  
  void  handle();

// protected: this would be uncommented but the python binding generator can't handle protected variables
  PacketTimerRecord *record_;

};

/** Timer for DCD interval */
class WimaxDCDTimer : public Wimax2Timer {
 public:
  WimaxDCDTimer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for UCD interval */
class WimaxUCDTimer : public Wimax2Timer {
 public:
  WimaxUCDTimer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for initial ranging regions interval */
class WimaxRngIntTimer : public Wimax2Timer {
 public:
  WimaxRngIntTimer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for Lost DL-MAP interval */
class WimaxLostDLMAPTimer : public Wimax2Timer {
 public:
  WimaxLostDLMAPTimer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for Lost UL-MAP interval */
class WimaxLostULMAPTimer : public Wimax2Timer {
 public:
  WimaxLostULMAPTimer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for T1 : wait for DCD timeout */
class WimaxT1Timer : public Wimax2Timer {
 public:
  WimaxT1Timer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for T2 : wait for broadcast ranging timeout */
class WimaxT2Timer : public Wimax2Timer {
 public:
  WimaxT2Timer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for T3 : ranging response timeout */
class WimaxT3Timer : public Wimax2Timer {
 public:
  WimaxT3Timer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for T6 : wait for registration response */
class WimaxT6Timer : public Wimax2Timer {
 public:
  WimaxT6Timer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for T9 : registration timeout */
class WimaxT9Timer : public Wimax2Timer {
 public:
  WimaxT9Timer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for T12 : wait for UCD descriptor */
class WimaxT12Timer : public Wimax2Timer {
 public:
  WimaxT12Timer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for T16 : wait for bw request grant */
class WimaxT16Timer : public Wimax2Timer {
 public:
  WimaxT16Timer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for T17 : wait for SS to register */
class WimaxT17Timer : public Wimax2Timer {
 public:
  WimaxT17Timer(Ptr<Wimax2NetDevice> m, Mac48Address peerIndex);
  
  void	handle();
 private:
  Mac48Address peerIndex_;
};

/** Timer for T21 : time the station searches for DL-MAP on a channel */
class WimaxT21Timer : public Wimax2Timer {
 public:
  WimaxT21Timer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for T44 : wait for BS to send MOB_SCN-REP */
class WimaxT44Timer : public Wimax2Timer {
 public:
  WimaxT44Timer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for scan interval : timer for scanning */
class WimaxScanIntervalTimer : public Wimax2Timer {
 public:
  WimaxScanIntervalTimer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for neighbor advertisement interval */
class WimaxMobNbrAdvTimer : public Wimax2Timer {
 public:
  WimaxMobNbrAdvTimer(Ptr<Wimax2NetDevice> m);
  
  void	handle();
};

/** Timer for rendez-vous with target BSs */
class WimaxRdvTimer : public Wimax2Timer {
 public:
  WimaxRdvTimer(Ptr<Wimax2NetDevice> m, int channel);
  
  void	handle();
 private:
  int channel_;
};

/** Timer to indicate a new Downlink frame */
class DlTimer : public Timer {
 public:
  DlTimer(Ptr<Wimax2NetDevice> m);// : Timer(REMOVE_ON_DESTROY) {m_ = m;}
  void resched(double d);
  void	expire();
 private:
  Ptr<Wimax2NetDevice> m_;
}; 

/** Timer to indicate a new uplink frame */
class UlTimer : public Timer {
 public:
  UlTimer(Ptr<Wimax2NetDevice> m);// : Timer(REMOVE_ON_DESTROY) {m_=m;}
  void resched (double d);
  void	expire();
 private:
  Ptr<Wimax2NetDevice> m_;
}; 

/** Timer to indicate a new uplink frame */
class InitTimer : public Timer {
 public:
  InitTimer(Ptr<Wimax2NetDevice> m);// : Timer(REMOVE_ON_DESTROY) {m_=m;}
  
  void	expire(/**/);
 private:
  Ptr<Wimax2NetDevice> m_;
}; 

/** ArqTimer Class */
class Connection;
class ARQTimer : public Timer {

 public: 
 ARQTimer(Connection *connection): Timer(REMOVE_ON_DESTROY) { connection_ = connection;}
 void expire();
 inline Connection* getConnection() { return connection_; }
 inline void setConnection(Connection *connection) { connection_ = connection; }
 
 inline void sched(double s) {Schedule(Seconds(s));}
   
// protected: this would be uncommented but the python binding generator can't handle protected variables
 Connection *connection_;

};

// This struct stores info about a packt currently being recvd. These structs are dynamically stored in a doubly linked list during simulation.  
struct PacketTimerRecord {
        PacketTimerRecord( Ptr<Wimax2NetDevice> m );
        WimaxRxTimer timer;
        PacketTimerRecord *prev;
        PacketTimerRecord *next;
        Ptr<Packet> p;
};
} //namespace ns3

#endif
