/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#include "wimaxneighborentry.h"
namespace ns3 {
/** 
 * Constructor
 */
WimaxNeighborEntry::WimaxNeighborEntry (Mac48Address id) : nbr_adv_(0), dcd_(0), ucd_ (0), rng_rsp_(0)
{
  id_ = id;
  memset (&state_, 0, sizeof (struct sched_state_info));
  detected_ = false;
}

/**
 * Destructor
 */
WimaxNeighborEntry::~WimaxNeighborEntry ()
{
  free (nbr_adv_);
  free (dcd_);
  free (ucd_);
  free (rng_rsp_);
}


/**
 * Return the address of the neighbor of this entry
 * @return the address of the neighbor of this entry
 */
Mac48Address WimaxNeighborEntry::getID ()
{
  return id_;
}



/**
 * Set the neighbor advertisement message
 * @param frame The advertisment message
 */
void WimaxNeighborEntry::setNbrAdvMessage (mac802_16_nbr_adv_info *frame)
{
  nbr_adv_ = frame;
}
  
/**
 * Return the neighbor advertisement message
 * @param frame The advertisment message
 */
mac802_16_nbr_adv_info * WimaxNeighborEntry::getNbrAdvMessage ()
{
  return nbr_adv_;
}

/**
 * Set the DCD message received during scanning
 * @param dcd the DCD message received
 */
void WimaxNeighborEntry::setDCD (DCD_FRAME_HEADER *frame)
{
  dcd_ = frame;
}

/**
 * Get the DCD message received during scanning
 * @return the DCD message received
 */
DCD_FRAME_HEADER * WimaxNeighborEntry::getDCD ()
{
  return dcd_;
}

/**
 * Set the UCD message received during scanning
 * @param dcd the DCD message received
 */
void WimaxNeighborEntry::setUCD (UCD_FRAME_DATA *frame)
{
  ucd_ = frame;
}

/**
 * Get the DCD message received during scanning
 * @return the DCD message received
 */
UCD_FRAME_DATA * WimaxNeighborEntry::getUCD ()
{
  return ucd_;
}

/**
 * Set the MAC state associated with this neighbor
 * @param state
 */
/*void WimaxNeighborEntry::setState (sched_state_info *state)
{
  state_ = state;
}*/

/**
 * Get the MAC state associated with this neighbor
 * @return the MAC state associated with this neighbor
 */
sched_state_info * WimaxNeighborEntry::getState ()
{
  return &state_;
}

/**
 * Mark the neighbor as being detected
 * @param detected indicate if the neighbor has been detected
 */
void WimaxNeighborEntry::setDetected (bool detected)
{
  detected_ = detected;
}

/**
 * Indicates the neighbor as being detected
 * @return indication if the neighbor has been detected
 */
bool WimaxNeighborEntry::isDetected ()
{
  return detected_;
}

/**
 * Set the UCD message received during scanning
 * @param dcd the DCD message received
 */
void WimaxNeighborEntry::setRangingRsp (RANGING_RESPONSE_FRAME *frame)
{
  rng_rsp_ = frame;
}

/**
 * Get the DCD message received during scanning
 * @return the DCD message received
 */
RANGING_RESPONSE_FRAME *WimaxNeighborEntry::getRangingRsp ()
{
  return rng_rsp_;
}

} //namespace ns3
