/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ns3/core-module.h"
#include "ns3/simulator-module.h"
#include "ns3/node-module.h"
#include "ns3/helper-module.h"
//#include "ns3/wimax2-net-device.h"
//#include "ns3/wimax2-helper.h"
//#include "ns3/wimax2-bs-net-device.h"
#include "ns3/wimax2-module.h"
#include "ns3/global-router-interface.h"
#include "ns3/ipv4-list-routing.h"
#include "ns3/ipv4-global-routing.h"
#include "ns3/ipv4-static-routing-helper.h"
#include "ns3/ipv4-global-routing-helper.h"
#include "ns3/mobility-module.h"
#include "ns3/random-variable.h"
#include "ns3/data-rate.h"
#include "ns3/inet-socket-address.h"
#include <iostream>

using namespace ns3;

NS_LOG_COMPONENT_DEFINE ("Wimax2ScriptExample");


/** 
	Note that output from this script is sent to std::cerr instead of std::cout
	This to allow it to be easiliy seperated from the wimax model's output
	routing standard output to a file will still print error output to stdout
	so we can see script-level output but everything else is sent to a file
*/

int 
main (int argc, char *argv[])
{
  bool verbose = true;
  //uint32_t nCsma = 3;
  uint32_t nWimax= 12;



  CommandLine cmd;
  //cmd.AddValue ("nCsma", "Number of \"extra\" CSMA nodes/devices", nCsma);
  cmd.AddValue ("nWimax", "Number of \"extra\" WiMAX nodes/devices", nWimax);
  cmd.AddValue ("verbose", "Tell echo applications to log if true", verbose);

  cmd.Parse (argc,argv);

//  if (verbose)
//    {
//      LogComponentEnable("UdpEchoClientApplication", LOG_LEVEL_INFO);
//      LogComponentEnable("UdpEchoServerApplication", LOG_LEVEL_INFO);
//    }

  //nCsma = nCsma == 0 ? 1 : nCsma;

  printf ("Creating nodes...\n");

  NodeContainer wmxNodes;
  wmxNodes.Create (nWimax);
  NodeContainer wmxBSNode;
  wmxBSNode.Create(1);

//  NodeContainer csmaNodes;
//  csmaNodes.Add (wmxBSNode.Get (0));
//  csmaNodes.Create (nCsma);


  printf ("finished creating nodes...\n");

	printf ("about to start wimax device creation\n");

	Mac802_16MIB *mmib = new Mac802_16MIB();
	Phy802_16MIB *pmib = new Phy802_16MIB();
	Wimax2Helper wm2;

	pmib->rtg = 20;
	pmib->ttg = 20;
	pmib->fbandwidth = 10e6;
	mmib->frame_duration = 0.005;
	mmib->ITU_PDP_ = 2;
	wm2.setDlRatio(0.3);

	wm2.setMacMIB(mmib);
	wm2.setPhyMIB(pmib);

	Ptr<Wimax2Channel> chan = CreateObject<Wimax2Channel>();

	NetDeviceContainer wmxDevices;
	NetDeviceContainer wmxBSDevice;
	wmxBSDevice = wm2.InstallBS(wmxBSNode, chan);
	wmxDevices = wm2.Install(wmxNodes, chan);

	printf ("finished wimax device creation\n");


  MobilityHelper mob;
  SequentialVariable seqvar(0.0, 7.283, 6.283/(double)nWimax, 1);
  mob.SetPositionAllocator("ns3::RandomDiscPositionAllocator", "X", DoubleValue(0.0),
  							"Y", DoubleValue(0.0), "Rho", RandomVariableValue(ConstantVariable(35.0)),
							"Theta", RandomVariableValue(seqvar));
  mob.SetMobilityModel("ns3::ConstantPositionMobilityModel");

  mob.Install(wmxNodes);

  mob.SetPositionAllocator("ns3::GridPositionAllocator", "MinX", DoubleValue(0.0), "MinY", DoubleValue(0.0), "DeltaX", DoubleValue(10.0), "DeltaY", DoubleValue(10.0), "GridWidth", UintegerValue(1), "LayoutType", StringValue("RowFirst"));
  mob.Install(wmxBSNode);



//  CsmaHelper csma;
//  csma.SetChannelAttribute ("DataRate", StringValue ("100Mbps"));
//  csma.SetChannelAttribute ("Delay", TimeValue (NanoSeconds (6560)));


 // NetDeviceContainer csmaDevices;
 // csmaDevices = csma.Install (csmaNodes);

  InternetStackHelper stack;
  stack.Install (wmxNodes);
  stack.Install(wmxBSNode);
  //stack.Install (csmaNodes);

  Ipv4AddressHelper address;
  Ipv4Address wimaxBaseNetwork("10.1.1.0");
  address.SetBase (wimaxBaseNetwork, "255.255.255.0");
  Ipv4InterfaceContainer wmxInterfaces, wmxBSInterface;
  wmxBSInterface = address.Assign(wmxBSDevice);
  wmxInterfaces = address.Assign (wmxDevices);

//  address.SetBase ("10.1.2.0", "255.255.255.0");
//  Ipv4InterfaceContainer csmaInterfaces;
//  csmaInterfaces = address.Assign (csmaDevices);

 // UdpEchoServerHelper echoServer (9);
 
 
 Time end_of_simulation = Seconds ( nWimax + 5);

// ApplicationContainer serverApps = echoServer.Install (wmxNodes.Get(nWimax-1));
// serverApps.Start (Seconds (1));
// serverApps.Stop (end_of_simulation);
  PacketSinkHelper sink ("ns3::UdpSocketFactory", InetSocketAddress (wmxBSInterface.GetAddress(0), 8080));
  ApplicationContainer sinks = sink.Install(wmxNodes);
  sinks.Start(Seconds(.5));
  sinks.Stop (end_of_simulation);

 UniformVariable unfv;
 ExponentialVariable expv( 500, 1500);

 //we set each node to send 1-512 packets of 1 KB each second
 for (uint32_t i = 0; i < nWimax; i++) {
 	OnOffHelper oof("ns3::UdpSocketFactory", InetSocketAddress(wmxInterfaces.GetAddress(i), 8080));
	oof.SetAttribute("OnTime", RandomVariableValue(ConstantVariable (ceil(end_of_simulation.GetSeconds())))); //just leave it on until the end of the simulation
	oof.SetAttribute("OffTime", RandomVariableValue(ConstantVariable (0))); //never
	uint32_t pkt_size = expv.GetInteger();
	oof.SetAttribute("PacketSize", UintegerValue(pkt_size));
	DataRate dr(unfv.GetInteger(200000, 600000));
	oof.SetAttribute("DataRate", DataRateValue(dr));
 	ApplicationContainer newApp = oof.Install(wmxBSNode);

	newApp.Start(Seconds( i + 1));
	newApp.Stop(end_of_simulation);
	std::cerr << "BS will send " << pkt_size << "bytes to " << wmxInterfaces.GetAddress(i) << " often enough to maintain " << dr.GetBitRate()/1000 << " Kb/s, starting at " << i+1 << " Seconds\n";
}
 

  Ipv4GlobalRoutingHelper::PopulateRoutingTables ();
  
  Ptr<Ipv4> ipv4 = wmxNodes.Get(0)->GetObject<Ipv4> ();
  int16_t a = 0; if (a);
  
  for (uint32_t i = 0; i < wmxNodes.GetN(); i++) {
        int16_t pri = 0;
        Ptr<Ipv4ListRouting> lr = wmxNodes.Get(i)->GetObject<Ipv4>()->GetRoutingProtocol()->GetObject<Ipv4ListRouting>();
        Ptr<Ipv4StaticRouting> sr = lr->GetRoutingProtocol(0,pri)->GetObject<Ipv4StaticRouting>();
        sr->SetDefaultRoute(wmxBSInterface.GetAddress(0,0),1,0);

        for (uint32_t j = 0; j < sr->GetNRoutes(); j++) {
                //the routing helper used to install the static/list routing systems automatically sets a 
                //network route assuming all devices on the same channel can communicate directly, which is
                //not the case here, as they must go through the BS.
                if (sr->GetRoute(j).IsNetwork() && sr->GetRoute(j).GetDestNetwork() == wimaxBaseNetwork) {
                        sr->RemoveRoute(j);
                }   
        }   

	Ptr<NetDevice> nd = wmxNodes.Get(i)->GetDevice(0);
  	ServiceFlowHandler *sfh = nd->GetObject<Wimax2NetDevice>()->getServiceHandler();
  	
  	sfh->addStaticFlow(true, 10000, SERVICE_BE, 275, 2, 0, 0.05, 1024, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  	sfh->addStaticFlow(false, 10000, SERVICE_BE, 275, 2, 0, 0.05, 1024, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  	
  }
  
//  csma.EnablePcapAll ("second");
  wm2.EnablePcapAll ("second");

  Simulator::Stop(end_of_simulation);
  Simulator::Run ();

  std::cerr << "Finished Simulation\n";

  std::cerr << "Printing Mapping Stats:\n";
  Wimax2BSScheduler *sch = reinterpret_cast<Wimax2BSScheduler*>(wmxBSNode.Get(0)->GetDevice(0)->GetObject<Wimax2BSNetDevice>()->getScheduler());
  for (uint16_t i = 1; i < sch->getMaximumNumberDlBursts(); i++) 
	fprintf(stderr, "FINAL STATS FOR %d BURSTS: Number of Frames: %llu, Total Overallocation: %llu, Total Unmapped Blocks: %llu, Total Unmapped Bursts %llu, Total Empty Blocks (that were needed for a burst): %llu\n", i, sch->getNumberFramesWithNumberBursts(i), sch->getTotalOverallocations(i), sch->getTotalUnmappedBlocks(i), sch->getTotalUnmappedBursts(i), sch->getTotalEmptyNeededBlocks(i));

  Simulator::Destroy ();
  return 0;
}

